/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime;

import java.awt.Color;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Locale;
import java.util.Properties;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.xpath.XPathExpressionException;

import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.utility.XMLUtil;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * T2Mapのプロパティファイルを取得するクラスです。
 */
public class T2MapProperties extends Properties
{
    /**
     * シリアルバーションID
     */
    private static final long serialVersionUID = 1L;

    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(T2MapProperties.class, T2Map.LOCALE);

    /**
     * 一般設定ファイルパス
     */
    private static final String fname = "/T2Map/t2map.properties";

    /**
     * 表示設定ファイルパス
     */
    private static final String stylePropertiesFname = "/T2Map/default_temp.xml";

    /**
     * 一般設定ファイル
     */
    private File file;

    /**
     * 表示設定ファイル
     */
    private File stylePropertiesFile;

    /**
     * 表示設定
     */
    private Document styleProperties;

    private void copyDefaultProperties(String inName, File outFile) throws IOException
    {
        InputStream in = getClass().getClassLoader().getResourceAsStream(inName);
        FileOutputStream out = new FileOutputStream(outFile);
        int read = 0;
        int pos = 0;
        int len = 36;
        byte[] buf = new byte[len];
        while ((read = in.read(buf, 0, len)) != -1)
        {
            out.write(buf, 0, read);
            pos = pos + read;
        }
        in.close();
        out.flush();
        out.close();
    }

    /**
     * プロパティファイルの内容を取得します。
     * 
     * @throws T2MapPropertiesException
     */
    public T2MapProperties() throws T2MapPropertiesException
    {
        try
        {
            file = new File(System.getProperty("user.home") + fname);
            if (!file.exists() && (file.getParentFile().exists() || file.getParentFile().mkdirs()))
            {
                copyDefaultProperties("t2map.properties", file);
            }
            load(new FileInputStream(file));

            if (getProperty("jp.nihu.gttime.T2Map.PLUGIN_DIR") == null)
            {
                copyDefaultProperties("t2map.properties", file);
                load(new FileInputStream(file));
            }

            try
            {
                stylePropertiesFile = new File(System.getProperty("user.home") + stylePropertiesFname);
                if (!stylePropertiesFile.exists() && (stylePropertiesFile.getParentFile().exists() || stylePropertiesFile.getParentFile().mkdirs()))
                {
                    copyDefaultProperties("default_temp.xml", stylePropertiesFile);
                }

                DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
                factory.setNamespaceAware(true);
                DocumentBuilder builder = null;
                builder = factory.newDocumentBuilder();
                styleProperties = builder.parse(new FileInputStream(stylePropertiesFile));
            }
            catch (Exception e)
            {
                throw new T2MapPropertiesException(e.getMessage(), e);
            }

        }
        catch (IOException e)
        {
            throw new T2MapPropertiesException(rmgr.format("error.message", fname), e);
        }
    }

    public void setStyleValue(String xpath, String newValue) throws XPathExpressionException
    {
        if (XMLUtil.getElement(styleProperties, xpath, XMLUtil.getGtsNsContext()) == null) {
            String parentXpath = xpath.substring(0, xpath.lastIndexOf("/"));
            String elemName = xpath.substring(xpath.lastIndexOf("/")+1);
            Element elem = XMLUtil.getElement(styleProperties, parentXpath, XMLUtil.getGtsNsContext());
            Element ne = styleProperties.createElement(elemName);
            ne.setTextContent(newValue);
            elem.appendChild(ne);
            
        } else {
            XMLUtil.getElement(styleProperties, xpath, XMLUtil.getGtsNsContext()).setTextContent(newValue);
        }
    }

    public String getStyleValue(String xpath) throws XPathExpressionException
    {
        return XMLUtil.getValue(styleProperties, xpath, XMLUtil.getGtsNsContext());
    }
    
    public NodeList getStyleValues(String xpath) throws XPathExpressionException
    {
        return XMLUtil.getElements(styleProperties, xpath, XMLUtil.getGtsNsContext());
    }

    public void setLocale(Locale locale)
    {
        rmgr = PluginResourceManager.getManager(T2MapProperties.class, locale);
    }

    /**
     * 適切なフォーマットで出力ストリームに書き込みます。
     * 
     * @throws FileNotFoundException
     * @throws IOException
     * @throws T2MapPropertiesException 
     */
    public void store() throws FileNotFoundException, IOException
    {
        super.store(new FileOutputStream(file), null);
        XMLUtil.saveXml(styleProperties, stylePropertiesFile);
        
        try {
            DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
            factory.setNamespaceAware(true);
            DocumentBuilder builder = null;
            builder = factory.newDocumentBuilder();
            styleProperties = builder.parse(new FileInputStream(stylePropertiesFile));
        }
        catch (Exception e)
        {
            throw new IOException(e.getMessage(), e);
        }
    }

    /**
     * Propertyを取得します。
     * 
     * @param name Propertyの名前
     * @return トークンに分割された文字列
     */
    public String[] getProperties(String name)
    {
        String pdirs = getProperty(name);
        String[] values = pdirs.split(";");
        return values;
    }

    /**
     * プロパティから色を取得します。
     * 
     * @param key Propertyの名前
     * @return 現在設定されている色
     */
    public Color getColor(String key)
    {
        Color color = null;
        String value = getProperty(key);
        if (value != null)
        {
            String[] vals = value.split(" ");
            int n = vals.length;
            if (n == 3)
            {
                color = new Color(Integer.parseInt(vals[0]), Integer.parseInt(vals[1]), Integer.parseInt(vals[2]));
            }
            else if (n == 4)
            {
                color = new Color(Integer.parseInt(vals[0]), Integer.parseInt(vals[1]), Integer.parseInt(vals[2]), Integer.parseInt(vals[3]));
            }
        }
        return color;
    }

    public int getInt(String key, int defaultValue) throws NumberFormatException
    {
        String value = getProperty(key);
        int n = 0;
        try
        {
            n = Integer.parseInt(value);
        }
        catch (NumberFormatException e)
        {
            n = defaultValue;
            setProperty(key, Integer.toString(defaultValue));
            try
            {
                store();
            }
            catch (FileNotFoundException e1)
            {
            }
            catch (IOException e1)
            {
            }
        }
        return n;
    }

    public boolean getBoolean(String key) throws NumberFormatException
    {
        String value = getProperty(key);
        return Boolean.valueOf(value);
    }

    /**
     * @see java.util.Hashtable#equals(java.lang.Object)
     */
    public boolean equals(Object o)
    {
        return super.equals(o);
    }
}
