/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.data;

import java.io.IOException;
import java.math.BigDecimal;
import java.util.List;

import jp.nihu.gttime.data.property.PropertiesIF;
import jp.nihu.gttime.plugin.OpenPlugin;

import org.semarglproject.rdf.ParseException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.hp.hpl.jena.rdf.model.Model;

public interface DatasetIF
{
    public List<String> getGroups();
    
    public void setGroups(List<String> groups);
    
    public void setGroup(String group);
    
    public PropertiesIF getProperties();

    public void setProperties(PropertiesIF properties);

    public void setProperties(Element time);

    public void reloadProperties(String layerName);
    
    public String getRdfBase();
    
    public Model getModel() throws ParseException, IOException;

    /**
     * データセットファイルのDocumentを取得します。
     * 
     * @return データセットファイルのDocument
     */
    public Document getDomDataset();

    /**
     * データセットファイルのDocumentを設定します。
     * 
     * @param dom データセットファイルのDocument
     */
    public void setDomDataset(Document doc);

    /**
     * 指定されたデータセットをロードします。
     * 
     * @param op プラグイン名
     * @param path ディレクトリパス
     * @param base ファイル名
     * @throws DatasetException
     */
    public void loadDataset(OpenPlugin op, String path, String base) throws DatasetException;

    /**
     * 指定されたデータセットをロードします。
     * 
     * @param op プラグイン名
     * @param pathx ディレクトリパス
     * @param base ファイル名
     * @param uniquePath 他のTMMファイルと同じでないファイル名
     * @throws DatasetException
     */
    public void loadDataset(OpenPlugin op, String pathx, String base, String uniquePath) throws DatasetException;

    /**
     * レコードが存在するかチェックします。
     * 
     * @return レコードが存在する、しない
     */
    public boolean hasRecords();

    /**
     * レコードを配列で取得します。
     * 
     * @return レコードの配列
     */
    public Record[] getRecords();
    
    public Record[] getRecords(String group);

    /**
     * 指定された範囲内にあるレコードを配列で取得します。
     * 
     * @param s 開始日時
     * @param e 終了日時
     * @return レコードの配列
     */
    public Record[] getRecords(BigDecimal s, BigDecimal e);

    public Record[] getHitRecords();

    /**
     * レイヤ表示設定をプロジェクト（のDOM）に保存する。
     */
    public void applyStyleChanges(String layerTypeName, String layerStyleName);

    public void setFragmentDatasetNodeOfProject(Element fragment);

    /**
     * レイヤタイプを設定します。
     * 
     * @param layertype レイヤタイプ名
     */
    public void setLayerType(String layertype);

    /**
     * レイヤタイプを取得します。
     * 
     * @return レイヤタイプ名
     * @throws DatasetException
     */
    public String getLayerType();

    /**
     * データセットのタイトルを取得します。
     * 
     * @return タイトル
     * @throws DatasetException
     */
    public String getTitle();

    /**
     * データセットのタイトルを設定します。
     * 
     * @param title タイトル
     */
    public void setTitle(String title);

    /**
     * パスを取得します。
     * 
     * @return パス
     */
    public String getPath();
    
    public void setPath(String path);
    
    public String getAbsolutePath();
    
    public void setAbsolutePath(String absolutePath);

    /**
     * 開始日時を取得します。
     * 
     * @return 開始日時
     * @throws DatasetException
     */
    public BigDecimal getStartDatetime() throws DatasetException;

    /**
     * 終了日時を取得します。
     * 
     * @return 終了日時
     * @throws DatasetException
     */
    public BigDecimal getEndDatetime() throws DatasetException;

    /**
     * @see java.lang.Object#toString()
     */
    public String toString();

    public boolean isVisible();

    /**
     * @param visible the visible to set
     */
    public void setVisible(boolean visible);
    
    public String getRecordPath();
}
