/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.data;

import java.io.IOException;
import java.io.StringReader;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.property.PropertiesIF;
import jp.nihu.gttime.plugin.OpenPlugin;
import jp.nihu.gttime.plugin.OpenPluginException;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.rdf.RDFaReader;
import jp.nihu.gttime.time.utility.ISO8601Exception;
import jp.nihu.gttime.time.utility.JulianGregorianDate;
import jp.nihu.gttime.utility.Utilities;
import jp.nihu.gttime.utility.XMLUtil;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.semarglproject.rdf.ParseException;
import org.semarglproject.source.StreamProcessor;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;

/**
 * データセットの内容を設定、取得するクラスです。
 */
public class DatasetTmm implements DatasetIF
{
    /**
     * シリアルバーションID
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(DatasetTmm.class, T2Map.LOCALE);

    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(DatasetTmm.class);

    /*
     * (non-Javadoc)
     * 
     * @see jp.nihu.gttime.t2map.data.DatasetIF#applyStyleChanges()
     */
    public void applyStyleChanges(String layerTypeName, String layerStyleName)
    {
        // DO NOTHING.
    }

    // /**
    // * プロジェクトXMLのこのデータセットに関する記述がされている要素
    // */
    // private Element fragmentOfProject = null;

    /*
     * (non-Javadoc)
     * 
     * @see jp.nihu.gttime.data.DatasetIF#setFragmentDatasetNodeOfProject(org.w3c.dom.Element)
     */
    public void setFragmentDatasetNodeOfProject(Element fragment)
    {
        // DO NOTHING.
    }

    /**
     * Documentオブジェクト
     */
    private Document doc;

    /**
     * レコード配列のオブジェクト
     */
    private Record[] records;
    
    /**
     * groupsのオブジェクト
     */
    private List<String> groups;
    
    private Model model;
    
    public String getRdfBase() {
    	return rdfBase;
    }
    
    @Override
    public Model getModel() throws ParseException, IOException {
    	if (model == null && getDomDataset() != null) {
//    		String basePath = T2Map.RRF_BASE_URI;
    		String basePath = Utilities.getUri(getRdfBase());
    		model = ModelFactory.createDefaultModel();
    		RDFaReader reader = new RDFaReader();
    		StreamProcessor streamProcessor = reader.getStreamProcessor(model);
    		streamProcessor.process(XMLUtil.getByteInputStream(getDomDataset()), basePath);
    		
    		if (0 < records.length) {
    			Node n = records[0].getNode();
    			if (n != null) {
    				streamProcessor.process(XMLUtil.getByteInputStream(n.getOwnerDocument()), basePath);
    			}
    		}
    	}
    	return model;
    }
    
    public List<String> getGroups() {
        if (groups == null) {
            for (int i = 0; i < records.length; i++) {
                setGroup(records[i].getGroup());
            }
        }
        return groups;
    }
    
    public void setGroup(String group) {
        if (group == null){
            group = "";
        }
        if (!groups.contains(group)) {
            groups.add(group);
        }
    }
    
    public void setGroups(List<String> groups) {
        this.groups = groups;
    }

    /**
     * TMMファイルのパスを格納します。
     */
    private String path;

    /**
     * レイヤタイプオブジェクト
     */
    private String layertype = null;

    private boolean visible = true;

    /**
     * @return the visible
     */
    public boolean isVisible()
    {
        return visible;
    }

    /**
     * @param visible the visible to set
     */
    public void setVisible(boolean visible)
    {
        this.visible = visible;
    }

    private PropertiesIF properties = null;

    public PropertiesIF getProperties()
    {
        return properties;
    }

    public void setProperties(PropertiesIF properties)
    {
        this.properties = properties;
    }

    public void setProperties(Element time)
    {
        properties = null;
    }

    public void reloadProperties(String layerName)
    {
        properties = null;
    }

    /**
     * デフォルトのコンストラクタです。
     */
    public DatasetTmm()
    {
    }

    /**
     * 新規データセットを作成する情報を取得します。
     * 
     * @param title データセットのタイトル
     * @param layerType データセットのレイヤタイプ
     * @param records レコードの配列
     * @throws DatasetException
     */
    public DatasetTmm(String title, String layerType, Record[] records) throws DatasetException
    {
        this("検索結果(" + title + ")", layerType, "NIHU TIMESPACE XML", records);
    }

    /**
     * 新規データセットを作成する情報を取得します。
     * 
     * @param title title データセットのタイトル
     * @param layerType データセットのレイヤタイプ
     * @param dataFormat データフォーマット
     * @param records レコードの配列
     * @throws DatasetException
     */
    public DatasetTmm(String title, String layerType, String dataFormat, Record[] records) throws DatasetException
    {
        if (records != null && records.length > 0)
        {
            if (layerType == null || layerType.length() <= 0)
            {
                layerType = "DefaultLayer";
            }
            this.records = records;
            String t = Utilities.escapeFileName(getTitle());
            if (t.length() > 10) {
            	t = t.substring(0, 10);
            }
            String newFilename = t + "_" + Utilities.getUniqueString();
            

            this.path = newFilename + ".tmm";

            String[] replaceKey = { "-T2MAP_TITLE-", "-T2MAP_DESC-", "-T2MAP_DATE-", "-T2MAP_FORMAT-", "-T2MAP_EARLY-", "-T2MAP_LATE-", "-T2MAP_LAYERTYPE-" };
            String sDate = "";
            if (getStartDatetime() != null)
            {
                sDate = JulianGregorianDate.toGregorian(getStartDatetime(), "Z");
            }

            String eDate = "";
            if (getEndDatetime() != null)
            {
                eDate = JulianGregorianDate.toGregorian(getEndDatetime(), "Z");
            }

            String[] replace = { title, title, newFilename, dataFormat, sDate, eDate, layerType };
            String tmmXml = getDatasetXML();
            Pattern p2 = null;
            Matcher m = null;
            for (int i = 0; i < replace.length; i++)
            {
                p2 = Pattern.compile(replaceKey[i]);
                m = p2.matcher(tmmXml);
                if (m.find())
                {
                    tmmXml = m.replaceFirst(replace[i]);
                }
            }

            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = null;
            try
            {
                db = dbf.newDocumentBuilder();
            }
            catch (ParserConfigurationException e)
            {
                throw new DatasetException(rmgr.format("error.parser.configuration", rmgr.getString("word.kensaku")), e);
            }
            try
            {
                doc = db.parse(new InputSource(new StringReader(tmmXml)));
            }
            catch (SAXException e)
            {
                throw new DatasetException(rmgr.format("error.sax", rmgr.getString("word.kensaku")), e);
            }
            catch (IOException e)
            {
                throw new DatasetException(rmgr.format("error.io", rmgr.getString("word.kensaku")), e);
            }
        }
    }

    /**
     * Datasetの初期化をします。
     * 
     * @return Datasetの初期設定
     */
    private String getDatasetXML()
    {
        StringBuffer sb = new StringBuffer();

        sb.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
        sb.append("<METADATA>\n");
        sb.append("<META NAME=\"dc.title\" TMCODE=\"92\" SCHEMECODE=\"7\" CONTENT=\"-T2MAP_TITLE-\"/>\n");
        sb.append("<META NAME=\"dc.description\" TMCODE=\"24\" SCHEMECODE=\"7\" CONTENT=\"-T2MAP_DESC-\"/>\n");
        sb.append("<META NAME=\"dc.pulisher\" TMCODE=\"89\" SCHEMECODE=\"7\" CONTENT=\"T2Map\"/>\n");
        sb.append("<META NAME=\"dc.date\" TMCODE=\"39\" SCHEMECODE=\"14\" CONTENT=\"-T2MAP_DATE-\"/>\n");
        sb.append("<META NAME=\"dc.type\" TMCODE=\"38\" SCHEMECODE=\"6\" CONTENT=\"Dataset\"/>\n");
        sb.append("<META NAME=\"dc.format\" TMCODE=\"7\" SCHEMECODE=\"5\" CONTENT=\"-T2MAP_FORMAT-\"/>\n");
        sb.append("<META NAME=\"dc.language\" TMCODE=\"6\" SCHEMECODE=\"26\" CONTENT=\"Ja\"/>\n");
        sb.append("<META NAME=\"dc.coverage.x.min\" TMCODE=\"5\" SCHEMECODE=\"27\" CONTENT=\"-180\"/>\n");
        sb.append("<META NAME=\"dc.coverage.x.max\" TMCODE=\"4\" SCHEMECODE=\"27\" CONTENT=\"180\"/>\n");
        sb.append("<META NAME=\"dc.coverage.y.min\" TMCODE=\"19\" SCHEMECODE=\"27\" CONTENT=\"-90\"/>\n");
        sb.append("<META NAME=\"dc.coverage.y.max\" TMCODE=\"16\" SCHEMECODE=\"27\" CONTENT=\"90\"/>\n");
        sb.append("<META NAME=\"dc.coverage.t.early\" TMCODE=\"15\" SCHEMECODE=\"14\" CONTENT=\"-T2MAP_EARLY-\"/>\n");
        sb.append("<META NAME=\"dc.coverage.t.late\" TMCODE=\"14\" SCHEMECODE=\"14\" CONTENT=\"-T2MAP_LATE-\"/>\n");
        sb.append("<META NAME=\"dc.rights\" TMCODE=\"10\" SCHEMECODE=\"7\" CONTENT=\"デフォルト著作権\"/>\n");
        sb.append("<META NAME=\"dc.subject.domain\" TMCODE=\"8\" SCHEMECODE=\"32\" CONTENT=\"\"/>\n");
        sb.append("<META NAME=\"ecai.team\" TMCODE=\"168\" SCHEMECODE=\"33\" CONTENT=\"\"/>\n");
        sb.append("<META NAME=\"tm.Server.Type\" TMCODE=\"90\" SCHEMECODE=\"50\" CONTENT=\"\"/>\n");
        sb.append("<META NAME=\"tm.Spatial.MapObjType\" TMCODE=\"36\" SCHEMECODE=\"17\" CONTENT=\"Polygon\"/>\n");
        sb.append("<META NAME=\"tm.Time.TimeVarying\" TMCODE=\"34\" SCHEMECODE=\"18\" CONTENT=\"NoDateFields\"/>\n");
        sb.append("<META NAME=\"tm.Filter.AccessibleFields\" TMCODE=\"37\" SCHEMECODE=\"7\" CONTENT=\"\"/>\n");
        sb.append("<META NAME=\"tm.TableData.TableName\" TMCODE=\"49\" SCHEMECODE=\"7\" CONTENT=\"\"/>\n");
        sb.append("<META NAME=\"tm.TableData.KeyIDfield\" TMCODE=\"51\" SCHEMECODE=\"7\" CONTENT=\"\"/>\n");
        sb.append("<META NAME=\"tm.TableData.Labelfield\" TMCODE=\"172\" SCHEMECODE=\"7\" CONTENT=\"\"/>\n");
        sb.append("<META NAME=\"tm.sys.Local.GISType\" TMCODE=\"80\" SCHEMECODE=\"8\" CONTENT=\"xml\"/>\n");
        sb.append("<META NAME=\"t2map.layertype\" CONTENT=\"-T2MAP_LAYERTYPE-\"/>\n");
        sb.append("</METADATA>\n");

        return sb.toString();
    }

    /**
     * tmmファイルのDocumentを取得します。
     * 
     * @return tmmファイルのDocument
     */
    public Document getDomDataset()
    {
        return doc;
    }

    /**
     * tmmファイルのDocumentを設定します。
     * 
     * @param domTmm tmmファイルのDocument
     */
    public void setDomDataset(Document doc)
    {
        this.doc = doc;
    }

    /**
     * 指定されたデータセットをロードします。
     * 
     * @param op プラグイン名
     * @param path ディレクトリパス
     * @param base ファイル名
     * @throws DatasetException
     */
    public void loadDataset(OpenPlugin op, String path, String base) throws DatasetException
    {
        loadDataset(op, path, base, path);
    }
    
    private String absolutePath;
    
    public String getAbsolutePath() {
    	return absolutePath;
    }
    
    public void setAbsolutePath(String absolutePath) {
    	this.absolutePath = absolutePath;
    }
    
    private String rdfBase;

    /**
     * 指定されたデータセットをロードします。
     * 
     * @param op プラグイン名
     * @param pathx ディレクトリパス
     * @param base ファイル名
     * @param uniquePath 他のTMMファイルと同じでないファイル名
     * @throws DatasetException
     */
    public void loadDataset(OpenPlugin op, String pathx, String base, String uniquePath) throws DatasetException
    {
        log.debug("loadDataset");
        this.path = uniquePath;
        this.absolutePath = op.getAbsolutePath(uniquePath, base);
        
        String realPath = op.getAbsolutePath(pathx, base);
        
        this.rdfBase = realPath;
        
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder = null;
        try
        {
            builder = factory.newDocumentBuilder();
        }
        catch (ParserConfigurationException e)
        {
            throw new DatasetException(rmgr.format("error.parser.configuration", ""), e);
        }
        try
        {
            doc = builder.parse(op.getDatasetInputStream(realPath));
        }
        catch (SAXException e)
        {
            throw new DatasetException(rmgr.format("error.sax", realPath), e);
        }
        catch (IOException e)
        {
            throw new DatasetException(rmgr.format("error.io", realPath), e);
        }
        catch (OpenPluginException e)
        {
            throw new DatasetException(e.getMessage(), e);
        }
        setRecords(op, realPath);
    }

    /**
     * レコードをセットします。
     * 
     * @param op プラグイン名
     * @param realPath パス
     * @throws DatasetException
     */
    private void setRecords(OpenPlugin op, String realPath) throws DatasetException
    {
        try
        {
            String format = XMLUtil.getValue(doc, "/METADATA/META[@NAME='dc.format']/@CONTENT", null);
            records = op.getRecords(realPath, format);
            setGroups(op.getGroups());
        }
        catch (XPathExpressionException e)
        {
            // throw new DatasetException(rmgr.format("open.plugin.error",
            // e.getMessage()), e);
            throw new DatasetException(e.getMessage(), e);
        }
        catch (OpenPluginException e)
        {
            // throw new DatasetException(rmgr.format("open.plugin.error",
            // e.getMessage()), e);
            throw new DatasetException(e.getMessage(), e);
        }
    }

    /**
     * レコードが存在するかチェックします。
     * 
     * @return レコードが存在する、しない
     */
    public boolean hasRecords()
    {
        if (records != null && records.length > 0)
        {
            return true;
        }
        return false;
    }

    /**
     * レコードを配列で取得します。
     * 
     * @return レコードの配列
     */
    public Record[] getRecords()
    {
        return records;
    }
    
    public Record[] getRecords(String group)
    {
        List<Record> list = new ArrayList<Record>();
        for (Record r : records) {
            if (r.getGroup().equals(group)) {
                list.add(r);
            }
        }
        return list.toArray(new Record[0]);
    }

    /**
     * 指定された範囲内にあるレコードを配列で取得します。
     * 
     * @param s 開始日時
     * @param e 終了日時
     * @return レコードの配列
     */
    public Record[] getRecords(BigDecimal s, BigDecimal e)
    {
        ArrayList<Record> v = new ArrayList<Record>();
        int al = records.length;
        for (int i = 0; i < al; i++)
        {
            boolean doAdd = false;
            JulianDate jsdate = records[i].getStartDatetime();
            JulianDate jedate = records[i].getEndDatetime();
            if (jsdate == null && jedate == null)
            {
                doAdd = true;
            }
            else if (jedate == null && e.compareTo(jsdate.getDatetime()) >= 0)
            {
                doAdd = true;
            }
            else if (jsdate == null && s.compareTo(jedate.getDatetime()) <= 0)
            {
                doAdd = true;
            }
            else if (e.compareTo(jsdate.getDatetime()) >= 0 && s.compareTo(jedate.getDatetime()) <= 0)
            {
                doAdd = true;
            }

            if (doAdd)
            {
                v.add(records[i]);
            }
        }
        Record[] rs = new Record[v.size()];
        rs = v.toArray(rs);// v.copyInto(rs);
        return rs;
    }

    public Record[] getHitRecords()
    {
        ArrayList<Record> v = new ArrayList<Record>();
        int al = records.length;
        for (int i = 0; i < al; i++)
        {
            if (records[i].isHit())
            {
                v.add(records[i]);
            }
        }
        Record[] rs = new Record[v.size()];
        rs = v.toArray(rs);// v.copyInto(rs);
        return rs;
    }

    /**
     * レイヤタイプを設定します。
     * 
     * @param layertype レイヤタイプ名
     */
    public void setLayerType(String layertype)
    {
        this.layertype = layertype;
    }

    /**
     * レイヤタイプを取得します。
     * 
     * @return レイヤタイプ名
     * @throws DatasetException
     */
    public String getLayerType()
    {
        String layer = null;
        String xpath = "/METADATA/META[@NAME='t2map.layertype']/@CONTENT";
        try
        {
            layer = XMLUtil.getValue(doc, xpath, null);
        }
        catch (XPathExpressionException e)
        {
            // never catch
            log.fatal(xpath + " is not valid xpath.", e);
        }
        // if ((layertype != layer) && (layertype != null))
        if (layertype != null && !layertype.equals(layer))
        {
            return layertype;
        }
        else
        {
            return layer;
        }
    }

    /**
     * データセットのタイトルを取得します。
     * 
     * @return タイトル
     * @throws DatasetException
     */
    public String getTitle()
    {
        if (title == null)
        {
            String xpath = "/METADATA/META[@NAME='dc.title']/@CONTENT";
            try
            {
                title = XMLUtil.getValue(doc, xpath, null);
            }
            catch (XPathExpressionException e)
            {
                // never catch
                log.fatal(xpath + " is not valid xpath.", e);
            }
        }
        return title;
    }

    /**
     * データセットのタイトル
     */
    private String title = null;

    /**
     * データセットのタイトルを設定します。
     * 
     * @param title タイトル
     */
    public void setTitle(String title)
    {
        this.title = title;
        try
        {
            XPathFactory factory = XPathFactory.newInstance();
            javax.xml.xpath.XPath xpath = factory.newXPath();
            XPathExpression expr = xpath.compile("/METADATA/META[@NAME='dc.title']");
            Element meta = (Element) expr.evaluate(doc, XPathConstants.NODE);
            meta.setAttribute("dc.title", title);
        }
        catch (XPathExpressionException e)
        {
            // never catch
            log.fatal("not valid xpath.", e);
        }
    }

    /**
     * パスを取得します。
     * 
     * @return パス
     */
    public String getPath()
    {
        return path;
    }
    
    public void setPath(String path) {
    	this.path = path;
    }

    // /**
    // * パスを設定します。
    // *
    // * @param path パス
    // */
    // public void setPath(String path)
    // {
    // this.path = path;
    // }

    /**
     * 開始日時を取得します。
     * 
     * @return 開始日時
     * @throws DatasetException
     */
    public BigDecimal getStartDatetime() throws DatasetException
    {
        if (startDate == null)
        {
            BigDecimal jdate = null;
            try
            {
                if (doc != null)
                {
                    String date = XMLUtil.getValue(doc, "/METADATA/META[@NAME='dc.coverage.t.early']/@CONTENT", null);
                    if (date != null && date.length() > 0)
                    {
                        jdate = JulianGregorianDate.toJulian(date);
                    }
                }
                Record[] records = getRecords();
                if (records != null)
                {
                    for (int i = 0; i < records.length; i++)
                    {
                        JulianDate jsd = records[i].getStartDatetime();
                        if (jsd != null)
                        {
                            BigDecimal rdate = jsd.getDatetime();
                            if (jdate == null || rdate.compareTo(jdate) == -1)
                            {
                                jdate = rdate;
                            }
                        }
                    }
                }
            }
            catch (XPathExpressionException e)
            {
                throw new DatasetException(e.getMessage(), e);
            }
            catch (ISO8601Exception e)
            {
                throw new DatasetException(e.getMessage(), e);
            }
            startDate = jdate;
        }
        return startDate;
    }

    /**
     * データセットの時間範囲終了日
     */
    private BigDecimal endDate = null;

    /**
     * データセットの時間範囲開始日
     */
    private BigDecimal startDate = null;

    /**
     * 終了日時を取得します。
     * 
     * @return 終了日時
     * @throws DatasetException
     */
    public BigDecimal getEndDatetime() throws DatasetException
    {
        if (endDate == null)
        {
            BigDecimal jdate = null;
            try
            {
                if (doc != null)
                {
                    String date = XMLUtil.getValue(doc, "/METADATA/META[@NAME='dc.coverage.t.late']/@CONTENT", null);
                    if (date != null && date.length() > 0)
                    {
                        jdate = JulianGregorianDate.toJulian(date);
                    }
                }
                Record[] records = getRecords();
                if (records != null)
                {
                    for (int i = 0; i < records.length; i++)
                    {
                        JulianDate jsd = records[i].getEndDatetime();
                        if (jsd != null)
                        {
                            BigDecimal rdate = jsd.getDatetime();
                            if (jdate == null || rdate.compareTo(jdate) == 1)
                            {
                                jdate = rdate;
                            }
                        }
                    }
                }
            }
            catch (XPathExpressionException e)
            {
                throw new DatasetException(e.getMessage(), e);
            }
            catch (ISO8601Exception e)
            {
                throw new DatasetException(e.getMessage(), e);
            }
            endDate = jdate;
        }
        return endDate;
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
        return getTitle();
    }
    
    public String getRecordPath()
    {
    	int x = getAbsolutePath().lastIndexOf(".");
        String xmlPath = getAbsolutePath().substring(0, x) + ".xml";
        // log.info("read dataset: " + xmlPath);
        return xmlPath;
    }
}