/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.data;

import jp.nihu.gttime.utility.GIO;
import jp.nihu.gttime.TimeMapGeograhicCoordinates;

/**
 * 位置情報を設定、取得するクラスです。
 */
public class Location
{
    /**
     * 点を表す定数。
     */
    public static final int POINT = 0;

    /**
     * 線を表す定数。
     */
    public static final int LINE = 1;

    /**
     * 面を表す定数。
     */
    public static final int PLANE = 2;

    /**
     * 位置情報のタイプ。
     */
    private int type = POINT;

    private TimeMapGeograhicCoordinates[] pointsLatiLongs;

    /**
     * UTMPointオブジェクト。
     */
    private UTMPoint[] points;

    /**
     * UTMPointを取得します。
     * 
     * @return UTMPoint
     */
    public UTMPoint[] getPoints()
    {
        return points;
    }

    /**
     * UTMPointを設定します。
     * 
     * @param points UTMPoint
     */
    public void setPoints(UTMPoint[] points)
    {
        this.points = points;
        if (pointsLatiLongs == null)
        {
            pointsLatiLongs = new TimeMapGeograhicCoordinates[points.length];
            for (int i = 0; i < points.length; i++)
            {
                double[] latilong = GIO.getWGS84LatLon(points[i].getX(), points[i].getY(), points[i].getZone());
                pointsLatiLongs[i] = new TimeMapGeograhicCoordinates(latilong[0], latilong[1]);
            }
        }
    }

    /**
     * 位置情報のタイプを取得します。
     * 
     * @return 位置情報のタイプ
     */
    public int getType()
    {
        return type;
    }

    /**
     * 位置情報のタイプを設定します。
     * 
     * @param type 位置情報のタイプ
     */
    public void setType(int type)
    {
        this.type = type;
    }

    /**
     * 緯度、経度を取得します。
     * 
     * @return 緯度、経度
     */
    public TimeMapGeograhicCoordinates[] getPointsLongLaties()
    {
        return pointsLatiLongs;
    }

    /**
     * 緯度、経度を設定します。
     * 
     * @param pointsLatiLongs 緯度、経度
     */
    public void setPointsLongLaties(TimeMapGeograhicCoordinates[] pointsLatiLongs)
    {
        this.pointsLatiLongs = pointsLatiLongs;
        if (points == null)
        {
            points = new UTMPoint[pointsLatiLongs.length];
            for (int i = 0; i < pointsLatiLongs.length; i++)
            {
                double[] utm = GIO.getWGS84UTM(pointsLatiLongs[i].getLatitude(), pointsLatiLongs[i].getLongitude());
                points[i] = new UTMPoint();
                points[i].setX((float) utm[0]);
                points[i].setY((float) utm[1]);
                points[i].setZone((int) utm[2]);
            }
        }
    }
}
