/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.data;

import java.awt.Font;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.math.BigDecimal;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.property.LineIF;
import jp.nihu.gttime.data.property.PolygonIF;
import jp.nihu.gttime.data.property.PropertiesIF;
import jp.nihu.gttime.data.property.SymbolIF;
import jp.nihu.gttime.plugin.OpenPlugin;
import jp.nihu.gttime.plugin.OpenPluginException;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.rdf.RDFaReader;
import jp.nihu.gttime.time.utility.ISO8601Exception;
import jp.nihu.gttime.time.utility.JulianGregorianDate;
import jp.nihu.gttime.utility.GTM;
import jp.nihu.gttime.utility.GTS;
import jp.nihu.gttime.utility.Utilities;
import jp.nihu.gttime.utility.XMLUtil;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.semarglproject.rdf.ParseException;
import org.semarglproject.source.StreamProcessor;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;

public class ProjectGts implements ProjectIF
{
	private boolean root = false;
	public boolean isRoot() {
		return root;
	}
	public void setRoot(boolean root) {
		this.root = root;
	}
	
	public String getRdfBase() {
    	return rdfBase;
    }
	
	private Model model;
    
    @Override
    public Model getModel() throws ParseException, IOException {
    	if (model == null && getDomProject() != null) {
//    		String basePath = T2Map.RRF_BASE_URI;
    		String basePath = Utilities.getUri(getRdfBase());
//    		System.out.println("ProjectGts: " + basePath);
    		model = ModelFactory.createDefaultModel();
    		RDFaReader reader = new RDFaReader();
    		StreamProcessor streamProcessor = reader.getStreamProcessor(model);
    		streamProcessor.process(XMLUtil.getByteInputStream(getDomProject()), basePath);
    		
    		for (int i = 0; i < children.size(); i++) {
    			Object obj = children.get(i);
    			Model m = null;
    			if (obj instanceof ProjectIF) {
    				m = ((ProjectIF)obj).getModel();
    			} else if (obj instanceof DatasetIF) {
    				m = ((DatasetIF)obj).getModel();
    			}
    			if (m != null) {
    				model.add(m);
    			}
    		}
    	}
    	return model;
    }
	
    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(ProjectGts.class);

    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(ProjectGts.class, T2Map.LOCALE);

    /**
     * ドキュメントオブジェクト
     */
    private Document domGts = null;

    /**
     * データセットオブジェクト
     */
//    private ArrayList<DatasetIF> datasets = null;

    /**
     * プロジェクトオブジェクト
     */
//    private ArrayList<ProjectIF> projects = null;
    
    /**
     * 子のプロジェクトオブジェクトおよびデータセットオブジェクト
     */
    private ArrayList<Object> children = null;

    /**
     * パスオブジェクト
     */
    private String path = "UNKNOWN";

    /**
     * プロジェクトのタイトル
     */
    private String title = null;

    private boolean visible = true;

    private boolean open = true;

    /**
     * @return the visible
     */
    public boolean isVisible()
    {
        return visible;
    }

    /**
     * @param visible the visible to set
     */
    public void setVisible(boolean visible)
    {
        this.visible = visible;
    }

    /**
     * @return the open
     */
    public boolean isOpen()
    {
        return open;
    }

    /**
     * @param open the open to set
     */
    public void setOpen(boolean open)
    {
        this.open = open;
    }

    /**
     * デフォルトのコンストラクタです。
     */
    public ProjectGts()
    {
        T2Map.setProjectType(1);
    }

    /**
     * プロジェクトの子供にデータセットを追加します。
     * 
     * @param datasets データセット
     */
    public ProjectGts(DatasetIF[] datasets)
    {
        this();
        try
        {
            String pstr = getProjectXML(datasets);
            setDomGts(new ByteArrayInputStream(pstr.getBytes("UTF-8")));
        }
        catch (Exception e)
        {
            log.error(e.getMessage(), e);
        }

//        path = System.getProperty("user.dir") + File.separatorChar + Utilities.getUniqueString() + ".gts";
        String t = Utilities.escapeFileName(getTitle());
        path = t + "/" + t + "_" + Utilities.getUniqueString() + ".gts";
        if (datasets != null && datasets.length > 0)
        {
            for (int i = 0; i < datasets.length; i++)
            {
                try
                {
                    addDataset(datasets[i]);
                }
                catch (ProjectException e)
                {
                    log.error(e.getMessage(), e);
                }
            }
        }
    }
    
    public ProjectGts(String sdate, String edate)
    {
        this();
        try
        {
            String pstr = getProjectXML(sdate, edate);
            setDomGts(new ByteArrayInputStream(pstr.getBytes("UTF-8")));
            String t = Utilities.escapeFileName(getTitle());
            path = t + "/" + t + "_" + Utilities.getUniqueString() + ".gts";
        }
        catch (Exception e)
        {
            log.error(e.getMessage(), e);
        }
    }

    /**
     * 開始日時、終了日時を取得します。
     * 
     * @param start 開始日時のジュリアン日
     * @param end 終了日時のジュリアン日
     * @return 時間
     */
    private String[] getDate(BigDecimal start, BigDecimal end)
    {
    	SimpleDateFormat uniqName = new SimpleDateFormat("yyyy-MM-dd");
        String edate = uniqName.format(new Date()) + "T00:00:00";
        String z = "Z";
        String[] date = { "0000-01-01T00:00:00", edate };
        if (start != null && end != null)
        {
            if (start.compareTo(end) < 0)
            {
                date[0] = JulianGregorianDate.toGregorian(start, z);
                date[1] = JulianGregorianDate.toGregorian(end, z);
            }
            else
            {
                date[1] = JulianGregorianDate.toGregorian(start, z);
                date[0] = JulianGregorianDate.toGregorian(end, z);
            }
        }
        else if (start != null && end == null)
        {
            date[0] = JulianGregorianDate.toGregorian(start, z);
            date[1] = JulianGregorianDate.toGregorian(start.add(BigDecimal.valueOf(10)), z);
        }
        else if (start == null && end != null)
        {
            date[0] = JulianGregorianDate.toGregorian(end.subtract(BigDecimal.valueOf(10)), z);
            date[1] = JulianGregorianDate.toGregorian(end, z);
        }
        return date;
    }
    
    private String getProjectXML(String sdate, String edate)
    {
        StringBuffer sb = new StringBuffer();
        sb.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
        sb.append("<hgis:gts version=\""+GTS.getVersion()+"\" xmlns:hgis=\"http://www.h-gis.org/hgis\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\">\n");
        sb.append("<hgis:title>New Project</hgis:title>\n");
        sb.append("<hgis:projection></hgis:projection>\n");
        sb.append("<hgis:coordinate></hgis:coordinate>\n");
        sb.append("<hgis:calendar>Gregorian</hgis:calendar>\n");
        sb.append("<hgis:displayedExtent>\n");
        sb.append("<hgis:westBoundLongitude></hgis:westBoundLongitude>\n");
        sb.append("<hgis:eastBoundLongitude></hgis:eastBoundLongitude>\n");
        sb.append("<hgis:southBoundLatitude></hgis:southBoundLatitude>\n");
        sb.append("<hgis:northBoundLatitude></hgis:northBoundLatitude>\n");
        sb.append("<hgis:begin>" + sdate + "</hgis:begin>\n");
        sb.append("<hgis:end>" + edate + "</hgis:end>\n");
        sb.append("</hgis:displayedExtent>\n");
        sb.append("</hgis:gts>\n");
        return sb.toString();
    }

    /**
     * データセットからドキュメントの開始、終了日時を設定します。
     * 
     * @param datasets データセット
     * @return XMLの文字列バッファ
     */
    private String getProjectXML(DatasetIF[] datasets)
    {
        String sdate = "0000-01-01T00:00:00";
        SimpleDateFormat uniqName = new SimpleDateFormat("yyyy-MM-dd");
        String edate = uniqName.format(new Date()) + "T00:00:00";
        if (datasets != null && datasets.length > 0)
        {
            try
            {
                BigDecimal start = getStartDatetime(datasets);
                BigDecimal end = getEndDatetime(datasets);
                String[] date = getDate(start, end);
                sdate = date[0];
                edate = date[1];
            }
            catch (DatasetException e)
            {
                log.error(e.getMessage(), e);
            }
        }
        return getProjectXML(sdate, edate);
    }

    /**
     * DOMをセットします。
     * 
     * @param in バイト入力ストリーム
     * @throws ParserConfigurationException
     * @throws SAXException
     * @throws IOException
     */
    private void setDomGts(InputStream in) throws ParserConfigurationException, SAXException, IOException
    {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        factory.setNamespaceAware(true);
        DocumentBuilder builder = null;
        builder = factory.newDocumentBuilder();
        domGts = builder.parse(in);
    }
    
    public void addProject(ProjectIF project) throws ProjectException
    {
    	log.debug("ADD PROJECT.");
    	try {
			String importProject = getProjectElement(project);
			Document addDoc = null;
			DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
			dbf.setNamespaceAware(true);
			DocumentBuilder db = null;
			try {
				db = dbf.newDocumentBuilder();
			} catch (ParserConfigurationException e) {
				throw new ProjectException(rmgr.format(
						"error.parser.configuration",
						rmgr.getString("word.kensaku")), e);
			}
			try {
				addDoc = db.parse(new InputSource(new StringReader(
						importProject)));
			} catch (SAXException e) {
				throw new ProjectException(rmgr.format("error.sax",
						rmgr.getString("word.kensaku")), e);
			} catch (IOException e) {
				throw new ProjectException(rmgr.format("error.io",
						rmgr.getString("word.kensaku")), e);
			}

			Node newNode = addDoc.getElementsByTagNameNS(
					"http://www.h-gis.org/hgis", "project").item(0);
			Node dataNode = domGts.getDocumentElement();
			Node addNode = domGts.importNode(newNode, true);
			dataNode.appendChild(addNode);
			if (children == null) {
            	children = new ArrayList<Object>();
            }
			children.add(project);
		} catch (ISO8601Exception e) {
			log.error(e.getMessage(), e);
		} catch (DatasetException e) {
			log.error(e.getMessage(), e);
		}
    }

    public void addDataset(DatasetIF dataset) throws ProjectException
    {
//    	log.debug("XXXXXXXXXXXXXXXXXx 1");
        Record[] records = null;
        if (dataset != null)
        {
//        	log.debug("XXXXXXXXXXXXXXXXXx 2");
            records = dataset.getRecords();
        }
        if (records != null && records.length > 0)
        {
//        	log.debug("XXXXXXXXXXXXXXXXXx 3");
//            DatasetIF[] addDs = getDatasets();
//            DatasetIF[] newDs = null;
//            if (addDs != null)
//            {
//                newDs = new DatasetIF[addDs.length + 1];
//                System.arraycopy(addDs, 0, newDs, 0, addDs.length);
//            }
//            else
//            {
//                newDs = new DatasetIF[1];
//            }
//            newDs[newDs.length - 1] = dataset;
//            setDatasets(newDs);

            String importDataset = getDatasetXML(dataset);
            
//            log.debug("XXXXXXXXXXXXXXXXXx " + importProject);
            
            
            Document addDoc = null;
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            dbf.setNamespaceAware(true);
            DocumentBuilder db = null;
            try
            {
                db = dbf.newDocumentBuilder();
            }
            catch (ParserConfigurationException e)
            {
                throw new ProjectException(rmgr.format("error.parser.configuration", rmgr.getString("word.kensaku")), e);
            }
            try
            {
                addDoc = db.parse(new InputSource(new StringReader(importDataset)));
            }
            catch (SAXException e)
            {
                throw new ProjectException(rmgr.format("error.sax", rmgr.getString("word.kensaku")), e);
            }
            catch (IOException e)
            {
                throw new ProjectException(rmgr.format("error.io", rmgr.getString("word.kensaku")), e);
            }

            Node newNode = addDoc.getElementsByTagNameNS("http://www.h-gis.org/hgis", "layer").item(0);
//            int len = 0;
//            len = domGts.getElementsByTagNameNS("http://www.h-gis.org/hgis", "gts").getLength();
//            Node dataNode = domGts.getElementsByTagNameNS("http://www.h-gis.org/hgis", "gts").item(len - 1);
            Node dataNode = domGts.getDocumentElement();
            Node addNode = domGts.importNode(newNode, true);
            dataNode.appendChild(addNode);
            dataset.setFragmentDatasetNodeOfProject((Element) addNode);
//            try {
//				log.debug("XXXXXXXXXXXXXXXXXx " + Utilities.toXmlString(domGts));
//			} catch (IOException e) {
//				// TODO Auto-generated catch block
//				e.printStackTrace();
//			}
            if (children == null) {
            	children = new ArrayList<Object>();
            }
            children.add(dataset);
        }
    }
    
    private String getProjectElement(ProjectIF project) throws ISO8601Exception, DatasetException {
    	String sdate = JulianGregorianDate.toGregorian(project.getStartDatetime(), "Z");
    	String edate = JulianGregorianDate.toGregorian(project.getEndDatetime(), "Z");
    	
    	StringBuffer sb = new StringBuffer();
    	sb.append("<hgis:project xmlns:hgis=\"http://www.h-gis.org/hgis\">");
    	// TODO: プロジェクトを新規に作成した場合のファイルパスの決定方法を考える。
    	String path = project.getPath();
    	if (path != null && path.length() > 0) {
    		sb.append("<hgis:resourceFileName>").append(path).append("</hgis:resourceFileName>");
    	}
    	sb.append("<hgis:title>New ProjectC1</hgis:title>");
    	sb.append("<hgis:projection/>");
    	sb.append("<hgis:coordinate/>");
    	sb.append("<hgis:calendar>Gregorian</hgis:calendar>");
    	sb.append("<hgis:displayedExtent>");
    	sb.append("<hgis:westBoundLongitude/>");
    	sb.append("<hgis:eastBoundLongitude/>");
    	sb.append("<hgis:southBoundLatitude/>");
    	sb.append("<hgis:northBoundLatitude/>");
    	sb.append("<hgis:begin>").append(sdate).append("</hgis:begin>");
    	sb.append("<hgis:end>").append(edate).append("</hgis:end>");
    	sb.append("</hgis:displayedExtent>");
    	sb.append("</hgis:project>");
    	return sb.toString();
    }

    /**
     * データセットからドキュメントの開始、終了日時を設定します。
     * 
     * @param ds データセット
     * @return XMLの文字列バッファ
     */
    private String getDatasetXML(DatasetIF ds)
    {
        StringBuffer sb = new StringBuffer();

        PropertiesIF properties = ds.getProperties();

        sb.append("<hgis:layer xmlns:hgis=\"http://www.h-gis.org/hgis\">\n");
        sb.append("<hgis:resourceFileName>").append(ds.getPath()).append("</hgis:resourceFileName>\n");
        sb.append("<hgis:drawOrder>1</hgis:drawOrder>\n");

        sb.append("<hgis:isVisible>true</hgis:isVisible>\n");

        sb.append("<hgis:resourceStatus version=\""+GTM.getVersion()+"\">\n");
        sb.append("<hgis:common />\n");
        sb.append("<hgis:time version=\""+GTM.getVersion()+"\">\n");

        sb.append("<hgis:layerType>").append(ds.getLayerType()).append("</hgis:layerType>\n");
        if (properties != null)
        {
            Map<String, LineIF> lines = properties.getLines();
            if (lines != null)
            {
                Iterator<String> it = lines.keySet().iterator();
                while (it.hasNext()) {
                    String group = it.next();
                    sb.append("<hgis:line");
                    if (!group.equals("")) {
                        sb.append(" group=\"");
                        sb.append(group);
                        sb.append("\"");
                    }
                    sb.append(">\n");
                    LineIF l = properties.getLine(group);
                    sb.append("<hgis:thin>").append(l.getThin()).append("</hgis:thin>\n");
                    sb.append("<hgis:lineColor>").append(Utilities.getColor(l.getLineColor())).append("</hgis:lineColor>\n");
                    sb.append("</hgis:line>\n");
                }
            }
            Map<String, PolygonIF> polygons = properties.getPolygons();
            if (polygons != null)
            {
                Iterator<String> it = polygons.keySet().iterator();
                while (it.hasNext()) {
                    String group = it.next();
                    sb.append("<hgis:polygon");
                    if (!group.equals("")) {
                        sb.append(" group=\"");
                        sb.append(group);
                        sb.append("\"");
                    }
                    sb.append(">\n");
                    PolygonIF p = properties.getPolygon(group);
                    sb.append("<hgis:fillColor>").append(Utilities.getColor(p.getFillColor())).append("</hgis:fillColor>\n");
                    sb.append("<hgis:thin>").append(p.getThin()).append("</hgis:thin>\n");
                    sb.append("<hgis:lineColor>").append(Utilities.getColor(p.getLineColor())).append("</hgis:lineColor>\n");
                    sb.append("</hgis:polygon>\n");
                }
            }
            Map<String, SymbolIF> symbols = properties.getSymbols();
            if (symbols != null)
            {
                Iterator<String> it = symbols.keySet().iterator();
                while (it.hasNext()) {
                    String group = it.next();
                    sb.append("<hgis:symbol");
                    if (!group.equals("")) {
                        sb.append(" group=\"");
                        sb.append(group);
                        sb.append("\"");
                    }
                    sb.append(">\n");
                    SymbolIF s = properties.getSymbol(group);
                    sb.append("<hgis:symbol>").append(s.getSymbol()).append("</hgis:symbol>\n");
                    sb.append("<hgis:size>").append(s.getSize()).append("</hgis:size>\n");
                    sb.append("<hgis:fillColor>").append(Utilities.getColor(s.getFillColor())).append("</hgis:fillColor>\n");
                    sb.append("<hgis:thin>").append(s.getThin()).append("</hgis:thin>\n");
                    sb.append("<hgis:lineColor>").append(Utilities.getColor(s.getLineColor())).append("</hgis:lineColor>\n");
                    sb.append("</hgis:symbol>\n");
                }
            }
            Map<String, Font> fonts = properties.getFonts();
            if (fonts != null)
            {
                Iterator<String> it = fonts.keySet().iterator();
                while (it.hasNext()) {
                    String group = it.next();
                    Font f = properties.getFont(group);
                	if (f != null) {
                        sb.append("<hgis:font");
                        if (!group.equals("")) {
                            sb.append(" group=\"");
                            sb.append(group);
                            sb.append("\"");
                        }
                        sb.append(">\n");
                        sb.append("<hgis:font-family>").append(f.getFamily()).append("</hgis:font-family>\n");
                        sb.append("<hgis:font-name>").append(f.getName()).append("</hgis:font-name>\n");
                        sb.append("<hgis:font-size>").append(f.getSize()).append("pt</hgis:font-size>\n");
                        if (properties.getFontColor(group) != null)
                        {
                            sb.append("<hgis:font-color>").append(Utilities.getColor(properties.getFontColor(group))).append("</hgis:font-color>\n");
                        }
                        sb.append("<hgis:font-style>").append(Utilities.getFontStyle(f.getStyle())).append("</hgis:font-style>\n");
                        sb.append("</hgis:font>\n");
                	}
                }
            }
            if (properties.getSelectedColor() != null)
            {
                sb.append("<hgis:selectedColor>").append(Utilities.getColor(properties.getSelectedColor())).append("</hgis:selectedColor>\n");
            }
            if (properties.getLayerHeight() != null && properties.getLayerHeight() > 70)
            {
                sb.append("<hgis:layerHight>").append(properties.getLayerHeight()).append("</hgis:layerHight>\n");
            }
            else
            {
                sb.append("<hgis:layerHight>100</hgis:layerHight>\n");
            }
            if (properties.getScaleMain() != null)
            {
                sb.append("<hgis:scaleMain>\n");
                sb.append("<hgis:fillColor>").append(Utilities.getColor(properties.getScaleMain().getFillColor())).append("</hgis:fillColor>\n");
                sb.append("<hgis:thin>").append(properties.getScaleMain().getThin()).append("</hgis:thin>\n");
                sb.append("<hgis:lineColor>").append(Utilities.getColor(properties.getScaleMain().getLineColor())).append("</hgis:lineColor>\n");
                sb.append("</hgis:scaleMain>\n");
            }
            if (properties.getScaleSub() != null)
            {
                sb.append("<hgis:scaleSub>\n");
                sb.append("<hgis:fillColor>").append(Utilities.getColor(properties.getScaleSub().getFillColor())).append("</hgis:fillColor>\n");
                sb.append("<hgis:thin>").append(properties.getScaleSub().getThin()).append("</hgis:thin>\n");
                sb.append("<hgis:lineColor>").append(Utilities.getColor(properties.getScaleSub().getLineColor())).append("</hgis:lineColor>\n");
                sb.append("</hgis:scaleSub>\n");
            }
            if (lines != null || polygons != null || symbols != null || fonts != null) {
                sb.append("<hgis:showGroupName>").append(properties.isShowGroupName()).append("</hgis:showGroupName>\n");
            }
            if (properties.getDetailFontName() != null) {
                sb.append("<hgis:detailFontName>").append(properties.getDetailFontName()).append("</hgis:detailFontName>\n");
            }
            if (properties.getDataReductionNum() != null) {
                sb.append("<hgis:dataReductionEnableRecordNum>").append(properties.getDataReductionNum()).append("</hgis:dataReductionEnableRecordNum>\n");
            }
            sb.append("<hgis:dataReduction>").append(properties.isDataReduction()).append("</hgis:dataReduction>\n");
        }

        sb.append("</hgis:time>\n");
        sb.append("</hgis:resourceStatus>\n");
        sb.append("</hgis:layer>\n");

        return sb.toString();
    }

    public DatasetIF[] getDatasets()
    {
//        if (datasets != null)
//        {
//            DatasetIF[] ds = new DatasetGtm[datasets.size()];
//            ds = datasets.toArray(ds);
//            return ds;
//        }
        
        Object[] objs = getNodes();
        if (objs != null)
        {
        	ArrayList<DatasetIF> pa = new ArrayList<DatasetIF>();
        	for (int i = 0; i < objs.length; i++) {
        		if (objs[i] instanceof DatasetIF) {
        			pa.add((DatasetIF)objs[i]);
				}
        	}
        	DatasetIF[] ps = new DatasetGtm[pa.size()];
            ps = pa.toArray(ps);
            return ps;
        }
        
        return null;
    }

    public Document getDomProject()
    {
        // DatasetIF[] ds = getDatasets();
        // NodeList nl =
        // domGts.getElementsByTagNameNS("http://www.h-gis.org/hgis", "layer");
        // for (int i = 0; i < nl.getLength(); i++)
        // {
        // Element layer = (Element) nl.item(i);
        // DatasetIF dataset = ds[i];
        // Element isVisible = getElement(layer, "./hgis:isVisible");
        // isVisible.setTextContent(Boolean.toString(dataset.isVisible()));
        // }
        return domGts;
    }

    public BigDecimal getEndDatetime() throws DatasetException, ISO8601Exception
    {
        BigDecimal jdate = null;
        if (domGts != null)
        {
            NodeList nl = domGts.getElementsByTagNameNS("http://www.h-gis.org/hgis", "end");
            if (nl.getLength() > 0)
            {
                Element elem = (Element) nl.item(0);
                String date = elem.getTextContent();
                if (date != null && date.length() > 0)
                {
                    jdate = JulianGregorianDate.toJulian(date);
                }
            }
        }
        ProjectIF[] projects = getProjects();
        if (projects != null)
        {
            for (int i = 0; i < projects.length; i++)
            {
                BigDecimal date = projects[i].getEndDatetime();
                if (date != null && (jdate == null || date.compareTo(jdate) == 1))
                {
                    jdate = date;
                }
            }
        }
        DatasetIF[] datasets = getDatasets();
        if (datasets != null)
        {
            BigDecimal date = getEndDatetime(datasets);
            if (date != null && (jdate == null || date.compareTo(jdate) == 1))
            {
                jdate = date;
            }
        }
        return jdate;
    }

    /**
     * データセットから終了日時を取得します。
     * 
     * @param datasets データセット
     * @return 終了日時
     * @throws DatasetException
     */
    private BigDecimal getEndDatetime(DatasetIF[] datasets) throws DatasetException
    {
        BigDecimal jdate = null;
        for (int i = 0; i < datasets.length; i++)
        {
            BigDecimal date = datasets[i].getEndDatetime();
            if (date != null && (jdate == null || date.compareTo(jdate) == 1))
            {
                jdate = date;
            }
        }
        return jdate;
    }

    public String getPath()
    {
        return path;
    }
    
    public void setPath(String path) {
    	this.path = path;
    }

    public ProjectIF[] getProjects()
    {
//        if (projects != null)
//        {
//            ProjectIF[] ps = new ProjectGts[projects.size()];
//            ps = projects.toArray(ps);
//            return ps;
//        }
        
        Object[] objs = getNodes();
        if (objs != null)
        {
        	ArrayList<ProjectIF> pa = new ArrayList<ProjectIF>();
        	for (int i = 0; i < objs.length; i++) {
        		if (objs[i] instanceof ProjectIF) {
        			pa.add((ProjectIF)objs[i]);
				}
        	}
            ProjectIF[] ps = new ProjectGts[pa.size()];
            ps = pa.toArray(ps);
            return ps;
        }
        
        return null;
    }
    
    public Object[] getNodes()
    {
        if (children != null)
        {
        	Object[] objs = new Object[children.size()];
        	objs = children.toArray(objs);
            return objs;
        }
        return null;
    }
    
    

    public BigDecimal getStartDatetime() throws DatasetException, ISO8601Exception
    {
        BigDecimal jdate = null;
        if (domGts != null)
        {
            NodeList nl = domGts.getElementsByTagNameNS("http://www.h-gis.org/hgis", "begin");
            if (nl.getLength() > 0)
            {
                Element elem = (Element) nl.item(0);
                String date = elem.getTextContent();
                if (date != null && date.length() > 0)
                {
                    jdate = JulianGregorianDate.toJulian(date);
                }
            }
        }
        ProjectIF[] projects = getProjects();
        if (projects != null)
        {
            for (int i = 0; i < projects.length; i++)
            {
                BigDecimal date = projects[i].getStartDatetime();
                if (date != null && (jdate == null || date.compareTo(jdate) == -1))
                {
                    jdate = date;
                }
            }
        }
        DatasetIF[] datasets = getDatasets();
        if (datasets != null)
        {
            BigDecimal date = getStartDatetime(datasets);
            if (date != null && (jdate == null || date.compareTo(jdate) == -1))
            {
                jdate = date;
            }
        }
        return jdate;
    }

    /**
     * データセットから開始日時を取得します。
     * 
     * @param datasets データセット
     * @return 開始日時
     * @throws DatasetException
     */
    private BigDecimal getStartDatetime(DatasetIF[] datasets) throws DatasetException
    {
        BigDecimal jdate = null;
        for (int i = 0; i < datasets.length; i++)
        {
            BigDecimal date = datasets[i].getStartDatetime();
            if (date != null && (jdate == null || date.compareTo(jdate) == -1))
            {
                jdate = date;
            }
        }
        return jdate;
    }

    public String getTitle()
    {
        if (title == null)
        {
            if (domGts != null)
            {
                Element name = getTitleNode();
                title = name.getTextContent();
            }
        }
        return title;
    }
    
    private String absolutePath;
    
    public String getAbsolutePath() {
    	return absolutePath;
    }
    
    public void setAbsolutePath(String absolutePath) {
    	this.absolutePath = absolutePath;
    }
    
    private String rdfBase;

    public void loadProject(OpenPlugin op, String path, String base) throws ProjectException
    {
        String realPath = op.getAbsolutePath(path, base);
        
        this.rdfBase = realPath;
        
        try
        {
            setDomGts(op.getProjectInputStream(realPath));
        }
        catch (ParserConfigurationException e)
        {
            throw new ProjectException(rmgr.format("error.parser.configuration", ""), e);
        }
        catch (SAXException e)
        {
            throw new ProjectException(rmgr.format("error.sax", realPath), e);
        }
        catch (IOException e)
        {
            throw new ProjectException(rmgr.format("error.io", realPath), e);
        }
        catch (OpenPluginException e)
        {
            throw new ProjectException(e.getMessage(), e);
        }
        
        // TODO: 以下、要検討
        if (path.endsWith("tms"))
        {
            path = path.replaceAll(".tms", ".gts");
        }
        this.path = path;
        if (realPath.endsWith("tms"))
        {
        	realPath = realPath.replaceAll(".tms", ".gts");
        }
        this.absolutePath = realPath;
        base = op.getBasePath(realPath);
//        loadChildDataset(op, base);
//        loadChildProject(op, base);
        loadChildNodes(op, base);
    }

    /**
     * @param op プラグイン名
     * @param base ファイル名
     * @throws ProjectException
     */
//    private void loadChildDataset(OpenPlugin op, String base) throws ProjectException
//    {
//        NodeList nl = getChildDatasetNodes();
//        DatasetIF dataset = null;
//        datasets = new ArrayList<DatasetIF>();
//        for (int i = 0; i < nl.getLength(); i++)
//        {
//            Element elem = (Element) nl.item(i);
//            String gtmPath = getElementContent(elem, "./hgis:resourceFileName/text()");
//            dataset = new DatasetGtm(elem);
//            // dataset.setFragmentDatasetNodeOfProject(elem);
//            NodeList time = elem.getElementsByTagNameNS("http://www.h-gis.org/hgis", "time");
//            if (time.getLength() > 0)
//            {
//                dataset.setLayerType(getElementContent((Element) time.item(0), "./hgis:layerType"));
//                dataset.setProperties((Element) time.item(0));
//            }
//
//            try
//            {
//                dataset.loadDataset(op, gtmPath, base);
//                if (gtmPath.endsWith("tmm"))
//                {
//                    gtmPath = gtmPath.replaceAll(".tmm", ".gtm");
//                    getElement(elem, "./hgis:resourceFileName").setTextContent(gtmPath);
//                }
//                if (dataset.hasRecords())
//                {
//                    datasets.add(dataset);
//                }
//            }
//            catch (DatasetException e)
//            {
//                log.fatal(e.getMessage(), e);
//                // throw new ProjectException(e.getMessage(), e);
//            }
//        }
//    }

    private Element getElement(Object obj, String xpathStr)
    {
        Element elem = null;
        try
        {
            XPathFactory factory = XPathFactory.newInstance();
            XPath xpath = factory.newXPath();
            xpath.setNamespaceContext(XMLUtil.getGtsNsContext());
            XPathExpression expr = xpath.compile(xpathStr);
            elem = (Element) expr.evaluate(obj, XPathConstants.NODE);
        }
        catch (XPathExpressionException e)
        {
            // never catch
            log.fatal("not valid xpath.", e);
        }
        return elem;
    }

    private String getElementContent(Element elem, String xpathStr)
    {
        String str = null;
        try
        {
            XPathFactory factory = XPathFactory.newInstance();
            XPath xpath = factory.newXPath();
            xpath.setNamespaceContext(XMLUtil.getGtsNsContext());
            XPathExpression expr = xpath.compile(xpathStr);
            str = (String) expr.evaluate(elem, XPathConstants.STRING);
        }
        catch (XPathExpressionException e)
        {
            // never catch
            log.fatal("not valid xpath.", e);
        }
        return str;
    }

//    private NodeList getChildDatasetNodes()
//    {
//        NodeList name = null;
//        try
//        {
//            XPathFactory factory = XPathFactory.newInstance();
//            XPath xpath = factory.newXPath();
//            xpath.setNamespaceContext(jp.nihu.gttime.utility.XMLUtil.getGtsNsContext());
//            XPathExpression expr = xpath.compile("/hgis:gts/hgis:layer");
//            name = (NodeList) expr.evaluate(domGts, XPathConstants.NODESET);
//            // name = domGms.getElementsByTagNameNS("http://www.h-gis.org/hgis",
//            // "resourceFileName");
//        }
//        catch (XPathExpressionException e)
//        {
//            // never catch
//            log.fatal("not valid xpath.", e);
//        }
//        log.debug(name.getLength());
//        return name;
//    }

    /**
     * @param op プラグイン名
     * @param base ファイル名
     * @throws ProjectException
     */
//    private void loadChildProject(OpenPlugin op, String base) throws ProjectException
//    {
//        NodeList nl = getChildProjectNodes();
//        ProjectIF project = null;
//        projects = new ArrayList<ProjectIF>();
//        for (int i = 0; i < nl.getLength(); i++)
//        {
//            Element elem = (Element) nl.item(i);
//            String gtsPath = elem.getTextContent().trim();
//            project = new ProjectGts();
//            project.loadProject(op, gtsPath, base);
//            if (gtsPath.endsWith("tms"))
//            {
//                gtsPath = gtsPath.replaceAll(".tms", ".gts");
//                elem.setTextContent(gtsPath);
//            }
//            projects.add(project);
//        }
//        // projects = new Project[v.size()];
//        // v.copyInto(projects);
//    }
//
//    private NodeList getChildProjectNodes()
//    {
//        NodeList name = null;
//        try
//        {
//            XPathFactory factory = XPathFactory.newInstance();
//            XPath xpath = factory.newXPath();
//            xpath.setNamespaceContext(jp.nihu.gttime.utility.XMLUtil.getGtsNsContext());
//            XPathExpression expr = xpath.compile("/hgis:gts/hgis:project/hgis:resourceFileName");
//            name = (NodeList) expr.evaluate(domGts, XPathConstants.NODESET);
//            // name = domGms.getElementsByTagNameNS("http://www.h-gis.org/hgis",
//            // "resourceFileName");
//        }
//        catch (XPathExpressionException e)
//        {
//            // never catch
//            log.fatal("not valid xpath.", e);
//        }
//        return name;
//    }
    
    private void loadChildNodes(OpenPlugin op, String base) throws ProjectException
    {
    	children = new ArrayList<Object>();
    	NodeList nl = getChildNodes();
    	for (int i = 0; i < nl.getLength(); i++)
        {
            Element elem = (Element) nl.item(i);
            String path = getElementContent(elem, "./hgis:resourceFileName/text()").trim();
            if (path.endsWith("tms") || path.endsWith("gts")) {
            	ProjectIF project = new ProjectGts();
                project.loadProject(op, path, base);
                if (path.endsWith("tms"))
                {
                    path = path.replaceAll(".tms", ".gts");
                    getElement(elem, "./hgis:resourceFileName").setTextContent(path);
                }
                children.add(project);
            } else if (path.endsWith("tmm") || path.endsWith("gtm")) {
            	DatasetIF dataset = new DatasetGtm(elem);
                // dataset.setFragmentDatasetNodeOfProject(elem);
                NodeList time = elem.getElementsByTagNameNS("http://www.h-gis.org/hgis", "time");
                if (time.getLength() > 0)
                {
                    dataset.setLayerType(getElementContent((Element) time.item(0), "./hgis:layerType"));
                    dataset.setProperties((Element) time.item(0));
                }

                try
                {
//                	log.debug("path : " + path);
//                	log.debug("base : " + base);
                    dataset.loadDataset(op, path, base);
                    if (path.endsWith("tmm"))
                    {
                        path = path.replaceAll(".tmm", ".gtm");
                        getElement(elem, "./hgis:resourceFileName").setTextContent(path);
                    }
                    if (dataset.hasRecords())
                    {
                    	children.add(dataset);
                    }
                }
                catch (DatasetException e)
                {
                    log.fatal(e.getMessage(), e);
                    // 正常に読み込めたレコードを画面表示するため（処理を継続するた）スローしない
                    //throw new ProjectException(e.getMessage(), e);
                }
            } else {
            	String message = "System can not process the file name : " + path + ". The extension shuld be .tms, .gts, .tmm or .gtm";
            	log.fatal(message);
            	// 正常に読み込めたレコードを画面表示するため（処理を継続するた）スローしない
            	//throw new ProjectException(message);
            }
        }
    	
    }
    
    private NodeList getChildNodes()
    {
    	NodeList name = null;
        try
        {
            XPathFactory factory = XPathFactory.newInstance();
            XPath xpath = factory.newXPath();
            xpath.setNamespaceContext(jp.nihu.gttime.utility.XMLUtil.getGtsNsContext());
            XPathExpression expr = xpath.compile("/hgis:gts/hgis:project | /hgis:gts/hgis:layer");
            name = (NodeList) expr.evaluate(domGts, XPathConstants.NODESET);
            // name = domGms.getElementsByTagNameNS("http://www.h-gis.org/hgis",
            // "resourceFileName");
        }
        catch (XPathExpressionException e)
        {
            // never catch
            log.fatal("not valid xpath.", e);
        }
        return name;
    }

    public void moveDataset(DatasetIF dataset, Node node, int pos)
    {
        // Node inode1 = domTms.adoptNode(node);
        // 又は
        // Node inode2 = domTms.importNode(node, true);
        // を実行し、戻り値のノードを所定の位置（pos）に挿入する。
//        Node inode = domGts.adoptNode(node);
//        Element root = domGts.getDocumentElement();
//        NodeList nl = root.getElementsByTagNameNS("http://www.h-gis.org/hgis", "layer");
//        if (nl.getLength() == pos)
//        {
//            NodeList nlp = root.getElementsByTagNameNS("http://www.h-gis.org/hgis", "project");
//            if (nlp == null || nlp.getLength() < 1)
//            {
//                root.appendChild(inode);
//            }
//            else
//            {
//                root.insertBefore(inode, nlp.item(0));
//            }
//            datasets.add(dataset);
//        }
//        else
//        {
//            root.insertBefore(inode, nl.item(pos));
//            datasets.add(pos, dataset);
//        }
    	
    	if (children == null) {
        	children = new ArrayList<Object>();
        }
    	
    	Node inode = domGts.adoptNode(node);
        Element root = domGts.getDocumentElement();
        NodeList nl = getChildNodes();
        if (nl.getLength() == pos)
        {
            root.appendChild(inode);
            children.add(dataset);
        }
        else
        {
            root.insertBefore(inode, nl.item(pos));
            children.add(pos, dataset);
        }
    }

    public void moveProject(ProjectIF project, Node node, int pos)
    {
        // Node inode1 = domTms.adoptNode(node);
        // 又は
        // Node inode2 = domTms.importNode(node, true);
        // を実行し、戻り値のノードを所定の位置（pos）に挿入する。
//        pos = pos - datasets.size();
//        Node inode = domGts.adoptNode(node);
//        Element root = domGts.getDocumentElement();
//        NodeList nl = root.getElementsByTagNameNS("http://www.h-gis.org/hgis", "project");
//        if (nl.getLength() == pos)
//        {
//            root.appendChild(inode);
//            projects.add(project);
//        }
//        else
//        {
//            root.insertBefore(inode, nl.item(pos));
//            projects.add(pos, project);
//        }
        
        Node inode = domGts.adoptNode(node);
        Element root = domGts.getDocumentElement();
        NodeList nl = getChildNodes();
        if (nl.getLength() == pos)
        {
            root.appendChild(inode);
            children.add(project);
        }
        else
        {
            root.insertBefore(inode, nl.item(pos));
            children.add(pos, project);
        }
    }

    public Node removeDataset(DatasetIF childDataset)
    {
        // domTmsの所定の位置のDatasetのノードを削除する。
        // datasetsの所定の位置（pos）のdatasetを削除する。
//        int pos = datasets.indexOf(childDataset);
//        boolean removed = datasets.remove(childDataset);
//        if (removed)
//        {
//            NodeList nl = domGts.getElementsByTagNameNS("http://www.h-gis.org/hgis", "layer");
//            return nl.item(pos);
//        }
//        else
//        {
//            log.error("can not remove: " + childDataset.getTitle() + " dataset");
//        }
        
        int pos = children.indexOf(childDataset);
        boolean removed = children.remove(childDataset);
        if (removed)
        {
            return removeNode(pos);
        }
        else
        {
            log.error("can not remove: " + childDataset.getTitle() + " project");
        }
        
        return null;// 削除したdomTms内のDatasetのノード（のコピー）を返却
    }

    public void removeDataset(Node datasetNode)
    {
        Node node = domGts.adoptNode(datasetNode);
        if (node != null)
        {
            log.debug("dataset node removed.");
        }
        else
        {
            log.error("dataset node could not removed.");
        }
    }

    public Node removeNode(int p)
    {
        // projectとdatasetを混合して扱う時にここを実装する
    	NodeList nl = getChildNodes();
        return nl.item(p);
    }

    public Node removeProject(ProjectIF childProject)
    {
        // domTmsの所定の位置のProjectのノードを削除する。
        // projectsの所定の位置（pos）にprojectを削除する。
//        int pos = projects.indexOf(childProject);
//        boolean removed = projects.remove(childProject);
//        if (removed)
//        {
//            NodeList nl = domGts.getElementsByTagNameNS("http://www.h-gis.org/hgis", "project");
//            return nl.item(pos);
//        }
//        else
//        {
//            log.error("can not remove: " + childProject.getTitle() + " project");
//        }
    	
    	int pos = children.indexOf(childProject);
        boolean removed = children.remove(childProject);
        if (removed)
        {
            return removeNode(pos);
        }
        else
        {
            log.error("can not remove: " + childProject.getTitle() + " project");
        }
    	
        return null;// 削除したdomTms内のProjectのノード（のコピー）を返却
    }

    public void removeProject(Node projectNode)
    {
        Node node = domGts.adoptNode(projectNode);
        if (node != null)
        {
            log.debug("project node removed.");
        }
        else
        {
            log.error("project node could not removed.");
        }
    }

//    public void setDatasets(DatasetIF[] ds)
//    {
////        datasets = new ArrayList<DatasetIF>();
////        for (int i = 0; i < ds.length; i++)
////        {
////            datasets.add(ds[i]);
////        }
//    	children = new ArrayList<Object>();
//        for (int i = 0; i < ds.length; i++)
//        {
//        	children.add(ds[i]);
//        }
//    }

//    public void setProjects(ProjectIF[] ps)
//    {
////        projects = new ArrayList<ProjectIF>();
////        for (int i = 0; i < ps.length; i++)
////        {
////            projects.add(ps[i]);
////        }
//    	children = new ArrayList<Object>();
//    	for (int i = 0; i < ps.length; i++)
//        {
//    		children.add(ps[i]);
//        }
//    }

    public void setTitle(String title)
    {
        this.title = title;
        Element name = getTitleNode();
        name.setTextContent(title);
    }

    private Element getTitleNode()
    {
        Element name = null;
        try
        {
            XPathFactory factory = XPathFactory.newInstance();
            XPath xpath = factory.newXPath();
            xpath.setNamespaceContext(jp.nihu.gttime.utility.XMLUtil.getGtsNsContext());
            XPathExpression expr = xpath.compile("/hgis:gts/hgis:title");
            name = (Element) expr.evaluate(domGts, XPathConstants.NODE);
        }
        catch (XPathExpressionException e)
        {
            // never catch
            log.fatal("not valid xpath.", e);
        }
        return name;
    }

    public String toString()
    {
        return getTitle();
    }
}
