/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.data;

import java.io.IOException;
import java.math.BigDecimal;

import jp.nihu.gttime.plugin.OpenPlugin;
import jp.nihu.gttime.time.utility.ISO8601Exception;

import org.semarglproject.rdf.ParseException;
import org.w3c.dom.Document;
import org.w3c.dom.Node;

import com.hp.hpl.jena.rdf.model.Model;

public interface ProjectIF
{
	public boolean isRoot();
	public void setRoot(boolean root);
	public Model getModel() throws ParseException, IOException;
	
    /**
     * ノードを削除します。
     * 
     * @param pos DOMノードの位置
     * @return null
     */
    public Node removeNode(int pos);

    /**
     * データセットオブジェクトを削除します。
     * 
     * @param childDataset 削除するデータセット
     * @return 削除するノード
     */
    public Node removeDataset(DatasetIF childDataset);

    public void removeDataset(Node datasetNode);

    /**
     * データセットを追加します。
     * 
     * @param dataset 追加するデータセットオブジェクト
     * @param node 移動するデータセットノード
     * @param pos DOMノードの位置
     */
    public void moveDataset(DatasetIF dataset, Node node, int pos);

    /**
     * プロジェクトオブジェクトを削除します。
     * 
     * @param childProject 削除するプロジェクト
     * @return 削除するノード
     */
    public Node removeProject(ProjectIF childProject);

    public void removeProject(Node projectNode);

    /**
     * プロジェクトを追加します。
     * 
     * @param project 追加するプロジェクトオブジェクト
     * @param node 移動するプロジェクトノード
     * @param pos DOMノードの位置
     */
    public void moveProject(ProjectIF project, Node node, int pos);

    /**
     * tmsドキュメントを取得します。
     * 
     * @return tmsドキュメント
     */
    public Document getDomProject();
    
    /**
     * プロジェクトを追加します。
     * 
     * @param project 追加するプロジェクト
     * @throws ProjectException
     */
    public void addProject(ProjectIF project) throws ProjectException;

    /**
     * データセットを追加します。
     * 
     * @param dataset 追加するデータセット
     * @throws ProjectException
     */
    public void addDataset(DatasetIF dataset) throws ProjectException;

    /**
     * プロジェクトをロードします。
     * 
     * @param op プラグイン名
     * @param path ディレクトリパス
     * @param base ファイル名
     * @throws ProjectException
     */
    public void loadProject(OpenPlugin op, String path, String base) throws ProjectException;

    /**
     * プロジェクトのタイトルを取得します。
     * 
     * @return タイトル
     */
    public String getTitle();

    /**
     * プロジェクトのタイトルを設定します。
     * 
     * @param title タイトル
     */
    public void setTitle(String title);

    /**
     * データセットを取得します。
     * 
     * @return データセット
     */
    public DatasetIF[] getDatasets();

//    /**
//     * データセットを設定します。
//     * 
//     * @param ds データセット
//     */
//    public void setDatasets(DatasetIF[] ds);

    /**
     * プロジェクトを取得します。
     * 
     * @return プロジェクト
     */
    public ProjectIF[] getProjects();

//    /**
//     * プロジェクトを設定します。
//     * 
//     * @param ps プロジェクト
//     */
//    public void setProjects(ProjectIF[] ps);
    
    /**
     * プロジェクトとデータセットを取得します。
     * 
     * @return プロジェクト
     */
    public Object[] getNodes();

    /**
     * パスを取得します。
     * 
     * @return パス
     */
    public String getPath();
    
    public void setPath(String path);
    
    public String getRdfBase();
    
    public String getAbsolutePath();
    
    public void setAbsolutePath(String absolutePath);

    /**
     * 開始日時を取得します。
     * 
     * @return 開始日時
     * @throws DatasetException
     * @throws ISO8601Exception
     */
    public BigDecimal getStartDatetime() throws DatasetException, ISO8601Exception;

    /**
     * 終了日時を取得します。
     * 
     * @return 終了日時
     * @throws DatasetException
     * @throws ISO8601Exception
     */
    public BigDecimal getEndDatetime() throws DatasetException, ISO8601Exception;

    /**
     * @see java.lang.Object#toString()
     */
    public String toString();

    /**
     * @return the visible
     */
    public boolean isVisible();

    /**
     * @param visible the visible to set
     */
    public void setVisible(boolean visible);

    /**
     * @return the open
     */
    public boolean isOpen();

    /**
     * @param open the open to set
     */
    public void setOpen(boolean open);

}
