/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.data;

import java.io.ByteArrayInputStream;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.math.BigDecimal;
import java.util.ArrayList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.plugin.OpenPlugin;
import jp.nihu.gttime.plugin.OpenPluginException;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.rdf.RDFaReader;
import jp.nihu.gttime.time.utility.ISO8601Exception;
import jp.nihu.gttime.time.utility.JulianGregorianDate;
import jp.nihu.gttime.utility.Utilities;
import jp.nihu.gttime.utility.XMLUtil;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.semarglproject.rdf.ParseException;
import org.semarglproject.source.StreamProcessor;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;

/**
 * Projectの内容を設定、取得するクラスです。
 */
public class ProjectTms implements ProjectIF
{
	private boolean root = false;
	public boolean isRoot() {
		return root;
	}
	public void setRoot(boolean root) {
		this.root = root;
	}
	
	public String getRdfBase() {
    	return rdfBase;
    }
	
	private Model model;
    
    @Override
    public Model getModel() throws ParseException, IOException {
    	if (model == null && getDomProject() != null) {
//    		String basePath = T2Map.RRF_BASE_URI;
    		String basePath = Utilities.getUri(getRdfBase());
    		model = ModelFactory.createDefaultModel();
    		RDFaReader reader = new RDFaReader();
    		StreamProcessor streamProcessor = reader.getStreamProcessor(model);
    		streamProcessor.process(XMLUtil.getByteInputStream(getDomProject()), basePath);
    		for (int i = 0; i < projects.size(); i++) {
    			ProjectIF obj = projects.get(i);
    			Model m = obj.getModel();
    			model.add(m);
    		}
    		for (int i = 0; i < datasets.size(); i++) {
    			DatasetIF obj = datasets.get(i);
    			Model m = obj.getModel();
    			model.add(m);
    		}
    	}
    	return model;
    }
	
    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(ProjectTms.class);

    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(ProjectTms.class, T2Map.LOCALE);

    /**
     * ドキュメントオブジェクト
     */
    private Document domTms = null;

    /**
     * データセットオブジェクト
     */
    private ArrayList<DatasetIF> datasets = null;

    /**
     * プロジェクトオブジェクト
     */
    private ArrayList<ProjectIF> projects = null;

    /**
     * パスオブジェクト
     */
    private String path = "UNKNOWN";

    /**
     * プロジェクトのタイトル
     */
    private String title = null;

    private boolean visible = true;

    private boolean open = true;

    /**
     * @return the visible
     */
    public boolean isVisible()
    {
        return visible;
    }

    /**
     * @param visible the visible to set
     */
    public void setVisible(boolean visible)
    {
        this.visible = visible;
    }

    /**
     * @return the open
     */
    public boolean isOpen()
    {
        return open;
    }

    /**
     * @param open the open to set
     */
    public void setOpen(boolean open)
    {
        this.open = open;
    }

    /**
     * デフォルトのコンストラクタです。
     */
    public ProjectTms()
    {
        T2Map.setProjectType(0);
    }

    /**
     * プロジェクトの子供にデータセットを追加します。
     * 
     * @param datasets データセット
     */
    public ProjectTms(DatasetIF[] datasets)
    {
        this();
        if (datasets != null && datasets.length > 0)
        {
            try
            {
                String pstr = getProjectXML(datasets);
                setDomTms(new ByteArrayInputStream(pstr.getBytes("UTF-8")));
            }
            catch (Exception e)
            {
                log.error(e.getMessage(), e);
            }

            path = System.getProperty("user.dir") + File.separatorChar + Utilities.getUniqueString() + ".tms";

            for (int i = 0; i < datasets.length; i++)
            {
                try
                {
                    addDataset(datasets[i]);
                }
                catch (ProjectException e)
                {
                    log.error(e.getMessage(), e);
                }
            }
        }
    }

    /**
     * ノードを削除します。
     * 
     * @param pos DOMノードの位置
     * @return null
     */
    public Node removeNode(int pos)
    {
        return null;
    }

    /**
     * データセットオブジェクトを削除します。
     * 
     * @param childDataset 削除するデータセット
     * @return 削除するノード
     */
    public Node removeDataset(DatasetIF childDataset)
    {
        // domTmsの所定の位置のDatasetのノードを削除する。
        // datasetsの所定の位置（pos）のdatasetを削除する。
        int pos = datasets.indexOf(childDataset);
        boolean removed = datasets.remove(childDataset);
        if (removed)
        {
            NodeList nl = domTms.getElementsByTagName("Dataset");
            return nl.item(pos);
        }
        else
        {
            log.error("can not remove: " + childDataset.getTitle() + " dataset");
        }
        return null;// 削除したdomTms内のDatasetのノード（のコピー）を返却
    }

    public void removeDataset(Node datasetNode)
    {
        Node node = domTms.adoptNode(datasetNode);
        if (node != null)
        {
            log.debug("dataset node removed.");
        }
        else
        {
            log.error("dataset node could not removed.");
        }
    }

    /**
     * データセットを追加します。
     * 
     * @param dataset 追加するデータセットオブジェクト
     * @param node 移動するデータセットノード
     * @param pos DOMノードの位置
     */
    public void moveDataset(DatasetIF dataset, Node node, int pos)
    {
        // Node inode1 = domTms.adoptNode(node);
        // 又は
        // Node inode2 = domTms.importNode(node, true);
        // を実行し、戻り値のノードを所定の位置（pos）に挿入する。
        Node inode = domTms.adoptNode(node);
        Element root = domTms.getDocumentElement();
        NodeList nl = root.getElementsByTagName("Dataset");
        if (nl.getLength() == pos)
        {
            NodeList nlp = root.getElementsByTagName("MapspaceFile");
            if (nlp == null || nlp.getLength() < 1)
            {
                root.appendChild(inode);
            }
            else
            {
                root.insertBefore(inode, nlp.item(0));
            }
            datasets.add(dataset);
        }
        else
        {
            root.insertBefore(inode, nl.item(pos));
            datasets.add(pos, dataset);
        }
        // datasetsの所定の位置（pos）にdatasetを挿入する。
        // datasets.add(pos, dataset);
    }

    /**
     * プロジェクトオブジェクトを削除します。
     * 
     * @param childProject 削除するプロジェクト
     * @return 削除するノード
     */
    public Node removeProject(ProjectIF childProject)
    {
        // domTmsの所定の位置のProjectのノードを削除する。
        // projectsの所定の位置（pos）にprojectを削除する。
        int pos = projects.indexOf(childProject);
        boolean removed = projects.remove(childProject);
        if (removed)
        {
            NodeList nl = domTms.getElementsByTagName("MapspaceFile");
            return nl.item(pos);
        }
        else
        {
            log.error("can not remove: " + childProject.getTitle() + " project");
        }
        return null;// 削除したdomTms内のProjectのノード（のコピー）を返却
    }

    public void removeProject(Node projectNode)
    {
        Node node = domTms.adoptNode(projectNode);
        if (node != null)
        {
            log.debug("project node removed.");
        }
        else
        {
            log.error("project node could not removed.");
        }
    }

    /**
     * プロジェクトを追加します。
     * 
     * @param project 追加するプロジェクトオブジェクト
     * @param node 移動するプロジェクトノード
     * @param pos DOMノードの位置
     */
    public void moveProject(ProjectIF project, Node node, int pos)
    {
        pos = pos - datasets.size();
        // Node inode1 = domTms.adoptNode(node);
        // 又は
        // Node inode2 = domTms.importNode(node, true);
        // を実行し、戻り値のノードを所定の位置（pos）に挿入する。
        Node inode = domTms.adoptNode(node);
        Element root = domTms.getDocumentElement();
        NodeList nl = root.getElementsByTagName("MapspaceFile");
        if (nl.getLength() == pos)
        {
            root.appendChild(inode);
            projects.add(project);
        }
        else
        {
            root.insertBefore(inode, nl.item(pos));
            projects.add(pos, project);
        }
    }

    /**
     * tmsドキュメントを取得します。
     * 
     * @return tmsドキュメント
     */
    public Document getDomProject()
    {
        return domTms;
    }
    
    public void addProject(ProjectIF project) throws ProjectException
    {
    	log.debug("ADD PROJECT.");
    }

    /**
     * データセットを追加します。
     * 
     * @param dataset 追加するデータセット
     * @throws ProjectException
     */
    public void addDataset(DatasetIF dataset) throws ProjectException
    {
        Record[] records = null;
        if (dataset != null)
        {
            records = dataset.getRecords();
        }
        if (records != null && records.length > 0)
        {
            DatasetIF[] addDs = getDatasets();
            DatasetIF[] newDs = null;
            if (addDs != null)
            {
                newDs = new DatasetTmm[addDs.length + 1];
                System.arraycopy(addDs, 0, newDs, 0, addDs.length);
            }
            else
            {
                newDs = new DatasetTmm[1];
            }
            newDs[newDs.length - 1] = dataset;
            setDatasets(newDs);

            String importProject = getDatasetXML(dataset);
            Document addDoc = null;
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = null;
            try
            {
                db = dbf.newDocumentBuilder();
            }
            catch (ParserConfigurationException e)
            {
                throw new ProjectException(rmgr.format("error.parser.configuration", rmgr.getString("word.kensaku")), e);
            }
            try
            {
                addDoc = db.parse(new InputSource(new StringReader(importProject)));
            }
            catch (SAXException e)
            {
                throw new ProjectException(rmgr.format("error.sax", rmgr.getString("word.kensaku")), e);
            }
            catch (IOException e)
            {
                throw new ProjectException(rmgr.format("error.io", rmgr.getString("word.kensaku")), e);
            }

            Node newNode = addDoc.getElementsByTagName("Dataset").item(0);

            int len = 0;
            len = domTms.getElementsByTagName("Mapspace").getLength();
            Node dataNode = domTms.getElementsByTagName("Mapspace").item(len - 1);
            Node addNode = domTms.importNode(newNode, true);
            dataNode.appendChild(addNode);
        }
    }

    /**
     * 開始日時、終了日時を取得します。
     * 
     * @param start 開始日時のジュリアン日
     * @param end 終了日時のジュリアン日
     * @return 時間
     */
    private String[] getDate(BigDecimal start, BigDecimal end)
    {
        String[] date = { "0000-01-01T00:00:00", "2004-01-01T00:00:00" };
        if (start != null && end != null)
        {
            if (start.compareTo(end) < 0)
            {
                date[0] = JulianGregorianDate.toGregorian(start, "Z");
                date[1] = JulianGregorianDate.toGregorian(end, "Z");
            }
            else
            {
                date[1] = JulianGregorianDate.toGregorian(start, "Z");
                date[0] = JulianGregorianDate.toGregorian(end, "Z");
            }
        }
        else if (start != null && end == null)
        {
            date[0] = JulianGregorianDate.toGregorian(start, "Z");
            date[1] = JulianGregorianDate.toGregorian(start.add(BigDecimal.valueOf(10)), "Z");
        }
        else if (start == null && end != null)
        {
            date[0] = JulianGregorianDate.toGregorian(end.subtract(BigDecimal.valueOf(10)), "Z");
            date[1] = JulianGregorianDate.toGregorian(end, "Z");
        }
        return date;
    }

    /**
     * データセットからドキュメントの開始、終了日時を設定します。
     * 
     * @param datasets データセット
     * @return XMLの文字列バッファ
     */
    private String getProjectXML(DatasetIF[] datasets)
    {
        String sdate = "0000-01-01T00:00:00";
        String edate = "2004-01-01T00:00:00";
        try
        {
            BigDecimal start = getStartDatetime(datasets);
            BigDecimal end = getEndDatetime(datasets);
            String[] date = getDate(start, end);
            sdate = date[0];
            edate = date[1];
        }
        catch (DatasetException e)
        {
            log.error(e.getMessage(), e);
        }
        StringBuffer sb = new StringBuffer();
        sb.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
        sb.append("<Mapspace>\n");
        sb.append("<TMSVersion>3.0</TMSVersion>\n");
        sb.append("<TMVersion>2.2.0.70</TMVersion>\n");
        sb.append("<Name>New Project</Name>\n");
        sb.append("<DatasetType>Dataset</DatasetType>\n");
        sb.append("<DatasetFormat>Timemap Dataset</DatasetFormat>\n");
        sb.append("<CoordSys>Latitude/Longitude</CoordSys>\n");
        sb.append("<TimeEnabled>Yes</TimeEnabled>\n");
        sb.append("<Domain>\n");
        sb.append("<Name>MapSpace domain</Name>\n");
        sb.append("<MinX>-180</MinX>\n");
        sb.append("<MaxX>180</MaxX>\n");
        sb.append("<MinY>-90</MinY>\n");
        sb.append("<MaxY>90</MaxY>\n");
        sb.append("<StartT>" + sdate + "</StartT>\n");// 0000-01-01T00:00:00
        sb.append("<EndT>" + edate + "</EndT>\n");// 2004-01-01T00:00:00
        sb.append("<TimeDetail>0</TimeDetail>\n");
        sb.append("</Domain>\n");
        sb.append("<Domain>\n");
        sb.append("<Name>Startup extent</Name>\n");
        sb.append("<MinX>122.933</MinX>\n");
        sb.append("<MaxX>153.986</MaxX>\n");
        sb.append("<MinY>24.041</MinY>\n");
        sb.append("<MaxY>45.555</MaxY>\n");
        sb.append("<StartT>" + sdate + "</StartT>\n");// 0000-01-01T00:00:00
        sb.append("<EndT>" + edate + "</EndT>\n");// 2004-01-01T00:00:00
        sb.append("<TimeDetail>0</TimeDetail>\n");
        sb.append("</Domain>\n");
        sb.append("<AutoLayerAdd>\n");
        sb.append("<Enabled>No</Enabled>\n");
        sb.append("<SearchAlways>Yes</SearchAlways>\n");
        sb.append("</AutoLayerAdd>\n");
        sb.append("<Domain>\n");
        sb.append("<Name>New Domain</Name>\n");
        sb.append("</Domain>\n");
        sb.append("</Mapspace>\n");
        return sb.toString();
    }

    /**
     * DOMをセットします。
     * 
     * @param in バイト入力ストリーム
     * @throws ParserConfigurationException
     * @throws SAXException
     * @throws IOException
     */
    private void setDomTms(InputStream in) throws ParserConfigurationException, SAXException, IOException
    {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder = null;
        builder = factory.newDocumentBuilder();
        domTms = builder.parse(in);
    }

    /**
     * データセットからドキュメントの開始、終了日時を設定します。
     * 
     * @param ds データセット
     * @return XMLの文字列バッファ
     */
    private String getDatasetXML(DatasetIF ds)
    {
        String sdate = "0000-01-01T00:00:00";
        String edate = "2004-01-01T00:00:00";
        try
        {
            BigDecimal start = ds.getStartDatetime();
            BigDecimal end = ds.getEndDatetime();
            String[] date = getDate(start, end);
            sdate = date[0];
            edate = date[1];
        }
        catch (DatasetException e)
        {
            log.error(e.getMessage(), e);
        }
        StringBuffer sb = new StringBuffer();
        sb.append("<Dataset>\n");
        sb.append("<LocalFilePath>" + ds.getPath() + "</LocalFilePath>\n");
        sb.append("<Layer>");
        sb.append("<Title>search result</Title>\n");

        sb.append("<Handle>search result</Handle>\n");
        sb.append("<DrawOrder>1</DrawOrder>\n");
        sb.append("<DateMin>" + sdate + "</DateMin>\n");// 0000-01-01T00:00:00
        sb.append("<DateMax>" + edate + "</DateMax>\n");// 2004-01-01T00:00:00
        sb.append("<Symbol>\n");
        sb.append("<MapObjectTypeCode>0</MapObjectTypeCode>\n");
        sb.append("<FlagDefault>Yes</FlagDefault>\n");
        sb.append("<Fitted>No</Fitted>\n");
        sb.append("<Size>10</Size>\n");
        sb.append("<ForeColor>0:0:255</ForeColor>\n");
        sb.append("<ForeColorDis>0:0:0</ForeColorDis>\n");
        sb.append("<StyleCode>3</StyleCode>\n");
        sb.append("<UseOutline>Yes</UseOutline>\n");
        sb.append("<OutlineColor>0:0:255</OutlineColor>\n");
        sb.append("</Symbol>\n");

        // sb.append("<Symbol/>\n");
        sb.append("<Renderer/>\n");
        sb.append("</Layer>\n");
        sb.append("</Dataset>\n");
        return sb.toString();
    }
    
    private String absolutePath;
    
    public String getAbsolutePath() {
    	return absolutePath;
    }
    
    public void setAbsolutePath(String absolutePath) {
    	this.absolutePath = absolutePath;
    }
    
    private String rdfBase;

    /**
     * プロジェクトをロードします。
     * 
     * @param op プラグイン名
     * @param path ディレクトリパス
     * @param base ファイル名
     * @throws ProjectException
     */
    public void loadProject(OpenPlugin op, String path, String base) throws ProjectException
    {
        this.path = path;
        String realPath = op.getAbsolutePath(path, base);
        this.absolutePath = realPath;
        
        this.rdfBase = realPath;
        
        try
        {
            setDomTms(op.getProjectInputStream(realPath));
        }
        catch (ParserConfigurationException e)
        {
            throw new ProjectException(rmgr.format("error.parser.configuration", ""), e);
        }
        catch (SAXException e)
        {
            throw new ProjectException(rmgr.format("error.sax", realPath), e);
        }
        catch (IOException e)
        {
            throw new ProjectException(rmgr.format("error.io", realPath), e);
        }
        catch (OpenPluginException e)
        {
            throw new ProjectException(e.getMessage(), e);
        }
        base = op.getBasePath(realPath);
        loadChildDataset(op, base);
        loadChildProject(op, base);
    }

    /**
     * @param op プラグイン名
     * @param base ファイル名
     * @throws ProjectException
     */
    private void loadChildDataset(OpenPlugin op, String base) throws ProjectException
    {
        NodeList nl = domTms.getElementsByTagName("LocalFilePath");
        DatasetIF dataset = null;
        datasets = new ArrayList<DatasetIF>();
        for (int i = 0; i < nl.getLength(); i++)
        {
            Element elem = (Element) nl.item(i);
            String tmmPath = elem.getTextContent().trim();
            dataset = new DatasetTmm();
            try
            {
                dataset.loadDataset(op, tmmPath, base);
                if (dataset.hasRecords())
                {
                    datasets.add(dataset);
                }
            }
            catch (DatasetException e)
            {
                log.fatal(e.getMessage(), e);
                // throw new ProjectException(e.getMessage(), e);
            }
        }
    }

    /**
     * @param op プラグイン名
     * @param base ファイル名
     * @throws ProjectException
     */
    private void loadChildProject(OpenPlugin op, String base) throws ProjectException
    {
        NodeList nl = domTms.getElementsByTagName("MapspaceFile");
        ProjectIF project = null;
        projects = new ArrayList<ProjectIF>();
        for (int i = 0; i < nl.getLength(); i++)
        {
            Element elem = (Element) nl.item(i);
            String tmsPath = elem.getTextContent().trim();
            project = new ProjectTms();
            project.loadProject(op, tmsPath, base);
            projects.add(project);
        }
        // projects = new Project[v.size()];
        // v.copyInto(projects);
    }

    /**
     * プロジェクトのタイトルを取得します。
     * 
     * @return タイトル
     */
    public String getTitle()
    {
        if (title == null)
        {
            if (domTms != null)
            {
                NodeList nl = domTms.getElementsByTagName("Name");
                Element elem = (Element) nl.item(0);
                title = elem.getTextContent();
            }
        }
        return title;
    }

    /**
     * プロジェクトのタイトルを設定します。
     * 
     * @param title タイトル
     */
    public void setTitle(String title)
    {
        this.title = title;
        try
        {
            XPathFactory factory = XPathFactory.newInstance();
            javax.xml.xpath.XPath xpath = factory.newXPath();
            XPathExpression expr = xpath.compile("/Mapspace/Name");
            Element name = (Element) expr.evaluate(domTms, XPathConstants.NODE);
            name.setTextContent(title);
        }
        catch (XPathExpressionException e)
        {
            // never catch
            log.fatal("not valid xpath.", e);
        }
    }

    /**
     * データセットを取得します。
     * 
     * @return データセット
     */
    public DatasetIF[] getDatasets()
    {
        if (datasets != null)
        {
            DatasetIF[] ds = new DatasetTmm[datasets.size()];
            ds = datasets.toArray(ds);
            return ds;
        }
        return null;
    }

    /**
     * データセットを設定します。
     * 
     * @param ds データセット
     */
    public void setDatasets(DatasetIF[] ds)
    {
        datasets = new ArrayList<DatasetIF>();
        for (int i = 0; i < ds.length; i++)
        {
            datasets.add(ds[i]);
        }
    }

    /**
     * プロジェクトを取得します。
     * 
     * @return プロジェクト
     */
    public ProjectIF[] getProjects()
    {
        if (projects != null)
        {
            ProjectIF[] ps = new ProjectTms[projects.size()];
            ps = projects.toArray(ps);// projects.copyInto(ps);
            return ps;
        }
        return null;
    }

    /**
     * プロジェクトを設定します。
     * 
     * @param ps プロジェクト
     */
    public void setProjects(ProjectIF[] ps)
    {
        projects = new ArrayList<ProjectIF>();
        for (int i = 0; i < ps.length; i++)
        {
            projects.add(ps[i]);
        }
    }
    
    public Object[] getNodes()
    {
//        if (children != null)
//        {
//        	Object[] objs = new Object[children.size()];
//        	objs = children.toArray(objs);
//            return objs;
//        }
        return null;
    }

    /**
     * パスを取得します。
     * 
     * @return パス
     */
    public String getPath()
    {
        return path;
    }
    
    public void setPath(String path) {
    	this.path = path;
    }

    /**
     * 開始日時を取得します。
     * 
     * @return 開始日時
     * @throws DatasetException
     * @throws ISO8601Exception
     */
    public BigDecimal getStartDatetime() throws DatasetException, ISO8601Exception
    {
        BigDecimal jdate = null;
        if (domTms != null)
        {
            NodeList nl = domTms.getElementsByTagName("StartT");
            if (nl.getLength() > 0)
            {
                Element elem = (Element) nl.item(0);
                String date = elem.getTextContent();
                jdate = JulianGregorianDate.toJulian(date);
            }
        }
        ProjectIF[] projects = getProjects();
        if (projects != null)
        {
            for (int i = 0; i < projects.length; i++)
            {
                BigDecimal date = projects[i].getStartDatetime();
                if (date != null && (jdate == null || date.compareTo(jdate) == -1))
                {
                    jdate = date;
                }
            }
        }
        DatasetIF[] datasets = getDatasets();
        if (datasets != null)
        {
            BigDecimal date = getStartDatetime(datasets);
            if (date != null && (jdate == null || date.compareTo(jdate) == -1))
            {
                jdate = date;
            }
        }
        return jdate;
    }

    /**
     * データセットから開始日時を取得します。
     * 
     * @param datasets データセット
     * @return 開始日時
     * @throws DatasetException
     */
    private BigDecimal getStartDatetime(DatasetIF[] datasets) throws DatasetException
    {
        BigDecimal jdate = null;
        for (int i = 0; i < datasets.length; i++)
        {
            BigDecimal date = datasets[i].getStartDatetime();
            if (jdate == null || (date != null && date.compareTo(jdate) == -1))
            {
                jdate = date;
            }
        }
        return jdate;
    }

    /**
     * 終了日時を取得します。
     * 
     * @return 終了日時
     * @throws DatasetException
     * @throws ISO8601Exception
     */
    public BigDecimal getEndDatetime() throws DatasetException, ISO8601Exception
    {
        BigDecimal jdate = null;
        if (domTms != null)
        {
            NodeList nl = domTms.getElementsByTagName("EndT");
            if (nl.getLength() > 0)
            {
                Element elem = (Element) nl.item(0);
                String date = elem.getTextContent();
                jdate = JulianGregorianDate.toJulian(date);
            }
        }
        ProjectIF[] projects = getProjects();
        if (projects != null)
        {
            for (int i = 0; i < projects.length; i++)
            {
                BigDecimal date = projects[i].getEndDatetime();
                if (date != null && (jdate == null || date.compareTo(jdate) == 1))
                {
                    jdate = date;
                }
            }
        }
        DatasetIF[] datasets = getDatasets();
        if (datasets != null)
        {
            BigDecimal date = getEndDatetime(datasets);
            if (date != null && (jdate == null || date.compareTo(jdate) == 1))
            {
                jdate = date;
            }
        }
        return jdate;
    }

    /**
     * データセットから終了日時を取得します。
     * 
     * @param datasets データセット
     * @return 終了日時
     * @throws DatasetException
     */
    private BigDecimal getEndDatetime(DatasetIF[] datasets) throws DatasetException
    {
        BigDecimal jdate = null;
        for (int i = 0; i < datasets.length; i++)
        {
            BigDecimal date = datasets[i].getEndDatetime();
            if (jdate == null || (date != null && date.compareTo(jdate) == 1))
            {
                jdate = date;
            }
        }
        return jdate;
    }

    /**
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
        return getTitle();
    }
}
