/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.data;

import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import jp.nihu.gttime.utility.Utilities;

import org.w3c.dom.Attr;
import org.w3c.dom.DOMException;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * レコードの内容を設定、取得するクラスです。
 */
public class Record
{
    /**
     * IDオブジェクト
     */
    private String id = null;

    /**
     * レベルオブジェクト
     */
    private int level = -1;

    /**
     * ロケーションオブジェクト
     */
    private Location location;

    /**
     * 開始日時オブジェクト
     */
    private JulianDate startDatetime;

    /**
     * 終了日時オブジェクト
     */
    private JulianDate endDatetime;

    /**
     * イベントオブジェクト
     */
    private Event[] events = new Event[0];
    
    /**
     * レコードがXMLの場合、recod要素以下のnodeをここに格納する。
     * レコードを保存する際にnoeがnullでない場合はこの値を利用してXMLを出力する。
     */
    private Node node = null;
    
    public void setNode(Node node) {
    	this.node = node;
    }
    
    public Node getNode() {
    	return node;
    }


    /**
     * 検索フラグ
     */
    private boolean hit = false;

    public boolean isEndEdgeOpen()
    {
        if (endDatetime != null && endDatetime.getEdge() != null && endDatetime.getEdge().equals("close"))
        {
            return false;
        }
        return true;
    }

    public boolean isStartEdgeOpen()
    {
        if (startDatetime != null && startDatetime.getEdge() != null && startDatetime.getEdge().equals("open"))
        {
            return true;
        }
        return false;
    }

    /**
     * レコード表示解像度下限（指定時間幅以上のときに表示）。日数で指定。
     */
    private BigDecimal resolutionMin = null;

    /**
     * レコード表示解像度上限（指定時間幅以下のときに表示）。日数で指定。
     */
    private BigDecimal resolutionMax = null;

    private BigDecimal maxVisibleTResolution = null;

    private BigDecimal minVisibleTResolution = null;

    public void setMaxVisibleTResolution(String maxVisibleTResolution)
    {
        this.maxVisibleTResolution = Utilities.getDurationVal(maxVisibleTResolution);
    }

    public BigDecimal getMaxVisibleTResolution()
    {
        return maxVisibleTResolution;
    }

    public void setMinVisibleTResolution(String minVisibleTResolution)
    {
        this.minVisibleTResolution = Utilities.getDurationVal(minVisibleTResolution);
    }

    public BigDecimal getMinVisibleTResolution()
    {
        return minVisibleTResolution;
    }

    /**
     * @return the resolutionMin
     */
    public BigDecimal getResolutionMin()
    {
        return resolutionMin;
    }

    /**
     * @param resolutionMin the resolutionMin to set
     */
    public void setResolutionMin(BigDecimal resolutionMin)
    {
        this.resolutionMin = resolutionMin;
    }

    /**
     * @return the resolutionMax
     */
    public BigDecimal getResolutionMax()
    {
        return resolutionMax;
    }

    /**
     * @param resolutionMax the resolutionMax to set
     */
    public void setResolutionMax(BigDecimal resolutionMax)
    {
        this.resolutionMax = resolutionMax;
    }

    /**
     * 検索フラグを設定します。
     * 
     * @param hit ヒットした、しない
     */
    public void setHit(boolean hit)
    {
        this.hit = hit;
    }

    /**
     * 検索フラグを取得します。
     * 
     * @return 検索フラグ
     */
    public boolean isHit()
    {
        return hit;
    }

    /**
     * 終了日時を取得します。
     * 
     * @return 終了日時
     */
    public JulianDate getEndDatetime()
    {
        return endDatetime;
    }

    /**
     * 終了日時を設定します。
     * 
     * @param endDatetime 終了日時
     */
    public void setEndDatetime(JulianDate endDatetime) throws RecordException
    {
        this.endDatetime = endDatetime;
        // if (startDatetime == null || compare(startDatetime, endDatetime) > 0)
        // {
        // this.startDatetime = endDatetime;
        // }
        if (startDatetime != null && endDatetime != null && compare(startDatetime, endDatetime) > 0)
        {
            throw new RecordException("invalid end date value.");
        }
    }

    /**
     * イベントを取得します。
     * 
     * @return イベント
     */
    public Event[] getEvents()
    {
        // return (events != null)? events: new Event[0];
        return events;
    }

    /**
     * イベントを設定します。
     * 
     * @param events イベント
     */
    public void setEvents(Event[] events)
    {
        this.events = events;
    }

    /**
     * ロケーションを取得します。
     * 
     * @return ロケーション
     */
    public Location getLocation()
    {
        return location;
    }

    /**
     * ロケーションを設定します。
     * 
     * @param location ロケーション
     */
    public void setLocation(Location location)
    {
        this.location = location;
    }

    /**
     * 開始日時を取得します。
     * 
     * @return 開始日時
     */
    public JulianDate getStartDatetime()
    {
        return startDatetime;
    }

    /**
     * 開始日時を設定します。
     * 
     * @param startDatetime 開始日時
     */
    public void setStartDatetime(JulianDate startDatetime) throws RecordException
    {
        this.startDatetime = startDatetime;
        // if (endDatetime == null || compare(startDatetime, endDatetime) > 0)
        // {
        // this.endDatetime = startDatetime;
        // }
        if (startDatetime != null && endDatetime != null && compare(startDatetime, endDatetime) > 0)
        {
            throw new RecordException("invalid start date value.");
        }
    }

    /**
     * 開始日時と終了日時を比較します。
     * 
     * @param startDatetime 開始日時
     * @param endDatetime 終了日時
     * @return 比較した値
     */
    private int compare(JulianDate startDatetime, JulianDate endDatetime)
    {
        BigDecimal start = startDatetime.getDatetime();
        BigDecimal end = endDatetime.getDatetime();
        return start.compareTo(end);
    }

    /**
     * レベルを取得します。
     * 
     * @return レベル
     */
    public int getLevel()
    {
        return level;
    }

    /**
     * レベルを設定します。
     * 
     * @param level レベル
     */
    public void setLevel(int level)
    {
        this.level = level;
    }

    /**
     * IDを取得します。
     * 
     * @return ID
     */
    public String getId()
    {
        return id;
    }

    /**
     * IDを設定します。
     * 
     * @param id ID
     */
    public void setId(String id)
    {
        this.id = id;
    }
    
    private String group;
    
    public String getGroup() {
        if (group == null) {
            group = "";
        }
        return group;
    }
    
    public void setGroup(String group) {
        this.group = group;
    }
    
    public void createRdfSameAsEvent(String uri) {
    	if (isIdWithoutProtocol()) {
        	NodeList nl = node.getChildNodes();
        	for (int i = 0; i < nl.getLength(); i++) {
        		Node cn = nl.item(i);
        		if (cn.getNodeName().equals("event")) {
        			try {
        				String sameAsId = uri + "#" + id;
        				Element item = node.getOwnerDocument().createElementNS("http://www.nihu.jp/ns/spacetime/2006/11", "item");
            			item.setAttribute("property", "owl:sameAs");
						item.setAttribute("resource", sameAsId);
            			item.setAttribute("name", "Same As");
            			item.setTextContent(sameAsId);
            			cn.appendChild(item);
            			List<Event> es = new ArrayList<Event>(Arrays.asList(getEvents()));
            			Event ev = new Event();
            			ev.setName("Same As");
            			ev.setData(sameAsId);
            			es.add(ev);
            			setEvents(es.toArray(new Event[0]));
					} catch (DOMException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
        		}
        	}
    	}
    }
    
    public void setAttributeToRootElement(Element currentRoot, Element newRoot) {
    	// 属性をチェックしてまだ新しいルート要素にない属性があったら新しいルート要素に追加する
    	NamedNodeMap attrs = currentRoot.getAttributes();
    	for (int i = 0; i < attrs.getLength(); i++) {
    		Attr attr = (Attr)attrs.item(i);
    		if (!hasAttribute(newRoot, attr)) {
    			newRoot.setAttribute(attr.getName(), attr.getValue());
    		}
    	}
    }
    
    private boolean hasAttribute(Element elem, Attr attr) {
    	NamedNodeMap attrs = elem.getAttributes();
    	for (int i = 0; i < attrs.getLength(); i++) {
    		Attr a0 = (Attr)attrs.item(i);
    		if (a0.getName().equals(attr.getName()) && a0.getValue().equals(attr.getValue())) {
    			return  true;
    		}
    	}
    	return  false;
    }
    
    public boolean isIdWithoutProtocol() {
    	return (getId() != null
    			&& 0 < getId().length()
    			&& !getId().startsWith("file:/")
    			&& !getId().startsWith("http://")
    			&& !getId().startsWith("https://"));
    }
}
