/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.layer;

import java.awt.Color;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.GridBagLayout;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;

import javax.swing.BoxLayout;
import javax.swing.JComponent;
import javax.swing.JLayeredPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.DatasetIF;
import jp.nihu.gttime.plugin.LayerPlugin;
import jp.nihu.gttime.plugin.PluginManager;
import jp.nihu.gttime.plugin.PluginManagerException;
import jp.nihu.gttime.tree.TreeCheckNode;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * レイヤパネルの初期表示となるパネルを生成するクラスです。
 */
public class LayerViewPart extends JScrollPane
{
    /**
     * シリアルバーションID
     */
    private static final long serialVersionUID = 1L;

    /**
     * ログ用オブジェクト
     */
    private static final Log log = LogFactory.getLog(LayerViewPart.class);

    private LayeredPanel layeredPane = null;

    /**
     * レイヤの初期表示を設定します。
     */
    public LayerViewPart()
    {
        super();
        getVerticalScrollBar().setUnitIncrement(25);
        setHorizontalScrollBarPolicy(HORIZONTAL_SCROLLBAR_NEVER);
        setVerticalScrollBarPolicy(VERTICAL_SCROLLBAR_ALWAYS);

        JPanel emptyPanel = new JPanel();
        emptyPanel.setLayout(new GridBagLayout());
        emptyPanel.setBackground(Color.gray);
        setViewportView(emptyPanel);
    }

    public JPanel getLayerList()
    {
        return layeredPane.getLayerList();
    }

    /**
     * レイヤを描画して、追加する
     * 
     * @param node TreeCheckNodeオブジェクト
     * @param pos 追加する位置
     */
    public void addLayer(TreeCheckNode node, int pos)
    {
        addLayer(createLayer(node), pos);
    }

    /**
     * レイヤを描画して、追加する
     * 
     * @param node TreeCheckNodeオブジェクト
     */
    public void addLayer(TreeCheckNode node)
    {
        addLayer(createLayer(node));
    }

    public void addLayer(LayerPlugin layer, int pos)
    {
        layeredPane.addLayer(layer, pos);
    }

    public void addLayer(LayerPlugin layer)
    {
        layeredPane.addLayer(layer);
    }

    public void removeLayer(int i)
    {
        layeredPane.removeLayer(i);
    }
    
    public LayerPlugin getLayer(int i)
    {
    	if (layeredPane != null) {
    		return layeredPane.getLayer(i);
    	} else {
    		return null;
    	}
    }

    public BigDecimal[][] getDurations()
    {
        return layeredPane.getDurations();
    }

    public void setForcePaint()
    {
        layeredPane.getLayerList().setForcePaint();
    }

    private LayerPlugin createLayer(TreeCheckNode node)
    {
        LayerPlugin layer = null;
        try
        {
            PluginManager pluginManager = T2Map.getPluginManager();

            DatasetIF dataset = (DatasetIF) node.getUserObject();
            String layertype = dataset.getLayerType();
            Hashtable<String, String> plugin = pluginManager.getPlugins("LayerPlugin", layertype);
            if (plugin == null)
            {
                plugin = pluginManager.getPlugins("LayerPlugin", "DefaultLayer");
            }
            layer = (LayerPlugin) pluginManager.newInstance(plugin.get("class"));
            layer.init(getT2Map().getTimeAxisData(), node);
            node.setIcon(layer.getIcon());
            node.setLayer(layer);
        }
        catch (PluginManagerException e)
        {
            log.fatal(e.getMessage(), e);
        }
        return layer;
    }

    /**
     * T2Mapオブジェクトを取得します。
     * 
     * @return T2Mapオブジェクト
     */
    private T2Map getT2Map()
    {
        for (Container cont = getParent(); cont != null; cont = cont.getParent())
        {
            if (cont instanceof T2Map)
            {
                return (T2Map) cont;
            }
        }
        return null;
    }

    public void createLayerList()
    {
        layeredPane = new LayeredPanel();
        setViewportView(layeredPane);
        layeredPane.layeredPanel();
    }

    private class LayeredPanel extends JLayeredPane
    {
        private static final long serialVersionUID = 1L;

        private List<JComponent> parts = new ArrayList<JComponent>();

        private LayerListPanel layerList;

        private TimeDurationSelector timeDurationSelector;

        private HighLighter highLighter;

        public LayeredPanel()
        {
        }

        public BigDecimal[][] getDurations()
        {
            return timeDurationSelector.getDurations();
        }

        public LayerListPanel getLayerList()
        {
            return layerList;
        }

        public void addLayer(LayerPlugin layer)
        {
            layerList.addLayer(layer, parts);
        }

        public void addLayer(LayerPlugin layer, int pos)
        {
            layerList.addLayer(layer, parts, pos);
        }

        public void removeLayer(int i)
        {
            layerList.removeLayer(i);
        }
        
        public LayerPlugin getLayer(int i)
        {
            return layerList.getLayer(i);
        }

        @Override
        public Dimension getPreferredSize()
        {
            if (layerList != null)
            {
                return layerList.getPreferredSize();
            }
            return super.getPreferredSize();
        }

        public void layeredPanel()
        {
            TimeCursorLine line = new TimeCursorLine();
            parts.add(line);
            add(line, new Integer(50), 0);

            DateLabel dateLabel = new DateLabel();
            parts.add(dateLabel);
            add(dateLabel, new Integer(50), 0);

            ResizeLine resizeLine = new ResizeLine();
            parts.add(resizeLine);
            add(resizeLine, new Integer(50), 0);

            timeDurationSelector = new TimeDurationSelector();
            parts.add(timeDurationSelector);
            timeDurationSelector.init(getT2Map().getTimeAxisData());
            highLighter = new HighLighter();
            parts.add(highLighter);

            layerList = new LayerListPanel();
            add(layerList, new Integer(2), 0);

            add(timeDurationSelector, new Integer(30), 0);
            add(highLighter, new Integer(20), 0);
        }

        /**
         * @see javax.swing.JComponent#getWidth()
         */
        // @Override
        // public int getWidth()
        // {
        // return getParent().getWidth();
        // }
        /**
         * @see java.awt.Component#setBounds(int, int, int, int)
         */
        @Override
        public void setBounds(int x, int y, int width, int height)
        {
            super.setBounds(x, y, width, height);
            if (layerList == null)
            {
            	layeredPanel();
            }
            layerList.setBounds(0, 0, getWidth(), layerList.getHeight());
            timeDurationSelector.setBounds(0, 0, getWidth(), layerList.getHeight());
            highLighter.setBounds(0, 0, getWidth(), layerList.getHeight());
        }
    }

    private class LayerListPanel extends JPanel
    {
        /**
         * シリアルバージョンID
         */
        private static final long serialVersionUID = 1L;

        /**
         * コンストラクタ
         */
        public LayerListPanel()
        {
            super();
            BoxLayout layout = new BoxLayout(this, BoxLayout.Y_AXIS);
            setLayout(layout);
        }

        public void addLayer(LayerPlugin layer, List<JComponent> parts)
        {
            addListeners(layer, parts);
            add(layer);
        }

        public void addLayer(LayerPlugin layer, List<JComponent> parts, int pos)
        {
            addListeners(layer, parts);
            add(layer, pos);
        }

        private void addListeners(LayerPlugin layer, List<JComponent> parts)
        {
            Iterator<JComponent> it = parts.iterator();
            while (it.hasNext())
            {
                JComponent comp = it.next();
                layer.addMouseListener((MouseListener) comp);
                layer.addMouseMotionListener((MouseMotionListener) comp);
            }
        }
        
        public LayerPlugin getLayer(int i)
        {
        	if (this.getComponentCount() > 0) {
        		return (LayerPlugin)getComponent(i);
        	} else {
        		return null;
        	}
        }

        public void removeLayer(int i)
        {
            remove(i);
        }

        public void setForcePaint()
        {
            Component[] comps = getComponents();
            for (int i = 0; i < comps.length; i++)
            {
                if (comps[i] instanceof LayerPlugin)
                {
                    ((LayerPlugin) comps[i]).setForcePaint();
                }
            }
        }

        /**
         * @see javax.swing.JComponent#getWidth()
         */
        // @Override
        // public int getWidth()
        // {
        // return getParent().getWidth();
        // }
        /**
         * @see javax.swing.JComponent#getHeight()
         */
        @Override
        public int getHeight()
        {
            return getParent().getHeight();
        }
    }

    // private JSplitPane getSplitPane(Container c)
    // {
    // if (c == null || c instanceof JRootPane)
    // {
    // return null;
    // }
    // else
    // {
    // Container p = c.getParent();
    // if (p instanceof JSplitPane)
    // {
    // return (JSplitPane)p;
    // }
    // else
    // {
    // return getSplitPane(p);
    // }
    // }
    // }
    //  
    // public int getWidth()
    // {
    // JSplitPane sp = getSplitPane(this);
    // if (sp != null)
    // {
    // return sp.getWidth() - sp.getLeftComponent().getWidth();
    // }
    // else
    // {
    // return super.getWidth();
    // }
    // }
}
