/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.layer;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.math.BigDecimal;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Iterator;
import java.util.TreeMap;

/**
 * レイヤパネル内で数値データ（年輪値など）を表示する際に、<br/> 縦軸線について管理するクラスです。
 */
public class ScaleBar
{
    /**
     * スケール値の設定
     */
    private static final double LOG10SCALE = 1 / Math.log(10);

    /**
     * 値の自然対数値を返します
     * 
     * @param val
     * @return 自然対数値
     */
    private static double log10(double val)
    {
        return Math.log(val) * LOG10SCALE;
    }

    /**
     * 縦軸線を描画します。
     * 
     * @param maxTics 目盛の最大数
     * @param ticLength 目盛の長さ
     * @param lowVal 目盛の最小値
     * @param highVal 目盛の最大値
     * @param screenStart 画面の左端の位置
     * @param screenEnd 画面の右端の位置
     * @param screenOffset 画面のオフセット
     * @param screenHeight 画面の高さ
     * @param screenWidth 画面の幅
     * @param g 描画対象のグラフィックス
     */
    public static void drawScale(int maxTics, int ticLength, float lowVal, float highVal, int screenStart, int screenEnd, int screenOffset, int screenHeight, int screenWidth, Graphics g)
    {
        g.drawLine(screenOffset, screenStart, screenOffset, screenEnd);
        if (lowVal == highVal)
        {
            String ticStr = Float.toString(lowVal);
            if (ticStr.contains("."))
            {
                if (ticStr.length() > 5) {
                    ticStr = ticStr.substring(0, 5);
                }
            }
            Font font = new Font("SansSerif", Font.PLAIN, 9);
            g.setFont(font);

            //float ticVal = Float.parseFloat(ticStr);
            int ticCoord = screenStart;
            Dimension strSize = stringSize(ticStr, g);
            //ticCoord += plotValue(ticVal, lowVal, highVal, screenStart, screenEnd, screenHeight);

            //ticCoord = screenHeight - ticCoord;
            g.drawLine(screenOffset - ticLength, ticCoord, screenWidth, ticCoord);

            g.setColor(Color.black);
            g.drawString(ticStr, screenOffset - ticLength - strSize.width - 1, ticCoord + strSize.height / 3);
            g.setColor(Color.lightGray);
            //lastLabelEnd = ticCoord - strSize.height;
        }
        else
        {
            TreeMap<Double, String> tics = ScaleBar.computeTicks(lowVal, highVal, maxTics);
            int lastLabelEnd = 88888;
            for (Iterator<Double> e = tics.keySet().iterator(); e.hasNext();)
            {
                Double ticD = e.next();
                String ticStr = tics.get(ticD);
                if (ticStr.contains("."))
                {
                    if (ticStr.length() > 5) {
                        ticStr = ticStr.substring(0, 5);
                    }
                }
                Font font = new Font("SansSerif", Font.PLAIN, 9);
                g.setFont(font);

                //float ticVal = Float.parseFloat(ticStr);
                BigDecimal bd = new BigDecimal(ticD);
                float ticVal = bd.floatValue();
                int ticCoord = screenStart;
                Dimension strSize = stringSize(ticStr, g);
                ticCoord += plotValue(ticVal, lowVal, highVal, screenStart, screenEnd, screenHeight);

                ticCoord = screenHeight - ticCoord;
                g.drawLine(screenOffset - ticLength, ticCoord, screenWidth, ticCoord);

                if (ticCoord - strSize.height / 3 < lastLabelEnd)
                {
                    g.setColor(Color.black);
                    g.drawString(ticStr, screenOffset - ticLength - strSize.width - 1, ticCoord + strSize.height / 3);
                    g.setColor(Color.lightGray);
                    lastLabelEnd = ticCoord - strSize.height;
                }
            }
        }
    }

    /**
     * 目盛りの表示可能なラベルを取得します。
     * 
     * @param ticMinVal 線の最小値
     * @param ticMaxVal 線の最大値
     * @param maxTicks 線の表示可能数
     * @return 目盛りの表示可能なラベル
     */
    private static TreeMap<Double, String> computeTicks(double ticMinVal, double ticMaxVal, int maxTicks)
    {
        double xStep = roundUp((ticMaxVal - ticMinVal) / maxTicks);
        int numfracdigits = numFracDigits(xStep);

        double xStart = xStep * Math.ceil(ticMinVal / xStep);
        TreeMap<Double, String> labels = new TreeMap<Double, String>();
        for (double xpos = xStart; xpos <= ticMaxVal; xpos += xStep)
        {
            //log.debug(formatNum(xpos, numfracdigits) + " <- " + xpos + " + " + numfracdigits);
            labels.put(xpos,formatNum(xpos, numfracdigits));
        }
        return labels;
    }

    /**
     * 目盛りの描画位置を取得します。
     * 
     * @param val 目盛の値
     * @param lowVal 目盛の最小値
     * @param highVal 目盛の最大値
     * @param screenStart 画面の左端の位置
     * @param screenEnd 画面の右端の位置
     * @param screenHeight 画面の高さ
     * @return 目盛りの描画位置
     */
    private static int plotValue(float val, float lowVal, float highVal, int screenStart, int screenEnd, int screenHeight)
    {
        int screenRange = screenEnd - screenStart;
        float valueRange = highVal - lowVal;
        float pixelsPerUnit = screenRange / valueRange;
        return (int) ((val - lowVal) * pixelsPerUnit + .5);
    }

    /**
     * 目盛りと値を考慮した最適な表示間隔を取得します。
     * 
     * @param val 目盛の値
     * @return 目盛りと値を考慮した表示間隔
     */
    private static double roundUp(double val)
    {
        int exponent = (int) Math.floor(log10(val));
        val *= Math.pow(10, -exponent);
        if (val > 5.0)
        {
            val = 10.0;
        }
        else if (val > 2.0)
        {
            val = 5.0;
        }
        else if (val > 1.0)
        {
            val = 2.0;
        }
        val *= Math.pow(10, exponent);
        return val;
    }

    /**
     * computeTicksメソッドの計算時に使用します。
     * 
     * @param num 目盛の数
     * @return computeTicksメソッドの計算時
     */
    private static int numFracDigits(double num)
    {
        int numdigits = 0;
        while (numdigits <= 15 && num != Math.floor(num))
        {
            num *= 10.0;
            numdigits += 1;
        }
        return numdigits;
    }

    /**
     * 値をフォーマットします。
     * 
     * @param num 目盛の値
     * @param numfracdigits 桁数
     * @return フォーマット結果
     */
    private static String formatNum(double num, int numfracdigits)
    {
        if (num >= 10000 || (num < 0.001 && num > -0.001 && num != 0) || num <= -1000)
        {
            DecimalFormat df = new DecimalFormat();
            df.applyLocalizedPattern("0.#E0");
            return df.format(num);
        }
        else
        {
            NumberFormat numberFormat = NumberFormat.getInstance();
            numberFormat.setGroupingUsed(false);
            numberFormat.setMinimumFractionDigits(numfracdigits);
            numberFormat.setMaximumFractionDigits(numfracdigits);
            return numberFormat.format(num);
        }
    }

    /**
     * 目盛り値（2007等）を幅値に変換して取得します。
     * 
     * @param str 目盛の値の文字列
     * @param g 描画対象のグラフィックス
     * @return 目盛り値（2007等）を幅値に変換した結果
     */
    private static Dimension stringSize(String str, Graphics g)
    {
        if (g instanceof Graphics2D)
        {
            java.awt.geom.Rectangle2D bounds = g.getFont().getStringBounds(str, ((Graphics2D) g).getFontRenderContext());
            return new Dimension((int) (bounds.getWidth() + .5), (int) (bounds.getHeight() + .5));
        }
        else
        {
            return new Dimension(g.getFontMetrics().stringWidth(str), g.getFontMetrics().getHeight());
        }
    }
}
