/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.DatasetIF;
import jp.nihu.gttime.data.ProjectException;
import jp.nihu.gttime.data.ProjectIF;
import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.time.TimeAxisIF;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * 解析プラグインについて管理するクラスである。 <br/>また、検索結果の情報をリセットするメソッドが存在する。
 */
public abstract class AnalizePlugin implements Plugin
{
    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(AnalizePlugin.class);

    /**
     * T2Mapオブジェクト
     */
    private T2Map t2map;

    /**
     * 初期化します。<br/>
     * <ul>
     * <li>T2Mapオブジェクトを自クラスのT2Mapオブジェクトに指定します</li>
     * <li>removeHitメソッドを呼び出しProjectオブジェクトを初期化します</li>
     * </ul>
     * 
     * @param t2map
     * @param project 初期化するProjectオブジェクト
     * @see jp.nihu.gttime.plugin.AnalizePlugin#removeHit(org.h_gis.launcher.gttime.data.Project)
     */
    public void init(T2Map t2map, ProjectIF project)
    {
        this.t2map = t2map;
        DatasetIF[] ds = analize(removeHit(project));
        if (ds != null && ds[0] != null)
        {
            for (int i = 0; i < ds.length; i++)
            {
                try
                {
                    project.addDataset(ds[i]);
                }
                catch (ProjectException e)
                {
                    log.fatal(e.getMessage(), e);
                }
            }
            getT2Map().updateProject(project);
        }
        else
        {
            if (t2map != null && getT2Map().getTree() != null)
            {
                t2map.setForcePaint();
                t2map.getTree().rebuildLayerList();
            }
        }
    }

    /**
     * T2Mapオブジェクトを取得します。
     * 
     * @return T2Mapオブジェクト
     */
    protected T2Map getT2Map()
    {
        return t2map;
    }

    /**
     * PluginManagerオブジェクトを取得します。
     * 
     * @return PluginManagerオブジェクト
     */
    protected PluginManager getPluins()
    {
        return T2Map.getPluginManager();
    }

    /**
     * 検索結果についての情報をリセットします。
     * 
     * @param project 検索結果情報を初期化したいProjectオブジェクト
     * @return 検索結果についての情報をリセットしたProjectオブジェクト
     */
    protected ProjectIF removeHit(ProjectIF project)
    {
        log.debug("Remove hit flags.");
        if (project != null)
        {
            TimeAxisIF data = t2map.getTimeAxisData();
            // BigDecimal sd = data.getMin();
            // BigDecimal ed = data.getMax();
            DatasetIF[] ds = project.getDatasets();
            for (int i = 0; i < ds.length; i++)
            {
                // Record[] recors = ds[i].getRecords(sd, ed);
                Record[] recors = ds[i].getHitRecords();
                for (int j = 0; j < recors.length; j++)
                {
                    // Event[] event = recors[j].getEvents();
                    // for (int k = 0; k < event.length; k++)
                    // {
                    recors[j].setHit(false);
                    // }
                }
            }
            ProjectIF[] projects = project.getProjects();
            if (projects != null)
            {
                for (int a = 0; a < projects.length; a++)
                {
                    removeHit(projects[a]);
                }
            }
        }
        return project;
    }

    /**
     * 解析処理をするダイアログを実装して下さい。
     * 
     * @param project 解析処理に使用するProjectオブジェクト
     */
    protected abstract DatasetIF[] analize(ProjectIF project);

}
