/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin;

import java.io.File;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import javax.swing.JPanel;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.ProjectIF;
import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.utility.ExtensionFileFilter;

/**
 * CSVファイルをインポートするクラスです。
 */
public abstract class ImportPlugin implements Plugin
{
    /**
     * T2Mapオブジェクト
     */
    private T2Map parent;

    /**
     * PluginManagerオブジェクト
     */
    private PluginManager plugins;

    private IOError ioError = null;

    /**
     * T2Map、PluginManagerを設定します。
     * 
     * @param parent T2Mapオブジェクト
     * @param plugins PluginManagerオブジェクト
     */
    public void init(T2Map parent, PluginManager plugins)
    {
        this.parent = parent;
        this.plugins = plugins;
        this.ioError = new IOError();
    }

    /**
     * PluginManagerオブジェクトを取得します。
     * 
     * @return PluginManagerオブジェクト
     */
    protected PluginManager getPlugins()
    {
        return plugins;
    }

    public void processError()
    {
        ioError.processError();
    }

    /**
     * T2Mapオブジェクトを取得します。
     * 
     * @return T2Mapオブジェクト
     */
    protected T2Map getParentComponent()
    {
        return parent;
    }

    /**
     * レコードを取得します。
     * 
     * @param path パス
     * @param format レコードのファイルフォーマット（XML, CSV...）
     * @param code 文字コード
     * @return レコード
     * @throws OpenPluginException
     */
    public Record[] getRecords(String path, String format, String code) throws OpenPluginException
    {
        Record[] records = null;
        RecordIOFilter filter = getRecordIOFilter(format);
        if (filter != null)
        {
            String rpath = "";
            try
            {
                rpath = filter.getPath(path);
                records = filter.read(getRecordsInputStream(rpath), code, ioError);
                groups = filter.getGroups();
            }
            catch (RecordIOFilterException e)
            {
                throw new OpenPluginException(e.getMessage() + "(" + rpath + ")", e);
            }
        }
        getParentComponent().changeInitialMinMax(records);
        return records;
    }
    
    private List<String> groups;
    
    public List<String> getGroups() {
        return groups;
    }

    /**
     * RecordIOFilterを取得します。
     * 
     * @param format フォーマット
     * @return RecordIOFilter
     * @throws OpenPluginException
     */
    private RecordIOFilter getRecordIOFilter(String format) throws OpenPluginException
    {
        RecordIOFilter f = null;
        ArrayList<Hashtable<String, String>> filters = plugins.getPlugins("RecordIOFilter");
        for (int i = 0; i < filters.size(); i++)
        {
            if (filters.get(i).get("name").equals(format))
            {
                try
                {
                    f = (RecordIOFilter) plugins.newInstance(filters.get(i).get("class"));
                }
                catch (PluginManagerException e)
                {
                    throw new OpenPluginException(e.getMessage(), e);
                }
            }
        }
        return f;
    }

    /**
     * ファイルをオープンします。
     * 
     * @return プロジェクト
     * @throws OpenPluginException
     */
    public abstract ProjectIF open() throws OpenPluginException;
    
    /**
     * XMLファイルを読み込みます。
     * 
     * @param file 選択したファイル
     * @return project Projectオブジェクト
     * @throws OpenPluginException
     */
    public abstract ProjectIF open(File file, ProjectIF insertToProject) throws OpenPluginException;
    
    /**
     * ファイルチューザー用の拡張子フィルターを取得します。
     * @return
     */
    public abstract ExtensionFileFilter getExtensionFileFilter();
    
    /**
     * ファイルチューザー用のアクセサリパネルを取得します。
     * @return
     */
    public abstract JPanel getJPanel();

    /**
     * ファイルを読み込みます。
     * 
     * @param path パス
     * @return InputStream
     * @throws OpenPluginException
     */
    public abstract InputStream getRecordsInputStream(String path) throws OpenPluginException;
}
