/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin;

import java.awt.Color;
import java.awt.Container;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.Rectangle;
import java.awt.Shape;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.SortedSet;
import java.util.Timer;
import java.util.TimerTask;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.ImageIcon;
import javax.swing.JMenuItem;
import javax.swing.JRootPane;
import javax.swing.SwingUtilities;
import javax.xml.xpath.XPathExpressionException;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.T2MapProperties;
import jp.nihu.gttime.TimeMapGeograhicCoordinates;
import jp.nihu.gttime.TimePanel;
import jp.nihu.gttime.data.DatasetGtm;
import jp.nihu.gttime.data.DatasetIF;
import jp.nihu.gttime.data.Event;
import jp.nihu.gttime.data.JulianDate;
import jp.nihu.gttime.data.Location;
import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.data.UTMPoint;
import jp.nihu.gttime.data.property.LineGts;
import jp.nihu.gttime.data.property.LineIF;
import jp.nihu.gttime.data.property.PolygonGts;
import jp.nihu.gttime.data.property.PolygonIF;
import jp.nihu.gttime.data.property.PropertiesIF;
import jp.nihu.gttime.data.property.SymbolGts;
import jp.nihu.gttime.data.property.SymbolIF;
import jp.nihu.gttime.layer.LayerPopupMenu;
import jp.nihu.gttime.layer.ResizeHeightDialog;
import jp.nihu.gttime.layer.ResizeLine;
import jp.nihu.gttime.layer.TimeDurationSelector;
import jp.nihu.gttime.plugin.layer.NumericLayerPlugin;
import jp.nihu.gttime.preferences.BaseLayerPluginPreferences;
import jp.nihu.gttime.time.TimeAxisIF;
import jp.nihu.gttime.time.utility.JulianGregorianDate;
import jp.nihu.gttime.tree.TreeCheckNode;
import jp.nihu.gttime.utility.Utilities;
import jp.nihu.gttime.utility.XMLUtil;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * レイヤについてのプラグインを管理するクラスである。<br/>レイヤ上で左クリックした際のメニュー表示、「歴表示」・「画像出力」・「クリップボードにコピー」・「計測」の表示管理をします。
 */
public abstract class LayerPlugin extends TimePanel implements Plugin, MouseMotionListener, MouseListener
{
    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(LayerPlugin.class);

    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(LayerPlugin.class, T2Map.LOCALE);

    /**
     * レコードに適用するグループ情報を格納する
     */
    protected Map<String, Object> groups = new HashMap<String, Object>();

    /**
     * Dataset用オブジェクト
     */
    protected DatasetIF dataset;

    public DatasetIF getDataset()
    {
        return dataset;
    }

    /**
     * ブラウザ起動パスを設定ファイルから取得します。
     */
    protected String browser = T2Map.properties.getProperty("jp.nihu.gttime.plugin.LayerPlugin.WEBBROWSER_PATH");

    /**
     * レイヤの大きさ変更時のフラグとして使用する。
     */
    private boolean resize = false;

    /**
     * レイヤ上で左クリックした際のメニュー表示、<br/>「歴表示」・「画像出力」・「クリップボードにコピー」・「計測」に使用します。
     */
    protected LayerPopupMenu popupMenu;

    /**
     * レイヤパネルの初期値となる高さを定義しています。
     */
    private static int MIN_HEIGHT = 70;// 200

    /**
     * レイヤの初期化処理を実装して下さい。
     */
    public abstract void initLayer();

    /**
     * レイヤにオブジェクト情報を描画する際に実装して下さい。
     * 
     * @param g
     */
    public abstract void paintLayer(Graphics g);

    protected abstract void setLayerProperties();

    public abstract ImageIcon getIcon();

    public abstract String getName();

    public abstract String getPluginName();

    public abstract String getStyleName();

    public abstract BaseLayerPluginPreferences getLayerPluginPreferencesDialog();

    /**
     * コンストラクター
     */
    public LayerPlugin()
    {
        super();
        addMouseListener(this);
        addMouseMotionListener(this);
    }

    // #############################################################################NEW
    private String getBasePathForStyleProperties()
    {
        return "/gt:gtm/hgis:hgisApplications/hgis:time[@xsi:type='" + getStyleName() + "']";
    }

    public PropertiesIF getProperties()
    {
        return dataset.getProperties();
    }

    public T2MapProperties getT2MapProperties()
    {
        return T2Map.properties;
    }

    protected void setPolygonProperties(String _group) throws XPathExpressionException
    {
        PropertiesIF properties = getProperties();
        T2MapProperties p = getT2MapProperties();
        String prefix = getBasePathForStyleProperties();
        NodeList nl = p.getStyleValues(prefix + "/hgis:polygon");
        if (nl != null) {
            for (int i = 0; i < nl.getLength(); i++) {
                Element elem = (Element)nl.item(i);
//                String group = elem.getAttribute("group");
//                if (group == null) {
//                    group = "";
//                }
                String polygonColor = XMLUtil.getValue(elem, "hgis:fillColor/text()", XMLUtil.getGtsNsContext());
                String polygonThin = XMLUtil.getValue(elem, "hgis:thin/text()", XMLUtil.getGtsNsContext());
                String polygonLineColor = XMLUtil.getValue(elem, "hgis:lineColor/text()", XMLUtil.getGtsNsContext());
                PolygonIF polygon = new PolygonGts();
                polygon.setFillColor(Utilities.getColor(polygonColor));
                polygon.setThin(Integer.parseInt(polygonThin.trim()));
                polygon.setLineColor(Utilities.getColor(polygonLineColor));
                properties.setPolygon(polygon, _group);
            }
        }
    }

    protected void setLineProperties(String _group) throws XPathExpressionException
    {
        PropertiesIF properties = getProperties();
        T2MapProperties p = getT2MapProperties();
        String prefix = getBasePathForStyleProperties();
        NodeList nl = p.getStyleValues(prefix + "/hgis:line");
        if (nl != null) {
            for (int i = 0; i < nl.getLength(); i++) {
                Element elem = (Element)nl.item(i);
//                String group = elem.getAttribute("group");
//                if (group == null) {
//                    group = "";
//                }
                String lineThin = XMLUtil.getValue(elem, "hgis:thin/text()", XMLUtil.getGtsNsContext());
                String lineColor = XMLUtil.getValue(elem, "hgis:lineColor/text()", XMLUtil.getGtsNsContext());
                LineIF line = new LineGts();
                line.setLineColor(Utilities.getColor(lineColor));
                line.setThin(Integer.parseInt(lineThin));
                properties.setLine(line, _group);
            }
        }
    }

    protected void setSymbolProperties(String _group) throws XPathExpressionException
    {
        PropertiesIF properties = getProperties();
        T2MapProperties p = getT2MapProperties();
        String prefix = getBasePathForStyleProperties();
        NodeList nl = p.getStyleValues(prefix + "/hgis:symbol");
        if (nl != null) {
            for (int i = 0; i < nl.getLength(); i++) {
                Element elem = (Element)nl.item(i);
//                String group = elem.getAttribute("group");
//                if (group == null) {
//                    group = "";
//                }
                String symbolType = XMLUtil.getValue(elem, "hgis:symbol/text()", XMLUtil.getGtsNsContext());
                String symbolSize = XMLUtil.getValue(elem, "hgis:size/text()", XMLUtil.getGtsNsContext());
                String symbolColor = XMLUtil.getValue(elem, "hgis:fillColor/text()", XMLUtil.getGtsNsContext());
                String symbolThin = XMLUtil.getValue(elem, "hgis:thin/text()", XMLUtil.getGtsNsContext());
                String symbolLineColor = XMLUtil.getValue(elem, "hgis:lineColor/text()", XMLUtil.getGtsNsContext());
                SymbolIF symbol = new SymbolGts();
                symbol.setSymbol(symbolType);
                symbol.setSize(Integer.parseInt(symbolSize.trim()));
                symbol.setFillColor(Utilities.getColor(symbolColor));
                symbol.setThin(Integer.parseInt(symbolThin.trim()));
                symbol.setLineColor(Utilities.getColor(symbolLineColor));
                properties.setSymbol(symbol, _group);
            }
        }
    }

    protected void setFontProperties(String _group) throws XPathExpressionException
    {
        PropertiesIF properties = getProperties();
        T2MapProperties p = getT2MapProperties();
        String prefix = getBasePathForStyleProperties();
        NodeList nl = p.getStyleValues(prefix + "/hgis:font");
        if (nl != null) {
            for (int i = 0; i < nl.getLength(); i++) {
                Element elem = (Element)nl.item(i);
//                String group = elem.getAttribute("group");
//                if (group == null) {
//                    group = "";
//                }
                String fontFamily = XMLUtil.getValue(elem, "hgis:font-family/text()", XMLUtil.getGtsNsContext());
                String fontName = XMLUtil.getValue(elem, "hgis:font-name/text()", XMLUtil.getGtsNsContext());
                String fontSize = XMLUtil.getValue(elem, "hgis:font-size/text()", XMLUtil.getGtsNsContext());
                String fontStyle = XMLUtil.getValue(elem, "hgis:font-styl/text()", XMLUtil.getGtsNsContext());
                Font font = new Font(fontFamily, Utilities.getFontStyle(fontStyle), Integer.parseInt(fontSize.trim()));
                properties.setFont(font, _group);
            }
        }
    }

    protected void setFontColorProperties(String _group) throws XPathExpressionException
    {
        PropertiesIF properties = getProperties();
        T2MapProperties p = getT2MapProperties();
        String prefix = getBasePathForStyleProperties();
        NodeList nl = p.getStyleValues(prefix + "/hgis:font");
        if (nl != null) {
            for (int i = 0; i < nl.getLength(); i++) {
                Element elem = (Element)nl.item(i);
//                String group = elem.getAttribute("group");
//                if (group == null) {
//                    group = "";
//                }
                String fontColor = XMLUtil.getValue(elem, "hgis:font-color/text()", XMLUtil.getGtsNsContext());
                properties.setFontColor(Utilities.getColor(fontColor), _group);
            }
        }
    }

    protected void setSelectedColorProperties() throws XPathExpressionException
    {
        PropertiesIF properties = getProperties();
        T2MapProperties p = getT2MapProperties();
        String prefix = getBasePathForStyleProperties();
        String selectedColor = p.getStyleValue(prefix + "/hgis:selectedColor/text()");
        properties.setSelectedColor(Utilities.getColor(selectedColor));
    }

    protected void setBackgroundColorProperties() throws XPathExpressionException
    {
        PropertiesIF properties = getProperties();
        T2MapProperties p = getT2MapProperties();
        String prefix = getBasePathForStyleProperties();
        String backgroundColor = p.getStyleValue(prefix + "/hgis:backgroundColor/text()");
        properties.setBackgroundColor(Utilities.getColor(backgroundColor));
    }

    protected void setLayerHeightProperties() throws XPathExpressionException
    {
        PropertiesIF properties = getProperties();
        T2MapProperties p = getT2MapProperties();
        String prefix = getBasePathForStyleProperties();
        String layerHight = p.getStyleValue(prefix + "/hgis:layerHight/text()");
        if (layerHight != null && !layerHight.trim().equals("")) {
        	properties.setLayerHeight(Integer.parseInt(layerHight.trim()));
        }
    }

    protected void setDataReductionProperties() throws XPathExpressionException
    {
        PropertiesIF properties = getProperties();
        T2MapProperties p = getT2MapProperties();
        String prefix = getBasePathForStyleProperties();
        String dataReduction = p.getStyleValue(prefix + "/hgis:dataReduction/text()");
        properties.setDataReduction(Boolean.parseBoolean(dataReduction));
    }

    protected void setDataReductionNumProperties() throws XPathExpressionException
    {
        PropertiesIF properties = getProperties();
        T2MapProperties p = getT2MapProperties();
        String prefix = getBasePathForStyleProperties();
        String dataReductionEnableRecordNum = p.getStyleValue(prefix + "/hgis:dataReductionEnableRecordNum/text()");
        properties.setDataReductionNum(Integer.parseInt(dataReductionEnableRecordNum.trim()));
    }

    protected void setDetailFontNameProperties() throws XPathExpressionException
    {
        PropertiesIF properties = getProperties();
        T2MapProperties p = getT2MapProperties();
        String prefix = getBasePathForStyleProperties();
        String detailFontName = p.getStyleValue(prefix + "/hgis:detailFontName/text()");
        properties.setDetailFontName(detailFontName.trim());
    }
    
    protected void setShowGroupNameProperties() throws XPathExpressionException
    {
        PropertiesIF properties = getProperties();
        T2MapProperties p = getT2MapProperties();
        String prefix = getBasePathForStyleProperties();
        String showGroupName = p.getStyleValue(prefix + "/hgis:showGroupName/text()");
        if (showGroupName != null && showGroupName.length() > 0) {
            properties.setShowGroupName(Boolean.parseBoolean(showGroupName.trim()));
        } else {
            properties.setShowGroupName(false);
        }
    }
    
    protected void setScaleMainProperties() throws XPathExpressionException
    {
        PropertiesIF properties = getProperties();
        T2MapProperties p = getT2MapProperties();
        String prefix = getBasePathForStyleProperties();
        String scaleMainColor = p.getStyleValue(prefix + "/hgis:scaleMain/hgis:fillColor/text()");
        String scaleMainThin = p.getStyleValue(prefix + "/hgis:scaleMain/hgis:thin/text()");
        String scaleMainLineColor = p.getStyleValue(prefix + "/hgis:scaleMain/hgis:lineColor/text()");
        PolygonIF scaleMain = new PolygonGts();
        scaleMain.setFillColor(Utilities.getColor(scaleMainColor));
        scaleMain.setThin(Integer.parseInt(scaleMainThin.trim()));
        scaleMain.setLineColor(Utilities.getColor(scaleMainLineColor));
        properties.setScaleMain(scaleMain);
    }

    protected void setScaleSubProperties() throws XPathExpressionException
    {
        PropertiesIF properties = getProperties();
        T2MapProperties p = getT2MapProperties();
        String prefix = getBasePathForStyleProperties();
        String scaleSubColor = p.getStyleValue(prefix + "/hgis:scaleSub/hgis:fillColor/text()");
        String scaleSubThin = p.getStyleValue(prefix + "/hgis:scaleSub/hgis:thin/text()");
        String scaleSubLineColor = p.getStyleValue(prefix + "/hgis:scaleSub/hgis:lineColor/text()");
        PolygonIF scaleSub = new PolygonGts();
        scaleSub.setFillColor(Utilities.getColor(scaleSubColor));
        scaleSub.setThin(Integer.parseInt(scaleSubThin.trim()));
        scaleSub.setLineColor(Utilities.getColor(scaleSubLineColor));
        properties.setScaleSub(scaleSub);
    }

    // #############################################################################NEW

    private int previouseWidth = 0;

    private int previouseHeight = 0;

    protected ArrayList<RecordPlot> shapes = new ArrayList<RecordPlot>();

    private ArrayList<RecordPlot> mouseOveredRecords = new ArrayList<RecordPlot>();

    protected Image highLightImage = null;

    // protected Color highlightColor = Color.BLUE;

    public ArrayList<RecordPlot> getMouseOveredRecords()
    {
        return mouseOveredRecords;
    }

    public Image getHighLightImage()
    {
        return highLightImage;
    }

    public Color getHighLightColor()
    {
        PropertiesIF properties = dataset.getProperties();
        if (properties != null)
        {
            Color hColor = properties.getSelectedColor();
            if (hColor != null)
            {
                return hColor;
            }
        }
        return Color.BLUE;
    }

    private boolean isWidthChanged()
    {
        if (previouseWidth != getWidth())
        {
            return true;
        }
        return false;
    }

    private boolean isHeightChanged()
    {
        if (previouseHeight != getHeight())
        {
            return true;
        }
        return false;
    }

    private void resizeHeight(int m)
    {
        // setPreferredSize(new Dimension(getWidth(), m));//has problem
        setPreferredSize(new Dimension(0, m));
        setMaximumSize(new Dimension(20000000, m));
        // setMinimumSize(new Dimension(0, m));//doesn't work
        dataset.getProperties().setLayerHeight(m);
        dataset.applyStyleChanges(getPluginName(), getStyleName());
        // node.setHeight(m);
    }

    /**
     * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
     */
    public void mouseDragged(MouseEvent e)
    {
        if (resize && MIN_HEIGHT < e.getY() + 1)
        {
            resizeHeight(e.getY() + 1);
        }
        else if (resize && MIN_HEIGHT >= e.getY() + 1)
        {
            resizeHeight(MIN_HEIGHT);
        }
    }

    protected void setMinHeight(int height)
    {
        MIN_HEIGHT = height;
    }

    protected int getMinHeight()
    {
        return MIN_HEIGHT;
    }

    /**
     * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
     */
    public void mouseMoved(MouseEvent e)
    {
        JRootPane root = SwingUtilities.getRootPane(this);
        mouseOveredRecords.clear();
        int x = e.getX();
        int y = e.getY();
        boolean hasMouseOver = false;
        for (int i = 0; i < shapes.size(); i++)
        {
            RecordPlot shape = shapes.get(i);
            if (shape.contains(x, y))
            {
                // log.debug("MouseOver : " + x + " : " + y);
                hasMouseOver = true;
                mouseOveredRecords.add(shape);
            }
        }

        // プロット上のカーソル
        if (hasMouseOver)
        {
            root.setCursor(new Cursor(Cursor.HAND_CURSOR));
        }
        else
        {
            root.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
        }

        // レイヤの高さ調節時のカーソル
        if (getHeight() - 10 <= e.getY() && e.getY() <= getHeight() + 10)
        {
            getTimeDurationSelector().setEnabled(false);
            getResizeLine().setEnabled(true);
            getResizeLine().setMinimumHeight(getMinHeight());
            if (isResizable())
            {
                resize = true;
                if (root.getCursor().getType() == Cursor.DEFAULT_CURSOR)
                {
                    root.setCursor(new Cursor(Cursor.N_RESIZE_CURSOR));
                }
            }
        }
        else
        {
            getTimeDurationSelector().setEnabled(true);
            getResizeLine().setEnabled(false);
            resize = false;
            if (root.getCursor().getType() == Cursor.N_RESIZE_CURSOR)
            {
                root.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
            }
        }
    }

    private TimeDurationSelector timeDurationSelector = null;

    public TimeDurationSelector getTimeDurationSelector()
    {
        if (timeDurationSelector == null)
        {
            MouseMotionListener[] listeners = getMouseMotionListeners();
            for (int i = 0; i < listeners.length; i++)
            {
                if (listeners[i] instanceof TimeDurationSelector)
                {
                    timeDurationSelector = (TimeDurationSelector) listeners[i];
                    break;
                }
            }
        }
        return timeDurationSelector;
    }

    private ResizeLine resizeLine = null;

    private ResizeLine getResizeLine()
    {
        if (resizeLine == null)
        {
            MouseMotionListener[] listeners = getMouseMotionListeners();
            for (int i = 0; i < listeners.length; i++)
            {
                if (listeners[i] instanceof ResizeLine)
                {
                    resizeLine = (ResizeLine) listeners[i];
                    break;
                }
            }
        }
        return resizeLine;
    }

    public int popupPosX = 0;

    /**
     * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
     */
    public void mouseClicked(MouseEvent e)
    {
        if (SwingUtilities.isRightMouseButton(e))
        {
            showPopupMenu(e);
        }
        log.debug("mouseClicked x=" + e.getX() + " y=" + e.getY());
    }

    public void showPopupMenu(MouseEvent e)
    {
    	popupMenu.fromTree = false;
        popupPosX = e.getX();
        popupMenu.setVisibleMeasureMenuItem(getTimeDurationSelector().isOnSelected(pixelToJulian(popupPosX, data.getCurrentMin(), data.getCurrentMax())));
        popupMenu.show(LayerPlugin.this, e.getX(), e.getY());
    }

    /**
     * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
     */
    public void mouseEntered(MouseEvent e)
    {
    }

    /**
     * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
     */
    public void mouseExited(MouseEvent e)
    {
        mouseOveredRecords.clear();
        JRootPane root = SwingUtilities.getRootPane(this);
        root.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
    }

    /**
     * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
     */
    public void mousePressed(MouseEvent e)
    {
        int x = e.getX();
        int y = e.getY();
        log.debug("MouseClicked : " + x + " : " + y);
    }

    /**
     * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
     */
    public void mouseReleased(MouseEvent e)
    {
        JRootPane root = SwingUtilities.getRootPane(this);
        if (root.getCursor().getType() == Cursor.N_RESIZE_CURSOR)
        {
            root.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
        }
        if (resize)
        {
            resize = false;
            updateUI();
        }
    }

    // #############################################################################NEW

    // ##############################################################################################
    private Timer timer = new Timer();

    private TimerTask timerTask = null;

    private boolean delayPaint = true;

    protected boolean forcePaint = false;

    public void setForcePaint()
    {
        forcePaint = true;
    }

    /**
     * @see jp.nihu.gttime.TimePanel#valueChanged()
     */
    public void valueChanged()
    {
        forcePaint = true;
        // スクロール画面上の表示されている領域にある場合に実行
        if (getVisibleRect().getWidth() > 0)
        {
            TimeAxisIF data = getData();
            if (prevStart == null) {
//            	log.debug("3 --> " + prevStart);
            	prevStart = data.getCurrentMin();
            }
            if (prevEnd == null) {
//            	log.debug("4 --> " + prevStart);
            	prevEnd = data.getCurrentMax();
            }
            int s = data.getCurrentMin().compareTo(prevStart);
            int e = data.getCurrentMax().compareTo(prevEnd);
            // 遅延描画を実行
            if (delayPaint)
            {
                if (timerTask != null)
                {
                    timerTask.cancel();
                    timer.purge();
                    timerTask = null;
                }
                timerTask = new TimerTask()
                {
                    public void run()
                    {
                        paintMode = PAINT_MODE_REPAINT;
                        valueChangedWorker();
                        cancel();
                    }
                };
                timer.schedule(timerTask, 100);
                if (s != 0 && e != 0)
                {
                    paintMode = PAINT_MODE_MOVE;
                    valueChangedWorker();
                }
                else if (s != 0 || e != 0)
                {
                    paintMode = PAINT_MODE_RESIZE;
                    valueChangedWorker();
                }
            }
            else
            {
                if (s != 0 || e != 0)
                {
                    valueChangedWorker();
                }
            }
        }
    }

    private void valueChangedWorker()
    {
        super.valueChanged();
    }

    // ##############################################################################################

    /**
     * TreeCheckNodeを取得します。
     * 
     * @return TreeCheckNode
     */
    public TreeCheckNode getTreeCheckNode()
    {
        return node;
    }

    /**
     * TreeCheckNodeオブジェクト（ツリーのチェックボックス内の選択情報などを取得する）
     */
    protected TreeCheckNode node;

    /**
     * レイヤ内で利用されるデータ（TimeAxisIF、TreeCheckNode）を登録する。<br/>画面上のマウス操作によって起動される機能を初期化します。
     * 
     * @param data TimeAxisIFオブジェクト
     * @param node TreeCheckNodeオブジェクト
     */
    public void init(TimeAxisIF data, final TreeCheckNode node)
    {
        super.init(data);
        this.dataset = (DatasetIF) node.getUserObject();
        this.setLayerProperties();
        this.node = node;

        popupMenu = new LayerPopupMenu(this);
        
        node.getTreeView().addRenameMenuItem(popupMenu, node);
        
//        popupMenu.add(new JMenuItem("【将来拡張】切り取り"));
//        popupMenu.add(new JMenuItem("【将来拡張】コピー"));
        
        node.getTreeView().addDeleteNodeMenuItem(popupMenu, node);
        
        popupMenu.addSeparator();
        
        addReduceSwitchMenu();
        
//        popupMenu.addDefaultMenus();
        
        
        if (canDrawTimeAxis()) {
        	popupMenu.createDrawTimeAxisMenu();
        }
        
//        JMenu em = new JMenu("【将来拡張】レコード編集");
//        em.add(new JMenuItem("【将来拡張】暦変換"));
//        popupMenu.add(em);
        popupMenu.createZoomLayerMenu();
        
//        popupMenu.createExportMenu();
        popupMenu.createMeasureMenu();
        
        popupMenu.addSeparator();
        
        popupMenu.createSaveMenu();
        
        popupMenu.createSaveExportMenu();
        
        popupMenu.createExportMenu();
        
        popupMenu.addSeparator();
        
        addResizeHeightMenu();
        
        popupMenu.createPropertiesMenu();
        
        if (dataset instanceof DatasetGtm) {
	        popupMenu.addSeparator();
	        popupMenu.createShowMetadaMenu();
        }
        
        setVisible(node.isSelected());
        
        Integer height = dataset.getProperties().getLayerHeight();
        if (height == null || 0 >= height)
        {
        	height = getMinHeight();
        }
        else if (getMinHeight() > height)
        {
            height = getMinHeight();
        }
//        setPreferredSize(new Dimension(getWidth(), height));// dataset.getProperties().getLayerHeight()
        setMinimumSize(new Dimension(getWidth(), height));// dataset.getProperties().getLayerHeight()
        if (dataset.getProperties().getLayerHeight() != null) {//目盛り（線、幅）レイヤ以外のレイヤでnullの場合は高さを画面いっぱいで表示するので値をセットしない。
        	this.setMaximumSize(new Dimension(20000000, height));
        }
        this.setPreferredSize(new Dimension(getWidth(), height));
        
        initLayer();
    }

    public void initForPrint(TimeAxisIF data, final TreeCheckNode node)
    {
        super.init(data);
        this.dataset = (DatasetIF) node.getUserObject();
        this.setLayerProperties();
        this.node = node;
        Integer height = dataset.getProperties().getLayerHeight();
        if (height == null || 0 >= height)
        {
            height = 200;
            setMinHeight(70);
        }
        else if (getMinHeight() > height)
        {
            height = getMinHeight();
        }
        setPreferredSize(new Dimension(getWidth(), height));// dataset.getProperties().getLayerHeight()
        setMinimumSize(new Dimension(getWidth(), height));// dataset.getProperties().getLayerHeight()
        // setLayerPluginProperties();

        initLayer();
    }

    protected void addReduceSwitchMenu()
    {
        final JMenuItem item = new JMenuItem();
        item.setText(rmgr.getString("reduce.record"));
        if (dataset.getProperties().isDataReduction())
        {
            // item.setText(rmgr.getString("reduce.record"));
            item.setIcon(popupMenu.getCheckIcon());
        }
        else
        {
            // item.setText(rmgr.getString("reduce.record"));
            item.setIcon(null);
        }
        item.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                PropertiesIF properties = dataset.getProperties();
                if (properties.isDataReduction())
                {
                    // item.setText(rmgr.getString("reduce.record"));
                    item.setIcon(null);
                }
                else
                {
                    // item.setText(rmgr.getString("reduce.record"));
                    item.setIcon(popupMenu.getCheckIcon());
                }
                dataset.getProperties().setDataReduction(!properties.isDataReduction());
                dataset.getProperties().setChanged(true);
                setForcePaint();
                repaint();
            }
        });
        popupMenu.add(item);
    }

    protected void addResizeHeightMenu()
    {
        if (isResizable())
        {
            final JMenuItem item = new JMenuItem();
            item.setText(rmgr.getString("height.change"));
            item.addActionListener(new ActionListener()
            {
                public void actionPerformed(ActionEvent e)
                {
                    ResizeHeightDialog dialog = new ResizeHeightDialog(getT2Map(), true, getHeight(), getMinHeight());
                    dialog.setVisible(true);
                    int h = dialog.getChangedHeight();
                    if (h >= getMinHeight() && h != getHeight())
                    {
                        resizeHeight(h);
                        setForcePaint();
                        updateUI();
                    }
                    dialog.dispose();
                    dialog = null;
                }
            });
            popupMenu.add(item);
        }
    }

    /**
     * レイヤの高さが変更可能か不可能か
     * 
     * @return 可能
     */
    protected boolean isResizable()
    {
        return true;
    }
    
    /**
     * 時間軸を表示可能か不可能か
     * 
     * @return 可能
     */
    protected boolean canDrawTimeAxis()
    {
        return true;
    }

    /**
     * ポップアップメニューオブジェクトを取得します。
     * 
     * @return メニューオブジェクト
     */
    public LayerPopupMenu getMenu()
    {
    	if (popupMenu != null) {
        	boolean showTimeMeasureMenu = false;
        	if (getTimeDurationSelector() != null) {
            	BigDecimal[][] duls = getTimeDurationSelector().getDurations();
            	if (duls != null && duls.length > 0)
                {
            		showTimeMeasureMenu = true;
                }
        	}
            popupMenu.setVisibleMeasureMenuItem(showTimeMeasureMenu);
    	}
    	return popupMenu;
    }

    /**
     * 高さの余白を設定します。<br/> この余白情報を利用して「データセットのタイトル文字」を加えます。
     */
    private int HEIGHT_INSETS = 30;

    /**
     * 高さの余白を取得します。
     * 
     * @return 高さの余白
     */
    protected int getHeightInsets()
    {
        return HEIGHT_INSETS;
    }

    private Image offscreen = null;

    public int paintMode = PAINT_MODE_REPAINT;

    private static final int PAINT_MODE_REPAINT = 0;

    private static final int PAINT_MODE_MOVE = 1;

    private static final int PAINT_MODE_RESIZE = 2;

    private static final int PAINT_MODE_AS_NOT_IMAGE = 3;

    public void setPaintAsNotImage()
    {
        paintMode = PAINT_MODE_AS_NOT_IMAGE;
    }

    private BigDecimal prevStart = null;

    private BigDecimal prevEnd = null;

    /**
     * @see java.awt.Component#createImage(int, int)
     */
    @Override
    public Image createImage(int width, int height)
    {
        return new BufferedImage(width, height, BufferedImage.TYPE_INT_ARGB);
    }

    public void drawBackground(Graphics g)
    {
        int w = getWidth();
        int h = getHeight();

        // g.setColor(Color.WHITE);
        g.setColor(dataset.getProperties().getBackgroundColor());
        g.fillRect(0, 0, w, h);
        String fontName = null;
//        if (dataset.getProperties().getFont("") != null) {
//            fontName = dataset.getProperties().getFont("").getFontName();
//        }
        g.setFont(new Font(fontName, Font.PLAIN, 12));
        g.setColor(Color.BLACK);

        if (node.isDrawTimeAxis())
        {
            drawAxis(data.getCurrentMin(), data.getCurrentMax(), g);
        }

        String title = dataset.getTitle();
        if (reduce)
        {
            title = title + rmgr.getString("record.deducted.message");
        }
        Dimension strSize = stringSize(title, g);
        g.drawString(title, 0 + 5, strSize.height + 1);
        g.setColor(Color.black);
        g.drawLine(1, h - 1, w - 1, h - 1);
    }

    private void paintPlot()
    {
        int w = getWidth();
        int h = getHeight();
        prevStart = data.getCurrentMin();
        prevEnd = data.getCurrentMax();
        if (offscreen != null)
        {
            offscreen.flush();
        }
        offscreen = createImage(w, h);
        Graphics offg = offscreen.getGraphics();
        paintLayer(offg);
    }

    /**
     * レイヤ上に描画します
     * 
     * @param g
     */
    public void paintComponent(Graphics g)
    {
        super.paintComponent(g);
        switch (paintMode)
        {
            case PAINT_MODE_REPAINT:
                PropertiesIF properties = dataset.getProperties();

                boolean propChanged = properties.isChanged();

                if (offscreen == null || isWidthChanged() || isHeightChanged() || propChanged)
                {
                    forcePaint = true;
                }

                if (propChanged)
                {
                    int h = properties.getLayerHeight();
                    if (h >= getMinHeight() && h != getHeight())
                    {
                        setPreferredSize(new Dimension(0, h));
                        setMaximumSize(new Dimension(20000000, h));
                        updateUI();
                    }
                }

                if (forcePaint)
                {
                    shapes.clear();
                    paintPlot();
                    previouseWidth = getWidth();
                    previouseHeight = getHeight();
                    drawBackground(g);
                    g.drawImage(offscreen, 0, 0, this);
                }
                else
                {
                    drawBackground(g);
                    g.drawImage(offscreen, 0, 0, this);
                }
                forcePaint = false;
                properties.setChanged(false);
                paintMode = PAINT_MODE_REPAINT;
                break;
            case PAINT_MODE_MOVE:
                int x1 = julianToPixel(prevStart, data.getCurrentMin(), data.getCurrentMax()) - LINE_INSETS;
                drawBackground(g);
                g.drawImage(offscreen, x1, 0, this);
                break;
            case PAINT_MODE_RESIZE:
                int x11 = julianToPixel(prevStart, data.getCurrentMin(), data.getCurrentMax()) - LINE_INSETS;
                int x12 = julianToPixel(prevEnd, data.getCurrentMin(), data.getCurrentMax()) + LINE_INSETS;
                Image n = offscreen.getScaledInstance(x12 - x11, getHeight(), Image.SCALE_FAST);
                int a = (LINE_INSETS * (x12 - x11 - getWidth())) / getWidth();
                if (x11 == 0)
                {
                    x11 = x11 - a;
                }
                else
                {
                    x11 = x11 + a;
                }
                drawBackground(g);
                g.drawImage(n, x11, 0, null);
                break;
            case PAINT_MODE_AS_NOT_IMAGE:
                drawBackground(g);
                paintLayer(g);
                paintMode = PAINT_MODE_REPAINT;
                break;
            default:
                drawBackground(g);
                break;
        }
    }

    public class RecordComparatorStart implements Comparator<RecordPlot>
    {
        public int compare(RecordPlot o1, RecordPlot o2)
        {
            //log.debug("--------------------------------------AAAAAAAAAAAAAAAAA------------------------------------");
            if (o1.jdStart != null && o2.jdStart != null)
            {
                BigDecimal v1 = o1.jdStart.getDatetime();
                BigDecimal v2 = o2.jdStart.getDatetime();
                if (o1.title != null && o1.title.equals(COMP_START))
                {
                    int r = v1.compareTo(v2);
                    if (r == 1)
                    {
                        if (o2.jdEnd != null)
                        {
                            v2 = o2.jdEnd.getDatetime();
                            int f = v1.compareTo(v2);
                            if (f == 0)
                            {
                                f = 1;
                            }
                            return f;
                        }
                    }
                    return -1;
                }
                else if (o1.title != null && o1.title.equals(COMP_END))
                {
                    int f = v1.compareTo(v2);
                    if (f == 0)
                    {
                        f = 1;
                    }
                    return f;
                }
                else if (o2.title != null && o2.title.equals(COMP_START))
                {
                    int r = v1.compareTo(v2);
                    if (r == -1)
                    {
                        if (o1.jdEnd != null)
                        {
                            v1 = o1.jdEnd.getDatetime();
                            int f = v1.compareTo(v2);
                            if (f == 0)
                            {
                                f = 1;
                            }
                            return f;
                        }
                    }
                    return 1;
                }
                else if (o2.title != null && o2.title.equals(COMP_END))
                {
                    int f = v1.compareTo(v2);
                    if (f == 0)
                    {
                        f = 1;
                    }
                    return f;
                }
                int f = v1.compareTo(v2);
                if (f == 0)
                {
                    f = 1;
                }
                return f;
            }
            else
            {
                // never call.
                return -1;
            }
        }
    }

    public class RecordComparatorNullStart implements Comparator<RecordPlot>
    {
        public int compare(RecordPlot o1, RecordPlot o2)
        {
            BigDecimal v1 = o1.jdEnd.getDatetime();
            BigDecimal v2 = o2.jdEnd.getDatetime();
            int r = v1.compareTo(v2);
            if (o1.title != null && o1.title.equals(COMP_START) && o2.title != null && o2.title.equals(COMP_END))
            {
                if (r == 0)
                {
                    r = 1;
                }
                return r;
                //return v1.compareTo(v2);
            }
            else if (o1.title != null && o1.title.equals(COMP_END))
            {
                return 1;
            }
            else if (o2.title != null && o2.title.equals(COMP_END))
            {
                return -1;
            }
            // log.debug(o1.title + " : " + o2.title + " : " + r);
            if (r == 0)
            {
                r = 1;
            }
            return r;
        }
    }

    public class RecordComparatorNullEnd implements Comparator<RecordPlot>
    {
        public int compare(RecordPlot o1, RecordPlot o2)
        {
            BigDecimal v1 = o1.jdStart.getDatetime();
            BigDecimal v2 = o2.jdStart.getDatetime();
            int r = v1.compareTo(v2);
            if (o1.title != null && o1.title.equals(COMP_START) && o2.title != null && o2.title.equals(COMP_END))
            {
                if (r == 0)
                {
                    r = 1;
                }
                return r;
                //return v1.compareTo(v2);
            }
            else if (o1.title != null && o1.title.equals(COMP_START))
            {
                return -1;
            }
            else if (o2.title != null && o2.title.equals(COMP_START))
            {
                return 1;
            }
            // log.debug(o1.title + " : " + o2.title + " : " + r);
            if (r == 0)
            {
                r = 1;
            }
            return r;
        }
    }

    /**
     * プロットを新規作成します。
     * 
     * @param record Recordオブジェクト
     * @param data TimeAxisIFオブジェクト
     * @return RecordPlot
     */
    protected abstract RecordPlot createRecordPlot(Record record, TimeAxisIF data);

    protected TreeSet<RecordPlot> treeSetStart = new TreeSet<RecordPlot>(new RecordComparatorStart());

    protected TreeSet<RecordPlot> treeSetNullStart = new TreeSet<RecordPlot>(new RecordComparatorNullStart());

    protected TreeSet<RecordPlot> treeSetNullEnd = new TreeSet<RecordPlot>(new RecordComparatorNullEnd());

    protected ArrayList<RecordPlot> treeSetNullBoth = new ArrayList<RecordPlot>();

    private static final String COMP_START = "#######FOR COMP. START#######";

    private static final String COMP_END = "#######FOR COMP. END#######";

    protected void addRecordPlot(RecordPlot rp)
    {
        JulianDate s = rp.jdStart;
        JulianDate e = rp.jdEnd;

        if (s == null && e == null)
        {
            treeSetNullBoth.add(rp);
        }
        else if (s == null && e != null)
        {
            treeSetNullStart.add(rp);
        }
        else if (s != null && e == null)
        {
            treeSetNullEnd.add(rp);
        }
        else
        {
            if (treeSetStart.add(rp))
            {
                //log.debug("--------------------------------------XXXXXXXXXX------------------------------------");
            }
            else
            {
                //log.debug("--------------------------------------YYYYYYYYYY------------------------------------");
            }
        }
    }

    protected boolean isPaintResolution(Record record)
    {
        BigDecimal cdul = data.getCurrentMax().subtract(data.getCurrentMin());
        if (isPaintResolutionMax(cdul, record) && isPaintResolutionMin(cdul, record))
        {
            return true;
        }
        return false;
    }

    protected boolean isPaintResolutionMax(BigDecimal cdul, Record record)
    {
        BigDecimal maxres = record.getMaxVisibleTResolution();
        if (maxres != null && cdul.compareTo(maxres) >= 0)
        {
            return false;
        }
        else
        {
            return true;
        }
    }

    protected boolean isPaintResolutionMin(BigDecimal cdul, Record record)
    {
        BigDecimal minres = record.getMinVisibleTResolution();
        if (minres != null && cdul.compareTo(minres) <= 0)
        {
            return false;
        }
        else
        {
            return true;
        }
    }
    
    private SortedSet<RecordPlot> getSubset(TreeSet<RecordPlot> set, RecordPlot a, RecordPlot b) {
        SortedSet<RecordPlot> res  =new TreeSet<RecordPlot>(new RecordComparatorStart());
        Iterator<RecordPlot> it = set.iterator();

        int g = getWidth() - 25 * 2;
        BigDecimal gap = ((data.getCurrentMax().subtract(data.getCurrentMin())).divide(BigDecimal.valueOf(g), 10, BigDecimal.ROUND_HALF_EVEN)).multiply(BigDecimal.valueOf(25));
        BigDecimal s = a.jdEnd.getDatetime().subtract(gap);
        BigDecimal e = b.jdStart.getDatetime().add(gap);
        
        while (it.hasNext())
        {
            RecordPlot rp = it.next();
            BigDecimal v1 = rp.jdStart.getDatetime();
            BigDecimal v2 = rp.jdEnd.getDatetime();
            if (s.compareTo(v2) <= 0 && v1.compareTo(e) <= 0)
            {
                res.add(rp);
            }
        }
        return res;
    }

    protected Iterator<RecordPlot> getProcRecordPlot()
    {
        try
        {
            Record ar = new Record();
            JulianDate ajs = new JulianDate();
            ajs.setDatetime(data.getCurrentMin().subtract(new BigDecimal("1")));
            ar.setStartDatetime(ajs);
            JulianDate aje = new JulianDate();
            aje.setDatetime(data.getCurrentMin());
            ar.setEndDatetime(aje);
            RecordPlot a = createRecordPlot(ar, data);
            a.title = COMP_START;

            Record br = new Record();
            JulianDate bjs = new JulianDate();
            bjs.setDatetime(data.getCurrentMax());
            br.setStartDatetime(bjs);
            JulianDate bje = new JulianDate();
            bje.setDatetime(data.getCurrentMax().add(new BigDecimal("1")));
            br.setEndDatetime(bje);
            RecordPlot b = createRecordPlot(br, data);
            b.title = COMP_END;

            Collection<RecordPlot> procPlots = new ArrayList<RecordPlot>();

            // if (this instanceof NumericLayerPlugin || this instanceof
            // TimeChartLayerPlugin)
            {
                //log.debug("--------------------------------------BBBBBBBBBBBBBBBBBBBBBBBB------------------------------------");
                //Iterator<RecordPlot> it0 = treeSetStart.subSet(a, b).iterator();
//                Iterator<RecordPlot> it0 = treeSetStart.iterator();
                Iterator<RecordPlot> it0 = getSubset(treeSetStart, a, b).iterator();
                while (it0.hasNext())
                {
                    procPlots.add(it0.next());
                }
                //log.debug("--------------------------------------CCCCCCCCCCCCCCCCCCCCCCCC------------------------------------");
                Iterator<RecordPlot> it1 = treeSetNullStart.subSet(a, b).iterator();
                while (it1.hasNext())
                {
                    procPlots.add(it1.next());
                }
                Iterator<RecordPlot> it2 = treeSetNullEnd.subSet(a, b).iterator();
                while (it2.hasNext())
                {
                    procPlots.add(it2.next());
                }
                Iterator<RecordPlot> it3 = treeSetNullBoth.iterator();
                while (it3.hasNext())
                {
                    procPlots.add(it3.next());
                }
            }
            reduce = doReduce(procPlots.size());
//            log.debug(reduce + " :: " + procPlots.size() + " :: " + dataset.getRecords().length);
            return procPlots.iterator();
        }
        catch (Exception e)
        {
            e.printStackTrace();
            return null;
        }
    }

    protected boolean reduce = true;

    protected abstract boolean doReduce(int ps);

    private static final Color DEFAULT_PLOT_HILIGHT_COLOR = Color.RED;

    public abstract class RecordPlot
    {
        /**
         * TimeAxisIFオブジェクト
         */
        protected TimeAxisIF data = null;

        /**
         * Recordオブジェクト
         */
        public Record record = null;

        /**
         * タイトル用
         */
        public String title = null;

        /**
         * ロケーション用
         */
        protected Location location = null;

        /**
         * 開始日
         */
        public JulianDate jdStart = null;

        /**
         * 終了日
         */
        public JulianDate jdEnd = null;

        /**
         * ハイライトフラグ
         */
        private boolean clicked = false;
        
        /**
         * グループ
         */
        private String group = null;

        public Color highlightColorx = DEFAULT_PLOT_HILIGHT_COLOR;

        /**
         * クリックフラグを設定します。
         * 
         * @param hit クリックされた、されていない
         */
        public void setClicked(boolean clicked)
        {
            if (clicked)
            {
                highlightColorx = getHighlightColor(100);
            }
            else
            {
                highlightColorx = DEFAULT_PLOT_HILIGHT_COLOR;
            }
            this.clicked = clicked;
        }

        /**
         * 検索フラグを取得します。
         * 
         * @return 検索フラグ
         */
        public boolean isClicked()
        {
            return clicked;
        }

        /**
         * プロット描画に関するコンストラクタ
         * 
         * @param record Recordオブジェクト
         * @param data TimeAxisIFオブジェクト
         */
        public RecordPlot(Record record, TimeAxisIF data)
        {
            this.record = record;
            this.data = data;
            this.title = getTitle();
            this.location = record.getLocation();
            this.group = record.getGroup();

            jdStart = record.getStartDatetime();
            jdEnd = record.getEndDatetime();
        }
        
        public String getGroup() {
            if (group == null) {
                group = "";
            }
            return group;
        }

        /**
         * タイトル情報の取得
         * 
         * @return タイトル
         */
        private String getTitle()
        {
            if (title == null)
            {
                Event[] events = record.getEvents();
                int el = events.length;
                for (int j = 0; j < el; j++)
                {
                    if (events[j].getName().equals("title"))
                    {
                        title = events[j].getData();
                        break;
                    }
                }
            }
            return title;
        }

        /**
         * イベント取得
         * 
         * @return イベント
         */
        protected Event[] getEvents()
        {
            return record.getEvents();
        }

        /**
         * UTM情報の取得
         * 
         * @return jp.nihu.gttime.data.UTMPoint
         */
        protected UTMPoint[] getUTMpoint()
        {
            if (location != null)
            {
                return location.getPoints();
            }
            return null;
        }

        private Color getHighlightColor(int alfa)
        {
            int r = (int) (Math.random() * 256);
            int g = (int) (Math.random() * 256);
            int b = (int) (Math.random() * 256);
            // return new Color(r, g, b, alfa);
            return new Color(r, g, b);
        }

        /**
         * 詳細表示の内容をHTMLとして取得する。
         * 
         * @return 詳細表示の内容HTML
         */
        public String getHtml()
        {
        	String htmlLi = "";
        	
        	String group = getGroup();
        	if (group != null && !group.equals("") && !group.equals("-")) {
        		htmlLi += "<b>[GROUP: " + group + "]</b><br>";
        	}
        	
            String startStr = "None";
            if (jdStart != null)
            {
                startStr = JulianGregorianDate.toGregorian(jdStart.getDatetime(), jdStart.getTimezone());
            }
            String endStr = "None";
            if (jdEnd != null)
            {
                endStr = JulianGregorianDate.toGregorian(jdEnd.getDatetime(), jdEnd.getTimezone());
            }
            htmlLi += "<b>" + startStr + " - " + endStr + "</b><br>";
            UTMPoint[] utmP = getUTMpoint();
            if (utmP != null)
            {
                for (int i = 0; i < utmP.length; i++)
                {
                    htmlLi += "zone: " + utmP[i].getZone() + "<br>";
                    htmlLi += "x: " + utmP[i].getX() + "<br>";
                    htmlLi += "y: " + utmP[i].getY() + "<br>";
                }
            }

            if (record.getLocation() != null)
            {
                TimeMapGeograhicCoordinates[] pointsLongLati = record.getLocation().getPointsLongLaties();
                if (pointsLongLati != null)
                {
                    for (int i = 0; i < pointsLongLati.length; i++)
                    {
                        htmlLi += "Latitude: " + pointsLongLati[i].getLatitude() + "<br>";
                        htmlLi += "Longitude: " + pointsLongLati[i].getLongitude() + "<br>";
                    }
                }
            }

            Event[] events = getEvents();
            if (events != null)
            {
                htmlLi += "<ul>";
                for (int i = 0; i < events.length; i++)
                {
//                    if (events[i].getName().equals("url"))
//                    {
//                        htmlLi += "<li>" + events[i].getName() + "）" + "<a href=\"" + events[i].getData() + "\">" + events[i].getData() + "</a></li>";
//                    }
//                    else if (events[i].getName().equals("number") && LayerPlugin.this instanceof NumericLayerPlugin)
                    if (events[i].getName().equals("number") && LayerPlugin.this instanceof NumericLayerPlugin)
                    {
                        htmlLi += "<li><span style=\"font-size:24pt\">" + events[i].getName() + "）" + events[i].getData() + "</span></li>";
                    }// <span style="font-size:24pt">
                    else
                    {
                    	String contents = events[i].getData();
                    	
//                    	if (!contents.contains("<")) {
//                    		String regex = "https?://[\\w!/:%#\\$&\\?\\(\\)~\\.=\\+\\-@,^ ]+";
//                            Pattern p = Pattern.compile(regex);
//                            Matcher m = p.matcher(contents);
//                            contents = m.replaceAll("<a href=\"$0\">$0</a>");
//                    		contents = contents.replaceAll("\n", "<br>");
//                    	} else {
//                    		contents = contents.replaceAll("<iframe", "&lt;iframe");
//                        	contents = contents.replaceAll("</iframe>", "&lt;/iframe&gt;");
//                    	}
                    	
                        htmlLi += "<li>" + events[i].getName() + "）" + contents + "</li>";
                    }
                }
                htmlLi += "</ul>";
            }
            return htmlLi;
        }

        protected void addShape(Shape shape)
        {
            this.shape = shape;
            shapes.add(this);
        }

        private Shape shape = null;

        public boolean contains(int x, int y)
        {
            return shape.contains(x, y);
        }

        public Rectangle getBounds()
        {
            return shape.getBounds();
        }

        public T2Map getT2Map()
        {
            return LayerPlugin.this.getT2Map();
        }

        public String getBrowser()
        {
            return browser;
        }
    }
    
    public String getBrowser()
    {
        return browser;
    }

    /**
     * 親クラスのオブジェクトを取得します。
     * 
     * @return 親クラスのオブジェクト
     */
    public T2Map getT2Map()
    {
        for (Container cont = getParent(); cont != null; cont = cont.getParent())
        {
            if (cont instanceof T2Map)
            {
                return (T2Map) cont;
            }
        }
        return null;
    }

    /**
     * @see java.awt.Component#toString()
     */
    public String toString()
    {
        return getName();
    }
}
