/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.ProjectIF;
import jp.nihu.gttime.data.Record;

/**
 * ファイルをオープンする為の抽象クラスです。
 */
public abstract class OpenPlugin implements Plugin
{
    /**
     * コンポーネントオブジェクト
     */
    private T2Map parent;

    /**
     * PluginManagerオブジェクト
     */
    private PluginManager plugins;

    private IOError ioError = null;

    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(OpenPlugin.class, T2Map.LOCALE);

    /**
     * コンポーネントとPluginManagerをセットします。
     * 
     * @param parent コンポーネントオブジェクト
     * @param plugins PluginManagerオブジェクト
     */
    public void init(T2Map parent, PluginManager plugins)
    {
        this.parent = parent;
        this.plugins = plugins;
        this.ioError = new IOError();
    }

    public void processError()
    {
        ioError.processError();
    }

    /**
     * コンポーネントオブジェクトを取得します。
     * 
     * @return コンポーネントオブジェクト
     */
    protected T2Map getParentComponent()
    {
        return parent;
    }

    /**
     * レコードを取得します。
     * 
     * @param path パス
     * @param tmm tmmドキュメント
     * @return レコード
     * @throws OpenPluginException
     */
    public Record[] getRecords(String path, String format) throws OpenPluginException
    {
        Record[] records = null;
        RecordIOFilter filter = getRecordIOFilter(format);
        if (filter != null)
        {
            String rpath = "";
            try
            {
                rpath = filter.getPath(path);
                records = filter.read(getRecordsInputStream(rpath), ioError);
                groups = filter.getGroups();
            }
            catch (RecordIOFilterException e)
            {
                throw new OpenPluginException(e.getMessage() + "(" + rpath + ")", e);
            }
        }
        getParentComponent().changeInitialMinMax(records);
        return records;
    }
    
    private List<String> groups;
    
    public List<String> getGroups() {
        return groups;
    }

    /**
     * RecordIOFilterを取得します。
     * 
     * @param format フォーマット
     * @return RecordIOFilter
     * @throws OpenPluginException
     */
    private RecordIOFilter getRecordIOFilter(String format) throws OpenPluginException
    {
        RecordIOFilter f = null;
        ArrayList<Hashtable<String, String>> filters = plugins.getPlugins("RecordIOFilter");
        for (int i = 0; i < filters.size(); i++)
        {
            if (filters.get(i).get("name").equals(format))
            {
                try
                {
                    f = (RecordIOFilter) plugins.newInstance(filters.get(i).get("class"));
                }
                catch (PluginManagerException e)
                {
                    throw new OpenPluginException(e.getMessage(), e);
                }
            }
        }
        return f;
    }

    /**
     * ファイル絶対パスを取得します。
     * 
     * @param path ファイルのディレクトリ
     * @param base ファイル名
     * @return 絶対パス
     */
    public abstract String getAbsolutePath(String path, String base);

    /**
     * ファイルのパスを取得します。
     * 
     * @param path パス
     * @return パス
     */
    public abstract String getBasePath(String path);

    /**
     * ファイルをオープンします。
     * 
     * @return プロジェクトオブジェクト
     * @throws OpenPluginException
     */
    public abstract ProjectIF open() throws OpenPluginException;

    /**
     * projectファイルを取得します。
     * 
     * @param path パス
     * @return バイト入力ストリーム
     * @throws OpenPluginException
     */
    public abstract InputStream getProjectInputStream(String path) throws OpenPluginException;

    /**
     * datasetファイルを取得します。
     * 
     * @param path パス
     * @return バイト入力ストリーム
     * @throws OpenPluginException
     */
    public abstract InputStream getDatasetInputStream(String path) throws OpenPluginException;

    /**
     * レコードファイルを取得します。
     * 
     * @param path パス
     * @return バイト入力ストリーム
     * @throws OpenPluginException
     */
    public abstract InputStream getRecordsInputStream(String path) throws OpenPluginException;
}
