/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.Hashtable;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import jp.nihu.gttime.T2Map;

import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * OpenPlugin、SavePlugin、SearchPlugin、LayerPlugin、AnalizePluginを管理するクラスです。
 */
public class PluginManager
{
    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(PluginManager.class, T2Map.LOCALE);

    /**
     * URLClassLoaderオブジェクト
     */
    private URLClassLoader loader;

    /**
     * ベクターオブジェクト
     */
    private ArrayList<Hashtable<String, String>> plugins = new ArrayList<Hashtable<String, String>>();

    /**
     * 「plugins.xml」に書かれている内容を取得します。
     * 
     * @param dirs
     * @throws PluginManagerException
     */
    public PluginManager(String[] dirs) throws PluginManagerException
    {
        readConfig(getClass().getClassLoader().getResourceAsStream("plugins.xml"));
        ArrayList<URL> jars = new ArrayList<URL>();
        if (dirs != null)
        {
            for (int i = 0; i < dirs.length; i++)
            {
                loadPlugins(jars, new File(dirs[i]));
                if (jars.size() > 0)
                {
                    URL[] urls = new URL[jars.size()];
                    urls = jars.toArray(urls);
                    loader = new URLClassLoader(urls, getClass().getClassLoader());
                }
                for (int j = 0; j < plugins.size(); j++)
                {
                    if (plugins.get(j).get("preferences") != null)
                    {
                        PluginPreferences prep = (PluginPreferences) newInstance(plugins.get(j).get("preferences"));
                        T2Map.preferences.add(prep);
                    }
                }
            }
        }
    }

    /**
     * ベクターに格納しているプラグインを取得します。
     * 
     * @return ベクター
     */
    public ArrayList<Hashtable<String, String>> getPlugins()
    {
        return plugins;
    }

    /**
     * ベクターに格納しているプラグインを取得します。
     * 
     * @param type
     * @return ベクター
     */
    public ArrayList<Hashtable<String, String>> getPlugins(String type)
    {
        ArrayList<Hashtable<String, String>> v = new ArrayList<Hashtable<String, String>>();
        for (int i = 0; i < plugins.size(); i++)
        {
            Hashtable<String, String> plugin = plugins.get(i);
            if (plugin.get("type").equals(type))
            {
                v.add(plugin);
            }
        }
        return v;
    }

    /**
     * Hashtableに格納しているプラグインを取得します。
     * 
     * @param type
     * @param name
     * @return ベクター
     */
    public Hashtable<String, String> getPlugins(String type, String name)
    {
        for (int i = 0; i < plugins.size(); i++)
        {
            Hashtable<String, String> plugin = plugins.get(i);
            if (plugin.get("type").equals(type))
            {
                if (plugin.get("name").equals(name))
                {
                    return plugin;
                }
            }
        }
        return null;
    }

    /**
     * インスタンスを新規作成します。
     * 
     * @param name プラグイン名
     * @return オブジェクト
     * @throws PluginManagerException
     */
    public Object newInstance(String className) throws PluginManagerException
    {
        Object obj = null;
        try
        {
            Class<?> clazz = null;
            if (loader != null)
            {
                clazz = loader.loadClass(className);

            }
            else
            {
                clazz = getClass().getClassLoader().loadClass(className);
            }
            if (clazz != null)
            {
                obj = clazz.newInstance();
            }
        }
        catch (InstantiationException e)
        {
            e.printStackTrace();
            throw new PluginManagerException(rmgr.format("instantiation.exception", className), e);
        }
        catch (IllegalAccessException e)
        {
            throw new PluginManagerException(rmgr.format("illegal.access.exception", className), e);
        }
        catch (ClassNotFoundException e)
        {
            throw new PluginManagerException(rmgr.format("class.not.found.exception", className), e);
        }
        return obj;
    }

    /**
     * プラグインをロードします。
     * 
     * @param jars ベクター
     * @param dir ファイル
     * @throws PluginManagerException
     */
    private void loadPlugins(ArrayList<URL> jars, File dir) throws PluginManagerException
    {
        File[] child = dir.listFiles();
        if (child != null)
        {
            for (int i = 0; i < child.length; i++)
            {
                if (child[i].isDirectory())
                {
                    loadPlugins(jars, child[i]);
                }
                else
                {
                    String name = child[i].getName();
                    if (name.equals("plugins.xml"))
                    {
                        readConfig(child[i]);
                    }
                    else if (name.endsWith(".jar"))
                    {
                        try
                        {
                            jars.add(child[i].toURI().toURL());
                        }
                        catch (MalformedURLException e)
                        {
                            throw new PluginManagerException(rmgr.format("malformed.url.exception", child[i].getPath()), e);
                        }
                    }
                }
            }
        }
    }

    /**
     * InputStreamからConfigファイルを読み込みます。
     * 
     * @param in InputStream
     * @throws PluginManagerException
     */
    private void readConfig(InputStream in) throws PluginManagerException
    {
        SAXParserFactory factory = SAXParserFactory.newInstance();
        SAXParser parser = null;
        try
        {
            parser = factory.newSAXParser();
        }
        catch (ParserConfigurationException e)
        {
            throw new PluginManagerException(rmgr.getString("parser.error"), e);
        }
        catch (SAXException e)
        {
            throw new PluginManagerException(rmgr.getString("parser.error"), e);
        }

        try
        {
            parser.parse(in, new ConfigParser());
        }
        catch (SAXException e)
        {
            throw new PluginManagerException(rmgr.getString("plugin.input.error"), e);
        }
        catch (IOException e)
        {
            throw new PluginManagerException(rmgr.getString("plugin.input.error"), e);
        }
    }

    /**
     * ファイルからConfigファイルを読み込みます。
     * 
     * @param file ファイル
     * @throws PluginManagerException
     */
    private void readConfig(File file) throws PluginManagerException
    {
        SAXParserFactory factory = SAXParserFactory.newInstance();
        SAXParser parser = null;
        try
        {
            parser = factory.newSAXParser();
        }
        catch (ParserConfigurationException e)
        {
            throw new PluginManagerException(rmgr.getString("parser.initialization.error"), e);
        }
        catch (SAXException e)
        {
            throw new PluginManagerException(rmgr.getString("parser.initialization.error"), e);
        }

        try
        {
            parser.parse(file, new ConfigParser());
        }
        catch (SAXException e)
        {
            throw new PluginManagerException(rmgr.format("file.analysis.error", file.getPath()), e);
        }
        catch (IOException e)
        {
            throw new PluginManagerException(rmgr.format("file.acquisition", file.getPath()), e);
        }
    }

    /**
     * Configをパースするクラスです。
     */
    class ConfigParser extends DefaultHandler
    {
        /**
         * StringBufferオブジェクト
         */
        private StringBuffer value = new StringBuffer();

        /**
         * Hashtableオブジェクト
         */
        private Hashtable<String, String> conf = null;

        /**
         * ベクターオブジェクト
         */
        private ArrayList<Hashtable<String, String>> confs = new ArrayList<Hashtable<String, String>>();

        /**
         * ベクターに文字列を追加します。
         * 
         * @param ch 追加する文字列
         * @param start 始点
         * @param length 追加する長さ
         * @throws SAXException
         */
        public void characters(char[] ch, int start, int length) throws SAXException
        {
            value.append(ch, start, length);
        }

        /**
         * Hashtableオブジェクトにマップします。
         * 
         * @param uri URI
         * @param localName
         * @param qName 名前
         * @throws SAXException
         */
        public void endElement(String uri, String localName, String qName) throws SAXException
        {
            if (qName.equals("plugin"))
            {
                confs.add(conf);
                conf = null;
            }
            else if (qName.equals("type"))
            {
                conf.put("type", value.toString().trim());
            }
            // else if (qName.equals("name"))
            // {
            // conf.put("name", value.toString().trim());
            // }
            else if (qName.equals("class"))
            {
                String className = value.toString().trim();
                conf.put("class", className);
                PluginResourceManager rm = PluginResourceManager.getManager(className, T2Map.LOCALE);
                conf.put("name", rm.getString("plugin.name").trim());
                rm = null;
            }
            else if (qName.equals("description"))
            {
                conf.put("description", value.toString().trim());
            }
            else if (qName.equals("preferences"))
            {
                conf.put("preferences", value.toString().trim());
            }
            else if (qName.equals("plugins"))
            {
                for (int i = 0; i < confs.size(); i++)
                {
                    plugins.add(confs.get(i));
                }
            }
        }

        /**
         * Hashtableオブジェクトを作成し、StringBufferオブジェクトを削除します。
         * 
         * @param uri URI
         * @param localName
         * @param qName 名前
         * @param attributes
         * @throws SAXException
         */
        public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
        {
            if (qName.equals("plugin"))
            {
                conf = new Hashtable<String, String>();
            }
            value.delete(0, value.length());
        }
    }
}
