/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin;

import java.text.MessageFormat;
import java.util.Locale;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * リソースバンドルの機能拡張用ラッパークラス
 */
public class PluginResourceManager
{

    private static Log log = LogFactory.getLog(PluginResourceManager.class);

    private ResourceBundle bundle;

    private PluginResourceManager(ResourceBundle bundle)
    {
        this.bundle = bundle;
    }

    /**
     * 指定された基底名の {@code PluginResourceManager} を取得する。
     * 
     * @param baseName リソースバンドル、完全指定されたクラス名の基底名
     * @return 指定された基底名の {@code PluginResourceManager}
     */
    public static PluginResourceManager getManager(String baseName)
    {
        try
        {
            return new PluginResourceManager(XMLResourceBundle.getXMLBundle(baseName));
            // return new
            // PluginResourceManager(ResourceBundle.getBundle(baseName));
        }
        catch (MissingResourceException e)
        {
            // Log [fatal]
            log.fatal(baseName + " resource not found.", e);
            throw e;
        }
    }

    /**
     * 指定された基底名の {@code PluginResourceManager} を取得する。
     * 
     * @param baseName リソースバンドル、完全指定されたクラス名の基底名
     * @param locale Locale
     * @return 指定された基底名の {@code PluginResourceManager}
     */
    public static PluginResourceManager getManager(String baseName, Locale locale)
    {
        return new PluginResourceManager(XMLResourceBundle.getXMLBundle(baseName, locale));
        // return new PluginResourceManager(ResourceBundle.getBundle(baseName,
        // locale));
    }

    /**
     * 指定されたクラス名の {@code PluginResourceManager} を取得する。
     * 
     * @param clazz リソースバンドル、完全指定されたクラス名
     * @return 指定されたクラス名の {@code PluginResourceManager}
     */
    // public static PluginResourceManager getManager(Class<?> clazz)
    // {
    // return getManager(clazz.getName());
    // }
    /**
     * 指定されたクラス名の {@code PluginResourceManager} を取得する。
     * 
     * @param clazz リソースバンドル、完全指定されたクラス名
     * @param locale Locale
     * @return 指定されたクラス名の {@code PluginResourceManager}
     */
    public static PluginResourceManager getManager(Class<?> clazz, Locale locale)
    {
        return getManager(clazz.getName(), locale);
    }

    /**
     * リソースバンドルを取得する。
     * 
     * @return {@code ResourceBundle} オブジェクト
     */
    public ResourceBundle getBundle()
    {
        return bundle;
    }

    /**
     * 指定されたキーのオブジェクトを取得する。
     * 
     * @param key 望ましいオブジェクトのキー
     * @return 指定されたキーのオブジェクト
     */
    public Object getObject(String key)
    {
    	Object val = getBundle().getObject(key);
    	if (val == null) {
    		log.warn("No resouce for " + key);
    	}
        return val;
    }

    /**
     * 指定されたキーの文字列を取得する。
     * 
     * @param key 望ましい文字列のキー
     * @return 指定されたキーの文字列
     */
    public String getString(String key)
    {
    	String val = getBundle().getString(key);
    	if (val == null) {
    		log.warn("No resouce for " + key);
    	}
        return val;
    }

    /**
     * 指定されたキーの文字列配列を取得する。 このメソッドを呼び出すことは、次を呼び出すことと同じ。
     * 
     * <pre>
     * getStrings(key, &quot;,&quot;)
     * </pre>
     * 
     * @param key 望ましい文字列のキー
     * @return 指定されたキーの文字列配列
     */
    public String[] getStrings(String key)
    {
        return getStrings(key, ",");
    }

    /**
     * 指定されたキーの文字列配列を取得する。 このメソッドを呼び出すことは、次を呼び出すことと同じ。
     * 
     * <pre>
     * getString(key).split(delimiter)
     * </pre>
     * 
     * @param key 望ましい文字列のキー
     * @param delimiter 文字区切子
     * @return 指定されたキーの文字列配列
     */
    public String[] getStrings(String key, String delimiter)
    {
        return getString(key).split(delimiter);
    }

    /**
     * 指定されたキーの {@code short} 値を取得する。
     * 
     * @param key 望ましい値のキー
     * @return 指定されたキーの {@code short} 値
     */
    public short getShort(String key)
    {
        return Short.parseShort(getString(key));
    }

    /**
     * 指定されたキーの {@code int} 値を取得する。
     * 
     * @param key 望ましい値のキー
     * @return 指定されたキーの {@code int} 値
     */
    public int getInt(String key)
    {
        return Integer.parseInt(getString(key));
    }

    /**
     * 指定されたキーの {@code long} 値を取得する。
     * 
     * @param key 望ましい値のキー
     * @return 指定されたキーの {@code long} 値
     */
    public long getLong(String key)
    {
        return Long.getLong(getString(key));
    }

    /**
     * 指定されたキーの {@code float} 値を取得する。
     * 
     * @param key 望ましい値のキー
     * @return 指定されたキーの {@code float} 値
     */
    public float getFloat(String key)
    {
        return Float.parseFloat(getString(key));
    }

    /**
     * 指定されたキーの {@code double} 値を取得する。
     * 
     * @param key 望ましい値のキー
     * @return 指定されたキーの {@code double} 値
     */
    public double getDouble(String key)
    {
        return Double.parseDouble(getString(key));
    }

    /**
     * 指定されたキーのパターンに対してフォーマットを行う。
     * 
     * @param key 望ましいパターンのキー
     * @param arguments フォーマットするオブジェクト群
     * @return フォーマットした文字列
     */
    public String format(String key, Object... arguments)
    {
        return MessageFormat.format(getString(key), arguments);
    }

    /**
     * 指定されたキーのリソースが存在するかを判定する。
     * 
     * @param key 望ましい値のキー
     * @return 存在する場合 {@code true}
     */
    public boolean hasResource(String key)
    {
        boolean b = false;
        try
        {
            getBundle().getObject(key);
            b = true;
        }
        catch (MissingResourceException e)
        {
            // Ignore;
        }
        return b;
    }
}
