/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.analyzer;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

import javax.swing.JDialog;
import javax.swing.JOptionPane;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.DatasetException;
import jp.nihu.gttime.data.DatasetGtm;
import jp.nihu.gttime.data.DatasetIF;
import jp.nihu.gttime.data.ProjectException;
import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.plugin.LayerData;
import jp.nihu.gttime.plugin.MaskAnalyzePlugin;
import jp.nihu.gttime.plugin.PluginResourceManager;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * マスク解析用プラグイン抽象クラス
 */
public abstract class AbstractMaskAnalyzePlugin implements MaskAnalyzePlugin
{

    private static Log log = LogFactory.getLog(AbstractMaskAnalyzePlugin.class);

    private static PluginResourceManager rmgr = PluginResourceManager.getManager(AbstractMaskAnalyzePlugin.class, T2Map.LOCALE);

    private static final String LAYER_MASK_TYPE = rmgr.getString("layer.mask.type");

    private T2Map t2map;

    private DatasetIF dataset;

    /**
     * オブジェクトの初期化を行う。
     */
    public void initPlugin(T2Map t2map)
    {
        this.t2map = t2map;
    }

    /**
     * {@code T2Map} オブジェクトを取得する。
     */
    protected T2Map getT2Map()
    {
        return t2map;
    }

    /**
     * マスク用レイヤタイプを取得する。
     */
    public String getMaskLayerType()
    {
        return LAYER_MASK_TYPE;
    }

    /**
     * マスク解析結果を取得する。
     */
    protected DatasetIF getDataset()
    {
        return dataset;
    }

    /**
     * マスク解析結果を設定する。
     */
    protected void setDataset(DatasetIF dataset)
    {
        this.dataset = dataset;
    }

    /**
     * マスク解析結果として {@code Dataset} オブジェクトを生成する。 生成するレイヤタイプはマスク用となる。
     */
    protected DatasetIF createDataset(String title, Record[] records)
    {
        return createDataset(title, getMaskLayerType(), records);
    }

    /**
     * マスク解析結果として {@code Dataset} オブジェクトを生成する。
     */
    protected DatasetIF createDataset(String title, String type, Record[] records)
    {
        // Sort record.
        if (records != null)
        {
            Arrays.sort(records, new RecordComparator());
        }
        try
        {
            // int ptype = T2Map.getProjectType();
            // if (ptype == 0)
            // {
            // setDataset(new DatasetTmm(title, type, records));
            // }
            // else if (ptype == 1)
            // {
            setDataset(new DatasetGtm(title, type, records));
            // }
        }
        catch (DatasetException e)
        {
            // Log [error]
            log.error("Dataset instance creation error.", e);
        }
        return getDataset();
    }

    /**
     * マスク解析を実行する。
     */
    public DatasetIF analyze()
    {
        // Initialize.
        setDataset(null);

        if (T2Map.getProject() != null)
        {
            JDialog dialog = getDialog();
            dialog.setVisible(true);
        }
        else
        {
            JOptionPane.showMessageDialog(getT2Map(), rmgr.getString("mask.noproject.message"), rmgr.getString("mask.title"), JOptionPane.WARNING_MESSAGE);
        }
        return getDataset();
    }

    /**
     * 指定したオブジェクトをもとにマスク解析を実行する。
     */
    public abstract int analyze(Object... o);

    /**
     * マスク解析した {@code Dataset} オブジェクトを {@code T2Map} ツリーに 追加する。
     */
    public void addDataset(DatasetIF dataset)
    {
        if (dataset != null && dataset.hasRecords())
        {
            try
            {
            	T2Map.getProject().addDataset(dataset);
                getT2Map().updateProject(T2Map.getProject());
            }
            catch (ProjectException e)
            {
                // Log [error]
                log.error(e, e);
            }
        }
    }

    /**
     * マスク解析用ダイアログを取得する。
     */
    protected abstract JDialog getDialog();

    /**
     * 現在アプリケーション側で表示している {@code LayerData} オブジェクト配列 を取得する。
     */
    public LayerData[] getLayerDatas()
    {
        return DatasetList.getArray(T2Map.getProject(), getT2Map().getLayerPanel());
    }

    /**
     * 現在アプリケーション側で表示しているレイヤ情報よりマスクレイヤを除いた {@code LayerData} オブジェクト配列を取得する。
     */
    public LayerData[] getNomalLayerDatas()
    {
        return getOptionLayerDatas(getMaskLayerType(), false);
    }

    /**
     * 現在アプリケーション側で表示しているレイヤ情報よりマスクレイヤのみの {@code LayerData} オブジェクト配列を取得する。
     */
    public LayerData[] getMaskLayerDatas()
    {
        return getOptionLayerDatas(getMaskLayerType(), true);
    }

    /**
     * 指定したレイヤタイプより現在アプリケーション側で表示しているレイヤ情報 を{@code LayerData} オブジェクト配列として返却する。
     * 
     * @param type 取得するレイヤタイプ文字列
     * @param b {@code type} で指定したレイヤを含む場合は {@code true}
     */
    private LayerData[] getOptionLayerDatas(String type, boolean b)
    {
        List<LayerData> list = new ArrayList<LayerData>();
        if (type != null)
        {
            for (LayerData d : getLayerDatas())
            {
                if (b == type.equals(d.getType()))
                {
                    list.add(d);
                }
            }
        }
        return list.toArray(new LayerData[0]);
    }
}
