/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.analyzer;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.plugin.PluginResourceManager;

/**
 * データ解析用の値を定数として格納するユーティリティクラス。
 */
public class AnalyzeConstants
{

    private static PluginResourceManager rmgr = PluginResourceManager.getManager(AnalyzeConstants.class, T2Map.LOCALE);

    /**
     * データ解析に使用する論理演算子の列挙型
     */
    public static enum Query {
        /**
         * 論理積（AND）を表す。
         */
        AND(rmgr.getString("query.and.name")),
        /**
         * 論理和（OR）を表す。
         */
        OR(rmgr.getString("query.or.name")),
        /**
         * 否定（NOT）を表す。
         */
        NOT(rmgr.getString("query.not.name"));

        private String name;

        private Query(String name)
        {
            this.name = name;
        }

        /**
         * 列挙型定数の文字列表現を返す。
         */
        @Override
        public String toString()
        {
            return name;
        }
    }

    /**
     * データ解析に使用する関係演算子の列挙型
     */
    public static enum Operator {
        /**
         * 等号（{@literal ==}）を表す。
         */
        EQ(rmgr.getString("operator.eq.name")),
        /**
         * 不等号（{@literal <=}）を表す。
         */
        LE(rmgr.getString("operator.le.name")),
        /**
         * 不等号（{@literal <}）を表す。
         */
        LT(rmgr.getString("operator.lt.name")),
        /**
         * 不等号（{@literal >=}）を表す。
         */
        GE(rmgr.getString("operator.ge.name")),
        /**
         * 不等号（{@literal >}）を表す。
         */
        GT(rmgr.getString("operator.gt.name"));

        private String name;

        private Operator(String name)
        {
            this.name = name;
        }

        /**
         * 列挙型定数の文字列表現を返す。
         */
        @Override
        public String toString()
        {
            return name;
        }
    }

    private AnalyzeConstants()
    {
    }

    /**
     * 指定した比較演算子で 2 つの {@code short} 値を比較する。
     * 
     * @param o 比較演算子
     * @param s1 比較する最初の {@code short}
     * @param s2 比較する 2 番目の {@code short}
     * @return ({@code s1} 比較演算子 {@code s2}) が真の場合 {@code true}
     */
    public static boolean compare(Operator o, short s1, short s2)
    {
        return compare(o, Short.valueOf(s1).compareTo(s2));
    }

    /**
     * 指定した比較演算子で 2 つの {@code int} 値を比較する。
     * 
     * @param o 比較演算子
     * @param i1 比較する最初の {@code int}
     * @param i2 比較する 2 番目の {@code int}
     * @return ({@code i1} 比較演算子 {@code i2}) が真の場合 {@code true}
     */
    public static boolean compare(Operator o, int i1, int i2)
    {
        return compare(o, Integer.valueOf(i1).compareTo(i2));
    }

    /**
     * 指定した比較演算子で 2 つの {@code long} 値を比較する。
     * 
     * @param o 比較演算子
     * @param l1 比較する最初の {@code long}
     * @param l2 比較する 2 番目の {@code long}
     * @return ({@code l1} 比較演算子 {@code l2}) が真の場合 {@code true}
     */
    public static boolean compare(Operator o, long l1, long l2)
    {
        return compare(o, Long.valueOf(l1).compareTo(l2));
    }

    /**
     * 指定した比較演算子で 2 つの {@code float} 値を比較する。
     * 
     * @param o 比較演算子
     * @param f1 比較する最初の {@code float}
     * @param f2 比較する 2 番目の {@code float}
     * @return ({@code f1} 比較演算子 {@code f2}) が真の場合 {@code true}
     */
    public static boolean compare(Operator o, float f1, float f2)
    {
        return compare(o, Float.compare(f1, f2));
    }

    /**
     * 指定した比較演算子で 2 つの {@code double} 値を比較する。
     * 
     * @param o 比較演算子
     * @param d1 比較する最初の {@code double}
     * @param d2 比較する 2 番目の {@code double}
     * @return ({@code d1} 比較演算子 {@code d2}) が真の場合 {@code true}
     */
    public static boolean compare(Operator o, double d1, double d2)
    {
        return compare(o, Double.compare(d1, d2));
    }

    /**
     * 指定した比較演算子と 2 つの値の比較結果が合致するかを検証する。
     * 
     * @param o 比較演算子
     * @param compare 2 つの値の比較結果
     * @return 比較演算子と比較結果が合致する場合 {@code true}
     */
    private static boolean compare(Operator o, int compare)
    {
        boolean b = false;
        switch (o)
        {
            case EQ:
                b = (compare == 0);
                break;
            case LE:
                b = (compare <= 0);
                break;
            case LT:
                b = (compare < 0);
                break;
            case GE:
                b = (compare >= 0);
                break;
            case GT:
                b = (compare > 0);
                break;
        }
        return b;
    }
}
