/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.analyzer;

import java.net.MalformedURLException;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import javax.swing.JDialog;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.JulianDate;
import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.plugin.LayerData;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.utility.Utilities;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * マスクレコード抽出用プラグインクラス
 */
public class DefaultMaskAnalyzePlugin extends AbstractMaskAnalyzePlugin
{

    private static PluginResourceManager rmgr = PluginResourceManager.getManager(DefaultMaskAnalyzePlugin.class, T2Map.LOCALE);

    private RecordComparator comparator = new RecordComparator();

    /**
     * マスクレコード抽出用ダイアログを取得する。
     */
    @Override
    protected JDialog getDialog()
    {
        return new MaskAnalyzeDialog(getT2Map(), this, MaskAnalyzeDialog.Type.DEFAULT);
    }

    /**
     * 指定したオブジェクトをもとにマスク解析を実行する。
     */
    @Override
    public int analyze(Object... o)
    {
        LayerData layer = null;
        LayerData mask = null;
        Boolean b = null;
        if (o.length >= 3)
        {
            if (o[0] instanceof LayerData)
            {
                layer = (LayerData) o[0];
            }
            if (o[1] instanceof LayerData)
            {
                mask = (LayerData) o[1];
            }
            if (o[2] instanceof Boolean)
            {
                b = (Boolean) o[2];
            }
        }
        if (layer != null && mask != null && b != null)
        {
            return analyzeMask(layer, mask, b);
        }
        throw new IllegalArgumentException("Argument mismatched analyzeMask(LayerData, LayerData, Boolean) .");
    }

    /**
     * マスクによるレコード抽出を行う。
     * 
     * @param layer 抽出対象となるレイヤ情報
     * @param mask マスクレイヤ情報
     * @param b 部分的にマスクにかかるレコードを含める場合 {@code true}
     * @return レコード件数
     */
    public int analyzeMask(LayerData layer, LayerData mask, Boolean b)
    {
        // Create new instance.
        Set<Record> set = new HashSet<Record>();
        Record[] rs1 = null;
        Record[] rs2 = null;
        // Get record & sort.
        if (layer != null && layer.getDataset().hasRecords())
        {
            rs1 = layer.getDataset().getRecords(layer.getGroup());
            Arrays.sort(rs1, comparator);
        }
        if (mask != null && mask.getDataset().hasRecords())
        {
            rs2 = mask.getDataset().getRecords(mask.getGroup());
            Arrays.sort(rs2, comparator);
        }

        if (rs1 != null && rs2 != null)
        {
            String type = layer.getType();
            boolean isMask = getMaskLayerType().equalsIgnoreCase(type);
            for (Record r1 : rs1)
            {
                for (Record r2 : rs2)
                {
                    boolean judge = contains(r2, r1, !isMask);
                    if (!judge && b)
                    {
                        judge = intersects(r2, r1);
                    }
                    if (judge)
                    {
                        set.add(r1);
                        break;
                    }
                }
            }
            if (set.size() > 0)
            {
            	Record[] results = RecordCloner.clone(set);
            	if (0 < results.length) {
                	DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
                	dbf.setNamespaceAware(true);
                	DocumentBuilder builder = null;
					try {
						builder = dbf.newDocumentBuilder();
					} catch (ParserConfigurationException e) {
						// TODO Auto-generated catch block
						e.printStackTrace();
					}
                	Document doc = builder.newDocument();
                	Element root = doc.createElementNS("http://www.nihu.jp/ns/spacetime/2006/11", "records");
//                	root.setAttribute("xmlns:owl", "http://www.w3.org/2002/07/owl#");
//                	root.setAttribute("xmlns:rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#");
//                	root.setAttribute("xmlns:dc", "http://www.dc.org");
            		doc.appendChild(root);
            		
                    for (Record r : results) {
                    	Node n = r.getNode();
                    	//属性をチェックしてまだ新しいルート要素にない属性があったら新しいルート要素に追加する
                    	Element root0 = n.getOwnerDocument().getDocumentElement();
                    	r.setAttributeToRootElement(root0, root);
                    	n = doc.adoptNode(n);
                    	r.setNode(n);
                    	root.appendChild(n);
						try {
							String uri = Utilities.getUri(layer.getDataset().getRdfBase());
							r.createRdfSameAsEvent(uri);
						} catch (MalformedURLException e) {
							// TODO Auto-generated catch block
							e.printStackTrace();
						}
                    }
                }
                createDataset(getTitle(layer, mask, b), type, results);
            }
        }
        return set.size();
    }

    /**
     * 最初の引数の境界内に 2 番めの引数があるかを判定する。
     * 
     * @param r1 最初の {@code Record} オブジェクト
     * @param r2 2 番目の {@code Record} オブジェクト
     * @param b 端点で接するものを含む場合 {@code true}
     */
    private boolean contains(Record r1, Record r2, boolean b)
    {
        // Get date.
        JulianDate ds1 = r1.getStartDatetime();
        JulianDate de1 = r1.getEndDatetime();
        JulianDate ds2 = r2.getStartDatetime();
        JulianDate de2 = r2.getEndDatetime();

        // Check contain.
        int c1 = comparator.compare(ds1, ds2);
        int c2 = comparator.compare(de1, de2);
        return (b) ? (c1 <= 0 && c2 >= 0) : (c1 < 0 && c2 > 0);
    }

    /**
     * 最初の引数と 2 番めの引数が交差するかを判定する。
     * 
     * @param r1 最初の {@code Record} オブジェクト
     * @param r2 2 番目の {@code Record} オブジェクト
     */
    private boolean intersects(Record r1, Record r2)
    {
        boolean b = false;
        // Get date.
        JulianDate ds1 = r1.getStartDatetime();
        JulianDate de1 = r1.getEndDatetime();
        JulianDate ds2 = r2.getStartDatetime();
        JulianDate de2 = r2.getEndDatetime();

        // Check intersect.
        if (comparator.compare(ds1, ds2) > 0)
        {
            b = (comparator.compare(ds1, de2) < 0);
        }
        else
        {
            b = (comparator.compare(de1, ds2) > 0);
        }
        return b;
    }

    /**
     * レイヤタイトルを取得する。
     */
    private String getTitle(LayerData layer, LayerData mask, Boolean b)
    {
        String note = (b) ? rmgr.getString("title.format.note") : "";
        return rmgr.format("title.format", layer.getName(), mask.getName(), note);
    }
}
