/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.analyzer;

import java.math.BigDecimal;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Set;

import javax.swing.JDialog;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.JulianDate;
import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.data.RecordException;
import jp.nihu.gttime.plugin.LayerData;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.plugin.analyzer.AnalyzeConstants.Query;

/**
 * マスク論理演算用プラグインクラス
 */
public class LogicalMaskAnalyzePlugin extends AbstractMaskAnalyzePlugin
{

    private static PluginResourceManager rmgr = PluginResourceManager.getManager(LogicalMaskAnalyzePlugin.class, T2Map.LOCALE);

    private RecordComparator comparator = new RecordComparator();

    /**
     * マスク論理演算用ダイアログを取得する。
     */
    @Override
    protected JDialog getDialog()
    {
        return new MaskAnalyzeDialog(getT2Map(), this, MaskAnalyzeDialog.Type.LOGICAL);
    }

    /**
     * 指定したオブジェクトをもとにマスク解析を実行する。
     */
    @Override
    public int analyze(Object... o)
    {
        LayerData mask1 = null;
        LayerData mask2 = null;
        Query query = null;
        if (o.length >= 3)
        {
            if (o[0] instanceof LayerData)
            {
                mask1 = (LayerData) o[0];
            }
            if (o[1] instanceof LayerData)
            {
                mask2 = (LayerData) o[1];
            }
            if (o[2] instanceof Query)
            {
                query = (Query) o[2];
            }
        }
        if (mask1 != null && query != null)
        {
            return analyzeMask(mask1, mask2, query);
        }
        throw new IllegalArgumentException("Argument mismatched analyzeMask(LayerData, LayerData, Query) .");
    }

    /**
     * マスクの論理演算を行う。
     * 
     * @param mask1 最初のマスクレイヤ情報
     * @param mask2 2 番目のマスクレイヤ情報
     * @param query 論理演算子
     * @return レコード件数
     */
    public int analyzeMask(LayerData mask1, LayerData mask2, Query query)
    {
        Record[] records = null;
        Record[] rs1 = null;
        Record[] rs2 = null;
        // Get record & sort.
        if (mask1 != null && mask1.getDataset().hasRecords())
        {
            rs1 = mask1.getDataset().getRecords(mask1.getGroup());
            Arrays.sort(rs1, comparator);
        }
        if (mask2 != null && mask2.getDataset().hasRecords())
        {
            rs2 = mask2.getDataset().getRecords(mask2.getGroup());
            Arrays.sort(rs2, comparator);
        }

        switch (query)
        {
            case AND:
                if (rs1 != null && rs2 != null)
                {
                    records = analyzeAndMask(rs1, rs2);
                }
                break;
            case OR:
                if (rs1 != null && rs2 != null)
                {
                    records = analyzeOrMask(rs1, rs2);
                }
                break;
            case NOT:
                if (rs1 != null)
                {
                    records = analyzeNotMask(rs1);
                }
                break;
        }
        if (records != null)
        {
            createDataset(getTitle(mask1, mask2, query), records);
        }
        return (records != null) ? records.length : 0;
    }

    /**
     * マスクの論理積 (AND) 演算を行う。
     * 
     * @param rs1 最初の {@code Record} オブジェクト配列
     * @param rs2 2 番目の {@code Record} オブジェクト配列
     * @return 結果 となる {@code Record} オブジェクト配列
     */
    public Record[] analyzeAndMask(Record[] rs1, Record[] rs2)
    {
        // Create new instance.
        Set<Record> set = new HashSet<Record>();
        for (Record r1 : rs1)
        {
            // Get date.
            JulianDate ds1 = r1.getStartDatetime();
            JulianDate de1 = r1.getEndDatetime();
            for (Record r2 : rs2)
            {
                // Get date.
                JulianDate ds = null;
                if (ds1 == null)
                {
                    ds = r2.getStartDatetime();
                }
                else if (r2.getStartDatetime() == null)
                {
                    ds = ds1;
                }
                else
                {
                    ds = comparator.max(ds1, r2.getStartDatetime());
                }
                JulianDate de = null;
                if (de1 == null)
                {
                    de = r2.getEndDatetime();
                }
                else
                {
                    de = comparator.min(de1, r2.getEndDatetime());
                }

                if (ds == null || comparator.compare(ds, de) < 0)
                {
                    // Create new instance.
                    Record r = new Record();

                    try
                    {
                        r.setStartDatetime(ds);
                        r.setEndDatetime(de);
                    }
                    catch (RecordException e)
                    {
                        e.printStackTrace();
                    }
                    set.add(r);
                }
            }
        }
        return set.toArray(new Record[0]);
    }

    /**
     * マスクの論理和 (OR) 演算を行う。
     * 
     * @param rs1 最初の {@code Record} オブジェクト配列
     * @param rs2 2 番目の {@code Record} オブジェクト配列
     * @return 結果 となる {@code Record} オブジェクト配列
     */
    public Record[] analyzeOrMask(Record[] rs1, Record[] rs2)
    {
        // Create new instance.
        Set<Record> set = new HashSet<Record>();
        // Join record.
        for (Record r : rs1)
        {
            set.add(r);
        }
        for (Record r : rs2)
        {
            set.add(r);
        }

        // Get joined record.
        Record current = null;
        Record[] records = set.toArray(new Record[0]);
        Arrays.sort(records, comparator);
        set.clear();
        try
        {
            for (int i = 0; i < records.length; i++)
            {
                if (current == null)
                {
                    // Create new instance.
                    current = new Record();
                    current.setStartDatetime(records[i].getStartDatetime());
                    current.setEndDatetime(records[i].getEndDatetime());
                    set.add(current);
                }
                // Merge record.
                for (int j = (i + 1); j < records.length; j++)
                {
                    // Get date.
                    JulianDate de1 = current.getEndDatetime();
                    JulianDate ds2 = records[j].getStartDatetime();
                    JulianDate de2 = records[j].getEndDatetime();
                    boolean startEndEdgeOpen = current.isEndEdgeOpen();
                    boolean endStartEdgeOpen = records[j].isStartEdgeOpen();
                    boolean isBothEdgeClose = (!endStartEdgeOpen && !startEndEdgeOpen);
//                    log.debug("startEndEdgeOpen : " + startEndEdgeOpen);
//                    log.debug("endStartEdgeOpen : " + endStartEdgeOpen);
//                    log.debug("isBothEdgeClose : " + isBothEdgeClose);
//                    log.debug("comparator.compare(de1, ds2) : " + comparator.compare(de1, ds2));
                    if (de1 == null)
                    {
                        if (ds2 == null)
                        {
                            current.setStartDatetime(ds2);
                        }
                        i++;
                    }
                    else if (ds2 == null)
                    {
                        current.setStartDatetime(ds2);
                        current.setEndDatetime(comparator.max(de1, de2));
                        i++;
                    }
                    else if (isBothEdgeClose ? comparator.compare(de1, ds2) >= 0 : comparator.compare(de1, ds2) > 0)
                    {
                        current.setEndDatetime(comparator.max(de1, de2));
                        i++;
                    }
                    else
                    {
                        break;
                    }
                }
                current = null;
            }
        }
        catch (RecordException e)
        {
            e.printStackTrace();
        }
        return set.toArray(new Record[0]);
    }

    /**
     * マスクの否定 (NOT) 演算を行う。
     * 
     * @param rs {@code Record} オブジェクト配列
     * @return 結果 となる {@code Record} オブジェクト配列
     */
    public Record[] analyzeNotMask(Record[] rs)
    {
        // Create new instance.
        Set<Record> set = new HashSet<Record>();
        Record current = null;
        JulianDate ds = null;
        JulianDate de = null;
        try
        {
            for (Record r : rs)
            {
                if (current == null)
                {
                    // Create new instance.
                    current = new Record();
                    // current.setStartDatetime(getStartDatetime( ));
                }
                ds = current.getStartDatetime();
                de = r.getStartDatetime();
                if (de == null)
                {

                }
                else if (ds == null)
                {
                    current.setEndDatetime(de);
                    set.add(current);
                }
                else if (comparator.compare(ds, de) < 0)
                {
                    current.setEndDatetime(de);
                    set.add(current);
                }

                // Create new instance.
                current = new Record();
                current.setStartDatetime(r.getEndDatetime());
            }
            if (current != null)
            {
                ds = current.getStartDatetime();
                de = getEndDatetime();
                if (ds != null && comparator.compare(ds, de) <= 0)
                {
                    // current.setEndDatetime(de);
                    set.add(current);
                }
            }
        }
        catch (RecordException e)
        {
            e.printStackTrace();
        }
        return set.toArray(new Record[0]);
    }

    /**
     * 開始日付を取得する。
     */
    private JulianDate getStartDatetime()
    {
        // Create new instance.
        JulianDate d = new JulianDate();
        // Project 側 開始日時取得
        try
        {
            // if (getT2Map( ).getProject( ) != null) {
            if (getT2Map().getTimeAxisData() != null)
            {

                // TimeAxisIF time = getT2Map().getTimeAxisData();
                // log.debug("CurrentMin: " +
                // time.getCurrentMin().toString());
                // log.debug("CurrentMax: " +
                // time.getCurrentMax().toString());
                // log.debug("ApplicationMin: " +
                // time.getApplicationMin().toString());
                // log.debug("ApplicationMax: " +
                // time.getApplicationMax().toString());
                // log.debug("InitialCurrentMin: " +
                // time.getInitialCurrentMin().toString());
                // log.debug("InitialCurrentMax: " +
                // time.getInitialCurrentMax().toString());
                // log.debug("InitialMin: " +
                // time.getInitialMin().toString());
                // log.debug("InitialMax: " +
                // time.getInitialMax().toString());
                // log.debug("Min: " + time.getMin().toString());
                // log.debug("Max: " + time.getMax().toString());
                //                
                // log.debug("ProjectMin: " + getT2Map( ).getProject(
                // ).getStartDatetime( ));
                // log.debug("ProjectMax: " + getT2Map( ).getProject(
                // ).getEndDatetime( ));

                // 表示されている範囲の値を利用するように変更
                // d.setDatetime(getT2Map( ).getProject( ).getStartDatetime( ));
                d.setDatetime(getT2Map().getTimeAxisData().getCurrentMin());
            }
        }
        catch (Exception e)
        {
            // Ignore.
        }
        // NULL の場合は Double 最小値を設定
        if (d.getDatetime() == null)
        {
            d.setDatetime(BigDecimal.valueOf(Double.MIN_VALUE));
        }
        return d;
    }

    /**
     * 終了日付を取得する。
     */
    private JulianDate getEndDatetime()
    {
        // Create new instance.
        JulianDate d = new JulianDate();
        // Project 側 終了日時取得
        try
        {
            // if (getT2Map( ).getProject( ) != null) {
            if (getT2Map().getTimeAxisData() != null)
            {
                // 表示されている範囲の値を利用するように変更
                // d.setDatetime(getT2Map( ).getProject( ).getEndDatetime( ));
                d.setDatetime(getT2Map().getTimeAxisData().getCurrentMax());
            }
        }
        catch (Exception e)
        {
            // Ignore.
        }
        // NULL の場合は Double 最大値を設定
        if (d.getDatetime() == null)
        {
            d.setDatetime(BigDecimal.valueOf(Double.MAX_VALUE));
        }
        return d;
    }

    /**
     * レイヤタイトルを取得する。
     */
    private String getTitle(LayerData mask1, LayerData mask2, Query query)
    {
        String s = (mask2 != null) ? mask2.getName() : "";
        return rmgr.format("title.format", mask1.getName(), s, query);
    }
}
