/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.analyzer;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.BoxLayout;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTextField;
import javax.swing.ListSelectionModel;
import javax.swing.border.TitledBorder;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.plugin.LayerData;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.plugin.analyzer.AnalyzeConstants.Operator;
import jp.nihu.gttime.plugin.analyzer.AnalyzeConstants.Query;
import jp.nihu.gttime.utility.TextValidator;
import jp.nihu.gttime.utility.TextValidator.NUMBER;
import jp.nihu.gttime.utility.TextValidator.REQUIRED;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * マスク解析用ダイアログクラス
 */
@SuppressWarnings("serial")
public class MaskAnalyzeDialog extends JDialog implements ActionListener, ItemListener, ListSelectionListener
{
    private static Log log = LogFactory.getLog(MaskAnalyzeDialog.class);

    private static PluginResourceManager rmgr = PluginResourceManager.getManager(MaskAnalyzeDialog.class, T2Map.LOCALE);

    /**
     * マスク解析タイプの列挙型
     */
    public static enum Type {
        DEFAULT, LOGICAL, NUMERIC
    }

    private static final int NUMERIC_ITEM_SIZE = 2;

    private AbstractMaskAnalyzePlugin plugin;

    private Type dialogType;

    private List<JButton> btnList;

    private List<JTextField> textList;

    private List<JComboBox> comboList;

    private JList layer1List;

    private JList layer2List;

    private ButtonGroup unionGrp;

    private JComboBox queryCmb;

    private List<Operator> numOpeList;

    private List<Double> numValList;

    /**
     * マスク解析用ダイアログを生成する。
     * 
     * @param owner ダイアログを表示する {@code Frame}
     * @param plugin 呼出元となるプラグインオブジェクト
     * @param type マスク解析のタイプ
     */
    public MaskAnalyzeDialog(Frame owner, AbstractMaskAnalyzePlugin plugin, Type type)
    {
        super(owner, true);
        initDialog(plugin, type);
    }

    /**
     * マスク解析用ダイアログを生成する。
     * 
     * @param owner ダイアログを表示する {@code Dialog}
     * @param plugin 呼出元となるプラグインオブジェクト
     * @param type マスク解析のタイプ
     */
    public MaskAnalyzeDialog(Dialog owner, AbstractMaskAnalyzePlugin plugin, Type type)
    {
        super(owner, true);
        initDialog(plugin, type);
    }

    /**
     * ダイアログの構築を行う。
     */
    private void initDialog(AbstractMaskAnalyzePlugin plugin, Type type)
    {
        // Set local variable.
        setPlugin(plugin);
        setType(type);

        // Create new instance.
        btnList = new ArrayList<JButton>();
        textList = new ArrayList<JTextField>();
        comboList = new ArrayList<JComboBox>();
        layer1List = new JList();
        layer2List = new JList();
        unionGrp = new ButtonGroup();
        queryCmb = new JComboBox();
        numOpeList = new ArrayList<Operator>();
        numValList = new ArrayList<Double>();
        // Set list.
        layer1List.addListSelectionListener(this);
        layer2List.addListSelectionListener(this);
        layer1List.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        layer2List.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

        // Create new instance.
        JPanel itemPane = new JPanel();
        JPanel buttonPane = new JPanel();
        // Set Border.
        itemPane.setBorder(BorderFactory.createTitledBorder(rmgr.getString(type + ".border.title")));
        switch (type)
        {
            case DEFAULT:
                setDefaultItemPane(itemPane);
                break;
            case LOGICAL:
                setLogicalItemPane(itemPane);
                break;
            case NUMERIC:
                setNumericItemPane(itemPane);
                break;
        }

        // Add Component. [buttonPane]
        JButton btn;
        btn = getButton(type + ".button", "exec");
        buttonPane.add(btn);
        btnList.add(btn);
        btn = getButton("button", "cancel");
        buttonPane.add(btn);
        setButtonEnabled(false);

        // Set Title.
        setTitle(rmgr.getString(type + ".dialog.title"));
        // Set Layout.
        getContentPane().setLayout(new BorderLayout());
        // Add Components.
        getContentPane().add(itemPane, BorderLayout.CENTER);
        getContentPane().add(buttonPane, BorderLayout.SOUTH);
        pack();
        setResizable(false);
        setLocationRelativeTo(getOwner());
    }

    /**
     * 呼出元となるプラグインオブジェクトを取得する。
     */
    private AbstractMaskAnalyzePlugin getPlugin()
    {
        return plugin;
    }

    /**
     * 呼出元となるプラグインオブジェクトを設定する。
     */
    private void setPlugin(AbstractMaskAnalyzePlugin plugin)
    {
        this.plugin = plugin;
    }

    /**
     * マスク解析タイプを取得する。
     */
    private Type getAnalyzeType()
    {
        return dialogType;
    }

    /**
     * リソースバンドルよりマスク解析タイプに応じた文字列を取得する。
     */
    private String getTypeString(String keySuffix)
    {
        return rmgr.getString(getAnalyzeType() + "." + keySuffix);
    }

    /**
     * マスク解析タイプを取得する。
     */
    private void setType(Type type)
    {
        dialogType = type;
    }

    /**
     * マスクレコード抽出用ダイアログパネルを構築する。
     */
    private void setDefaultItemPane(JPanel pane)
    {
        // Set component.
        layer1List.setListData(getPlugin().getNomalLayerDatas());
        layer2List.setListData(getPlugin().getMaskLayerDatas());

        // Create new instance.
        JPanel layerPane = new JPanel();
        JPanel radioPane = new JPanel();
        JRadioButton radioBtn;
        layerPane.setLayout(new FlowLayout());
        radioPane.setLayout(new FlowLayout());
        layerPane.add(getLayerScroll(layer1List, getTypeString("item.layer1.label")));
        layerPane.add(getLayerScroll(layer2List, getTypeString("item.layer2.label")));
        radioPane.add(new JLabel(getTypeString("item.union.label")));
        radioBtn = getRadioButton("item.union", "true");
        radioPane.add(radioBtn);
        unionGrp.add(radioBtn);
        radioBtn = getRadioButton("item.union", "false");
        radioPane.add(radioBtn);
        unionGrp.add(radioBtn);

        // Add Components.
        pane.setLayout(new BoxLayout(pane, BoxLayout.Y_AXIS));
        pane.add(layerPane);
        pane.add(radioPane);
    }

    /**
     * マスク論理演算用ダイアログパネルを構築する。
     */
    private void setLogicalItemPane(JPanel pane)
    {
        // Set component.
        layer1List.setListData(getPlugin().getMaskLayerDatas());
        layer2List.setListData(getPlugin().getMaskLayerDatas());

        // Add Components.
        pane.setLayout(new FlowLayout());
        pane.add(getLayerScroll(layer1List, getTypeString("item.layer1.label")));
        queryCmb.addItemListener(this);
        pane.add(getQueryCombo(queryCmb));
        pane.add(getLayerScroll(layer2List, getTypeString("item.layer2.label")));
    }

    /**
     * 数値マスク生成用ダイアログパネルを構築する。
     */
    private void setNumericItemPane(JPanel pane)
    {
        // Set component.
        layer1List.setListData(getPlugin().getNomalLayerDatas());

        // Create new instance.
        JPanel layerPane = getLayerScroll(layer1List, null);
        JPanel numPane = new JPanel();
        numPane.setLayout(new BoxLayout(numPane, BoxLayout.Y_AXIS));

        for (int i = 0; i < NUMERIC_ITEM_SIZE; i++)
        {
            comboList.add(getOperatorCombo());
            textList.add(new JTextField(8));
            JPanel p = new JPanel();
            p.setLayout(new FlowLayout(FlowLayout.LEFT));
            p.add(comboList.get(i));
            p.add(textList.get(i));
            if ((i + 1) != NUMERIC_ITEM_SIZE)
            {
                p.add(new JLabel(getTypeString("item.num.query.name")));
            }
            numPane.add(p);
        }

        // Add Components.
        pane.setLayout(new GridBagLayout());
        GridBagConstraints c = new GridBagConstraints();
        c.gridx = c.gridy = 0;
        c.ipadx = c.ipady = 2;
        c.anchor = GridBagConstraints.EAST;
        pane.add(new JLabel(getTypeString("item.layer1.label")), c);
        c.gridy++;
        pane.add(new JLabel(getTypeString("item.num.label")), c);
        c.gridx++;
        c.gridy = 0;
        c.anchor = GridBagConstraints.WEST;
        pane.add(layerPane, c);
        c.gridy++;
        pane.add(numPane, c);
    }

    /**
     * レイヤ選択用 {@code JPanel} を取得する。
     */
    private JPanel getLayerScroll(JList list, String label)
    {
        // Create new instance.
        JScrollPane layerScrll = new JScrollPane();
        JPanel scrollPane = new JPanel();
        // Set scroll panel.
        layerScrll.getViewport().add(list);
        layerScrll.setPreferredSize(new Dimension(rmgr.getInt("item.layer.width"), rmgr.getInt("item.layer.height")));
        if (label != null)
        {
            TitledBorder border = BorderFactory.createTitledBorder(BorderFactory.createEmptyBorder(), label);
            border.setTitleJustification(TitledBorder.CENTER);
            border.setTitleColor(Color.BLACK);
            scrollPane.setBorder(border);
        }
        scrollPane.add(layerScrll);
        return scrollPane;
    }

    /**
     * 関係演算子用 {@code JComboBox} を取得する。
     */
    private JComboBox getOperatorCombo()
    {
        JComboBox combo = new JComboBox();
        for (Operator o : Operator.values())
        {
            combo.addItem(o);
        }
        Dimension d = combo.getPreferredSize();
        d.width += rmgr.getInt("item.combo.width.inc");
        combo.setPreferredSize(d);
        return combo;
    }

    /**
     * 論理演算子用 {@code JComboBox} を取得する。
     */
    private JComboBox getQueryCombo(JComboBox combo)
    {
        if (combo == null)
        {
            combo = new JComboBox();
        }
        for (Query q : Query.values())
        {
            combo.addItem(q);
        }
        Dimension d = combo.getPreferredSize();
        d.width += rmgr.getInt("item.combo.width.inc");
        combo.setPreferredSize(d);
        return combo;
    }

    /**
     * 指定した条件で {@code JRadioButton} を生成する。
     */
    private JRadioButton getRadioButton(String keyPrefix, String cmd)
    {
        JRadioButton btn = new JRadioButton();
        StringBuffer buf = new StringBuffer(keyPrefix);
        buf.append(".").append(cmd).append(".name");
        btn.setText(getTypeString(buf.toString()));
        btn.setActionCommand(cmd);
        btn.setSelected(Boolean.getBoolean(cmd));
        btn.setSelected(getTypeString(keyPrefix + ".selected").equals(cmd));
        return btn;
    }

    /**
     * 指定した条件で {@code JButton} を生成する。
     */
    private JButton getButton(String keyPrefix, String cmd)
    {
        JButton btn = new JButton();
        StringBuffer buf = new StringBuffer(keyPrefix);
        buf.append(".").append(cmd);
        btn.setText(rmgr.getString(buf.toString()));
        btn.setActionCommand(cmd);
        btn.addActionListener(this);
        return btn;
    }

    /**
     * アクションが発生すると呼び出される。
     */
    public void actionPerformed(ActionEvent evt)
    {
        boolean b = false;
        if (!(b = "cancel".equals(evt.getActionCommand())))
        {
            int count = 0;
            switch (getAnalyzeType())
            {
                case DEFAULT:
                    count = analizeDefaultMask();
                    break;
                case LOGICAL:
                    count = analizeLogicalMask();
                    break;
                case NUMERIC:
                    count = analizeNumericMask();
                    break;
            }
            if (count == 0)
            {
                JOptionPane.showMessageDialog(this, getTypeString("norecord.message"), getTypeString("dialog.title"), JOptionPane.INFORMATION_MESSAGE);
            }
            b = (count > 0);
        }
        if (b)
        {
            dispose();
        }
    }

    /**
     * マスクレコード抽出を行う。
     */
    private int analizeDefaultMask()
    {
        // Get parameter.
        LayerData data1 = getLayerData(layer1List);
        LayerData data2 = getLayerData(layer2List);
        Boolean b = false;
        if (unionGrp.getSelection() != null)
        {
            b = Boolean.valueOf(unionGrp.getSelection().getActionCommand());
        }
        // Execute analyze.
        return getPlugin().analyze(data1, data2, b);
    }

    /**
     * マスク論理演算を行う。:
     */
    private int analizeLogicalMask()
    {
        // Get parameter.
        LayerData data1 = getLayerData(layer1List);
        LayerData data2 = getLayerData(layer2List);
        Query query = (Query) queryCmb.getSelectedItem();
        // Execute analyze.
        return getPlugin().analyze(data1, data2, query);
    }

    /**
     * 数値マスク生成を行う。
     */
    private int analizeNumericMask()
    {
        int count = -1;

        // Validate text.
        List<TextValidator> list = new ArrayList<TextValidator>();
        for (JTextField txt : textList)
        {
            list.add(new NUMBER(txt, "#.#"));
        }
        if (TextValidator.validate(list))
        {
            if (setNumericParameter())
            {
                // Get parameter.
                LayerData data = getLayerData(layer1List);
                Operator[] opes = numOpeList.toArray(new Operator[0]);
                Double[] vals = numValList.toArray(new Double[0]);
                // Execute analyze.
                count = getPlugin().analyze(data, opes, vals);
            }
            else
            {
                // Validate text.
                if (!textList.isEmpty())
                {
                    (new REQUIRED(textList.get(0))).validate();
                }
            }
        }
        return count;
    }

    /**
     * 解析対象となるレイヤを取得する。
     */
    private LayerData getLayerData(JList list)
    {
        LayerData data = null;
        if (!list.isSelectionEmpty())
        {
            Object o = list.getSelectedValue();
            if (o instanceof LayerData)
            {
                data = ((LayerData) o);
            }
        }
        return data;
    }

    /**
     * 数値マスク生成に使用するパラメータを設定する。
     */
    private boolean setNumericParameter()
    {
        numOpeList.clear();
        numValList.clear();
        for (int i = 0; i < comboList.size() && i < textList.size(); i++)
        {
            String s = textList.get(i).getText();
            if (s != null && s.trim().length() > 0)
            {
                try
                {
                    numValList.add(Double.parseDouble(s));
                    numOpeList.add((Operator) (comboList.get(i)).getSelectedItem());
                }
                catch (Exception e)
                {
                    // Log [warn]
                    log.warn("Not numeric value. [" + s + "]");
                }
            }
        }
        return !numValList.isEmpty();
    }

    /**
     * 選択範囲内の値が変更されたときに呼び出される。
     */
    public void valueChanged(ListSelectionEvent evt)
    {
        if (evt.getValueIsAdjusting())
        {
            Object o = evt.getSource();
            if (o instanceof JList)
            {
                listValueChanged();
            }
        }
    }

    /**
     * ユーザによって項目が選択または選択解除されたときに呼び出される。
     */
    public void itemStateChanged(ItemEvent evt)
    {
        if (evt.getStateChange() == ItemEvent.SELECTED)
        {
            Object o = evt.getItem();
            if (o instanceof Query)
            {
                boolean b = !o.equals(Query.NOT);
                if (b != layer2List.isEnabled())
                {
                    layer2List.setEnabled(b);
                    if (!b)
                    {
                        layer2List.clearSelection();
                    }
                    listValueChanged();
                }
            }
        }
    }

    /**
     * レイヤの選択状態に応じて {@code JButton} の状態を変更する。
     */
    private synchronized void listValueChanged()
    {
        boolean b = !layer1List.isSelectionEmpty();
        if (layer2List.isValid() && layer2List.isEnabled())
        {
            b &= !layer2List.isSelectionEmpty();
        }
        setButtonEnabled(b);
    }

    /**
     * {@code JButton} の使用可／不可を設定する。
     */
    private void setButtonEnabled(boolean b)
    {
        for (JButton button : btnList)
        {
            button.setEnabled(b);
        }
    }
}
