/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.analyzer;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.swing.JDialog;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.Event;
import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.data.RecordException;
import jp.nihu.gttime.plugin.LayerData;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.plugin.analyzer.AnalyzeConstants.Operator;

/**
 * 数値マスク生成用プラグインクラス
 */
public class NumericMaskAnalyzePlugin extends AbstractMaskAnalyzePlugin
{

    private static PluginResourceManager rmgr = PluginResourceManager.getManager(NumericMaskAnalyzePlugin.class, T2Map.LOCALE);

    private static final String EVENT_NUMBER_NAME = rmgr.getString("event.number.name");

    private RecordWithoutNullComparator comparator = new RecordWithoutNullComparator();

    /**
     * 数値マスク生成用ダイアログを取得する。
     */
    @Override
    protected JDialog getDialog()
    {
        return new MaskAnalyzeDialog(getT2Map(), this, MaskAnalyzeDialog.Type.NUMERIC);
    }

    /**
     * 指定したオブジェクトをもとにマスク解析を実行する。
     */
    @Override
    public int analyze(Object... o)
    {
        LayerData layer = null;
        Operator[] operators = null;
        Double[] values = null;
        if (o.length >= 3)
        {
            if (o[0] instanceof LayerData)
            {
                layer = (LayerData) o[0];
            }
            if (o[1] instanceof Operator[])
            {
                operators = (Operator[]) o[1];
            }
            if (o[2] instanceof Double[])
            {
                values = (Double[]) o[2];
            }
        }
        if (layer != null && operators != null && values != null)
        {
            return analyzeMask(layer, operators, values);
        }
        throw new IllegalArgumentException("Argument mismatched analyzeMask(LayerData, Operator[], Double[]) .");
    }

    /**
     * 数値レイヤよりマスク生成を行う。
     * 
     * @param layer 検索対象となるレイヤ情報
     * @param operators 比較演算子配列
     * @param values 比較する {@code Double} 値配列
     * @return レコード件数
     */
    public int analyzeMask(LayerData layer, Operator[] operators, Double[] values)
    {
        Set<Record> set = new HashSet<Record>();
        if (layer != null && layer.getDataset().hasRecords())
        {
            Record hit = null;
            Record[] records = layer.getDataset().getRecords(layer.getGroup());
            // Sort arrays.
            Arrays.sort(records, comparator);
            for (Record r : records)
            {
                boolean b = false;
                Event[] events = r.getEvents();
                if (events != null)
                {
                    for (Event evt : events)
                    {
                        if (b = compare(evt, operators, values))
                        {
                            break;
                        }
                    }
                }
                if (b)
                {
                    if (hit == null)
                    {
                        hit = new Record();
                        try
                        {
                            hit.setStartDatetime(r.getStartDatetime());
                        }
                        catch (RecordException e)
                        {
                            e.printStackTrace();
                        }
                        set.add(hit);
                    }
                    try
                    {
                        hit.setEndDatetime(comparator.max(hit.getEndDatetime(), r.getEndDatetime()));
                    }
                    catch (RecordException e)
                    {

                        e.printStackTrace();
                    }
                }
                else
                {
                    hit = null;
                }
            }
            // Create new instance.
            LogicalMaskAnalyzePlugin lm = new LogicalMaskAnalyzePlugin();
            createDataset(getTitle(layer, operators, values), lm.analyzeOrMask(set.toArray(new Record[0]), new Record[0]));
        }
        return set.size();
    }

    /**
     * {@code Event} オブジェクトが指定した条件に合致するかを判定する。
     * 
     * @param evt 比較対象となる {@code Event} オブジェクト
     * @param operators 比較演算子配列
     * @param values 比較する {@code Double} 値配列
     * @return 条件に合致刷る場合 {@code true}
     */
    private boolean compare(Event evt, Operator[] operators, Double[] values)
    {
        boolean b = false;
        if (EVENT_NUMBER_NAME.equalsIgnoreCase(evt.getName()))
        {
            try
            {
                double d = Double.parseDouble(evt.getData());
                for (int i = 0; i < operators.length && i < values.length; i++)
                {
                    b = AnalyzeConstants.compare(operators[i], d, values[i]);
                    if (!b)
                    {
                        break;
                    }
                }
            }
            catch (Exception e)
            {
                // Ignore.
            }
        }
        return b;
    }

    /**
     * レイヤタイトルを取得する。
     */
    private String getTitle(LayerData layer, Operator[] operators, Double[] values)
    {
        // Create new instance.
        StringBuffer buf = new StringBuffer();
        for (int i = 0; i < operators.length && i < values.length; i++)
        {
            String sep = "";
            if (buf.length() != 0)
            {
                sep = rmgr.getString("title.format.separator");
            }
            buf.append(rmgr.format("title.format.part", operators[i], values[i], sep));
        }
        return rmgr.format("title.format", layer.getName(), buf.toString());
    }

    /**
     * 現在アプリケーション側で表示しているレイヤ情報より、マスクレイヤを除く 数値データを保持する {@code LayerData}
     * オブジェクト配列を取得する。
     */
    @Override
    public LayerData[] getNomalLayerDatas()
    {
        List<LayerData> list = new ArrayList<LayerData>();
        for (LayerData layer : super.getNomalLayerDatas())
        {
            boolean b = false;
            Record[] rs = layer.getDataset().getRecords();
            for (Record r : rs)
            {
                Event[] evts = r.getEvents();
                if (evts != null)
                {
                    for (Event evt : evts)
                    {
                        String s = evt.getName();
                        if (b = EVENT_NUMBER_NAME.equalsIgnoreCase(s))
                        {
                            break;
                        }
                    }
                }
                if (b)
                    break;
            }
            if (b)
                list.add(layer);
        }
        return list.toArray(new LayerData[0]);
    }
}
