/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.analyzer;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;

import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.data.RecordException;

/**
 * {@code Record} オブジェクトのコピーを生成するユーティリティクラス。 尚、本クラスで複製される {@code Record}
 * オブジェクトは「シャローコピー」 であり、「ディープコピー」ではない。
 */
public class RecordCloner
{

    private RecordCloner()
    {
    }

    /**
     * {@code Record} オブジェクトのコピーを生成する。 このメソッドではオブジェクトの「シャローコピー」を生成する。
     * 
     * @param r 複製の対象となる {@code Record} オブジェクト
     * @return 指定したオブジェクトの複製
     */
    public static Record clone(Record r)
    {
        Record src = null;
        if (r != null)
        {
            // Create new instance.
            src = new Record();
            src.setId(r.getId());
            src.setGroup(r.getGroup());
            src.setLevel(r.getLevel());
            src.setHit(src.isHit());
            try
            {
                src.setStartDatetime(r.getStartDatetime());
                src.setEndDatetime(r.getEndDatetime());
            }
            catch (RecordException e)
            {
                // never catch
                e.printStackTrace();
            }
            src.setEvents(r.getEvents());
            src.setLocation(r.getLocation());
            src.setNode(r.getNode().cloneNode(true));
        }
        return src;
    }

    /**
     * {@code Record} オブジェクトのコピーを生成する。
     * 
     * @param records 複製の対象となる {@code Record} オブジェクト配列
     * @return 複製された{@code Record} オブジェクト配列
     */
    public static Record[] clone(Record[] records)
    {
        return (records != null) ? clone(Arrays.asList(records)) : null;
    }

    /**
     * {@code Record} オブジェクトのコピーを生成する。
     * 
     * @param c {@code Record} オブジェクトを要素にもつコレクション
     * @return 複製された{@code Record} オブジェクト配列
     */
    public static Record[] clone(Collection<Record> c)
    {
        List<Record> list = new ArrayList<Record>();
        if (c != null)
        {
            for (Record r : c)
            {
                list.add(clone(r));
            }
        }
        return list.toArray(new Record[0]);
    }
}