/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.analyzer;

import java.io.Serializable;
import java.math.BigDecimal;
import java.util.Comparator;

import jp.nihu.gttime.data.JulianDate;
import jp.nihu.gttime.data.Record;

/**
 * {@code Record} オブジェクトのコレクションで「全体順序付け」を行う比較関数。
 */
@SuppressWarnings("serial")
public class RecordComparator implements Comparator<Record>, Serializable
{

    /**
     * 順序付けのために 2 つの引数を比較する。 最初の引数が 2 番目の引数より小さい場合は負の整数、両方が等しい場合は {@code 0}、最初の引数が
     * 2 番目の引数より大きい場合は正の整数を返す。
     * 
     * @param r1 比較対象の最初のオブジェクト
     * @param r2 比較対象の 2 番目のオブジェクト
     * @return 最初の引数の開始日時が 2 番目の引数の開始日時より小さい場合、 または開始日時が等しく最初の引数の終了日時が 2 番目の引数の
     *         終了日時より小さい場合は負の整数、両方の開始日時／終了日時が 等しい場合は {@code 0}、最初の引数の開始日時が 2
     *         番目の引数の 開始日時より大きい場合、または開始日時が等しく最初の引数の終了 日時が 2
     *         番目の引数の終了日時より大きい場合は正の整数
     */
    public int compare(Record r1, Record r2)
    {
        int c = 0;
        if (notNull(r1, r2))
        {
            if (r1.getStartDatetime() == null)
            {
                c = -1;
            }
            else if (r2.getStartDatetime() == null)
            {
                c = 1;
            }
            else
            {
                c = compare(r1.getStartDatetime(), r2.getStartDatetime());
            }
            if (c == 0)
            {
                c = compare(r1.getEndDatetime(), r2.getEndDatetime());
            }
        }
        else
        {
            c = cmpNull(r1, r2);
        }
        return c;
    }

    /**
     * 指定した 2 つの引数を比較する。 最初の引数が 2 番目の引数より小さい場合は負の整数、両方が等しい場合は {@code 0}、最初の引数が
     * 2 番目の引数より大きい場合は正の整数を返す。
     * 
     * @param d1 比較対象の最初のオブジェクト
     * @param d2 比較対象の 2 番目のオブジェクト
     * @return 最初の引数の日時が 2 番目の引数の日時より小さい場合は負の整数、 両方の日時が等しい場合は {@code 0}、最初の引数の日時が
     *         2 番目 の引数の日時より大きい場合は正の整数
     */
    public int compare(JulianDate d1, JulianDate d2)
    {
        return (notNull(d1, d2)) ? compare(d1.getDatetime(), d2.getDatetime()) : cmpNull(d1, d2);
    }

    /**
     * 指定した 2 つの引数を比較する。 最初の引数が 2 番目の引数より小さい場合は負の整数、両方が等しい場合は {@code 0}、最初の引数が
     * 2 番目の引数より大きい場合は正の整数を返す。
     */
    private int compare(BigDecimal v1, BigDecimal v2)
    {
        return (notNull(v1, v2)) ? v1.compareTo(v2) : cmpNull(v1, v2);
    }

    /**
     * 指定した 2 つの引数を比較する。 本メソッドは引数のうち少なくとも一方が {@code null} の場合に呼び 出される。
     * 
     * @return 最初の引数が {@code null} の場合は負の整数、両方が{@code null} の場合は {@code 0}、 2
     *         番目の引数が {@code null} の場合は正の 整数
     */
    private int cmpNull(Record r1, Record r2)
    {
        return (r1 == null) ? ((r2 == null) ? 0 : -1) : 1;
    }

    private int cmpNull(BigDecimal v1, BigDecimal v2)
    {
        return (v1 == null) ? ((v2 == null) ? 0 : -1) : 1;
    }

    private int cmpNull(JulianDate d1, JulianDate d2)
    {
        return (d1 == null) ? ((d2 == null) ? 0 : 1) : -1;
    }

    /**
     * 指定したオブジェクト値が {@code null} かを判定する。
     * 
     * @return 最初の引数および 2 番目の引数の値が {@code null} でない 場合 {@code true}
     */
    private boolean notNull(Object o1, Object o2)
    {
        return (o1 != null && o2 != null);
    }

    /**
     * 2 つの {@code Record} オブジェクトのうち大きいほうを返却する。
     */
    public Record max(Record r1, Record r2)
    {
        return (compare(r1, r2) >= 0) ? r1 : r2;
    }

    /**
     * 2 つの {@code Record} オブジェクトのうち小さいほうを返却する。
     */
    public Record min(Record r1, Record r2)
    {
        return (compare(r1, r2) <= 0) ? r1 : r2;
    }

    /**
     * 2 つの {@code JulianDate} オブジェクトのうち大きいほうを返却する。
     */
    public JulianDate max(JulianDate d1, JulianDate d2)
    {
        return (compare(d1, d2) >= 0) ? d1 : d2;
    }

    /**
     * 2 つの {@code JulianDate} オブジェクトのうち小さいほうを返却する。
     */
    public JulianDate min(JulianDate d1, JulianDate d2)
    {
        return (compare(d1, d2) <= 0) ? d1 : d2;
    }
}
