/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.analyzer.searchrecord;

import java.awt.Shape;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.math.BigDecimal;
import java.net.MalformedURLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.TimeMapGeograhicCoordinates;
import jp.nihu.gttime.data.DatasetException;
import jp.nihu.gttime.data.DatasetGtm;
import jp.nihu.gttime.data.DatasetIF;
import jp.nihu.gttime.data.Event;
import jp.nihu.gttime.data.JulianDate;
import jp.nihu.gttime.data.Location;
import jp.nihu.gttime.data.ProjectIF;
import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.plugin.LayerData;
import jp.nihu.gttime.plugin.LayerPlugin;
import jp.nihu.gttime.plugin.PluginManager;
import jp.nihu.gttime.plugin.PluginManagerException;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.plugin.analyzer.AnalyzeConstants;
import jp.nihu.gttime.plugin.analyzer.AnalyzeConstants.Operator;
import jp.nihu.gttime.plugin.analyzer.DatasetList;
import jp.nihu.gttime.plugin.analyzer.LogicalMaskAnalyzePlugin;
import jp.nihu.gttime.plugin.analyzer.RecordCloner;
import jp.nihu.gttime.plugin.analyzer.RecordComparator;
import jp.nihu.gttime.utility.Utilities;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

/**
 * 検索機能を実装するクラスです。
 */
public class SearchService
{

    private static Log logger = LogFactory.getLog(SearchService.class);

    // リソースバンドルの管理
    protected static PluginResourceManager rmgr = PluginResourceManager.getManager(SearchRecordPlugin.class, T2Map.LOCALE);

    private List<DatasetIF> resultList = new ArrayList<DatasetIF>();

    private AffineTransform geoAf = AffineTransform.getScaleInstance(1d, -1d);

    /**
     * 検索結果となるした {@code Dataset} オブジェクトを取得する。
     * 
     * @return {@code Dataset} オブジェクト配列
     */
    public DatasetIF[] getResult()
    {
        // int ptype = T2Map.getProjectType();
        // if (ptype == 0)
        // {
        // return (!resultList.isEmpty( ))? resultList.toArray(new
        // DatasetTmm[0]): null;
        // }
        // else if (ptype == 1)
        // {
        return (!resultList.isEmpty()) ? resultList.toArray(new DatasetGtm[0]) : null;
        // }
        // else
        // {
        // return null;
        // }
    }

    private void setResult(Collection<DatasetIF> c)
    {
        // Initialized.
        resultList.clear();
        if (c != null && !c.isEmpty())
        {
            resultList.addAll(c);
        }
    }

    /**
     * 検索処理処理を実行する。
     * 
     * @param query 検索条件
     * @return 検索結果件数
     */
    public int search(Query query)
    {
        // 検索にヒットしたRecordオブジェクト
        List<Object[]> list = new ArrayList<Object[]>();//Record, String

        // Get search area.
        Rectangle2D[] rects = getSearchArea(query);
        // Get LayerData.
        LayerData[] layers = query.getLayers();
        if (layers == null)
        {
        	return 0;
        }
        for (LayerData layer : layers)
        {
            Record[] records = layer.getDataset().getRecords(layer.getGroup());
            if (records == null)
            {
            	break;
            }
            String uri = null;
			try {
				uri = Utilities.getUri(layer.getDataset().getRdfBase());
			} catch (MalformedURLException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
            for (Record r : records)
            {
                // ハイライト初期化
                r.setHit(false);
                // 検索実行
                if (checkSpace(r, rects) && checkTime(r, query) && checkEvent(r, query))
                {
                    list.add(new Object[]{(query.isHighlight() ? r : RecordCloner.clone(r)), uri});
                }
            }
        }
        
        if (query.isHighlight())
        {
            // ハイライト選択
            for (Object[] r : list)
            {
                ((Record)r[0]).setHit(true);
            }
        }
        else
        {
            // 新規レイヤ選択
            List<DatasetIF> dataList = new ArrayList<DatasetIF>();
            List<Record> results = new ArrayList<Record>();
            if (0 < list.size()) {
            	DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            	DocumentBuilder builder = null;
				try {
					builder = dbf.newDocumentBuilder();
				} catch (ParserConfigurationException e) {
					// TODO Auto-generated catch block
					e.printStackTrace();
				}
            	Document doc = builder.newDocument();
            	Element root = doc.createElementNS("http://www.nihu.jp/ns/spacetime/2006/11", "records");
//            	root.setAttribute("xmlns:owl", "http://www.w3.org/2002/07/owl#");
//            	root.setAttribute("xmlns:rdf", "http://www.w3.org/1999/02/22-rdf-syntax-ns#");
//            	root.setAttribute("xmlns:dc", "http://www.dc.org");
            	doc.appendChild(root);
            	
                for (Object[] os : list) {
                	Record r = (Record)os[0];
                	results.add(r);
                	Node n = r.getNode();
                	//属性をチェックしてまだ新しいルート要素にない属性があったら新しいルート要素に追加する
                	Element root0 = n.getOwnerDocument().getDocumentElement();
                	r.setAttributeToRootElement(root0, root);
                	n = doc.adoptNode(n);
                	r.setNode(n);
                	root.appendChild(n);
                	r.createRdfSameAsEvent((String)os[1]);
                }
            }
            dataList.add(createDataset(query.getQuery(), query.getCreateLayer(), results.toArray(new Record[0])));
            // マスクを作成する場合
            if (query.isCreateMask())
            {
                dataList.add(createMask(query.getQuery(), results.toArray(new Record[0])));
                // Log [debug]
                logger.debug("query.isCreated == true");
            }
            // 検索結果設定
            setResult(dataList);
        }
        return list.size();
    }

    /**
     * レコードが指定空間範囲に含まれるかをチェックします。
     * 
     * @param record 検索対象となる Record オブジェクト
     * @param rects 検索範囲を表す矩形
     * @return 合致するか
     */
    private boolean checkSpace(Record record, Rectangle2D[] rects)
    {
        boolean b = false;
        if (b = (rects != null))
        {
            Shape s = getRecordArea(record);
            if (s != null)
            {
                for (Rectangle2D r : rects)
                {
                    // 検索範囲が矩形の場合
                    if (b = s.intersects(r))
                    {
                        break;
                    }
                    else
                    {
                        // 検索範囲が点の場合
                        if (b = s.contains(r.getX(), r.getY()))
                        {
                            break;
                        }
                    }
                }
            }
        }
        return b;
    }

    private Rectangle2D[] getSearchArea(Query query)
    {
        List<Rectangle2D> list = new ArrayList<Rectangle2D>();
        Point2D p1 = query.getPoint1();
        Point2D p2 = query.getPoint2();
        if (p1 != null || p2 != null)
        {
            // 変換
            p1 = geoAf.transform(((p1 != null) ? p1 : p2), new Point2D.Double());
            p2 = geoAf.transform(((p2 != null) ? p2 : p1), new Point2D.Double());
            // 比較
            boolean b1 = (p1.getX() <= p2.getX());
            boolean b2 = (p1.getY() <= p2.getY());
            Rectangle2D r = new Rectangle2D.Double(p1.getX(), p1.getY(), 0, 0);
            if (!b1 || !b2)
            {
                r.add(((b1) ? p2.getX() : 180d), ((b2) ? p2.getY() : 90d));
                list.add(r);
                r = new Rectangle2D.Double(((b1) ? p1.getX() : -180d), ((b2) ? p1.getY() : -90d), 0, 0);
            }
            r.add(p2);
            list.add(r);
            if (!b1 && !b2)
            {
                r = new Rectangle2D.Double(p1.getX(), p2.getY(), 0, 0);
                r.add(180d, -90d);
                list.add(r);
                r = new Rectangle2D.Double(p2.getX(), p1.getY(), 0, 0);
                r.add(-180d, 90d);
                list.add(r);
            }
        }
        return list.toArray(new Rectangle2D[0]);
    }

    private Shape getRecordArea(Record record)
    {
        // Shape オブジェクトは暫定的に Rectangle2D.Double
        Rectangle2D r2d = null;
        Location loc = record.getLocation();
        if (loc != null)
        {
            TimeMapGeograhicCoordinates[] tmgs = loc.getPointsLongLaties();
            if (tmgs != null)
            {
                for (TimeMapGeograhicCoordinates tmg : tmgs)
                {
                    // 緯度・経度情報取得
                    Point2D p = new Point2D.Double(tmg.getLongitude(), tmg.getLatitude());
                    // 座標変換
                    p = geoAf.transform(p, new Point2D.Double());
                    if (r2d == null)
                    {
                        r2d = new Rectangle2D.Double(p.getX(), p.getY(), 0, 0);
                    }
                    else
                    {
                        r2d.add(p);
                    }
                }
            }
        }
        // Location 情報が点であった場合の処置
        if (r2d != null)
        {
            if (r2d.getWidth() == 0)
            {
                r2d.add(r2d.getX() + 0.1d, r2d.getY());
            }
            if (r2d.getHeight() == 0)
            {
                r2d.add(r2d.getX(), r2d.getY() + 0.1d);
            }
        }
        return r2d;
    }

    /**
     * レコードが指定時間範囲に含まれるかをチェックします。
     * 
     * @param record 検索対象となる Record オブジェクト
     * @param query 検索条件
     * @return 合致するか
     */
    private boolean checkTime(Record record, Query query)
    {
        boolean b = false;
        Iterator<BigDecimal[]> it = query.getTime().iterator();
        while (it.hasNext())
        {
            BigDecimal[] x = it.next();
            BigDecimal qs = x[0];
            BigDecimal qe = x[1];
            if (!(b = (qs == null && qe == null)))
            {
                JulianDate rs = record.getStartDatetime();
                JulianDate re = record.getEndDatetime();
                // if (rs != null && re != null) {
                // 値補正
                qs = (qs != null) ? qs : Query.DATE_MIN;
                qe = (qe != null) ? qe : Query.DATE_MAX;

                // 開区間
                // if (qs.compareTo(rs.getDatetime( )) <= 0) {
                // // パターン 2 or 3 の比較
                // b = (qe.compareTo(rs.getDatetime( )) > 0);
                // } else {
                // // パターン 1 or 4 の比較
                // b = (qs.compareTo(re.getDatetime( )) < 0);
                // }

                // 始まりが閉じている半開区間
                if (rs == null)
                {

                }

                // log.debug("-----------------------------------------------");
                // Event[] es = record.getEvents();
                // for (int i = 0; i < es.length; i++)
                // {
                // if (es[i].getName().equals("title"))
                // {
                // log.debug(es[i].getData());
                // }
                // }

                // log.debug("X --> " + (record.isStartEdgeOpen() ?
                // qs.compareTo(rs.getDatetime( )) < 0 :
                // qs.compareTo(rs.getDatetime( )) <= 0) + " :: " +
                // record.isStartEdgeOpen() + " ? " +
                // (qs.compareTo(rs.getDatetime( )) < 0) + " : " +
                // (qs.compareTo(rs.getDatetime( )) <= 0));
                // log.debug(qs + " :: " + rs.getDatetime( ));
                if (rs != null && (record.isStartEdgeOpen() ? qs.compareTo(rs.getDatetime()) < 0 : qs.compareTo(rs.getDatetime()) <= 0))
                {
                    // パターン 2 or 3 の比較
                    b = (record.isStartEdgeOpen() ? qe.compareTo(rs.getDatetime()) > 0 : qe.compareTo(rs.getDatetime()) >= 0);
                    // log.debug("A --> " + b + " :: " +
                    // record.isStartEdgeOpen() + " ? " +
                    // (qe.compareTo(rs.getDatetime( )) > 0) + " : " +
                    // (qe.compareTo(rs.getDatetime( )) >= 0));
                    // log.debug(qe + " :: " + rs.getDatetime( ));
                }
                else
                {
                    // パターン 1 or 4 の比較
                    if (re != null)
                    {
                        b = (record.isEndEdgeOpen() ? qs.compareTo(re.getDatetime()) < 0 : qs.compareTo(re.getDatetime()) <= 0);
                        // log.debug("B --> " + b + " :: " +
                        // record.isEndEdgeOpen() + " ? " +
                        // (qs.compareTo(re.getDatetime( )) < 0) + " : " +
                        // (qs.compareTo(re.getDatetime( )) <= 0));
                        // log.debug(qs + " :: " + re.getDatetime( ));
                    }
                    else
                    {
                        b = true;
                    }
                }
                // }
            }
            if (b)
            {
                break;
            }
        }
        return b;
    }

    /**
     * 文字列検索クエリとイベントが合致するかチェックします。
     * 
     * @param record
     * @param query
     * @return 合致するか
     */
    private boolean checkEvent(Record record, Query query)
    {
        Event[] events = record.getEvents();
        if (events == null)
            return false;

        return (checkHitString(events, query.getEventQuery1()) && checkHitString(events, query.getEventQuery2()) && checkHitString(events, query.getEventQuery3()));
    }

    private boolean checkHitString(Event[] events, EventQuery eq)
    {
        boolean b = false;
        if (events != null)
        {
            for (Event evt : events)
            {
                if (b = checkHitString(evt, eq))
                {
                    break;
                }
            }
        }
        return b;
    }

    /**
     * 文字列検索クエリとイベントが合致するかチェックします。
     * 
     * @param event 検索対象となる Event オブジェクト
     * @param eq 検索条件
     * @return 合致するか
     */
    private boolean checkHitString(Event event, EventQuery eq)
    {
        String eventName = event.getName(); // イベント属性
        String eventData = event.getData(); // イベント情報
        String type = eq.getType(); // 種別
        String key = eq.getKey(); // キー
        Operator ope = eq.getAssign();

        // 入力がないものはtrueとして扱う
        boolean b = (key == null || key.trim().length() == 0);
        // 入力値あり、かつ all または種別とイベント属性が合致
        if (!b && (type.equals("all") || type.equals(eventName)))
        {
            if (ope != null)
            {
                try
                {
                    float num = Float.parseFloat(eventData);
                    float val = Float.parseFloat(key);
                    b = AnalyzeConstants.compare(ope, num, val);
                }
                catch (Exception e)
                {
                    // Ignore.
                }
            }
            else
            {
                Pattern p = Pattern.compile(key);
                Matcher m = p.matcher(eventData);
                b = m.find();
            }
        }
        return b;
    }

    /**
     * 新規レイヤ選択時に指定するレイヤを取得します。
     */
    public static Map<String, String> getLayerNames(SearchRecordPlugin plugin)
    {
        Map<String, String> layerNames = new TreeMap<String, String>();
        PluginManager pm = plugin.getPluginManager();

        ArrayList<Hashtable<String, String>> plugins = pm.getPlugins("LayerPlugin");
        Iterator<Hashtable<String, String>> iter = plugins.iterator();
        while (iter.hasNext())
        {
            Hashtable<String, String> m = iter.next();
            String cname = (String) m.get("class");
            try
            {
                Object obj = pm.newInstance(cname);
                LayerPlugin lyr = (LayerPlugin) obj;
                layerNames.put(lyr.getName(), m.get("name"));
            }
            catch (PluginManagerException e)
            {
                e.printStackTrace();
            }
        }
        return layerNames;
    }

    /**
     * 文字列検索時の属性となるイベント名を取得します。
     * 
     * @param project
     * @return 取得したイベント名
     */
    public static Set<String> getEventNames(ProjectIF project)
    {
        Set<String> attributes = new TreeSet<String>();
        DatasetIF[] sets = project.getDatasets();
        attributes.add("all");
        for (DatasetIF s : sets)
        {
            Record[] rs = s.getRecords();
            for (Record r : rs)
            {
                Event[] es = r.getEvents();
                if (es != null)
                {
                    for (Event e : es)
                    {
                        attributes.add(e.getName());
                    }
                }
            }
        }
        ProjectIF[] ps = project.getProjects();
        if (ps != null)
        {
            for (ProjectIF p : ps)
            {
                attributes.addAll(getEventNames(p));
            }
        }
        return attributes;
    }

    /**
     * 演算子を取得します。
     * 
     * @return 演算子一覧
     */
    public static List<Operator> getNumberAssigns()
    {
        return Arrays.asList(Operator.values());
    }

    /**
     * レイヤ一覧を取得します。
     * 
     * @param project
     * @return レイヤ一覧
     */
    public static LayerData[] getLayerDatas(ProjectIF project)
    {
        return DatasetList.getArray(project);
    }

    /**
     * {@code Dataset} オブジェクトを生成する。
     * 
     * @param title レイヤタイトル
     * @param type レイヤ型
     * @param rs 検索結果 {@code Record} オブジェクト
     */
    private DatasetIF createDataset(String title, String type, Record[] rs)
    {
        DatasetIF dataset = null;
        try
        {
            Arrays.sort(rs, new RecordComparator());
            // int ptype = T2Map.getProjectType();
            // if (ptype == 0)
            // {
            // dataset = new DatasetTmm(title, type, rs);
            // }
            // else if (ptype == 1)
            // {
            dataset = new DatasetGtm(title, type, rs);
            // }
        }
        catch (DatasetException e)
        {
            // Log [error]
            logger.error("Dataset [" + title + ":" + type + "] create error.", e);
        }
        return dataset;
    }

    /**
     * マスクレイヤ用 {@code Dataset} オブジェクトを生成する。
     * 
     * @param title レイヤタイトル
     * @param records 検索結果 {@code Record} オブジェクト
     */
    private DatasetIF createMask(String title, Record[] records)
    {
        // Create new instance.
        LogicalMaskAnalyzePlugin lm = new LogicalMaskAnalyzePlugin();
        return createDataset(rmgr.format("query.format.mask", title), lm.getMaskLayerType(), lm.analyzeOrMask(records, new Record[0]));
    }
}
