/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.io;

import java.awt.Frame;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.concurrent.Callable;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;

import javax.swing.JFileChooser;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.filechooser.FileFilter;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.stream.StreamSource;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.T2MapFileChooser;
import jp.nihu.gttime.T2MapProgressBarDialog;
import jp.nihu.gttime.data.ProjectGts;
import jp.nihu.gttime.data.ProjectIF;
import jp.nihu.gttime.plugin.OpenPlugin;
import jp.nihu.gttime.plugin.OpenPluginException;
import jp.nihu.gttime.plugin.PluginResourceManager;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * ファイルを読み込む実行クラスです。
 */
public class FileOpenPlugin extends OpenPlugin
{
    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(FileOpenPlugin.class);

    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(FileOpenPlugin.class, T2Map.LOCALE);

    /**
     * ファイル選択画面を表示します。
     * 
     * @throws OpenPluginException
     */
    public ProjectIF open() throws OpenPluginException
    {
        ProjectIF project = null;
        UIManager.put("FileChooser.readOnly", Boolean.TRUE);
        JFileChooser chooser = new T2MapFileChooser();
        chooser.setFileFilter(new ProjectFileFilter());
        int returnVal = chooser.showOpenDialog(getParentComponent());
        if (returnVal == JFileChooser.APPROVE_OPTION)
        {
            project = open(chooser.getSelectedFile());
        }
        UIManager.put("FileChooser.readOnly", Boolean.FALSE);
        return project;
    }

    /**
     * ファイルを読み込みます。
     * 
     * @param file 選択したファイル
     * @return Projectオブジェクト
     * @throws OpenPluginException
     */
    public ProjectIF open(File file) throws OpenPluginException
    {
        log.info("open project file: " + file.getName());
        ProjectIF project = null;
        if (file.exists())
        {
            log.info("file exists: " + file.getAbsolutePath());
            Frame frame = (Frame) SwingUtilities.getRoot(getParentComponent());
            String title = rmgr.getString("progress.title");
            String message = rmgr.format("progress.message", file.getAbsolutePath());
            T2MapProgressBarDialog<Object> progress = new T2MapProgressBarDialog<Object>(frame, true, title, message);
            Callable<Object> callable = (Callable<Object>) new LoadProject(file.getPath(), file.getParentFile().getPath());
            FutureTask<Object> f = (FutureTask<Object>) progress.submit(callable);
            progress.setVisible(true);
            try
            {
                project = (ProjectIF) f.get();
            }
            catch (InterruptedException e)
            {
                throw new OpenPluginException(e.getMessage(), e);
            }
            catch (ExecutionException e)
            {
                e.printStackTrace();
                throw new OpenPluginException(e.getMessage(), e);
            }
            catch (CancellationException e)
            {
            }
        }
        else
        {
            log.info("not file exists: " + file.getAbsolutePath());
        }
        return project;
    }

    private class LoadProject implements Callable<Object>
    {
        private String filePath;

        private String parentFilePath;

        public LoadProject(String filePath, String parentFilePath)
        {
            this.filePath = filePath;
            this.parentFilePath = parentFilePath;
        }

        public ProjectIF call() throws Exception
        {
            ProjectIF project = null;
            T2Map.setProjectType(1);
            project = new ProjectGts();
            project.loadProject(FileOpenPlugin.this, filePath, parentFilePath);
            return project;
        }
    }

    /**
     * 親の抽象パス名を取得します。
     * 
     * @param path ファイルのディレクトリ
     * @param base ファイル名
     * @return 絶対パス
     */
    public String getAbsolutePath(String path, String base)
    {
    	File file = new File(path);
		if (!file.isAbsolute()) {
			if (path.startsWith("http://") || path.startsWith("https://")) {
				return path;
			} else {
				if (base.startsWith("http://") || base.startsWith("https://")) {
					try {
						URL url = new URL(new URL(base), path);
						return url.toString();
					} catch (MalformedURLException e) {
						log.fatal(e.getMessage(), e);
					}
					return null;
				} else {
					file = new File(base, path);
					return file.getAbsolutePath();
				}
			}
		} else {
			return file.getAbsolutePath();
		}
    }

    /**
     * ファイルのパスを取得します。
     * 
     * @param path ファイルのディレクトリ
     * @return パス
     */
    public String getBasePath(String path)
    {
    	if (path.startsWith("http://") || path.startsWith("https://")) {
        	return path.substring(0, path.lastIndexOf("/") + 1);
        } else {
        	return new File(path).getParentFile().getPath();
        }
    }

    private InputStream getGTMInputStream(String xmlFile, String xstFile) throws OpenPluginException
    {
        InputStream in = null;
        try
        {
            ByteArrayOutputStream bout = new ByteArrayOutputStream();
            TransformerFactory factory = TransformerFactory.newInstance();
            InputStream inXslt = getClass().getClassLoader().getResourceAsStream(xstFile);
            Transformer transformer = factory.newTransformer(new StreamSource(inXslt));
            File file = new File(xmlFile);
            String fileName = file.getName();
            if (fileName.endsWith("tmm"))
            {
                fileName = fileName.replaceAll(".tmm", ".xml");
                transformer.setParameter("FILE", fileName);
            }
            if (fileName.endsWith("tms"))
            {
                transformer.setParameter("BASE_PATH", file.getParentFile().getAbsolutePath());
            }
            transformer.transform(new StreamSource(new FileInputStream(file)), new StreamResult(bout));
            in = new ByteArrayInputStream(bout.toByteArray());
        }
        catch (TransformerConfigurationException e)
        {
            throw new OpenPluginException(e.getMessage(), e);
        }
        catch (TransformerException e)
        {
            throw new OpenPluginException(e.getMessage(), e);
        }
        catch (FileNotFoundException e)
        {
            throw new OpenPluginException(e.getMessage(), e);
        }
        return in;
    }

    /**
     * tmmファイルを取得します。
     * 
     * @param path パス
     * @return バイト入力ストリーム
     * @throws OpenPluginException
     */
    public InputStream getDatasetInputStream(String path) throws OpenPluginException
    {
        if ((T2Map.getProjectType() == -1 || T2Map.getProjectType() == 1) && path.endsWith("tmm"))
        {
            return getGTMInputStream(path, "tmm2gtm.xsl");
        }
        else
        {
        	return getFileInputStream(path);
        }
    }

    /**
     * tmsファイルを取得します。
     * 
     * @param path パス
     * @return バイト入力ストリーム
     * @throws OpenPluginException
     */
    public InputStream getProjectInputStream(String path) throws OpenPluginException
    {
        if (T2Map.getProjectType() == 1 && path.endsWith("tms"))
        {
            return getGTMInputStream(path, "tms2gts.xsl");
        }
        else
        {
        	return getFileInputStream(path);
        }
    }

    /**
     * レコードファイルを取得します。
     * 
     * @param path パス
     * @return バイト入力ストリーム
     * @throws OpenPluginException
     */
    public InputStream getRecordsInputStream(String path) throws OpenPluginException
    {
        return getFileInputStream(path);
    }

    /**
     * ファイルをバイト入力ストリームで取得します。
     * 
     * @param path パス
     * @return バイト入力ストリーム
     * @throws OpenPluginException
     */
    private InputStream getFileInputStream(String path) throws OpenPluginException
    {
        InputStream is = null;
        if (path.startsWith("http://") || path.startsWith("https://")) {
        	try {
				URL url = new URL(path);
				is = url.openStream();
			} catch (MalformedURLException e) {
				throw new OpenPluginException(e.getMessage(), e);
			} catch (IOException e) {
				throw new OpenPluginException("Failed connect \"" + path + "\". [MESSAGE]" + e.getMessage(), e);
			}
        } else {
            try
            {
                is = new FileInputStream(path);
            }
            catch (FileNotFoundException e)
            {
                throw new OpenPluginException(e.getMessage(), e);
            }
        }
        return is;
    }

    /**
     * Filterを実装するクラスです。
     */
    static class ProjectFileFilter extends FileFilter
    {
        /**
         * 抽象パス名が示すファイルがディレクトリであるかどうかを判定します。
         * 
         * @param file Fileオブジェクト
         */
        public boolean accept(File file)
        {
            if (file.isDirectory())
            {
                return true;
            }
            else
            {
                String extension = getExtension(file.getName()).toLowerCase();
                if (extension != null)
                {
                    if (extension.equals("tms"))
                    {
                        return true;
                    }
                    else if (extension.equals("gts"))
                    {
                        return true;
                    }
                }
            }
            return false;
        }

        /**
         * tmsファイルのdescriptionを取得する。
         * 
         * @return descriptionの文字列値
         */
        public String getDescription()
        {
            return "GTS / ECAI Project file";
        }

        /**
         * ファイルの拡張子を取得する。
         * 
         * @param name ファイル名
         * @return 拡張子
         */
        private String getExtension(String name)
        {
            int x = name.lastIndexOf('.') + 1;
            return name.substring(x);
        }
    }
}
