/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.io;

import java.awt.Frame;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.Map;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.concurrent.Callable;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;

import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.T2MapMenuBar;
import jp.nihu.gttime.T2MapProgressBarDialog;
import jp.nihu.gttime.data.DatasetException;
import jp.nihu.gttime.data.DatasetIF;
import jp.nihu.gttime.data.ProjectIF;
import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.plugin.ExportPlugin;
import jp.nihu.gttime.plugin.PluginManager;
import jp.nihu.gttime.plugin.PluginManagerException;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.plugin.RecordIOFilterException;
import jp.nihu.gttime.plugin.SavePlugin;
import jp.nihu.gttime.time.utility.JulianGregorianDate;
import jp.nihu.gttime.utility.ExtensionFileFilter;
import jp.nihu.gttime.utility.ExtensionFileSaveChooser;
import jp.nihu.gttime.utility.GTM;
import jp.nihu.gttime.utility.TMM;
import jp.nihu.gttime.utility.Utilities;
import jp.nihu.gttime.utility.XMLUtil;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.xerces.dom.DOMImplementationImpl;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.ls.DOMImplementationLS;
import org.w3c.dom.ls.LSOutput;
import org.w3c.dom.ls.LSSerializer;

/**
 * データセットなどをファイルに出力するクラスです。
 */
public class FileSavePlugin extends SavePlugin
{
    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(FileSavePlugin.class);

    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(FileSavePlugin.class, T2Map.LOCALE);
    private static PluginResourceManager rmgrmb = PluginResourceManager.getManager(T2MapMenuBar.class, T2Map.LOCALE);

    /**
     * T2Mapプロジェクト
     */
    private T2Map t2map = null;

    /**
     * Vectorオブジェクト
     */
    private ArrayList<String> done = new ArrayList<String>();
    
    public void save(T2Map t2map, DatasetIF dataset) {
    	// GTMを保存。以前に今後は「GTMをベースとする」としたのでTMMには対応しない。TMMを読み込んだ場合は同じディレクトリに拡張子をgtmにしてGTMを保存する。
    	log.debug("save gtm.");
    	log.debug("---------------------------------------------------------------> " + dataset.getPath());
    	log.debug("---------------------------------------------------------------:> " + dataset.getAbsolutePath());
    	
    	try {
			XMLUtil.printXml(dataset.getDomDataset());
			if (dataset.getAbsolutePath() != null) {
				// 元のファイルがTMMの場合、アラート表示。GTMがパスにあるか確認し、ない場合に（、TMMがあるか確認する。TMMがある場合に？）アラート表示。
				if (new File(dataset.getAbsolutePath()).exists()) {
					saveOverWriteDataset(dataset, dataset.getAbsolutePath());
				} else {
					String pathLowerCase = dataset.getAbsolutePath().toLowerCase();
					if (!pathLowerCase.startsWith("http://") && !pathLowerCase.startsWith("https://")) {
						showSaveNewMetadataAlert(rmgr.getString("alert.tmm.to.gtm"));
					}
					saveOverWriteDataset(dataset, dataset.getAbsolutePath());
				}
			} else {
				// 新規追加されたデータなのでレコードを保存する必要あり。
				if (parentProject == null) {
					//メタデータとデータを新規保存。SaveAsを実行の場合で親のGTSから保存を実行している場合は、親のGTS内のパスを変更して再保存する必要あり。
					showSaveAsDataset(dataset);
				} else {
					saveDataset(dataset, new File(parentProject.getAbsolutePath()).getParentFile(), dataset.getPath());
				}
			}
		} catch (IOException e) {
			log.error(e.getMessage(), e);
		}
    }
    
    private void showSaveAsDataset(DatasetIF dataset) {
    	PluginManager pluginManager = T2Map.getPluginManager();
    	
    	ArrayList<Hashtable<String, String>> plugins = pluginManager.getPlugins("ExportSavePlugin");
        
        final ArrayList<ExportPlugin> ps = new ArrayList<ExportPlugin>();
        for (int i = 0; i < plugins.size(); i++) {
        	try {
        		ExportPlugin plugin = (ExportPlugin) pluginManager.newInstance(plugins.get(i).get("class"));
        		plugin.init(t2map, pluginManager);
        		ps.add(plugin);
			} catch (PluginManagerException e1) {
				t2map.procError(e1);
			}
        }
    	
        String currentDirPath = T2Map.properties.getProperty("jp.nihu.gttime.T2Map.CURRENT_DIR");
        String path = currentDirPath + File.separator + dataset.getPath().substring(0, dataset.getPath().length() - 4);
        final ExtensionFileSaveChooser chooser = new ExtensionFileSaveChooser(path);
        for (int i = 0; i < ps.size(); i++)
        {
        	ExportPlugin plugin = ps.get(i);
			chooser.setFileFilter(plugin.getExtensionFileFilter());
			JPanel panel = plugin.getJPanel();
			if (panel != null) {
				chooser.setAccessory(panel);
			} else {
				chooser.setAccessory(new JPanel());
			}
        }
        chooser.addPropertyChangeListener(new PropertyChangeListener() {
			
			public void propertyChange(PropertyChangeEvent evt) {
				String prop = evt.getPropertyName();
				if (JFileChooser.FILE_FILTER_CHANGED_PROPERTY.equals(prop)) {
					Object o = evt.getNewValue();
					if (o != null) {
						if (o instanceof ExtensionFileFilter) {
							ExtensionFileFilter f = (ExtensionFileFilter)evt.getNewValue();
							for (int i = 0; i < ps.size(); i++)
			                {
								ExportPlugin plugin = ps.get(i);
								if (plugin.getExtensionFileFilter().getDescription().equals(f.getDescription())) {
									JPanel panel = plugin.getJPanel();
									if (panel != null) {
										chooser.setAccessory(panel);
									} else {
										chooser.setAccessory(new JPanel());
									}
									chooser.getAccessory().updateUI();
								}
			                }
						}
					}
				}
			}
		});
        
        int returnVal = chooser.showSaveDialog(t2map);
        if (returnVal == JFileChooser.APPROVE_OPTION)
        {
        	String ext = chooser.getFileFilter().getDescription();
        	for (int i = 0; i < ps.size(); i++)
            {
        		if (ps.get(i).getExtensionFileFilter().getDescription().equals(ext)) {
    				ExportPlugin plugin = ps.get(i);
                    plugin.init(t2map, pluginManager);
                    plugin.save(dataset, chooser.getSelectedFile());
        			break;
        		}
            }
        }
        // キャンセルボタンが押された時
        else
        {
            T2Map.properties.setProperty("jp.nihu.gttime.T2Map.CURRENT_DIR", chooser.getCurrentDirectory().toString());
        }
        UIManager.put("FileChooser.readOnly", Boolean.FALSE);
    }
    
    private void showSaveNewMetadataAlert(String message) {
    	if (!T2Map.properties.getBoolean("jp.nihu.gttime.T2Map.SHOW_ALERT_ON_SAVE")) {
        	JCheckBox cb = new JCheckBox(rmgr.getString("alert.switch.off"));
    		Object[] option = {message,cb};
    		JOptionPane.showMessageDialog(t2map, option, rmgrmb.getString("menu.file.save"), JOptionPane.WARNING_MESSAGE);
    		T2Map.properties.put("jp.nihu.gttime.T2Map.SHOW_ALERT_ON_SAVE", Boolean.toString(cb.isSelected()));
    	}
    }
    
    private ProjectIF parentProject = null;

    public void save(T2Map t2map, ProjectIF project) {
    	// GTSを保存。以前に今後は「GTSをベースとする」としたのでTMSには対応しない。TMSを読み込んだ場合は同じディレクトリに拡張子をgtsにしてGTSを保存する。
    	log.debug("save gts.");
    	log.debug("=========================================> " + project.getPath());
    	log.debug("=========================================:> " + project.getAbsolutePath());
    	try {
			XMLUtil.printXml(project.getDomProject());
			if (project.getAbsolutePath() != null) {
				// 元のファイルがTMSの場合、アラート表示。GTSがパスにあるか確認し、ない場合に（、TMSがあるか確認する。TMSがある場合に？）アラート表示。
				if (new File(project.getAbsolutePath()).exists()) {
					saveOverWriteProject(project, project.getAbsolutePath());
				} else {
					showSaveNewMetadataAlert(rmgr.getString("alert.tms.to.gts"));
					saveOverWriteProject(project, project.getAbsolutePath());
				}
			} else {
				if (parentProject == null) {
					//メタデータを新規保存。SaveAsを実行の場合で親のGTSから保存を実行している場合は、親のGTS内のパスを変更して再保存する必要あり。
					saveAsForOverWriteSave(t2map, project);
					t2map.setTitle(T2Map.APPLICATION_NAME + ":" + project.getAbsolutePath());
				} else {
					saveProject(project, new File(parentProject.getAbsolutePath()).getParentFile(), project.getPath());
				}
			}
		} catch (IOException e) {
			log.error(e.getMessage(), e);
		}
    }
    
    private int saveRemoteFile(String title, String url, String path) {
    	int option = JOptionPane.YES_OPTION;
    	if (!T2Map.properties.getBoolean("jp.nihu.gttime.T2Map.SAVE_REMOTE_FILE_ALWAYS")) {
			String dtitle = rmgr.getString("title.save.remote.file");
			String message = rmgr.getString("alert.save.remote.file").replace("{TITLE}", title).replace("{URL}", url).replace("{PATH}", path);
			JCheckBox cb = new JCheckBox(rmgr.getString("save.remote.faile.always"));
    		Object[] options = {message,cb};
    		option = JOptionPane.showConfirmDialog(t2map, options, dtitle, JOptionPane.YES_NO_OPTION);
    		T2Map.properties.put("jp.nihu.gttime.T2Map.SAVE_REMOTE_FILE_ALWAYS", Boolean.toString(cb.isSelected()));
		}
    	return option;
    }
    
    private void saveOverWriteProject(ProjectIF project, String path) {
        File file = new File(project.getAbsolutePath());
        if (!done.contains(file.getAbsolutePath()))
        {
            done.add(file.getAbsolutePath());
            log.info("SAVE : " + file.getAbsolutePath());

            Document doc = project.getDomProject();
            try
            {
                Object[] objs = project.getNodes();
                if (objs != null)
                {
                	for (int i = 0; i < objs.length; i++)
                    {
                		if (objs[i] instanceof ProjectIF)
                		{
                			ProjectIF p = (ProjectIF)objs[i];
                			String pt = p.getPath();
                			// 参照先がURLの場合の対応→パスの書き換えが必要、アラートを表示
                			if (pt.startsWith("http://") || pt.startsWith("https://")) {
                				String title = Utilities.escapeFileName(p.getTitle().trim());
                				title = title.equals("") ? "NO_TITLE" : title;
                        		String fname = title + ".gts";
                        		File base = new File(file.getParentFile(), title);//タイトル名でフォルダーを作成する
                        		String np = base.getAbsolutePath() + "/" + fname;
                        		int option = saveRemoteFile(title, pt, np);
        					    if (option == JOptionPane.YES_OPTION) {
                        			updateProjectFile(doc, i, np);
                        			saveProject(p, base, fname);
                        			p.setPath(np);
                        			p.setAbsolutePath(np);
        					    } else if (option == JOptionPane.NO_OPTION) {
        					    }
                			} else {
                				save(t2map, p);
                			}
                		}
                		else if (objs[i] instanceof DatasetIF)
                		{
                			DatasetIF d = (DatasetIF)objs[i];
                			String pt = d.getPath();
                			// 参照先がURLの場合の対応→パスの書き換えが必要、アラートを表示
                			if (pt.startsWith("http://") || pt.startsWith("https://")) {
                				String title = Utilities.escapeFileName(d.getTitle().trim());
                        		title = title.equals("") ? "NO_TITLE" : title;
                				String fname = title + ".gtm";
                        		File base = new File(file.getParentFile(), title);//タイトル名でフォルダーを作成する
                        		String np = base.getAbsolutePath() + "/" + fname;
                        		int option = saveRemoteFile(title, pt, np);
                        		if (option == JOptionPane.YES_OPTION) {
        					    	updateProjectFile(doc, i, np);
                        			saveDataset(d, base, fname);
                        			d.setPath(np);
                        			d.setAbsolutePath(np);
        					    } else if (option == JOptionPane.NO_OPTION) {
        					    }
                			} else {
                				save(t2map, d);
                			}
                		}
                    }
                }
                saveXml(doc, file);
            }
            catch (IOException e)
            {
                log.fatal(e.getMessage(), e);
            }
        }
    }
    
    private void saveOverWriteDataset(DatasetIF dataset, String path) {
        try
        {
            File file = new File(dataset.getAbsolutePath());
            if (!done.contains(file.getAbsolutePath()))
            {
                done.add(file.getAbsolutePath());
                log.info("SAVE Dataset : " + file.getAbsolutePath());

                Document doc = dataset.getDomDataset();
                String recordPath = dataset.getRecordPath();

                // 参照先がURLの場合の対応→パスの書き換えが必要、アラートを表示
				if (recordPath.startsWith("http://") || recordPath.startsWith("https://")) {
					String title = Utilities.escapeFileName(dataset.getTitle().trim());
            		title = title.equals("") ? "NO_TITLE" : title;
            		String fname = title + ".xml";
					String filePath = file.getParentFile().getAbsolutePath() + "/" + fname;
					int option = saveRemoteFile(title, recordPath, recordPath);
            		if (option == JOptionPane.YES_OPTION) {
				    	if (!done.contains(filePath)) {
							done.add(filePath);
							log.info("SAVE : " + filePath);
							NIHUTimeSpaceXmlIOFilter filter = new NIHUTimeSpaceXmlIOFilter();
							byte[] xml = filter.write(dataset.getRecords());
							saveBytedata(xml, new File(filePath));
						}

						try {
							XPathFactory factory = XPathFactory.newInstance();
							XPath xpath = factory.newXPath();
							xpath.setNamespaceContext(jp.nihu.gttime.utility.XMLUtil.getGtsNsContext());
							XPathExpression expr = xpath.compile("/gt:gtm/gt:resource/gt:filename");
							Element e = (Element) expr.evaluate(doc, XPathConstants.NODE);
							e.setTextContent(fname);
						} catch (XPathExpressionException e) {
							// never catch
							log.fatal("not valid xpath.", e);
						}
				    } else if (option == JOptionPane.NO_OPTION) {
				    }
				}
                
                saveXml(doc, file);
            }
        }
        catch (IOException e)
        {
            log.fatal(e.getMessage(), e);
        }
        catch (RecordIOFilterException e)
        {
            log.fatal(e.getMessage(), e);
        }
    }
    
    
    
    public void saveAsForOverWriteSave(T2Map t2map, ProjectIF project)
    {
        if (project != null)
        {
            this.t2map = t2map;
            ExtensionFileSaveChooser chooser = new ExtensionFileSaveChooser(T2Map.properties.getProperty("jp.nihu.gttime.T2Map.CURRENT_DIR") + File.separator
                    + project.getTitle());
            
            chooser.addChoosableFileFilter(new ExtensionFileFilter("gts"));
            chooser.addChoosableFileFilter(new ExtensionFileFilter("tms"));

            int returnVal = chooser.showSaveDialog(SwingUtilities.getRoot(t2map));
            // 保存ボタンが押された時
            if (returnVal == JFileChooser.APPROVE_OPTION)
            {
                File file = chooser.getSelectedFile();
                // saveProject(project, file.getParentFile(), file.getName());
                Frame frame = (Frame) SwingUtilities.getRoot(t2map);
                String title = rmgr.getString("progress.title");
                String message = rmgr.format("progress.message", file.getAbsolutePath());
                T2MapProgressBarDialog<Object> progress = new T2MapProgressBarDialog<Object>(frame, true, title, message);
                Callable<Object> callable = (Callable<Object>) new SaveAsProject(project, file.getParentFile(), file.getName(), true);
                FutureTask<Object> f = (FutureTask<Object>) progress.submit(callable);
                progress.setVisible(true);
                try
                {
                    f.get();
                }
                catch (CancellationException e)
                {
                }
                catch (InterruptedException e)
                {
                    log.fatal(e.getMessage(), e);
                }
                catch (ExecutionException e)
                {
                    log.fatal(e.getMessage(), e);
                }
            }
            // キャンセルボタンが押された時
            else
            {
                T2Map.properties.setProperty("jp.nihu.gttime.T2Map.CURRENT_DIR", chooser.getCurrentDirectory().toString());
            }
            try
            {
                T2Map.properties.store();
            }
            catch (FileNotFoundException e)
            {
                log.fatal(e.getMessage(), e);
            }
            catch (IOException e)
            {
                log.fatal(e.getMessage(), e);
            }
        }
    }
    
    
    

    /**
     * ファイル選択画面を表示します。
     * 
     * @param t2map T2Mapオブジェクト
     * @param project Projectオブジェクト
     */
    public void saveAs(T2Map t2map, ProjectIF project)
    {
        if (project != null)
        {
            this.t2map = t2map;

            ExtensionFileSaveChooser chooser = new ExtensionFileSaveChooser(T2Map.properties.getProperty("jp.nihu.gttime.T2Map.CURRENT_DIR") + File.separator
                    + project.getTitle());
            
            chooser.addChoosableFileFilter(new ExtensionFileFilter("gts"));
            chooser.addChoosableFileFilter(new ExtensionFileFilter("tms"));

            int returnVal = chooser.showSaveDialog(SwingUtilities.getRoot(t2map));
            // 保存ボタンが押された時
            if (returnVal == JFileChooser.APPROVE_OPTION)
            {
                File file = chooser.getSelectedFile();
                Frame frame = (Frame) SwingUtilities.getRoot(t2map);
                String title = rmgr.getString("progress.title");
                String message = rmgr.format("progress.message", file.getAbsolutePath());
                T2MapProgressBarDialog<Object> progress = new T2MapProgressBarDialog<Object>(frame, true, title, message);
                Callable<Object> callable = (Callable<Object>) new SaveAsProject(project, file.getParentFile(), file.getName(), false);
                FutureTask<Object> f = (FutureTask<Object>) progress.submit(callable);
                progress.setVisible(true);
                try
                {
                    f.get();
                }
                catch (CancellationException e)
                {
                }
                catch (InterruptedException e)
                {
                    log.fatal(e.getMessage(), e);
                }
                catch (ExecutionException e)
                {
                    log.fatal(e.getMessage(), e);
                }
            }
            // キャンセルボタンが押された時
            else
            {
                T2Map.properties.setProperty("jp.nihu.gttime.T2Map.CURRENT_DIR", chooser.getCurrentDirectory().toString());
            }
            try
            {
                T2Map.properties.store();
            }
            catch (FileNotFoundException e)
            {
                log.fatal(e.getMessage(), e);
            }
            catch (IOException e)
            {
                log.fatal(e.getMessage(), e);
            }
        }
    }

    /**
     * ファイルを出力します。
     * 
     * @param project Projectオブジェクト
     * @param base ディレクトリパス
     * @param relative ファイル名
     */
    private void saveProject(ProjectIF project, File base, String relative)
    {
    	if (project.getAbsolutePath() == null) {
    		//時間範囲を設定
    		setPeriod(project);
    	}
    	
        boolean doit = true;
        File file = new File(base, relative);
        if (!done.contains(file.getAbsolutePath()))
        {
            done.add(file.getAbsolutePath());
            if (done.size() > 1 && file.exists())
            {
                doit = doOverWrite(file);
            }
            else
            {
                doit = true;
            }
            if (doit)
            {
                log.info("SAVE : " + file.getAbsolutePath());

                boolean isTms = false;

                if (file.getName().endsWith("tms"))
                {
                    isTms = true;
                }

                Document doc = project.getDomProject();
                try
                {
                    Object[] objs = project.getNodes();
                    if (objs != null)
                    {
                    	for (int i = 0; i < objs.length; i++)
                        {
                    		if (objs[i] instanceof ProjectIF)
                    		{
                    			ProjectIF p = (ProjectIF)objs[i];
                    			String path = p.getPath();
                                File cpf = new File(path);
                                if (cpf.isAbsolute())
                                {
                                    base = cpf.getParentFile();
                                    relative = cpf.getName();
                                }
                                else
                                {
                            		base = file.getParentFile();
                                    relative = path;
                                }

                                if (isTms)
                                {
                                    relative = relative.replaceAll("\\..+$", ".tms");
                                }
                                saveProject(p, base, relative);
                    		}
                    		else if (objs[i] instanceof DatasetIF)
                    		{
                    			DatasetIF d = (DatasetIF)objs[i];
                    			String path = d.getPath();
                                File cdf = new File(path);
                                if (cdf.isAbsolute())
                                {
                                    base = cdf.getParentFile();
                                    relative = cdf.getName();
                                }
                                else
                                {
                                    base = file.getParentFile();
                                    relative = path;
                                }

                                if (isTms)
                                {
                                    relative = relative.replaceAll("\\..+$", ".tmm");
                                }
                                saveDataset(d, base, relative);
                    		}
                        }
                    }
                    
                    if (isTms)
                    {
                        // docをTMSに変換する。
                        saveXml(getTMSDocument(doc), file);
                    }
                    else
                    {
                        saveXml(doc, file);
                    }

                }
                catch (IOException e)
                {
                    log.fatal(e.getMessage(), e);
                }
            }
        }
    }
    

    /**
     * ファイルに出力する処理を実装します。
     * 
     * @param dataset Datasetオブジェクト
     * @param base Fileオブジェクト
     * @param relative ファイル名
     */
    private void saveDataset(DatasetIF dataset, File base, String relative)
    {
        try
        {
            boolean doit = true;
            File file = new File(base, relative);
            if (!done.contains(file.getAbsolutePath()))
            {
                done.add(file.getAbsolutePath());
                if (file.exists())
                {
                    doit = doOverWrite(file);
                }
                if (doit)
                {
                    log.info("SAVE Dataset : " + file.getAbsolutePath());

                    Document doc = null;

                    NIHUTimeSpaceXmlIOFilter filter = new NIHUTimeSpaceXmlIOFilter();
                    File filer = new File(filter.getPath(file.getAbsolutePath()));

                    if (file.getName().endsWith("tmm"))
                    {
                        // log.debug("############## TMM #################");
                        doc = TMM.create(dataset.getTitle(), dataset.getLayerType(), dataset, "NIHU TIMESPACE XML");
                    }
                    else
                    {
                        // log.debug("############## GTM #################");
                        doc = GTM.create(dataset.getTitle(), dataset.getLayerType(), dataset, "XML for HuTime", filer.getName());
                    }

                    saveXml(doc, file);

                    if (!done.contains(filer.getAbsolutePath()))
                    {
                        done.add(filer.getAbsolutePath());
                        if (filer.exists())
                        {
                            doit = doOverWrite(filer);
                        }
                        if (doit)
                        {
                        	log.debug("SAVE Records : " + filer.getAbsolutePath());
                        	Record[] records = dataset.getRecords();
                            log.info("SAVE : " + filer.getAbsolutePath());
                            byte[] xml = filter.write(records);
                            saveBytedata(xml, filer);
                        }
                    }
                }
            }
        }
        catch (IOException e)
        {
            log.fatal(e.getMessage(), e);
        }
        catch (RecordIOFilterException e)
        {
            log.fatal(e.getMessage(), e);
        }
    }

    private Document getTMSDocument(Document gts)
    {
        DOMResult res = null;
        try
        {
            TransformerFactory factory = TransformerFactory.newInstance();
            InputStream inXslt = getClass().getClassLoader().getResourceAsStream("gts2tms.xsl");
            Transformer transformer = factory.newTransformer(new StreamSource(inXslt));
            res = new DOMResult();
            transformer.transform(new DOMSource(gts), res);
        }
        catch (TransformerConfigurationException e)
        {
            log.fatal(e.getMessage(), e);
        }
        catch (TransformerException e)
        {
            log.fatal(e.getMessage(), e);
        }
        return (Document) res.getNode();
    }
    
    private class SaveAsProject implements Callable<Object>
    {
        private ProjectIF project;

        private File base;

        private String relative;
        
        private boolean needSetAbsolutepath;

        public SaveAsProject(ProjectIF project, File base, String relative, boolean needSetAbsolutepath)
        {
            this.project = project;
            this.base = base;
            this.relative = relative;
            this.needSetAbsolutepath = needSetAbsolutepath;
        }

        public Object call() throws Exception
        {
            saveAsProject(project, base, relative, needSetAbsolutepath);
            return project;
        }
    }
    
    private BigDecimal getRecordsBeginDate(ProjectIF p) {
    	BigDecimal jdate = null;
        ProjectIF[] projects = p.getProjects();
        if (projects != null)
        {
            for (int i = 0; i < projects.length; i++)
            {
            	jdate = getRecordsBeginDate(projects[i]);
            }
        }
        DatasetIF[] datasets = p.getDatasets();
        if (datasets != null)
        {
            for (int i = 0; i < datasets.length; i++)
            {
				try {
					BigDecimal date = datasets[i].getStartDatetime();
					if (date != null && (jdate == null || date.compareTo(jdate) == -1))
	                {
	                    jdate = date;
	                }
				} catch (DatasetException e) {
					log.error(e.getMessage(), e);
				}
            }
        }
    	return jdate;
    }
    
    private BigDecimal getRecordsEndDate(ProjectIF p) {
    	BigDecimal jdate = null;
        ProjectIF[] projects = p.getProjects();
        if (projects != null)
        {
            for (int i = 0; i < projects.length; i++)
            {
            	jdate = getRecordsEndDate(projects[i]);
            }
        }
        DatasetIF[] datasets = p.getDatasets();
        if (datasets != null)
        {
            for (int i = 0; i < datasets.length; i++)
            {
				try {
					BigDecimal date = datasets[i].getEndDatetime();
					if (date != null && (jdate == null || date.compareTo(jdate) == 1))
	                {
	                    jdate = date;
	                }
				} catch (DatasetException e) {
					log.error(e.getMessage(), e);
				}
            }
        }
    	return jdate;
    }
    
    private void setPeriod(ProjectIF project) {
    	Document doc = project.getDomProject();
		
		NodeList nl = doc.getElementsByTagNameNS("http://www.h-gis.org/hgis", "begin");
        if (nl.getLength() > 0)
        {
            Element elem = (Element) nl.item(0);
            BigDecimal jdate = getRecordsBeginDate(project);
            if (jdate != null) {
            	elem.setTextContent(JulianGregorianDate.toGregorian(jdate, "Z"));
            }
        }
		
		nl = doc.getElementsByTagNameNS("http://www.h-gis.org/hgis", "end");
        if (nl.getLength() > 0)
        {
            Element elem = (Element) nl.item(0);
            BigDecimal jdate = getRecordsEndDate(project);
            if (jdate != null) {
            	elem.setTextContent(JulianGregorianDate.toGregorian(jdate, "Z"));
            }
        }
    }
    
    
    private void saveAsProject(ProjectIF project, File base, String relative, boolean needSetAbsolutepath)
    {
    	if (project.getAbsolutePath() == null) {
    		//時間範囲を設定
    		setPeriod(project);
    	}
    	
    	if (needSetAbsolutepath) {
			File f = new File(base, relative);
			project.setAbsolutePath(f.getAbsolutePath());
    	}
    	
    	Map<String, String> fileNameAndOriginPathMap = new TreeMap<String, String>();
    	Map<String, Set<String>> filePathAndHash = new TreeMap<String, Set<String>>();
        boolean doit = true;
        File file = new File(base, relative);
        if (!done.contains(file.getAbsolutePath()))
        {
            done.add(file.getAbsolutePath());
            if (done.size() > 1 && file.exists())
            {
                doit = doOverWrite(file);
            }
            else
            {
                doit = true;
            }
            if (doit)
            {
                log.info("SAVE AS : " + file.getAbsolutePath());

                boolean isTms = false;

                if (file.getName().endsWith("tms"))
                {
                    isTms = true;
                }

                Document doc = (Document)project.getDomProject().cloneNode(true);
                
                try
                {
                    Object[] objs = project.getNodes();
                    if (objs != null)
                    {
                    	for (int i = 0; i < objs.length; i++)
                        {
                    		if (objs[i] instanceof ProjectIF)
                    		{
                    			ProjectIF p = (ProjectIF)objs[i];
                    			String t = Utilities.escapeFileName(p.getTitle());
                    			String dname = t;
                    			File childProjectDir = new File(file.getParent(), t);
                    			String childProjectFileName = t + ".gts";
                    			
                    			String hash = getProjectHash(p);
                    			boolean proc = false;
                    			if (fileNameAndOriginPathMap.containsKey(childProjectFileName)) {
                    				if (fileNameAndOriginPathMap.get(childProjectFileName).equals(p.getPath())) {
                    					//配下のメタデータをチェックする。
                    					boolean same = false;
                    					if (filePathAndHash.containsKey(t+p.getPath())) {
                            				Set<String> s = filePathAndHash.get(t+p.getPath());
                            				same = s.contains(hash);
                            			}
                    					if (!same) {
                    						//名前を変えて保存
                    						int num = getNewFileNameNum(t, "gts", fileNameAndOriginPathMap);
                    						dname = t + "(" + num + ")";
                    						childProjectDir = new File(file.getParent(), dname);
                    						childProjectFileName = dname + ".gts";
                    						fileNameAndOriginPathMap.put(childProjectFileName, p.getPath());
                    						proc = true;
                    					}
                    				} else {
                    					//名前を変えて保存
                    					int num = getNewFileNameNum(t, "gts", fileNameAndOriginPathMap);
                    					dname = t + "(" + num + ")";
                    					childProjectDir = new File(file.getParent(), dname);
                    					childProjectFileName = dname + ".gts";
                						fileNameAndOriginPathMap.put(childProjectFileName, p.getPath());
                    					proc = true;
                    				}
                    			} else {
                    				proc = true;
                    				fileNameAndOriginPathMap.put(childProjectFileName, p.getPath());
                    			}
                    			
                    			if (filePathAndHash.containsKey(t+p.getPath())) {
                    				Set<String> s = filePathAndHash.get(t+p.getPath());
                    				s.add(hash);
                    			} else {
                    				Set<String> s = new TreeSet<String>();
                    				s.add(hash);
                    				filePathAndHash.put(t+p.getPath(), s);
                    			}
                    			
                    			if (proc) {
                                    if (isTms)
                                    {
                                    	childProjectFileName = childProjectFileName.replaceAll("\\..+$", ".tms");
                                    }
                                    saveAsProject(p, childProjectDir, childProjectFileName, needSetAbsolutepath);
                    			}
                    			
                    			//docのファイルパスを変更する。
                    			String np = dname + "/" + childProjectFileName;
                    			updateProjectFile(doc, i, np);
                    		}
                    		else if (objs[i] instanceof DatasetIF)
                    		{
                    			DatasetIF d = (DatasetIF)objs[i];
                    			String t = Utilities.escapeFileName(d.getTitle());
                    			String childDatasetFileName = t + ".gtm";
                    			
                    			boolean proc = false;
                    			if (fileNameAndOriginPathMap.containsKey(childDatasetFileName)) {
                    				int num = getNewFileNameNum(t, "gtm", fileNameAndOriginPathMap);
                    				childDatasetFileName = t + "(" + num + ")" + ".gtm";
            						fileNameAndOriginPathMap.put(childDatasetFileName, d.getPath());
                					proc = true;
                    			} else {
                    				proc = true;
                    				fileNameAndOriginPathMap.put(childDatasetFileName, d.getPath());
                    			}
                    			
                    			if (proc) {
                        			if (isTms)
                                    {
                        				childDatasetFileName = childDatasetFileName.replaceAll("\\..+$", ".tmm");
                                    }
                                    saveAsDataset(d, new File(file.getParent()), childDatasetFileName);
                                    log.debug("Save Dataset: " + new File(file.getParent()).getAbsolutePath());
                    			}
                    			
                    			//docのファイルパスを変更する。
                    			String nd = childDatasetFileName;
                    			updateProjectFile(doc, i, nd);
                    		}
                        }
                    }
                    
                    if (isTms)
                    {
                        // docをTMSに変換する。
                        saveXml(getTMSDocument(doc), file);
                    }
                    else
                    {
                        saveXml(doc, file);
                    }
                    log.debug("Save Project: " + file.getAbsolutePath());

                }
                catch (IOException e)
                {
                    log.fatal(e.getMessage(), e);
                }
            }
        }
    }
    
    private void appendChildProject(MessageDigest md, ProjectIF p) {
		try {
			byte[] pbyte = XMLUtil.getByteXml(p.getDomProject());
			md.update(pbyte);
	    	Object[] os = p.getNodes();
	    	for (int i = 0; i < os.length; i++) {
	    		if (os[i] instanceof ProjectIF) {
	    			appendChildProject(md, (ProjectIF)os[i]);
	    		} else if (os[i] instanceof DatasetIF) {
	    			byte[] dbyte = XMLUtil.getByteXml(((DatasetIF)os[i]).getDomDataset());
	    			md.update(dbyte);
	    		}
	    	}
		} catch (IOException e) {
			log.fatal(e.getMessage(), e);
		}
    	
    }
    
    private String getProjectHash(ProjectIF p) {
    	StringBuffer s = new StringBuffer();
    	try {
			MessageDigest md = MessageDigest.getInstance("SHA-256");
			appendChildProject(md, p);
			byte[] digest = md.digest();
			for (int i = 0; i < digest.length; i++) {
				int d = digest[i];

				if (d < 0) { // byte 128-255
				d += 256;
				}
				if (d < 16) { //0-15 16
				s.append("0");
				}
				s.append(Integer.toString(d, 16));
				}
		} catch (NoSuchAlgorithmException e) {
			log.fatal(e.getMessage(), e);
		}
    	return s.toString();
    }
    
    private int getNewFileNameNum(String name, String ext, Map<String, String> names) {
    	int j = 1;
		String newFileName = name + "("+ j + ")" + "." + ext;
		while(names.containsKey(newFileName)) {
			j++;
			newFileName = name + "("+ j + ")" + "." + ext;
		}
    	return j;
    }
    
    private void updateProjectFile(Document doc, int pos, String newFilePath) {
    	NodeList nl = doc.getElementsByTagNameNS("http://www.h-gis.org/hgis", "resourceFileName");
		((Element)nl.item(pos)).setTextContent(newFilePath);
		log.debug("Update Project File: "  +newFilePath);
    }
    
    private void saveAsDataset(DatasetIF dataset, File base, String relative)
    {
        try
        {
            boolean doit = true;
            File file = new File(base, relative);
            if (!done.contains(file.getAbsolutePath()))
            {
                done.add(file.getAbsolutePath());
                if (file.exists())
                {
                    doit = doOverWrite(file);
                }
                if (doit)
                {
                    log.info("SAVE AS : " + file.getAbsolutePath());

                    Document doc = null;

                    NIHUTimeSpaceXmlIOFilter filter = new NIHUTimeSpaceXmlIOFilter();
                    File filer = new File(filter.getPath(file.getAbsolutePath()));

                    if (file.getName().endsWith("tmm"))
                    {
                        doc = TMM.create(dataset.getTitle(), dataset.getLayerType(), dataset, "NIHU TIMESPACE XML");
                    }
                    else
                    {
                        doc = GTM.create(dataset.getTitle(), dataset.getLayerType(), dataset, "XML for HuTime", filer.getName());
                    }

                    saveXml(doc, file);

                    Record[] records = dataset.getRecords();
                    if (!done.contains(filer.getAbsolutePath()))
                    {
                        done.add(filer.getAbsolutePath());
                        if (filer.exists())
                        {
                            doit = doOverWrite(filer);
                        }
                        if (doit)
                        {
                            log.info("SAVE : " + filer.getAbsolutePath());
                            byte[] xml = filter.write(records);
                            saveBytedata(xml, filer);
                        }
                    }
                }
            }
        }
        catch (IOException e)
        {
            log.fatal(e.getMessage(), e);
        }
        catch (RecordIOFilterException e)
        {
            log.fatal(e.getMessage(), e);
        }
    }
    

    private boolean yesAll = false;

    /**
     * 上書きの確認
     * 
     * @param file Fileオブジェクト
     * @return 上書きする、しない
     */
    private boolean doOverWrite(File file)
    {
        if (!yesAll)
        {
            String message = rmgr.getString("superscription.message1") + file.getAbsolutePath();

            Object[] options = { rmgr.getString("progress.yes"), rmgr.getString("progress.no"), rmgr.getString("progress.yes.all") };
            int answer = JOptionPane.showOptionDialog(t2map, message, rmgr.getString("superscription.message2"), JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.QUESTION_MESSAGE, null, options,
                            options[0]);
            if (answer == JOptionPane.YES_OPTION)
            {
                log.debug("JOptionPane.YES_OPTION");
            }
            else if (answer == JOptionPane.NO_OPTION)
            {
                log.debug("JOptionPane.NO_OPTION");
                return false;
            }
            else if (answer == JOptionPane.CANCEL_OPTION)
            {
                log.debug("JOptionPane.CANCEL_OPTION");
                yesAll = true;
            }

        }
        return true;
    }

    /**
     * バイトデータをファイル出力ストリームに書き込みます。
     * 
     * @param data バイトデータ
     * @param file Fileオブジェクト
     * @throws IOException
     */
    private void saveBytedata(byte[] data, File file) throws IOException
    {
        FileOutputStream out = new FileOutputStream(file);
        out.write(data);
        out.close();
    }

    /**
     * XMLファイルに記入する。
     * 
     * @param doc Documentオブジェクト
     * @param file Fileオブジェクト
     * @throws IOException
     */
    private void saveXml(Document doc, File file) throws IOException
    {
        File parent = file.getParentFile();
        if (parent.exists() || parent.mkdirs())
        {
            DOMImplementation domImpl = new DOMImplementationImpl();
            DOMImplementationLS domImplLS = (DOMImplementationLS) domImpl.getFeature("LS", "3.0");
            LSOutput lsOutput = domImplLS.createLSOutput();
            LSSerializer lsSerializer = domImplLS.createLSSerializer();
            FileOutputStream fos = new FileOutputStream(file);
            lsOutput.setByteStream(fos);
            lsSerializer.write(doc, lsOutput);
            fos.close();
        }
    }
}
