/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.io;

import java.io.IOException;
import java.io.InputStream;
import java.io.UnsupportedEncodingException;
import java.math.BigDecimal;
import java.util.ArrayList;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.Event;
import jp.nihu.gttime.data.JulianDate;
import jp.nihu.gttime.data.Location;
import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.data.RecordException;
import jp.nihu.gttime.data.UTMPoint;
import jp.nihu.gttime.plugin.IOError;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.plugin.RecordIOFilterException;
import jp.nihu.gttime.time.utility.ISO8601Exception;
import jp.nihu.gttime.time.utility.JulianGregorianDate;
import jp.nihu.gttime.utility.GIO;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.LocatorImpl;

/**
 * ファイルの内容を読み書きします。
 */
public class KMLIOFilter extends NIHUTimeSpaceXmlIOFilter
{
    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(KMLIOFilter.class);

    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(KMLIOFilter.class, T2Map.LOCALE);

    private static String xmlPath = null;

    /**
     * 出力先パスを取得します。
     * 
     * @param tmmPath tmmPath パス
     * @return パス
     * @throws RecordIOFilterException
     */
    public String getPath(String tmmPath) throws RecordIOFilterException
    {
        int x = tmmPath.lastIndexOf(".");
        xmlPath = tmmPath.substring(0, x) + ".kml";
        // log.info("read dataset: " + xmlPath);
        return xmlPath;
    }

    /**
     * xmlファイルを読み込みます。
     * 
     * @param is InputStreamオブジェクト
     * @return Record[]オブジェクト
     * @throws RecordIOFilterException
     */
    public Record[] read(InputStream is, IOError ioErrorx) throws RecordIOFilterException
    {
        ioError = ioErrorx;
        SAXParserFactory factory = SAXParserFactory.newInstance();
        factory.setNamespaceAware(true);
        SAXParser parser = null;
        try
        {
            parser = factory.newSAXParser();
        }
        catch (ParserConfigurationException e)
        {
            throw new RecordIOFilterException(rmgr.getString("xml.parser.error"), e);
        }
        catch (SAXException e)
        {
            throw new RecordIOFilterException(rmgr.getString("xml.parser.error"), e);
        }
        KMLParser xmlHandler = new KMLParser();
        try
        {
            parser.parse(is, xmlHandler);
        }
        catch (SAXException e)
        {
            throw new RecordIOFilterException(rmgr.getString("dataset.analysis.error"), e);
        }
        catch (IOException e)
        {
            throw new RecordIOFilterException(rmgr.getString("dataset.acquisition.error"), e);
        }
        return xmlHandler.getRecords();
    }

    /**
     * xmlファイルに内容を書き込みます。
     * 
     * @param records Record[]オブジェクト
     * @return byte[]
     * @throws RecordIOFilterException
     */
    public byte[] write(Record[] records) throws RecordIOFilterException
    {
        byte[] bxml = null;
        StringBuffer sb = new StringBuffer("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
        sb.append("<kml xmlns=\"http://www.opengis.net/kml/2.2\" xmlns:atom=\"http://www.w3.org/2005/Atom\">\n");
        sb.append("<Document>\n");
        sb.append("<Style id=\"polygonStyle\">\n");//polygonStyle
        sb.append("<LineStyle><color>770000ff</color><width>0.1</width></LineStyle>\n");
        sb.append("<PolyStyle><color>ff0000cc</color></PolyStyle>\n");
//        sb.append("<IconStyle><scale>1.3</scale><Icon><href>http://maps.google.com/mapfiles/kml/pushpin/ylw-pushpin.png</href></Icon></IconStyle>\n");
        sb.append("</Style>\n");
        sb.append("<Style id=\"pointStyle\">\n");//polygonStyle
        sb.append("<IconStyle><scale>1.3</scale><Icon><href>http://maps.google.com/mapfiles/kml/pushpin/ylw-pushpin.png</href></Icon></IconStyle>\n");
        sb.append("</Style>\n");
        for (int i = 0; i < records.length; i++)
        {
            Location location = records[i].getLocation();
            if (location != null)
            {
                sb.append("<Placemark>");
                Event[] events = records[i].getEvents();
                if (events != null)
                {
                    String descriptionVal = "";
                    for (int j = 0; j < events.length; j++)
                    {
                        String name = events[j].getName();
                        if (name.equals("title")) {
                            sb.append("<name>");
                            sb.append(escapeXmlChar(events[j].getData()));
                            sb.append("</name>\n");
                        } else if (name.equals("url")) {
                            sb.append("<atom:link href=\"");
                            sb.append(escapeXmlChar(events[j].getData()));
                            sb.append("\"/>\n");
                        } else {
                            if (descriptionVal.length() > 0) {
                                descriptionVal += ", ";
                            }
                            descriptionVal += escapeXmlChar(events[j].getData());
                        }
                    }
                    sb.append("<description>");
                    sb.append(descriptionVal);
                    sb.append("</description>\n");
                }
                int type = location.getType();
                if (type == Location.LINE || type == Location.PLANE)
                {
                    sb.append("<styleUrl>#polygonStyle</styleUrl>\n");
                    sb.append("<Polygon>");
//                    sb.append("<tessellate>1</tessellate>");
                    sb.append("<outerBoundaryIs><LinearRing>\n");
                    UTMPoint[] points = location.getPoints();
                    sb.append("<coordinates>");
                    for (int j = 0; j < points.length; j++)
                    {
                        if (j != 0) {
                            sb.append(" ");
                        }
                        double[] latlon = GIO.getWGS84LatLon(points[j].getX(), points[j].getY(), points[j].getZone());
                        log.debug("2 IN : " + latlon[1] + ":" + latlon[0]);
                        sb.append(latlon[1]);
                        sb.append(",");
                        sb.append(latlon[0]);
//                        sb.append(",");
//                        sb.append("0");
                    }
                    sb.append("</coordinates>\n");

                    sb.append("</LinearRing></outerBoundaryIs></Polygon>\n");
                }
                else if (type == Location.POINT)
                {
                    sb.append("<styleUrl>#pointStyle</styleUrl>\n");
                    sb.append("<Point><coordinates>");
                    UTMPoint[] points = location.getPoints();
                    double[] latlon = GIO.getWGS84LatLon(points[0].getX(), points[0].getY(), points[0].getZone());
                    sb.append(latlon[1]);
                    sb.append(",");
                    sb.append(latlon[0]);
                    sb.append("</coordinates></Point>\n");
                }
                
                JulianDate js = records[i].getStartDatetime();
                JulianDate je = records[i].getEndDatetime();
                if (js != null || je != null) {
                    sb.append("<TimeSpan>\n");

                    sb.append("<begin>");
                    if (js != null)
                    {
                        sb.append(JulianGregorianDate.toGregorian(js.getDatetime(), js.getTimezone()));
                    } else 
                    {
                        sb.append(JulianGregorianDate.toGregorian(je.getDatetime(), je.getTimezone()));
                    }
                    sb.append("</begin>\n");

                    sb.append("<end>");
                    if (je != null)
                    {
                        sb.append(JulianGregorianDate.toGregorian(je.getDatetime(), je.getTimezone()));
                    } else {
                        sb.append(JulianGregorianDate.toGregorian(js.getDatetime(), js.getTimezone()));
                    }
                    sb.append("</end>\n");
                    sb.append("</TimeSpan>\n");
                }

                sb.append("</Placemark>\n");
            }
        }
        sb.append("</Document>\n");
        sb.append("</kml>\n");
        try
        {
            bxml = sb.toString().getBytes("UTF-8");
        }
        catch (UnsupportedEncodingException e)
        {
            // never catch
            throw new RecordIOFilterException(e.getMessage(), e);
        }
        return bxml;
    }

    /**
     * 禁則文字を置き換えます。
     * 
     * @param str 文字
     * @return 置き換えられた文字
     */
    private String escapeXmlChar(String str)
    {
        str = str.trim();
        str = str.replaceAll("&", "&amp;");
        str = str.replaceAll("<", "&lt;");
        str = str.replaceAll(">", "&gt;");
        str = str.replaceAll("\"", "&quot;");
        str = str.replaceAll("\'", "&apos;");
        return str;
    }

    /**
     * レコードのチェックや設定を行うクラスです。
     */
    class KMLParser extends DefaultHandler
    {
        private Locator locator = null;

        public KMLParser()
        {
            setDocumentLocator(new LocatorImpl());
        }

        public void setDocumentLocator(Locator locator)
        {
            this.locator = locator;
        }

        public void warning(SAXParseException exception) throws SAXException
        {
            String message = "**Parsing Warning**\n" + " Line:    " + exception.getLineNumber() + "\n" + " URI:     " + xmlPath + "\n" + " Message: " + exception.getMessage();
            ioError.addWarnNum();
            ioError.add(message);
            log.warn(message);
            // throw new SAXException("Warning encountered");
        }

        public void error(SAXParseException exception) throws SAXException
        {
            String message = "**Parsing Error**\n" + " Line:    " + exception.getLineNumber() + "\n" + " URI:     " + xmlPath + "\n" + " Message: " + exception.getMessage();
            ioError.addWarnNum();
            ioError.add(message);
            log.error(message);
            // throw new SAXException("Error encountered");
        }

        public void fatalError(SAXParseException exception) throws SAXException
        {
            // Bring things to a crashing halt
            String message = "**Parsing Fatal Error**\n" + " Line:    " + exception.getLineNumber() + "\n" + " URI:     " + xmlPath + "\n" + " Message: " + exception.getMessage();
            ioError.addFatalNum();
            ioError.add(message);
            log.fatal(message);
            // throw new SAXException("Fatal Error encountered");
        }

        /**
         * StringBufferオブジェクト
         */
        private StringBuffer value = new StringBuffer();

        /**
         * Vectorオブジェクト
         */
        private ArrayList<Record> vrecord = new ArrayList<Record>();//全体
        
        private ArrayList<Record> mrecord = null;//For MultiGeometry

        /**
         * Vectorオブジェクト
         */
        private ArrayList<UTMPoint> vpoint = new ArrayList<UTMPoint>();

        /**
         * Vectorオブジェクト
         */
        private ArrayList<Event> vevent = new ArrayList<Event>();

        /**
         * Recordオブジェクト
         */
        private Record record = null;

        /**
         * Eventオブジェクト
         */
        private Event event = null;

        /**
         * JulianDateオブジェクト
         */
        private JulianDate sdate = null;
        private JulianDate edate = null;

        /**
         * Locationオブジェクト
         */
        private Location location = null;

        /**
         * UTMPointオブジェクト
         */
        private UTMPoint point = null;

        /**
         * StringBufferオブジェクトに指定された文字列を追加します。
         * 
         * @param ch 文字列
         * @param start 開始位置
         * @param length 長さ
         * @throws SAXException
         */
        public void characters(char[] ch, int start, int length) throws SAXException
        {
            value.append(ch, start, length);
        }

        /**
         * 開始タグが出現した場合の処理です。
         * 
         * @param uri
         * @param localName
         * @param qName タグ名
         * @param attributes
         * @throws SAXException
         */
        public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
        {
//            log.debug(uri + " : " + localName + " : " + qName);
            if (uri.equals("http://www.opengis.net/kml/2.2")) {
                if (localName.equals("Placemark"))
                {
                    record = new Record();
                    record.setGroup("");
                    setGroup("");
                    vevent = new ArrayList<Event>();
                }
                else if (localName.equals("name"))
                {
                    event = new Event();
                }
                else if (localName.equals("description"))
                {
                    event = new Event();
                }
                else if (localName.equals("begin"))
                {
                    sdate = new JulianDate();
                }
                else if (localName.equals("end"))
                {
                    edate = new JulianDate();
                }
                else if (localName.equals("Point"))
                {
                    if (multiValue) {
                        record = new Record();
                    }
                    location = new Location();
                    location.setType(Location.POINT);
                }
                else if (localName.equals("Polygon"))
                {
                    if (multiValue) {
                        record = new Record();
                    }
                    location = new Location();
                    vpoint = new ArrayList<UTMPoint>();
                    location.setType(Location.LINE);
                }
                else if (localName.equals("coordinates"))
                {
                    point = new UTMPoint();
                }
                else if (localName.equals("MultiGeometry"))
                {
                    multiValue = true;
                    mrecord = new ArrayList<Record>();
                }
            } else if (uri.equals("http://www.w3.org/2005/Atom")) {
                if (localName.equals("link"))
                {
                    event = new Event();
                    event.setName("url");
                    event.setData(escapeXmlChar(attributes.getValue("href").trim()));
                    vevent.add(event);
                    event = null;
                }
            }
        }
        
        private boolean multiValue = false;

        /**
         * 終了タグが出現した場合の処理です。
         * 
         * @param uri
         * @param localName
         * @param qName タグ名
         * @throws SAXException
         */
        public void endElement(String uri, String localName, String qName) throws SAXException
        {
            if (qName.equals("Placemark"))
            {
                if (sdate != null && edate != null) {
                    Event[] events = new Event[vevent.size()];
                    events = vevent.toArray(events);
                    if (multiValue) {
                        for(Record r : mrecord)
                        {
                            try
                            {
                                r.setStartDatetime(sdate);
                                r.setEndDatetime(edate);
                            }
                            catch (RecordException e)
                            {
                                warning(new SAXParseException(e.getMessage(), locator));
                                mrecord = null;
                                break;
                            }
                            r.setEvents(events);
                        }
                        if (mrecord != null) {
                            vrecord.addAll(mrecord);
                            mrecord = null;
                        }
                    } else {
                        try
                        {
                            record.setStartDatetime(sdate);
                            record.setEndDatetime(edate);
                        }
                        catch (RecordException e)
                        {
                            warning(new SAXParseException(e.getMessage(), locator));
                            record = null;
                        }
                        if (record != null) {
                            record.setEvents(events);
                            vrecord.add(record);
                            record = null;
                        }
                    }
                }
                sdate = null;
                edate = null;
                mrecord = null;
                record = null;
                vevent = null;
            }
            else if (qName.equals("name"))
            {
                event.setName("title");
                event.setData(escapeXmlChar(value.toString().trim()));
                vevent.add(event);
                event = null;
            }
            else if (qName.equals("link"))
            {
//                event.setName("url");
//                event.setData(escapeXmlChar(value.toString().trim()));
//                vevent.add(event);
//                event = null;
            }
            else if (qName.equals("description"))
            {
                event.setName("detail");
                event.setData(escapeXmlChar(value.toString().trim()));
                vevent.add(event);
                event = null;
            }
            else if (qName.equals("begin"))
            {
                String isoDateStr = getValue();
                if (isoDateStr != null && !isoDateStr.equals(""))
                {
                    BigDecimal d = getJulianDate(isoDateStr);
                    if (d != null && record != null)
                    {
                        sdate.setDatetime(d);
                        sdate.setTimezone(getTimezone(isoDateStr));
                    }
                }
            }
            else if (qName.equals("end"))
            {
                String isoDateStr = getValue();
                if (isoDateStr != null && !isoDateStr.equals(""))
                {
                    BigDecimal d = getJulianDate(isoDateStr);
                    if (d != null && record != null)
                    {
                        edate.setDatetime(d);
                        edate.setTimezone(getTimezone(isoDateStr));
                    }
                }
            }
            else if (qName.equals("Point"))
            {
                UTMPoint[] points = new UTMPoint[1];
                points[0] = point;
                location.setPoints(points);
                record.setLocation(location);
                if (multiValue) {
                    mrecord.add(record);
                    record = null;
                }
                location = null;
            }
            else if (qName.equals("Polygon"))
            {
                UTMPoint[] points = new UTMPoint[vpoint.size()];
                points = vpoint.toArray(points);
                
//                for (int i = 0; i < points.length; i++) {
//                    log.debug("AAA " + points[i].getX() + " : " + points[i].getY());
//                }
                
                location.setPoints(points);
                record.setLocation(location);
                if (multiValue) {
                    mrecord.add(record);
                    record = null;
                }
                vpoint = null;
                location = null;
            }
            else if (qName.equals("coordinates"))
            {
                String[] ps0 = getValue().trim().replaceAll("[\\s]+", " ").split(" ");
                for (String ps : ps0) {
                    String[] ps1 = ps.split(",");
//                    log.debug("1 IN : " + ps1[1] + ":" + ps1[0]);
                    double[] utm = GIO.getWGS84UTM(Double.parseDouble(ps1[1]), Double.parseDouble(ps1[0]));
//                    log.debug("OUT " + (float)utm[0] + ":" + (float)utm[1]);
                    point = new UTMPoint();
                    point.setX((float)utm[0]);
                    point.setY((float)utm[1]);
                    point.setZone((int)utm[2]);
                    if (vpoint != null) {
                        vpoint.add(point);
                    }
                }
//                point = null;
            }
            else if (qName.equals("MultiGeometry"))
            {
//                multiValue = false;
            }
            value.delete(0, value.length());
        }

        /**
         * Record[]オブジェクトを取得します。
         * 
         * @return Record[]オブジェクト
         */
        public Record[] getRecords()
        {
            Record[] records = new Record[vrecord.size()];
            records = vrecord.toArray(records);
            vrecord = null;
            return records;
        }

        /**
         * ジュリアン日を取得します。
         * 
         * @param iso8601 日時
         * @return ジュリアン日
         * @throws SAXException
         */
        private BigDecimal getJulianDate(String iso8601) throws SAXException
        {
            try
            {
                return JulianGregorianDate.toJulian(iso8601);
            }
            catch (ISO8601Exception e)
            {
                String message = e.getMessage();
                warning(new SAXParseException(message, locator));
                return null;
            }
        }

        /**
         * タイムゾーンを取得します。
         * 
         * @param iso8601 日時
         * @return タイムゾーン
         * @throws SAXException
         */
        private String getTimezone(String iso8601) throws SAXException
        {
            try
            {
                return JulianGregorianDate.getTimezone(iso8601);
            }
            catch (ISO8601Exception e)
            {
                String message = e.getMessage();
                warning(new SAXParseException(message, locator));
                return null;
            }
        }

        /**
         * 値を取得します。
         * 
         * @return 文字列のコピーを返します。
         */
        private String getValue()
        {
            return value.toString().trim();
        }
    }

    /**
     * ファイルを読み込みます。
     * 
     * @param is InputStreamオブジェクト
     * @param code サポートする charset の名前
     * @return Record[]オブジェクト
     * @throws RecordIOFilterException
     */
    @Override
    public Record[] read(InputStream is, String code, IOError ioError) throws RecordIOFilterException
    {
        return read(is, ioError);
    }
}