/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.io;

import java.io.IOException;
import java.io.InputStream;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.math.BigDecimal;
import java.util.ArrayList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.Event;
import jp.nihu.gttime.data.JulianDate;
import jp.nihu.gttime.data.Location;
import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.data.RecordException;
import jp.nihu.gttime.data.UTMPoint;
import jp.nihu.gttime.plugin.IOError;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.plugin.RecordIOFilter;
import jp.nihu.gttime.plugin.RecordIOFilterException;
import jp.nihu.gttime.time.utility.ISO8601Exception;
import jp.nihu.gttime.time.utility.JulianGregorianDate;
import jp.nihu.gttime.utility.Utilities;
import jp.nihu.gttime.utility.XMLUtil;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.Attributes;
import org.xml.sax.Locator;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;
import org.xml.sax.helpers.LocatorImpl;

/**
 * ファイルの内容を読み書きします。
 */
public class NIHUTimeSpaceXmlIOFilter extends RecordIOFilter
{
    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(NIHUTimeSpaceXmlIOFilter.class);

    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(NIHUTimeSpaceXmlIOFilter.class, T2Map.LOCALE);

    private static String xmlPath = null;

    /**
     * 出力先パスを取得します。
     * 
     * @param tmmPath tmmPath パス
     * @return パス
     * @throws RecordIOFilterException
     */
    public String getPath(String tmmPath) throws RecordIOFilterException
    {
        int x = tmmPath.lastIndexOf(".");
        xmlPath = tmmPath.substring(0, x) + ".xml";
        // log.info("read dataset: " + xmlPath);
        return xmlPath;
    }

    /**
     * xmlファイルを読み込みます。
     * 
     * @param is InputStreamオブジェクト
     * @return Record[]オブジェクト
     * @throws RecordIOFilterException
     */
//    public Record[] read(InputStream is, IOError ioErrorx) throws RecordIOFilterException
//    {
//        ioError = ioErrorx;
//        SAXParserFactory factory = SAXParserFactory.newInstance();
//        SAXParser parser = null;
//        try
//        {
//            parser = factory.newSAXParser();
//        }
//        catch (ParserConfigurationException e)
//        {
//            throw new RecordIOFilterException(rmgr.getString("xml.parser.error"), e);
//        }
//        catch (SAXException e)
//        {
//            throw new RecordIOFilterException(rmgr.getString("xml.parser.error"), e);
//        }
//        NIHUTimeSpaceXmlParser xmlHandler = new NIHUTimeSpaceXmlParser();
//        try
//        {
//            parser.parse(is, xmlHandler);
//        }
//        catch (SAXException e)
//        {
//            throw new RecordIOFilterException(rmgr.getString("dataset.analysis.error"), e);
//        }
//        catch (IOException e)
//        {
//            throw new RecordIOFilterException(rmgr.getString("dataset.acquisition.error"), e);
//        }
//        return xmlHandler.getRecords();
//    }
    
    private NIHUTimeSpaceXmlParser errorHandler = null;
    
    public Record[] read(InputStream is, IOError ioErrorx) throws RecordIOFilterException {
    	ioError = ioErrorx;
    	DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
    	dbf.setNamespaceAware(true);
	    DocumentBuilder db = null;
	    try {
	    	db = dbf.newDocumentBuilder();
	    	errorHandler = new NIHUTimeSpaceXmlParser();
	    	db.setErrorHandler(errorHandler);
        } catch (ParserConfigurationException e) {
            throw new RecordIOFilterException(rmgr.getString("xml.parser.error"), e);
        }
	    try{
    	    Document doc = db.parse(is);
    	    ArrayList<Record> list = procRecords(doc.getDocumentElement());
    	    Record[] records = new Record[list.size()];
            records = list.toArray(records);
            list = null;
            return records;
    	}catch(IOException e){
    		throw new RecordIOFilterException(rmgr.getString("dataset.acquisition.error"), e);
    	} catch (SAXException e) {
    		throw new RecordIOFilterException(rmgr.getString("dataset.analysis.error"), e);
		}
    }
    
    
    private ArrayList<Record> procRecords(Node node) throws RecordIOFilterException {
    	ArrayList<Record> list = new ArrayList<Record>();
    	NodeList nl = node.getChildNodes();
    	for (int i = 0; i < nl.getLength(); i++) {
    		int type = nl.item(i).getNodeType();
        	if (type == Node.ELEMENT_NODE) {
        		Record record = new Record();
        		procRecord(nl.item(i), record);
        		list.add(record);
        	}
		}
    	return list;
    }
    
    private void procRecord(Node node, Record record) throws RecordIOFilterException {

    	record.setNode(node);
    	
		Element e0 = (Element)node;//record
		//e0.getAttribute("level");
		String level = e0.getAttribute("level");
        if (level != null && level.length() > 0) {
            record.setLevel(Integer.parseInt(level));
        }
        String id = e0.getAttribute("id");
        if (id != null && id.length() > 0) {
//        	e0.setIdAttribute("id", true);
//        	System.out.println(id);
            record.setId(id);
        }
        String group = e0.getAttribute("group");
        if (group != null && group.length() > 0) {
            record.setGroup(group);
            setGroup(group);
        } else {
            setGroup("");
        }
        String maxVisibleTResolution = e0.getAttribute("maxVisibleTResolution");
        if (maxVisibleTResolution != null && maxVisibleTResolution.length() > 0) {
            record.setMaxVisibleTResolution(maxVisibleTResolution);
        }
        String minVisibleTResolution = e0.getAttribute("minVisibleTResolution");
        if (minVisibleTResolution != null && minVisibleTResolution.length() > 0) {
            record.setMinVisibleTResolution(minVisibleTResolution);
        }
        
        NodeList nl0 = e0.getChildNodes();
        for (int i = 0; i < nl0.getLength(); i++) {
        	if (nl0.item(i).getNodeType() == Node.ELEMENT_NODE) {
        		Element e1 = (Element)nl0.item(i);
        		//log.debug(e1 + " :: " + e1.getNodeName() + " :: " + e1.getNamespaceURI() + " :: " + e1.getPrefix() + " :: " + e1.getLocalName());
        		if (e1.getLocalName().equals("datetime")) {
        			NodeList nl1 = e1.getChildNodes();
        			for (int j = 0; j < nl1.getLength(); j++) {
        				if (nl1.item(j).getNodeType() == Node.ELEMENT_NODE) {
        					Element e2 = (Element)nl1.item(j);
        					try {
        		                if (e2.getLocalName().equals("from")) {
        		                	record.setStartDatetime(getJulianDate(e2));
        		                } else if (e2.getLocalName().equals("to")) {
        		                	record.setEndDatetime(getJulianDate(e2));
        		                }
    						} catch(Exception e) {
    							try {
									errorHandler.warning(new SAXParseException(e.getMessage(), errorHandler.locator));
								} catch (SAXException e3) {
									// never catch
									e3.printStackTrace();
								}
    						}
        				}
        			}
        		} else if (e1.getLocalName().equals("location")) {
        			Location location = new Location();
        			ArrayList<UTMPoint> pointl = new ArrayList<UTMPoint>();
        			
        			NodeList nl1 = e1.getChildNodes();
        			for (int j = 0; j < nl1.getLength(); j++) {
        				if (nl1.item(j).getNodeType() == Node.ELEMENT_NODE) {
        					Element e2 = (Element)nl1.item(j);
        					if (e2.getLocalName().equals("point")) {
        						UTMPoint point = getUTMPoint(e2);
        						pointl.add(point);
        		            } else if (e2.getLocalName().equals("line")) {
        		                location.setType(Location.LINE);
        		                NodeList nl2 = e1.getChildNodes();
                    			for (int k = 0; k < nl2.getLength(); k++) {
                    				if (nl1.item(k).getNodeType() == Node.ELEMENT_NODE) {
                    					Element e3 = (Element)nl2.item(j);
                    					if (e3.getLocalName().equals("point")) {
                    						UTMPoint point = getUTMPoint(e3);
                    						pointl.add(point);
                    		            }
                    				}
                    			}
        		            } else if (e2.getLocalName().equals("plane")) {
        		                location.setType(Location.PLANE);
        		                NodeList nl2 = e1.getChildNodes();
                    			for (int k = 0; k < nl2.getLength(); k++) {
                    				if (nl1.item(k).getNodeType() == Node.ELEMENT_NODE) {
                    					Element e3 = (Element)nl2.item(j);
                    					if (e3.getLocalName().equals("point")) {
                    						UTMPoint point = getUTMPoint(e3);
                    						pointl.add(point);
                    		            }
                    				}
                    			}
        		            }
        				}
        			}
        			
        			UTMPoint[] points = new UTMPoint[pointl.size()];
                    points = pointl.toArray(points);
                    location.setPoints(points);
                    record.setLocation(location);
        		} else if (e1.getLocalName().equals("event")) {
        			ArrayList<Event> eventl = new ArrayList<Event>();
        			NodeList nl1 = e1.getChildNodes();
        			for (int j = 0; j < nl1.getLength(); j++) {
        				if (nl1.item(j).getNodeType() == Node.ELEMENT_NODE) {
        					Element e2 = (Element)nl1.item(j);
        					if (e2.getLocalName().equals("item")) {
        						String name = e2.getAttribute("name");
        		                if (name != null && name.length() > 0)
        		                {
        		                	boolean hasEvent = false;
        		                	Event event = new Event();
        		                    event.setName(name);
        		                    
        		                    if (name.equals("number"))
        		                    {
        		                    	String val = e2.getTextContent().trim();
        		                        if (val.matches("^-?[0-9]+([.][0-9]+)?$"))
        		                        {
        		                    	    hasEvent = true;
        		                        	event.setData(val);
        		                        }
        		                        else
        		                        {
        		                        	String message = "### XML Reader ### value of (item/@name=number) is not number: " + val;
        		                        	try {
												errorHandler.warning(new SAXParseException(message, errorHandler.locator));
											} catch (SAXException e) {
												// TODO Auto-generated catch block
												e.printStackTrace();
											}
        		                        }
        		                    }
        		                    else
        		                    {
        		                    	hasEvent = true;
//        		                        event.setData(e2.getTextContent().trim());
        		                        event.setData(XMLUtil.getElementContentsAsString(e2));
        		                    }
        		                    
        		                    if (hasEvent) {
        		                    	eventl.add(event);
        		                    }
        		                }
        		                else
        		                {
        		                    String message = "name attribute must be specified for item element.";
        		                    try {
										errorHandler.warning(new SAXParseException(message, errorHandler.locator));
									} catch (SAXException e) {
										// TODO Auto-generated catch block
										e.printStackTrace();
									}
        		                }
        		            }
        				}
        			}
        			Event[] events = new Event[eventl.size()];
        			eventl.toArray(events);
                    record.setEvents(events);
         		}
        	}
        }
    }
    
    private JulianDate getJulianDate(Element elem) throws ISO8601Exception {
    	JulianDate date = new JulianDate();
    	String min = elem.getAttribute("min");
        if (min != null && min.length() > 0)
        {
            BigDecimal d = JulianGregorianDate.toJulian(min);
            if (d != null)
            {
                date.setMin(d);
                date.setTimezone(JulianGregorianDate.getTimezone(min));
            }
        }
        String max = elem.getAttribute("max");
        if (max != null && max.length() > 0)
        {
            BigDecimal d = JulianGregorianDate.toJulian(max);
            if (d != null)
            {
                date.setMax(d);
                date.setTimezone(JulianGregorianDate.getTimezone(max));
            }
        }
        String edge = elem.getAttribute("endpoint");
        if (edge != null && edge.length() > 0)
        {
            if (!edge.equals("open") && !edge.equals("close"))
            {
                String message = "the value of edge attribute is not correct: " + edge;
                throw new ISO8601Exception(message);
            }
            date.setEdge(edge);
        }
        
        String isoDateStr = elem.getTextContent().trim();
        // log.debug("from" + isoDateStr);
        if (isoDateStr != null && !isoDateStr.equals("")) {
            BigDecimal d = JulianGregorianDate.toJulian(isoDateStr);
            date.setDatetime(d);
            date.setTimezone(JulianGregorianDate.getTimezone(isoDateStr));
        } else {
        	String message = "value of " + elem.getLocalName() + " datetime is not correct: " + edge;
            throw new ISO8601Exception(message);
        }
        
    	return date;
    }
    
    private UTMPoint getUTMPoint(Element elem) {
    	UTMPoint point = new UTMPoint();
    	NodeList nl = elem.getChildNodes();
		for (int i = 0; i < nl.getLength(); i++) {
			if (nl.item(i).getNodeType() == Node.ELEMENT_NODE) {
				Element e = (Element)nl.item(i);
				if (e.getLocalName().equals("zone"))
		        {
		            if (e.getTextContent().length() > 0)
		            {
		                point.setZone(Integer.parseInt(e.getTextContent()));
		            }
		        }
		        else if (e.getLocalName().equals("x"))
		        {
		            if (e.getTextContent().length() > 0)
		            {
		                point.setX(Float.parseFloat(e.getTextContent()));
		            }
		        }
		        else if (e.getLocalName().equals("y"))
		        {
		            if (e.getTextContent().length() > 0)
		            {
		                point.setY(Float.parseFloat(e.getTextContent()));
		            }
		        }
			}
		}
    	return point;
    }
    
    private byte[] writeRecordDom(Record[] records) throws RecordIOFilterException {
    	byte[] bxml = null;
    	StringWriter writer = null;
    	try {
    		String xml = "";

    		Document document = records[0].getNode().getOwnerDocument();
    		
            writer = new StringWriter();
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.transform(new DOMSource(document), new StreamResult(writer));
            xml = writer.toString();
            bxml = xml.getBytes("UTF-8");
    	} catch (Exception e) {
    		e.printStackTrace();
    		throw new RecordIOFilterException(e.getMessage(), e);
    	} finally {
			if (writer != null) {
				try {
					writer.close();
				} catch (IOException e1) {
					e1.printStackTrace();
					throw new RecordIOFilterException(e1.getMessage(), e1);
				}
			}
		}
    	return bxml;
    }
    
    private byte[] writeRecord(Record[] records) throws RecordIOFilterException {
    	byte[] bxml = null;
    	StringBuffer sb = new StringBuffer("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
        sb.append("<records xmlns=\"http://www.hutime.org/ns/hutimedata/2015/03/\" dataVersion=\"1.0\">\n");
        for (int i = 0; i < records.length; i++)
        {
            sb.append("<record");
            int level = records[i].getLevel();
            if (level > -1)
            {
                sb.append(" level=\"" + level + "\"");
            }
            String id = records[i].getId();
            if (id != null)
            {
                sb.append(" id=\"" + id + "\"");
            }
            String group = records[i].getGroup();
            if (group != null)
            {
                sb.append(" group=\"" + group + "\"");
            }
            String minVisibleTResolution = Utilities.getDurationVal(records[i].getMinVisibleTResolution());
            if (minVisibleTResolution != null)
            {
                sb.append(" minVisibleTResolution=\"" + minVisibleTResolution + "\"");
            }
            String maxVisibleTResolution = Utilities.getDurationVal(records[i].getMaxVisibleTResolution());
            if (maxVisibleTResolution != null)
            {
                sb.append(" maxVisibleTResolution=\"" + maxVisibleTResolution + "\"");
            }
            sb.append(">\n");

            sb.append("<datetime>\n");

            JulianDate js = records[i].getStartDatetime();
            sb.append("<from");
            if (js != null)
            {
                String e = js.getEdge();
                if (e != null)
                {
                    sb.append(" endpoint=\"").append(e).append("\"");
                }
            }
            sb.append(">\n");
            if (js != null)
            {
                sb.append(JulianGregorianDate.toGregorian(js.getDatetime(), js.getTimezone()));
            }
            sb.append("</from>\n");

            JulianDate je = records[i].getEndDatetime();
            sb.append("<to");
            if (je != null)
            {
                String e = je.getEdge();
                if (e != null)
                {
                    sb.append(" endpoint=\"").append(e).append("\"");
                }
            }
            sb.append(">\n");
            if (je != null)
            {
                sb.append(JulianGregorianDate.toGregorian(je.getDatetime(), je.getTimezone()));
            }
            sb.append("</to>\n");
            sb.append("</datetime>\n");

            Location location = records[i].getLocation();
            if (location != null)
            {
                sb.append("<location>");
                int type = location.getType();
                String elemName = "point";
                if (type == Location.LINE)
                {
                    elemName = "line";
                }
                else if (type == Location.PLANE)
                {
                    elemName = "plane";
                }
                if (type != Location.POINT)
                {
                    sb.append("<" + elemName + ">\n");
                }

                UTMPoint[] points = location.getPoints();
                for (int j = 0; j < points.length; j++)
                {
                    sb.append("<point>\n");
                    sb.append("<zone>" + points[j].getZone() + "</zone>\n");
                    sb.append("<x>" + points[j].getX() + "</x>\n");
                    sb.append("<y>" + points[j].getY() + "</y>\n");
                    sb.append("</point>\n");
                }
                if (type != Location.POINT)
                {
                    sb.append("</" + elemName + ">\n");
                }
                sb.append("</location>");
            }

            sb.append("<event>\n");
            Event[] events = records[i].getEvents();
            if (events != null)
            {
                for (int j = 0; j < events.length; j++)
                {
                    sb.append("<item name=\"" + events[j].getName() + "\">" + escapeXmlChar(events[j].getData()) + "</item>\n");
                }
            }
            sb.append("</event>\n");

            sb.append("</record>\n");
        }
        sb.append("</records>\n");
        try
        {
            bxml = sb.toString().getBytes("UTF-8");
        }
        catch (UnsupportedEncodingException e)
        {
            // never catch
            throw new RecordIOFilterException(e.getMessage(), e);
        }
    	return bxml;
    }
    
    /**
     * xmlファイルに内容を書き込みます。
     * 
     * @param records Record[]オブジェクト
     * @return byte[]
     * @throws RecordIOFilterException
     */
    public byte[] write(Record[] records) throws RecordIOFilterException
    {
        byte[] bxml = null;
        
        Node nodeCheck = records[0].getNode();
        if (nodeCheck != null) {
        	bxml = writeRecordDom(records);
        } else {
        	bxml = writeRecord(records);
        }
        return bxml;
    }

    /**
     * 禁則文字を置き換えます。
     * 
     * @param str 文字
     * @return 置き換えられた文字
     */
    private String escapeXmlChar(String str)
    {
        str = str.trim();
        str = str.replaceAll("&", "&amp;");
        str = str.replaceAll("<", "&lt;");
        str = str.replaceAll(">", "&gt;");
        str = str.replaceAll("\"", "&quot;");
        str = str.replaceAll("\'", "&apos;");
        return str;
    }

    /**
     * レコードのチェックや設定を行うクラスです。
     */
    class NIHUTimeSpaceXmlParser extends DefaultHandler
    {
        public Locator locator = null;

        public NIHUTimeSpaceXmlParser()
        {
            setDocumentLocator(new LocatorImpl());
        }

        public void setDocumentLocator(Locator locator)
        {
            this.locator = locator;
        }

        public void warning(SAXParseException exception) throws SAXException
        {
            String message = "**Parsing Warning**\n" + " Line:    " + exception.getLineNumber() + "\n" + " URI:     " + xmlPath + "\n" + " Message: " + exception.getMessage();
            ioError.addWarnNum();
            ioError.add(message);
            log.warn(message);
            // throw new SAXException("Warning encountered");
        }

        public void error(SAXParseException exception) throws SAXException
        {
            String message = "**Parsing Error**\n" + " Line:    " + exception.getLineNumber() + "\n" + " URI:     " + xmlPath + "\n" + " Message: " + exception.getMessage();
            ioError.addWarnNum();
            ioError.add(message);
            log.error(message);
            // throw new SAXException("Error encountered");
        }

        public void fatalError(SAXParseException exception) throws SAXException
        {
            // Bring things to a crashing halt
            String message = "**Parsing Fatal Error**\n" + " Line:    " + exception.getLineNumber() + "\n" + " URI:     " + xmlPath + "\n" + " Message: " + exception.getMessage();
            ioError.addFatalNum();
            ioError.add(message);
            log.fatal(message);
            // throw new SAXException("Fatal Error encountered");
        }

        /**
         * StringBufferオブジェクト
         */
        private StringBuffer value = new StringBuffer();

        /**
         * Vectorオブジェクト
         */
        private ArrayList<Record> vrecord = new ArrayList<Record>();

        /**
         * Vectorオブジェクト
         */
        private ArrayList<UTMPoint> vpoint = new ArrayList<UTMPoint>();

        /**
         * Vectorオブジェクト
         */
        private ArrayList<Event> vevent = new ArrayList<Event>();

        /**
         * Recordオブジェクト
         */
        private Record record = null;

        /**
         * Eventオブジェクト
         */
        private Event event = null;

        /**
         * JulianDateオブジェクト
         */
        private JulianDate date = null;

        /**
         * Locationオブジェクト
         */
        private Location location = null;

        /**
         * UTMPointオブジェクト
         */
        private UTMPoint point = null;

        /**
         * StringBufferオブジェクトに指定された文字列を追加します。
         * 
         * @param ch 文字列
         * @param start 開始位置
         * @param length 長さ
         * @throws SAXException
         */
        public void characters(char[] ch, int start, int length) throws SAXException
        {
            value.append(ch, start, length);
        }

        /**
         * 開始タグが出現した場合の処理です。
         * 
         * @param uri
         * @param localName
         * @param qName タグ名
         * @param attributes
         * @throws SAXException
         */
        public void startElement(String uri, String localName, String qName, Attributes attributes) throws SAXException
        {
            if (qName.equals("record"))
            {
            	value.delete(0, value.length());
                record = new Record();
                String level = attributes.getValue("", "level");
                if (level != null && level.length() > 0)
                {
                    record.setLevel(Integer.parseInt(level));
                }
                String id = attributes.getValue("", "id");
                if (id != null && id.length() > 0)
                {
                    record.setId(id);
                }
                String group = attributes.getValue("", "group");
                if (group != null && group.length() > 0)
                {
                    record.setGroup(group);
                    setGroup(group);
                } else {
                    setGroup("");
                }

                String maxVisibleTResolution = attributes.getValue("", "maxVisibleTResolution");
                if (maxVisibleTResolution != null && maxVisibleTResolution.length() > 0)
                {
                    record.setMaxVisibleTResolution(maxVisibleTResolution);
                }

                String minVisibleTResolution = attributes.getValue("", "minVisibleTResolution");
                if (minVisibleTResolution != null && minVisibleTResolution.length() > 0)
                {
                    record.setMinVisibleTResolution(minVisibleTResolution);
                }
            }
            else if (qName.equals("datetime"))
            {
            	value.delete(0, value.length());
            }
            else if (qName.equals("location"))
            {
            	value.delete(0, value.length());
                location = new Location();
                vpoint = new ArrayList<UTMPoint>();
            }
            else if (qName.equals("event"))
            {
            	value.delete(0, value.length());
                vevent = new ArrayList<Event>();
            }
            else if (qName.equals("from") || qName.equals("to"))
            {
            	value.delete(0, value.length());
                date = new JulianDate();
                String min = attributes.getValue("min");
                if (min != null && min.length() > 0)
                {
                    BigDecimal d = getJulianDate(min);
                    if (d != null)
                    {
                        date.setMin(d);
                        date.setTimezone(getTimezone(min));
                    }
                }
                String max = attributes.getValue("max");
                if (max != null && max.length() > 0)
                {
                    BigDecimal d = getJulianDate(max);
                    if (d != null)
                    {
                        date.setMax(d);
                        date.setTimezone(getTimezone(max));
                    }
                }
                String edge = attributes.getValue("endpoint");
                if (edge != null && edge.length() > 0)
                {
                    if (!edge.equals("open") && !edge.equals("close"))
                    {
                        String message = "the value of edge attribute is not correct: " + edge;
                        warning(new SAXParseException(message, locator));
                    }
                    date.setEdge(edge);
                }
            }
            else if (qName.equals("point"))
            {
            	value.delete(0, value.length());
                point = new UTMPoint();
            }
            else if (qName.equals("line"))
            {
            	value.delete(0, value.length());
                location.setType(Location.LINE);
            }
            else if (qName.equals("plane"))
            {
            	value.delete(0, value.length());
                location.setType(Location.PLANE);
            }
            else if (qName.equals("zone"))
            {
            	value.delete(0, value.length());
            }
            else if (qName.equals("x"))
            {
            	value.delete(0, value.length());
            }
            else if (qName.equals("y"))
            {
            	value.delete(0, value.length());
            }
            else if (qName.equals("item"))
            {
            	value.delete(0, value.length());
                String name = attributes.getValue("", "name");
                if (name != null && name.length() > 0)
                {
                    event = new Event();
                    event.setName(name);
                }
                else
                {
                    event = null;
                    String message = "name attribute must be specified for item element.";
                    warning(new SAXParseException(message, locator));
                }
            }
        }

        /**
         * 終了タグが出現した場合の処理です。
         * 
         * @param uri
         * @param localName
         * @param qName タグ名
         * @throws SAXException
         */
        public void endElement(String uri, String localName, String qName) throws SAXException
        {
            if (qName.equals("record"))
            {
                if (record != null)
                {
                    // if (record.getStartDatetime() != null &&
                    // record.getEndDatetime() != null)
                    {
                        vrecord.add(record);
                    }
                    record = null;
                }
            }
            else if (qName.equals("datetime"))
            {
            }
            else if (qName.equals("location"))
            {
                if (record != null)
                {
                    UTMPoint[] points = new UTMPoint[vpoint.size()];
                    points = vpoint.toArray(points);
                    location.setPoints(points);
                    record.setLocation(location);
                }
                vpoint = null;
                location = null;

            }
            else if (qName.equals("event"))
            {
                if (record != null)
                {
                    Event[] events = new Event[vevent.size()];
                    vevent.toArray(events);
                    record.setEvents(events);
                }
                vevent = null;
            }
            else if (qName.equals("from"))
            {
                String isoDateStr = getValue();
                // log.debug("from" + isoDateStr);
                if (isoDateStr != null && !isoDateStr.equals(""))
                {
                    BigDecimal d = getJulianDate(isoDateStr);
                    if (d != null && record != null)
                    {
                        date.setDatetime(d);
                        date.setTimezone(getTimezone(isoDateStr));
                        try
                        {
                            record.setStartDatetime(date);
                        }
                        catch (RecordException e)
                        {
                            warning(new SAXParseException(e.getMessage(), locator));
                            record = null;
                        }
                        date = null;
                    }
                    else
                    {
                        record = null;
                    }
                }
            }
            else if (qName.equals("to"))
            {
                String isoDateStr = getValue();
                // log.debug("to" + isoDateStr);
                if (isoDateStr != null && !isoDateStr.equals(""))
                {
                    BigDecimal d = getJulianDate(isoDateStr);
                    if (d != null && record != null)
                    {
                        date.setDatetime(d);
                        date.setTimezone(getTimezone(isoDateStr));
                        try
                        {
                            record.setEndDatetime(date);
                        }
                        catch (RecordException e)
                        {
                            warning(new SAXParseException(e.getMessage(), locator));
                            record = null;
                        }
                        date = null;
                    }
                    else
                    {
                        record = null;
                    }
                }
            }
            else if (qName.equals("point"))
            {
                vpoint.add(point);
                point = null;
            }
            else if (qName.equals("line"))
            {
            }
            else if (qName.equals("plane"))
            {
            }
            else if (qName.equals("zone"))
            {
                if (getValue().length() > 0)
                {
                    point.setZone(Integer.parseInt(getValue()));
                }
            }
            else if (qName.equals("x"))
            {
                if (getValue().length() > 0)
                {
                    point.setX(Float.parseFloat(getValue()));
                }
            }
            else if (qName.equals("y"))
            {
                if (getValue().length() > 0)
                {
                    point.setY(Float.parseFloat(getValue()));
                }
            }
            else if (qName.equals("item"))
            {
                if (event != null)
                {
                    if (event.getName().equals("number"))
                    {
                        // if (getValue().matches("^-?[0-9]+([.][0-9]+)?$"))
                        // {
                        event.setData(getValue());
                        vevent.add(event);
                        // }
                        // else
                        // {
                        // String message = "### XML Reader ### value of
                        // (item/@name=number) is not number: " + getValue();
                        // warning(new SAXParseException(message, locator));
                        // }
                    }
                    else
                    {
                        event.setData(getValue());
                        vevent.add(event);
                    }
                    event = null;
                }
            }
        }

        /**
         * Record[]オブジェクトを取得します。
         * 
         * @return Record[]オブジェクト
         */
        public Record[] getRecords()
        {
            Record[] records = new Record[vrecord.size()];
            records = vrecord.toArray(records);
            vrecord = null;
            return records;
        }

        /**
         * ジュリアン日を取得します。
         * 
         * @param iso8601 日時
         * @return ジュリアン日
         * @throws SAXException
         */
        private BigDecimal getJulianDate(String iso8601) throws SAXException
        {
            try
            {
                return JulianGregorianDate.toJulian(iso8601);
            }
            catch (ISO8601Exception e)
            {
                String message = e.getMessage();
                warning(new SAXParseException(message, locator));
                return null;
            }
        }

        /**
         * タイムゾーンを取得します。
         * 
         * @param iso8601 日時
         * @return タイムゾーン
         * @throws SAXException
         */
        private String getTimezone(String iso8601) throws SAXException
        {
            try
            {
                return JulianGregorianDate.getTimezone(iso8601);
            }
            catch (ISO8601Exception e)
            {
                String message = e.getMessage();
                warning(new SAXParseException(message, locator));
                return null;
            }
        }

        /**
         * 値を取得します。
         * 
         * @return 文字列のコピーを返します。
         */
        private String getValue()
        {
            return value.toString().trim();
        }
    }

    /**
     * ファイルを読み込みます。
     * 
     * @param is InputStreamオブジェクト
     * @param code サポートする charset の名前
     * @return Record[]オブジェクト
     * @throws RecordIOFilterException
     */
    @Override
    public Record[] read(InputStream is, String code, IOError ioError) throws RecordIOFilterException
    {
        return read(is, ioError);
    }
}