/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.layer;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.math.BigDecimal;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import javax.swing.ImageIcon;
import javax.xml.xpath.XPathExpressionException;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.DatasetIF;
import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.data.property.PropertiesGts;
import jp.nihu.gttime.data.property.PropertiesIF;
import jp.nihu.gttime.plugin.LayerPlugin;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.preferences.BaseLayerPluginPreferences;
import jp.nihu.gttime.time.TimeAxisIF;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * レイヤについての年表形式による描画機能（プラグイン）を実装するクラスである。
 */
public class DefaultLayerPlugin extends LayerPlugin
{
    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(DefaultLayerPlugin.class, T2Map.LOCALE);

    public String getName()
    {
        return rmgr.getString("name");
    }

    public String getPluginName()
    {
        return rmgr.getString("plugin.name");
    }

    public String getStyleName()
    {
        return getPluginName() + ".type";
    }

    /**
     * シリアルバーションID
     */
    private static final long serialVersionUID = 1L;

    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(DefaultLayerPlugin.class);

    /**
     * レイヤ情報の初期化します。
     */
    public void initLayer()
    {
        log.debug("Init Default Layer.");
        // reduceSwitch = REDUCE_PLOT;
        // if (dataset.getProperties() != null)
        // {
        // dataset.getProperties().setDataReduction(REDUCE_PLOT);
        // }
        parseRecord(dataset, data);
        popupMenu = getMenu();
    }

    public BaseLayerPluginPreferences getLayerPluginPreferencesDialog()
    {
        BaseLayerPluginPreferences prep = new DefaultLayerPluginPreferences();
//        List<String> groups = dataset.getGroups();
//        for (int i = 0; i < groups.size(); i++) {
//            prep.initValueForLayer(dataset.getProperties(), groups.get(i));
//        }
        prep.initValueForLayer(dataset.getProperties());
        return prep;
    }

    protected void setLayerProperties()
    {
        PropertiesIF properties = dataset.getProperties();
        if (properties == null)
        {
            properties = new PropertiesGts();
            dataset.setProperties(properties);
        }
        try
        {
            List<String> groups = dataset.getGroups();
            for (int i = 0; i < groups.size(); i++) {
                String group = groups.get(i);
                if (properties.getPolygon(group) == null) {
                    setPolygonProperties(group);
                }
                if (properties.getFont(group) == null) {
                    setFontProperties(group);
                }
                if (properties.getFontColor(group) == null) {
                    setFontColorProperties(group);
                }
            }
            if (properties.getSelectedColor() == null)
            {
                setSelectedColorProperties();
            }
            if (properties.getBackgroundColor() == null)
            {
                setBackgroundColorProperties();
            }
            if (properties.getDetailFontName() == null)
            {
                setDetailFontNameProperties();
            }
            
            if (properties.isShowGroupName() == null)
            {
                setShowGroupNameProperties();
            }
            
            if (properties.getLayerHeight() == null)
            {
                setLayerHeightProperties();
            }
            if (properties.isDataReduction() == null)
            {
                setDataReductionProperties();
            }
            if (properties.getDataReductionNum() == null)
            {
                setDataReductionNumProperties();
            }
        }
        catch (XPathExpressionException e)
        {
            log.fatal(e.getMessage(), e);
        }
    }

    /**
     * 年表の表示位置を算出している。<br/> 一定の間隔を空けて下から上に向かって描画しており、<br/>
     * 上部に描画できなくなると下から描画するようにしている
     * 
     * @param g
     */
    public void paintLayer(Graphics g)
    {
    	Set<String> group = new HashSet<String>();
        Iterator<RecordPlot> it = getProcRecordPlot();
        while (it.hasNext())
        {
            DefaultRecordPlot plot = (DefaultRecordPlot) it.next();
            if (isPaintResolution(plot.record))
            {
                int y = getPositionY(plot.getNumber());
                if (!group.contains(plot.getGroup())) {
                	plot.printLegend(g, y);
                    group.add(plot.getGroup());
                }
                ((DefaultRecordPlot) plot).paint(g, y);
            }
        }
    }

    @Override
    protected boolean doReduce(int ps)
    {
        if (dataset.getProperties().getDataReductionNum() < ps && dataset.getProperties().isDataReduction())
        {
            return true;
        }
        else
        {
            return false;
        }
    }

    /**
     * イベントオブジェクトの1つずつに各リスナーを登録します。
     * 
     * @param dataset Datasetオブジェクト
     * @param data TimeAxisIFオブジェクト
     */
    protected void parseRecord(DatasetIF dataset, TimeAxisIF data)
    {
        // try
        // {
        // Record[] records = dataset.getRecords(dataset.getStartDatetime(),
        // dataset.getEndDatetime());
        Record[] records = dataset.getRecords();
        int rl = records.length;
        for (int i = 0; i < rl; i++)
        {
            DefaultRecordPlot rp = (DefaultRecordPlot) createRecordPlot(records[i], data);
            rp.setNumber(i);
            addRecordPlot(rp);
        }
        // }
        // catch (DatasetException e)
        // {
        // log.fatal(e.getMessage(), e);
        // }
    }

    /**
     * プロットを新規作成します。
     * 
     * @param record Recordオブジェクト
     * @param data TimeAxisIFオブジェクト
     * @return RecordPlot
     */
    @Override
    protected RecordPlot createRecordPlot(Record record, TimeAxisIF data)
    {
        return new DefaultRecordPlot(record, data);
    }

    /**
     * 年表オブジェクトのY軸表示位置が重ならないように30ピクセルずらす。
     * 
     * @param remain 縦の年表個数
     * @return 年表オブジェクトのY軸位置
     */
    protected int getPositionY(int position)
    {
        int itemCnt = (getHeight() - 40) / 30;
        int ypos = 1;
        if (itemCnt > 0)
        {
            ypos = (position + 1) % itemCnt;
            if (ypos == 0)
            {
                ypos = itemCnt;
            }
        }
        return ypos * 30;
    }

    /**
     * 年表の描画やイベントオブジェクトを押下した際の動作などを実装している
     */
    public class DefaultRecordPlot extends RecordPlot
    {
        /**
         * シリアルバーションID
         */
        private static final long serialVersionUID = 1L;
        
        private boolean isCircle = false;
        
        public boolean isCircle() {
        	return isCircle;
        }

        private int number = -1;

        public void setNumber(int number)
        {
            this.number = number;
        }

        public int getNumber()
        {
            return number;
        }

        /**
         * 年表描画に関するコンストラクタ
         * 
         * @param record Recordオブジェクト
         * @param data TimeAxisIFオブジェクト
         */
        public DefaultRecordPlot(Record record, TimeAxisIF data)
        {
            super(record, data);
        }

//        /**
//         * @see jp.nihu.gttime.plugin.LayerPlugin.RecordPlot#getHtml()
//         */
//        public String getHtml()
//        {
//            String startStr = "None";
//            if (jdStart != null)
//            {
//                startStr = JulianGregorianDate.toGregorian(jdStart.getDatetime(), jdStart.getTimezone());
//            }
//            String endStr = "None";
//            if (jdEnd != null)
//            {
//                endStr = JulianGregorianDate.toGregorian(jdEnd.getDatetime(), jdEnd.getTimezone());
//            }
//            String htmlLi = "<b>" + startStr + " - " + endStr + "</b><br>";
//            UTMPoint[] utmP = getUTMpoint();
//            if (utmP != null)
//            {
//                for (int i = 0; i < utmP.length; i++)
//                {
//                    htmlLi += "zone: " + utmP[i].getZone() + "<br>";
//                    htmlLi += "x: " + utmP[i].getX() + "<br>";
//                    htmlLi += "y: " + utmP[i].getY() + "<br>";
//                }
//            }
//
//            if (record.getLocation() != null)
//            {
//                TimeMapGeograhicCoordinates[] pointsLongLati = record.getLocation().getPointsLongLaties();
//                if (pointsLongLati != null)
//                {
//                    for (int i = 0; i < pointsLongLati.length; i++)
//                    {
//                        htmlLi += "Latitude: " + pointsLongLati[i].getLatitude() + "<br>";
//                        htmlLi += "Longitude: " + pointsLongLati[i].getLongitude() + "<br>";
//                    }
//                }
//            }
//
//            Event[] events = getEvents();
//            if (events != null)
//            {
//                htmlLi += "<ul>";
//                for (int i = 0; i < events.length; i++)
//                {
//                    if (events[i].getName().equals("url"))
//                    {
//                        htmlLi += "<li>" + events[i].getName() + "）" + "<a href=\"" + events[i].getData() + "\">" + events[i].getData() + "</a></li>";
//                    }
//                    else
//                    {
//                        htmlLi += "<li>" + events[i].getName() + "）" + events[i].getData().replaceAll("\n", "<br>") + "</li>";
//                    }
//                }
//                htmlLi += "</ul>";
//            }
//            return htmlLi;
//        }
        
        
        public void printLegend(Graphics g, int yy) {
            log.debug(dataset.getProperties());
            log.debug(dataset.getProperties().isShowGroupName());
            if (!getGroup().equals("") && dataset.getProperties().isShowGroupName()) {
                int y = yy + 18;//getPositionY(pos);

                BigDecimal sd = jdStart.getDatetime();
                //BigDecimal ed = jdEnd.getDatetime();
                BigDecimal csd = data.getCurrentMin();
                BigDecimal ced = data.getCurrentMax();
                //BigDecimal plotPoint = (sd.add(ed)).divide(new BigDecimal(2.0));
                //int x = julianToPixel(plotPoint, csd, ced);
                int x = julianToPixel(sd, csd, ced);
                
                String fontName = null;
//                if (dataset.getProperties().getFont("") != null) {
//                    fontName = dataset.getProperties().getFont("").getFontName();
//                }
                g.setFont(new Font(fontName, Font.PLAIN, 12));
                Dimension strSize = stringSize(getGroup(), g);
                //g.setColor(Color.black);
                g.setColor(dataset.getProperties().getPolygon(getGroup()).getFillColor());
                g.drawString(getGroup(), x, y - strSize.height);
                //g.drawLine(1, h - 1, w - 1, h - 1);
            }
        }

        /**
         * フォーカス時のプロットの高さ
         */
        // protected int fSize = 0;
        /**
         * 描画処理を実施します。
         * 
         * @param g
         * @param y Y軸の位置
         */
        public void paint(Graphics g, int yy)
        {
            PropertiesIF properties = dataset.getProperties();
            // PolygonIF polygon = properties.getPolygon();
            // if (polygon == null)
            // {
            // setLayerProperties();
            // }

            BigDecimal csd = data.getCurrentMin();
            BigDecimal ced = data.getCurrentMax();
            BigDecimal diff = (ced.subtract(csd)).divide(BigDecimal.valueOf(10), 10, BigDecimal.ROUND_HALF_EVEN);
            int xStart = 0;
            if (jdStart != null)
            {
                BigDecimal sd = jdStart.getDatetime();
                if (sd.compareTo(csd.subtract(diff)) < 0)
                {
                    sd = csd.subtract(diff);
                }
                xStart = julianToPixel(sd, csd, ced);
            }
            int xEnd = getWidth();
            if (jdEnd != null)
            {
                BigDecimal ed = jdEnd.getDatetime();
                if (ed.compareTo(ced.add(diff)) > 0)
                {
                    ed = ced.add(diff);
                }
                xEnd = julianToPixel(ed, csd, ced);
            }

            // if ((xEnd - xStart) < 1)
            // {
            // xEnd++;
            // }

            int paintIt = 0;

            // イベント幅が1px以下の場合は表示しない。
            if (reduce && !((xEnd - xStart) > 1))
            {
                paintIt = -1;
            }
            
            // 201312追加：イベント幅が2px以下の場合は○で表示する（タイトルなし）。
            if (!((xEnd - xStart) > 2) && paintIt != -1 && getPluginName().equals("DefaultLayer"))
            {
                paintIt = 1;
            }

            isCircle = false;//初期化
            if (paintIt == 1)
            {
            	isCircle = true;
            	int y = yy;
                int rx = xStart;
                // int ry = y - fSize;
                int ry = y;
                int rw = xEnd - xStart;
                Color fillColor = null;
                Color drawColor = null;
                if (record.isHit() || isClicked())
                {
                    fillColor = highlightColorx;// set random color
                    drawColor = highlightColorx;// set random color
                }
                else
                {
                    log.debug(getGroup());
                    log.debug(properties);
                    log.debug(properties.getPolygon(getGroup()));
                    fillColor = properties.getPolygon(getGroup()).getFillColor();
                    drawColor = properties.getPolygon(getGroup()).getLineColor();
                }
                
                int plotSize = 6;
                int cal = (plotSize / 2);
                g.setColor(fillColor);
                g.fillOval(rx - cal, y, plotSize, plotSize);
                g.setColor(drawColor);
                g.drawOval(rx - cal, y, plotSize, plotSize);
                
                if (title != null)
                {
                    Font pfont = properties.getFont(getGroup());
                    Font prevFont = null;
                    if (pfont != null)
                    {
                        g.setFont(pfont);
                        prevFont = g.getFont();
                    }
                    Color pfclor = properties.getFontColor(getGroup());
                    if (pfclor != null)
                    {
                        g.setColor(pfclor);
                    }
                    else
                    {
                        g.setColor(Color.black);
                    }
                    g.drawString(title, rx, y + 20);
                    if (pfont != null)
                    {
                        g.setFont(prevFont);
                    }
                }
                
                addShape(new Rectangle(rx - cal, y, plotSize, plotSize));
            }
            else if (paintIt == 0)
            {
                int y = yy;
                int rx = xStart;
                // int ry = y - fSize;
                int ry = y;
                int rw = xEnd - xStart;

                // int rh = 5 + fSize;
                int rh = 5;

                Color fillColor = null;
                Color drawColor = null;
                if (record.isHit() || isClicked())
                {
                    fillColor = highlightColorx;// set random color
                    drawColor = highlightColorx;// set random color
                }
                else
                {
                    log.debug(getGroup());
                    log.debug(properties);
                    log.debug(properties.getPolygon(getGroup()));
                    fillColor = properties.getPolygon(getGroup()).getFillColor();
                    drawColor = properties.getPolygon(getGroup()).getLineColor();
                }

                g.setColor(fillColor);
                if (jdStart == null && jdEnd == null)
                {
                    ((Graphics2D) g).setPaint(new GradientPaint(0, 0, getBackground(), getWidth() / 2, 0, g.getColor(), true));
                }
                else if (jdStart == null)
                {
                    ((Graphics2D) g).setPaint(new GradientPaint(0, 0, getBackground(), getWidth() / 4, 0, g.getColor()));
                }
                else if (jdEnd == null)
                {
                    ((Graphics2D) g).setPaint(new GradientPaint(getWidth() / 4 * 3, 0, g.getColor(), getWidth(), 0, getBackground()));
                }

                if (jdStart == null || jdEnd == null)
                {
                    rh++;
                }

                g.fillRect(rx, ry, rw, rh);
                g.setColor(drawColor);
                if (jdStart != null && jdEnd != null)
                {
                    g.drawRect(rx, ry, rw, rh);
                }

                // 開区間・閉区間表示（帯の幅が10px以上の時に表示される）
                if ((xEnd - xStart) > 10 && T2Map.properties.getBoolean("jp.nihu.gttime.T2Map.SHOW_OPEN_CLOSE"))
                {
                    if (jdStart != null)
                    {
                        if (record.isStartEdgeOpen())
                        {
                            g.setColor(Color.WHITE);
                        }
                        else
                        {
                            g.setColor(Color.BLACK);
                        }
                        g.fillRect(rx, ry, 5, rh);
                        g.setColor(fillColor);
                        if (jdEnd == null)
                        {
                            g.drawRect(rx, ry, 5, rh - 1);
                        }
                        else
                        {
                            g.drawRect(rx, ry, 5, rh);
                        }
                    }

                    if (jdEnd != null)
                    {
                        if (record.isEndEdgeOpen())
                        {
                            g.setColor(Color.WHITE);
                        }
                        else
                        {
                            g.setColor(Color.BLACK);
                        }
                        g.fillRect(rx + rw - 5, ry, 5, rh);
                        g.setColor(fillColor);
                        if (jdStart == null)
                        {
                            g.drawRect(rx + rw - 5, ry, 5, rh - 1);
                        }
                        else
                        {
                            g.drawRect(rx + rw - 5, ry, 5, rh);
                        }
                    }
                }

                int strPos = xStart;
                if (xStart < 0)
                {
                    strPos = 0;
                }
                if (title != null)
                {
                    Font pfont = properties.getFont(getGroup());
                    Font prevFont = null;
                    if (pfont != null)
                    {
                        g.setFont(pfont);
                        prevFont = g.getFont();
                    }
                    Color pfclor = properties.getFontColor(getGroup());
                    if (pfclor != null)
                    {
                        g.setColor(pfclor);
                    }
                    else
                    {
                        g.setColor(Color.black);
                    }
                    g.drawString(title, strPos, y + 20);
                    if (pfont != null)
                    {
                        g.setFont(prevFont);
                    }
                }

                addShape(new Rectangle(rx, ry, rw, rh));
            }
        }
    }

    private ImageIcon icon = new ImageIcon(getClass().getClassLoader().getResource("DefaultLayer.gif"));

    @Override
    public ImageIcon getIcon()
    {
        return icon;
    }
}
