/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.layer;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.GradientPaint;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.math.BigDecimal;
import java.util.Iterator;
import java.util.List;

import javax.swing.ImageIcon;
import javax.xml.xpath.XPathExpressionException;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.data.property.PropertiesGts;
import jp.nihu.gttime.data.property.PropertiesIF;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.preferences.BaseLayerPluginPreferences;
import jp.nihu.gttime.time.TimeAxisIF;
import jp.nihu.gttime.tree.TreeCheckNode;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * レイヤについてのマスク表示形式による描画機能（プラグイン）を実装するクラスである。
 */
public class MaskLayerPlugin extends DefaultLayerPlugin
{
    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(MaskLayerPlugin.class, T2Map.LOCALE);

    public String getName()
    {
        return rmgr.getString("name");
    }

    public String getPluginName()
    {
        return rmgr.getString("plugin.name");
    }

    public String getStyleName()
    {
        return getPluginName() + ".type";
    }

    /**
     * シリアルバーションID
     */
    private static final long serialVersionUID = 1L;

    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(MaskLayerPlugin.class);

    /**
     * レイヤ内で利用されるデータ（TimeAxisIF、TreeCheckNode）を登録する。<br/>画面上のマウス操作によって起動される機能を初期化します。
     * 
     * @param data TimeAxisIFオブジェクト
     * @param node TreeCheckNodeオブジェクト
     */
    public void init(TimeAxisIF data, final TreeCheckNode node)
    {
        log.debug("Initialize MaskLayerPlugin.");
        super.init(data, node);
        // node.setHeight(60);
        setMinHeight(60);
        dataset.getProperties().setLayerHeight(60);
        this.setMaximumSize(new Dimension(20000000, dataset.getProperties().getLayerHeight()));
        this.setPreferredSize(new Dimension(0, dataset.getProperties().getLayerHeight()));
        
        //20111129 Hayashi これより前の処理でpropertiyの値を取得して
        //簡易表示の設定をしているのいもかかわらずここでOFFにしていたので
        //コメントアウトした
//        dataset.getProperties().setDataReduction(false);
//        reduce = false;
    }

    public BaseLayerPluginPreferences getLayerPluginPreferencesDialog()
    {
        BaseLayerPluginPreferences prep = new MaskLayerPluginPreferences();
//        List<String> groups = dataset.getGroups();
//        for (int i = 0; i < groups.size(); i++) {
//            prep.initValueForLayer(dataset.getProperties(), groups.get(i));
//        }
        prep.initValueForLayer(dataset.getProperties());
        return prep;
    }

    protected void setLayerProperties()
    {
        log.debug("setLayerProperties()");
        PropertiesIF properties = dataset.getProperties();
        if (properties == null)
        {
            properties = new PropertiesGts();
            dataset.setProperties(properties);
        }
        try
        {
            List<String> groups = dataset.getGroups();
            for (int i = 0; i < groups.size(); i++) {
                String group = groups.get(i);
                if (properties.getPolygon(group) == null) {
                    setPolygonProperties(group);
                }
            }
            if (properties.getBackgroundColor() == null)
            {
                setBackgroundColorProperties();
            }
            if (properties.isDataReduction() == null)
            {
                log.debug("isDataReduction()");
                setDataReductionProperties();
            }
            if (properties.getDataReductionNum() == null)
            {
                setDataReductionNumProperties();
            }
            
            log.debug(dataset.getProperties().isDataReduction());
        }
        catch (XPathExpressionException e)
        {
            log.fatal(e.getMessage(), e);
        }
    }

    /**
     * レイヤの高さ変更が可能かどうか
     */
    protected boolean isResizable()
    {
        return false;
    }

    // private void setColors()
    // {
    // Color color = T2Map.properties.getColor(PLOT_COLOR_IN);
    // if (color != null)
    // {
    // plotColorIn = color;
    // }
    // color = T2Map.properties.getColor(PLOT_COLOR_OUT);
    // if (color != null)
    // {
    // plotColorOut = color;
    // }
    // }

    /**
     * マスクレイヤを描画します。
     * 
     * @param g
     */
    public void paintLayer(Graphics g)
    {
        // setColors();
        int y = (getHeight() / 2) + 10;
        g.setColor(Color.LIGHT_GRAY);
        Iterator<RecordPlot> it = getProcRecordPlot();
        // Iterator<RecordPlot> it = treeSet.iterator();
        while (it != null && it.hasNext())
        {
            MaskPlot plot = (MaskPlot) it.next();
            plot.paint(g, y);
        }
    }

    /**
     * プロットを新規作成します。
     * 
     * @param record Recordオブジェクト
     * @param data TimeAxisIFオブジェクト
     * @return RecordPlot
     */
    protected RecordPlot createRecordPlot(Record record, TimeAxisIF data)
    {
        return new MaskPlot(record, data);
    }

    /**
     * 年表の描画やイベントオブジェクトを押下した際の動作などを実装している。
     */
    public class MaskPlot extends DefaultRecordPlot
    {
        /**
         * シリアルバーションID
         */
        private static final long serialVersionUID = 1L;

        /**
         * 年表描画に関するコンストラクタ
         * 
         * @param record Recordオブジェクト
         * @param data TimeAxisIFオブジェクト
         */
        public MaskPlot(Record record, TimeAxisIF data)
        {
            super(record, data);
        }
        
        /**
         * 描画処理を実施します。
         * 
         * @param g
         * @param y Y軸の位置
         */
        public void paint(Graphics g, int y)
        {
            // BigDecimal sd = jdStart.getDatetime();
            // BigDecimal ed = jdEnd.getDatetime();
            BigDecimal csd = data.getCurrentMin();
            BigDecimal ced = data.getCurrentMax();

            // int xStart = julianToPixel(sd, csd, ced);
            // int xEnd = julianToPixel(ed, csd, ced);
            int xStart = 0;
            if (jdStart != null)
            {
                BigDecimal sd = jdStart.getDatetime();
                xStart = julianToPixel(sd, csd, ced);
            }
            int xEnd = getWidth();
            if (jdEnd != null)
            {
                BigDecimal ed = jdEnd.getDatetime();
                xEnd = julianToPixel(ed, csd, ced);
            }

            // if (sd.compareTo(ced) == -1 && ed.compareTo(csd) == 1)
            // {
            // g.setColor(dataset.getProperties().getPolygon().getFillColor());
            // g.fillRect(xStart, y - 5, xEnd - xStart, 15);
            // g.setColor(dataset.getProperties().getPolygon().getLineColor());
            // g.drawRect(xStart, y - 5, xEnd - xStart, 15);
            // }

            boolean paintIt = true;

            if (reduce && !((xEnd - xStart) > 1))
            {// イベント幅が1px以下の場合は表示しない。
                paintIt = false;
            }

            if (paintIt)
            {
                g.setColor(dataset.getProperties().getPolygon(getGroup()).getFillColor());
                if (jdStart == null && jdEnd == null)
                {
                    ((Graphics2D) g).setPaint(new GradientPaint(0, 0, getBackground(), getWidth() / 2, 0, g.getColor(), true));
                }
                else if (jdStart == null)
                {
                    ((Graphics2D) g).setPaint(new GradientPaint(0, 0, getBackground(), getWidth() / 4, 0, g.getColor()));
                }
                else if (jdEnd == null)
                {
                    ((Graphics2D) g).setPaint(new GradientPaint(getWidth() / 4 * 3, 0, g.getColor(), getWidth(), 0, getBackground()));
                }
                if (xEnd - xStart < 1) {
                    g.fillRect(xStart, y - 5, 1, 15);
                } else {
                    g.fillRect(xStart, y - 5, xEnd - xStart, 15);
                }

                // 開区間・閉区間表示（帯の幅が10px以上の時に表示される）
                if ((xEnd - xStart) > 10 && T2Map.properties.getBoolean("jp.nihu.gttime.T2Map.SHOW_OPEN_CLOSE"))
                {
                    if (jdStart != null)
                    {
                        if (record.isStartEdgeOpen())
                        {
                            g.setColor(Color.WHITE);
                        }
                        else
                        {
                            g.setColor(Color.BLACK);
                        }
                        g.fillRect(xStart, y - 5, 5, 15);
                        g.setColor(dataset.getProperties().getPolygon(getGroup()).getFillColor());
                        g.drawRect(xStart, y - 5, 5, 14);
                    }

                    if (jdEnd != null)
                    {
                        if (record.isEndEdgeOpen())
                        {
                            g.setColor(Color.WHITE);
                        }
                        else
                        {
                            g.setColor(Color.BLACK);
                        }
                        g.fillRect(xEnd - 5, y - 5, 5, 15);
                        g.setColor(dataset.getProperties().getPolygon(getGroup()).getFillColor());
                        g.drawRect(xEnd - 5, y - 5, 5, 14);
                    }
                }
            }
        }
    }
    
    private ImageIcon icon = new ImageIcon(getClass().getClassLoader().getResource("MaskLayer.gif"));

    @Override
    public ImageIcon getIcon()
    {
        return icon;
    }
}
