/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.layer;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Graphics;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JSeparator;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.T2MapDialog;
import jp.nihu.gttime.TimeMapGeograhicCoordinates;
import jp.nihu.gttime.data.DatasetIF;
import jp.nihu.gttime.data.Event;
import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.data.UTMPoint;
import jp.nihu.gttime.layer.LayerPopupMenu;
import jp.nihu.gttime.layer.ScaleBar;
import jp.nihu.gttime.layer.TimeDurationSelector;
import jp.nihu.gttime.plugin.LayerPlugin;
import jp.nihu.gttime.plugin.NumericMeasurePlugin;
import jp.nihu.gttime.plugin.PluginManager;
import jp.nihu.gttime.plugin.PluginManagerException;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.plugin.TimeMeasureDialog;
import jp.nihu.gttime.time.TimeAxisIF;
import jp.nihu.gttime.time.utility.JulianGregorianDate;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * レイヤについてのプロット形式による描画機能（プラグイン）を実装するクラスである。
 */
public abstract class NumericLayerPlugin extends LayerPlugin
{
    /**
     * シリアルバーションID
     */
    private static final long serialVersionUID = 1L;

    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(NumericLayerPlugin.class);

    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(NumericLayerPlugin.class, T2Map.LOCALE);

    @Override
    /**
     * レイヤ情報の初期化<br/>
     * <ul>
     * <li>イベントオブジェクトの1つずつに各リスナーを登録します。</li>
     * <li>数値情報用のメニューオブジェクトを作成します。</li>
     * </ul>
     * 
     * @param data TimeAxisIFオブジェクト
     * @param dataset Datasetオブジェクト
     */
    public void initLayer()
    {
        parseRecord(dataset, data);
        if (getMenu() != null)
        {
            popupMenu = createScaleMenu(getMenu());
        }
    }

    public void showPopupMenu(MouseEvent e)
    {
        yMeasure.setVisible(getTimeDurationSelector().isOnSelected(pixelToJulian(e.getX(), data.getCurrentMin(), data.getCurrentMax())));
        super.showPopupMenu(e);
    }

    /**
     * イベントオブジェクトの1つずつに各リスナーを登録します。
     * 
     * @param dataset Datasetオブジェクト
     * @param data TimeAxisIFオブジェクト
     */
    private void parseRecord(DatasetIF dataset, TimeAxisIF data)
    {
        Record[] records = dataset.getRecords();
        int numberNum = 0;
        Map<String, NumericRecordPlot> previousePlotMap = new HashMap<String, NumericRecordPlot>();
//        NumericRecordPlot previousePlot = null;
        for (int i = 0; i < records.length; i++)
        {
            Event[] events = records[i].getEvents();
            String group = records[i].getGroup();
            for (int j = 0; j < events.length; j++)
            {
                if (events[j] != null && events[j].getName() != null && events[j].getName().equals("number"))
                {
                    if (events[j].getData().matches("^-?[0-9]+([.][0-9]+)?$"))
                    {
                        float cnum = Float.parseFloat(events[j].getData());
                        setMinHandMaxH(numberNum, cnum);
                        numberNum++;
                        NumericRecordPlot rp = (NumericRecordPlot) createRecordPlot(records[i], data);
                        NumericRecordPlot previousePlot = previousePlotMap.get(group);
                        if (previousePlot != null)
                        {
                            rp.setPreviouseNumberVal(previousePlot.getNumberVal());
                            previousePlot.setNextNumberVal(rp.getNumberVal());
                            rp.previouseEnd = previousePlot.jdEnd.getDatetime();
                        }
                        rp.setNumberValIsNull(false);
                        previousePlot = rp;
                        addRecordPlot(rp);
                    }
                    else
                    {
                        // float cnum = Float.parseFloat(events[j].getData());
                        // setMinHandMaxH(numberNum, cnum);
                        // numberNum++;
                        NumericRecordPlot rp = (NumericRecordPlot) createRecordPlot(records[i], data);
                        NumericRecordPlot previousePlot = previousePlotMap.get(group);
                        if (previousePlot != null)
                        {
                            // rp.setPreviouseNumberVal(previousePlot.getNumberVal());
                            // previousePlot.setNextNumberVal(rp.getNumberVal());

                            // previousePlotにNextNumberValがnullであることを知らせる
                            previousePlot.setNextNumberValIsNull(true);
                        }
                        // 自分がnullであることを指定する
                        rp.setNumberValIsNull(true);
                        if (previousePlotMap.containsKey(group)) {
                            previousePlot = rp;
                        } else {
                            previousePlotMap.put(group, previousePlot);
                        }
                    }
                }
            }
        }

        Iterator<RecordPlot> it = this.treeSetStart.iterator();
        NumericRecordPlot pp = null;
        while (it.hasNext())
        {
            NumericRecordPlot p = (NumericRecordPlot) it.next();
            if (pp != null)
            {
                p.previouseEnd = pp.jdEnd.getDatetime();
            }
            pp = p;
        }

        node.setScaleMin(minH);
        node.setScaleMax(maxH);
    }

    /**
     * プロットの大きさを取得します。
     * 
     * @return プロットの大きさ
     */
    protected abstract int getPlotSize(String group);

    /**
     * X軸とY軸の線描画
     * 
     * @param g
     */
    private void drawScale(Graphics g)
    {
        g.setColor(Color.lightGray);
        int w = getWidth() - LINE_INSETS;
        int h = getHeight() - getHeightInsets();
        if (!node.isBorderDrow())
        {
            ScaleBar.drawScale(15, 0, minH, maxH, getHeightInsets(), h, LINE_INSETS, getHeight(), LINE_INSETS, g);
        }
        else
        {
            if (node.getSelectScale().equals("InputScale"))
            {
                if (node.getButtonScaleSts() == 1)
                {
                    minH = node.getInputMin();
                    maxH = node.getInputMax();
                }
            }
            ScaleBar.drawScale(15, 0, minH, maxH, getHeightInsets(), h, LINE_INSETS, getHeight(), w, g);
        }
        g.drawLine(LINE_INSETS, getHeightInsets(), w, getHeightInsets());
        g.drawLine(LINE_INSETS, h, w, h);
        g.drawLine(w, getHeightInsets(), w, h);
    }

    /**
     * 時間範囲内で再描画（「目盛り」-> 「時間範囲内」を選択）
     */
    private void rangeScaleFileter()
    {
        int numberNum = 0;
        Iterator<RecordPlot> it = getProcRecordPlot();
        while (it.hasNext())
        {
            NumericRecordPlot plot = (NumericRecordPlot) it.next();
            float cnum = plot.numberVal;
            setMinHandMaxH(numberNum, cnum);
            numberNum++;
        }
    }

    /**
     * Y軸の最小値と最大値を求める
     * 
     * @param numberPos カウント個数
     * @param value Y軸の値
     */
    protected void setMinHandMaxH(int numberPos, float value)
    {
        if (numberPos == 0)
        {
            minH = value;
            maxH = value;
        }
        else
        {
            if (value > maxH)
            {
                maxH = value;
            }
            else if (value < minH)
            {
                minH = value;
            }
            else if (value == maxH && value == minH)
            {
                minH = value;
                maxH = minH + 1.1f;
            }
        }
    }

    /**
     * Y軸の最小値
     */
    protected float minH = 0;

    /**
     * Y軸の最大値
     */
    protected float maxH = 0;

    /**
     * 選択されたY軸の最大値
     */
    private float selectMax = 0;

    /**
     * 選択されたY軸の最小値
     */
    private float selectMin = 0;

    /**
     * 測定時Y軸の最小値と最大値を求める
     * 
     * @param numberPos カウント個数
     * @param value Y軸の値
     */
    public void setSelectMaxMinH(int numberPos, float value)
    {
        if (numberPos == 0)
        {
            selectMin = value;
            selectMax = value;
        }
        else
        {
            if (value > selectMax)
            {
                selectMax = value;
            }
            else if (value < selectMin)
            {
                selectMin = value;
            }
            else if (value == selectMax && value == selectMin)
            {
                selectMin = value;
                selectMax = selectMin;
            }
        }
    }

    private JMenu yMeasure = null;

    /**
     * 数値情報用のメニューオブジェクトを作成します。
     * 
     * @param menu 数値情報用のJPopupMenuオブジェクト
     * @return 数値情報用のメニューオブジェクト
     */
    private LayerPopupMenu createScaleMenu(LayerPopupMenu rmenu)
    {
        // JSeparator separator = new JSeparator();
        // menu.add(separator);

        //rmenu.addSeparator();
        
        JMenu menu = new JMenu(rmgr.getString("graph.settings.name"));
        rmenu.insert(menu, rmenu.getComponentCount()-2);
        //rmenu.insert(new JSeparator(), rmenu.getComponentCount()-3);
//        JSeparator sep = new JSeparator();
//        
//        for (int i = 0;i < rmenu.getComponentCount(); i++) {
//        	Component c = rmenu.getComponent(i);
//        	if (c instanceof JSeparator) {
//        		sep.setUI(((JSeparator)c).getUI());
//        		break;
//        	}
//        }
//        rmenu.insert(sep, rmenu.getComponentCount()-3);

        PluginManager pluginManager = T2Map.getPluginManager();
        ArrayList<Hashtable<String, String>> plugins = null;

        yMeasure = new JMenu(rmgr.getString("numeric.measure"));
        plugins = pluginManager.getPlugins("NumericMeasurePlugin");
        for (int i = 0; i < plugins.size(); i++)
        {
            Hashtable<String, String> plugin = plugins.get(i);
            String name = (String) plugin.get("name");
            final String clazz = (String) plugin.get("class");
            JMenuItem measureHeightItem = new JMenuItem(name);
            measureHeightItem.addActionListener(new ActionListener()
            {
                public void actionPerformed(ActionEvent e)
                {
                    TimeDurationSelector timeDurationSelector = getTimeDurationSelector();

                	if (!popupMenu.fromTree)
                    {
                		BigDecimal[] se = timeDurationSelector.setFocus(popupPosX);
                        timeDurationSelector.repaint();
                        // 時間幅測定機能の表示
                        if (se != null)
                        {
                            Record[] records = dataset.getRecords();
                            NumericMeasurePlugin plugin = null;
                            try
                            {
                                plugin = (NumericMeasurePlugin) T2Map.pluginManager.newInstance(clazz);
                            }
                            catch (PluginManagerException e1)
                            {
                                log.fatal(e1.getMessage(), e1);
                            }
                            plugin.showDialog(records, null, se[0], se[1], NumericLayerPlugin.this.getParent(), timeDurationSelector);
                        }
                    } else {
                        timeDurationSelector.repaint();
                        // 時間幅測定機能の表示
                        Record[] records = dataset.getRecords();
                        NumericMeasurePlugin plugin = null;
                        try
                        {
                            plugin = (NumericMeasurePlugin) T2Map.pluginManager.newInstance(clazz);
                        }
                        catch (PluginManagerException e1)
                        {
                            log.fatal(e1.getMessage(), e1);
                        }
                        plugin.showDialog(records, null, NumericLayerPlugin.this.getParent(), timeDurationSelector);
                    }
                    
                    
                    
                }
            });

            yMeasure.add(measureHeightItem);
        }
        menu.add(yMeasure);

        final JMenuItem borderScale = new JMenuItem();
        menu.add(borderScale);
        borderScale.setText(rmgr.getString("supplementary.line"));
        if (node.isBorderDrow())
        {
            // borderScale.setText(rmgr.getString("supplementary.line"));
            borderScale.setIcon(popupMenu.getCheckIcon());
        }
        else
        {
            // borderScale.setText(rmgr.getString("supplementary.line"));
            borderScale.setIcon(null);
        }

        borderScale.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {

                if (node.isBorderDrow())
                {
                    node.setBorderDrow(false);
                    // borderScale.setText(rmgr.getString("supplementary.line"));
                    borderScale.setIcon(null);
                }
                else
                {
                    node.setBorderDrow(true);
                    // borderScale.setText(rmgr.getString("supplementary.line"));
                    borderScale.setIcon(popupMenu.getCheckIcon());
                }
                setForcePaint();
                paintComponent(getGraphics());
            }
        });
        menu.add(borderScale);

        JMenu scaleSubmenu = new JMenu(rmgr.getString("scale.name"));

        final JMenuItem allData = new JMenuItem(rmgr.getString("all.data"));
        allData.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                node.setSelectScale("DefaultScale");
                minH = node.getScaleMin();
                maxH = node.getScaleMax();
                setForcePaint();
                paintComponent(getGraphics());
            }
        });
        scaleSubmenu.add(allData);

        final JMenuItem rangeTime = new JMenuItem(rmgr.getString("time.range"));
        rangeTime.addActionListener(new ActionListener()
        {
            public void actionPerformed(ActionEvent e)
            {
                node.setSelectScale("RangeScale");
                setForcePaint();
                paintComponent(getGraphics());
            }
        });
        scaleSubmenu.add(rangeTime);

        final JMenuItem inputValue = new JMenuItem(rmgr.getString("arbitrary.input"));
        inputValue.addActionListener(new ActionListener()
        {
            private JTextField minTextH = null;

            private JTextField maxTextH = null;

            public void actionPerformed(ActionEvent e)
            {
                node.setSelectScale("InputScale");

                final T2MapDialog dialog = new T2MapDialog((Frame) SwingUtilities.getRoot(getParent()), false);
                dialog.setBackground(Color.white);
                dialog.setSize(new java.awt.Dimension(200, 100));
                int curPanelX = (int) getLocationOnScreen().getX();
                int curPanelY = (int) getLocationOnScreen().getY();
                dialog.setLocation((curPanelX + getWidth()) - 200, curPanelY);

                maxTextH = new JTextField();
                minTextH = new JTextField();

                JLabel maxLabel = new JLabel(rmgr.getString("max.value"));
                JLabel minLabel = new JLabel(rmgr.getString("min.value"));

                JButton btn_ok = new JButton(rmgr.getString("apply.button"));
                btn_ok.addActionListener(new ActionListener()
                {
                    public void actionPerformed(ActionEvent ae)
                    {
                        float fmin = formatChkNumber("min", minTextH.getText());
                        float fmax = formatChkNumber("max", maxTextH.getText());
                        if (fmin > fmax)
                        {
                            node.setInputMin(node.getScaleMin());
                            node.setInputMax(node.getScaleMax());
                        }
                        else if (fmin == fmax)
                        {
                            node.setInputMin(fmin);
                            node.setInputMax(fmin + 0.1f);
                        }
                        else
                        {
                            node.setInputMin(fmin);
                            node.setInputMax(fmax);
                        }
                        minH = node.getInputMin();
                        maxH = node.getInputMax();
                        node.setButtonScaleSts(1);
                        setForcePaint();
                        repaint();
                    }
                });

                JButton btn_no = new JButton(rmgr.getString("cancel.button"));
                btn_no.addActionListener(new ActionListener()
                {
                    public void actionPerformed(ActionEvent ae)
                    {
                        node.setButtonScaleSts(2);
                        repaint();
                        dialog.dispose();
                    }
                });

                dialog.setLayout(new GridLayout(0, 2));

                dialog.add(maxLabel);
                dialog.add(maxTextH);

                dialog.add(minLabel);
                dialog.add(minTextH);

                dialog.add(btn_ok);
                dialog.add(btn_no);

                dialog.setTitle(rmgr.getString("scale.value.input"));
                dialog.setVisible(true);

                // setForcePaint();
                // paintComponent(getGraphics());
            }
        });
        scaleSubmenu.add(inputValue);
        menu.add(scaleSubmenu);

        // 表示レイヤ変更メニュー
        JMenu graphChange = new JMenu(rmgr.getString("graph.change.name"));
        plugins = pluginManager.getPlugins("LayerPlugin");
        for (int i = 0; i < plugins.size(); i++)
        {
            Hashtable<String, String> plugin = plugins.get(i);

            try
            {
                Object obj = getClass().getClassLoader().loadClass((String) plugin.get("class")).newInstance();
                if (obj instanceof NumericLayerPlugin)
                {
                    NumericLayerPlugin lyr = (NumericLayerPlugin) obj;
                    String name = lyr.getName();
                    String pluginTypeName = lyr.getPluginName();
                    String clazz = (String) plugin.get("class");
                    JMenuItem changeLayerItem = new JMenuItem(name);
                    changeLayerItem.addActionListener(new GraphChangeListener(clazz, pluginTypeName));
                    graphChange.add(changeLayerItem);
                }
                else
                {
                    obj = null;
                }
            }
            catch (InstantiationException e)
            {
                log.fatal(e.getMessage(), e);
            }
            catch (IllegalAccessException e)
            {
                log.fatal(e.getMessage(), e);
            }
            catch (ClassNotFoundException e)
            {
                log.fatal(e.getMessage(), e);
            }
        }
        menu.add(graphChange);

        return rmenu;
    }

    @Override
    public void drawBackground(Graphics g)
    {
        super.drawBackground(g);
        drawScale(g);
    }

    @Override
    public void paintLayer(Graphics g)
    {
        if (groups != null)
        {
            groups.clear();
        }
        if (node.getSelectScale().equals("RangeScale"))
        {
            rangeScaleFileter();
        }
        try
        {
            int ps = treeSetStart.size();
            int i = 0;
            Set<String> group = new HashSet<String>();
            Iterator<RecordPlot> it = getProcRecordPlot();
            while (it.hasNext())
            {
                NumericRecordPlot plot = (NumericRecordPlot) it.next();
                if (isPaintResolution(plot.record))
                {
                    if (plot.doPaint() || i == 0 || i == ps - 1)
                    {
                        if (!group.contains(plot.getGroup())) {
                            plot.printLegend(g, minH, maxH);
                            group.add(plot.getGroup());
                        }
                        plot.paint(g, minH, maxH);
                    }
                }
                i++;
            }
        }
        catch (Exception e)
        {
            e.printStackTrace();
        }
    }

    /**
     * Y軸の目盛り値を任意入力した際の値が数値であるかのチェックとfloat型に変換し、取得します。
     * 
     * @param valueFlag 最小==min 最大==max
     * @param value 目盛り値の入力した値
     * @return 入力値のfloat型
     */
    public Float formatChkNumber(String valueFlag, String value)
    {
        if (valueFlag.equals("min"))
        {
            if (value == null || value.length() == 0)
                return node.getScaleMin();
        }

        if (valueFlag.equals("max"))
        {
            if (value == null || value.length() == 0)
                return node.getScaleMax();
        }

        try
        {
            return new Float(value);
        }
        catch (NumberFormatException e)
        {
            log.error("Invalid datatye: " + value);

            if (valueFlag.equals("min"))
                return node.getScaleMin();

            if (valueFlag.equals("max"))
                return node.getScaleMax();
        }
        return 0f;
    }

    /**
     * プロットの描画やイベントオブジェクトを押下した際の動作などを実装している
     */
    public abstract class NumericRecordPlot extends RecordPlot
    {
        /**
         * シリアルバーションID
         */
        private static final long serialVersionUID = 1L;

        /**
         * 数値用
         */
        protected float numberVal = 0;

        private float previouseNumberVal = 0;

        public BigDecimal previouseEnd = null;

        private float nextNumberVal = 0;

        protected boolean isValueNull = false;

        protected boolean isNextValueNull = false;

        public void setNumberValIsNull(boolean isNull)
        {
            isValueNull = isNull;
        }

        public void setNextNumberValIsNull(boolean isNull)
        {
            isNextValueNull = isNull;
        }

        protected boolean doPaint()
        {
            if (isValueNull)
            {
                return false;
            }
            else if (reduce)
            // if (reduce)
            {
                if ((previouseNumberVal < numberVal && numberVal > nextNumberVal) || (previouseNumberVal > numberVal && numberVal < nextNumberVal))
                {
                    return true;
                }
                else
                {
                    return false;
                }
            }
            else
            {
                return true;
            }
        }

        public void setNextNumberVal(float num)
        {
            nextNumberVal = num;
        }

        public void setPreviouseNumberVal(float num)
        {
            previouseNumberVal = num;
        }

        /**
         * プロット描画に関するコンストラクタ
         * 
         * @param record Recordオブジェクト
         * @param data TimeAxisIFオブジェクト
         */
        public NumericRecordPlot(Record record, TimeAxisIF data)
        {
            super(record, data);
            setNumber();
        }

        /**
         * 数値情報の取得
         * 
         * @return 数値情報
         */
        public float getNumberVal()
        {
            return numberVal;
        }

        private void setNumber()
        {
            Event[] events = record.getEvents();
            float number = 0;
            int el = events.length;
            for (int j = 0; j < el; j++)
            {
                if (events[j].getName().equals("number"))
                {
                    if (!events[j].getData().equals(""))
                    {
                        number = Float.valueOf(events[j].getData()).floatValue();
                    }
                    break;
                }
            }
            this.numberVal = number;
        }

//        /**
//         * @see jp.nihu.gttime.plugin.LayerPlugin.RecordPlot#getHtml()
//         */
//        public String getHtml()
//        {
//            String startStr = "None";
//            if (jdStart != null)
//            {
//                startStr = JulianGregorianDate.toGregorian(jdStart.getDatetime(), jdStart.getTimezone());
//            }
//            String endStr = "None";
//            if (jdEnd != null)
//            {
//                endStr = JulianGregorianDate.toGregorian(jdEnd.getDatetime(), jdEnd.getTimezone());
//            }
//            String htmlLi = "<b>" + startStr + " - " + endStr + "</b><br>";
//            UTMPoint[] utmP = getUTMpoint();
//            if (utmP != null)
//            {
//                for (int i = 0; i < utmP.length; i++)
//                {
//                    htmlLi += "zone: " + utmP[i].getZone() + "<br>";
//                    htmlLi += "x: " + utmP[i].getX() + "<br>";
//                    htmlLi += "y: " + utmP[i].getY() + "<br>";
//                }
//            }
//
//            if (record.getLocation() != null)
//            {
//                TimeMapGeograhicCoordinates[] pointsLongLati = record.getLocation().getPointsLongLaties();
//                if (pointsLongLati != null)
//                {
//                    for (int i = 0; i < pointsLongLati.length; i++)
//                    {
//                        htmlLi += "Latitude: " + pointsLongLati[i].getLatitude() + "<br>";
//                        htmlLi += "Longitude: " + pointsLongLati[i].getLongitude() + "<br>";
//                    }
//                }
//            }
//
//            Event[] events = getEvents();
//            if (events != null)
//            {
//                htmlLi += "<ul>";
//                for (int i = 0; i < events.length; i++)
//                {
//                    if (events[i].getName().equals("url"))
//                    {
//                        htmlLi += "<li>" + events[i].getName() + "）" + "<a href=\"" + events[i].getData() + "\">" + events[i].getData() + "</a></li>";
//                    }
//                    else if (events[i].getName().equals("number") && LayerPlugin.class.isInstance(NumericLayerPlugin.class))
//                    {
//                        htmlLi += "<li><span style=\"font-size:24pt\">" + events[i].getName() + "）" + events[i].getData() + "</span></li>";
//                    }// <span style="font-size:24pt">
//                    else
//                    {
//                        htmlLi += "<li>" + events[i].getName() + "）" + events[i].getData().replaceAll("\n", "<br>") + "</li>";
//                    }
//                }
//                htmlLi += "</ul>";
//            }
//            return htmlLi;
//        }

        /**
         * 描画処理を実施します。
         * 
         * @param g
         * @param minH Y軸の最小値
         * @param maxH Y軸の最大値
         */
        public abstract void paint(Graphics g, float minH, float maxH);
        
        /**
         * グループ属性が付与されているときにグループ名を表示します。
         * @param g
         * @param minH Y軸の最小値
         * @param maxH Y軸の最大値
         */
        public abstract void printLegend(Graphics g, float minH, float maxH);
    }

    /**
     * イベントのY位置を算出する
     * 
     * @param value 求めたい値
     * @param minH Y軸の最小値
     * @param maxH Y軸の最大値
     * @param screenStart 開始位置
     * @param screenEnd 終了位置
     * @return イベントのY位置
     */
    protected int getPositionY(float value, float minH, float maxH, int screenStart, int screenEnd)
    {
        int screenRange = screenEnd - screenStart;
        float valueRange = maxH - minH;
        float pixelsPerUnit = screenRange / valueRange;
        // return (int) ((maxH - value) * pixelsPerUnit + .5) +
        // getHeightInsets();
        return (int) ((maxH - value) * pixelsPerUnit) + getHeightInsets();
    }

    /**
     * 指定されたグラフに再描画する。
     * 
     * @param clazz クラス名
     * @param name レイヤ名
     */
    class GraphChangeListener implements ActionListener
    {
        /**
         * レイヤクラス名
         */
        private String clazz = null;

        /**
         * レイヤ名
         */
//        private String name = null;
        
        /**
         * レイヤタイプ名
         */
        private String pluginTypeName = null;

        /**
         * クラス名、画面表示名を設定します。
         * 
         * @param clazz クラス名
         * @param name レイヤ名
         */
        public GraphChangeListener(String clazz, String pluginTypeName)
        {
            this.clazz = clazz;
//            this.name = name;
            this.pluginTypeName = pluginTypeName;
        }

        /**
         * 再描画します。
         * 
         * @param event ActionEventオブジェクト
         */
        public void actionPerformed(ActionEvent event)
        {
            dataset.reloadProperties(pluginTypeName);
            T2Map t2map = getT2Map();
            log.debug("t2map " + t2map);
            getT2Map().getTree().changeLayerPlugin(dataset, node, pluginTypeName, clazz);
        }
    }

    private ImageIcon icon = new ImageIcon(getClass().getClassLoader().getResource("NumericLayer.gif"));

    @Override
    public ImageIcon getIcon()
    {
        return icon;
    }
}
