/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.layer;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.math.BigDecimal;
import java.util.Iterator;
import java.util.List;

import javax.swing.ImageIcon;
import javax.swing.JMenuItem;
import javax.xml.xpath.XPathExpressionException;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.DatasetIF;
import jp.nihu.gttime.data.Record;
import jp.nihu.gttime.data.property.PropertiesGts;
import jp.nihu.gttime.data.property.PropertiesIF;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.preferences.BaseLayerPluginPreferences;
import jp.nihu.gttime.time.TimeAxisIF;
import jp.nihu.gttime.tree.TreeCheckNode;
import jp.nihu.gttime.utility.Utilities;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * レイヤについての時間軸形式による描画機能（プラグイン）を実装するクラスである。
 */
public class TimeScaleLayerPlugin extends DefaultLayerPlugin
{
    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(TimeScaleLayerPlugin.class, T2Map.LOCALE);

    public String getName()
    {
        return rmgr.getString("name");
    }

    public String getStyleName()
    {
        return "TimeScaleLayer.type";
    }

    public String getPluginName()
    {
        return rmgr.getString("plugin.name");
    }

    /**
     * シリアルバーションID
     */
    private static final long serialVersionUID = 1L;

    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(TimeScaleLayerPlugin.class);

    /**
     * レイヤ内で利用されるデータ（TimeAxisIF、TreeCheckNode）を登録する。<br/>画面上のマウス操作によって起動される機能を初期化します。
     * 
     * @param data TimeAxisIFオブジェクト
     * @param node TreeCheckNodeオブジェクト
     */
    public void init(TimeAxisIF data, final TreeCheckNode node)
    {
        log.debug("Initialize TimeChartLayerPlugin.");
        super.init(data, node);
        // node.setHeight(70);
        dataset.getProperties().setLayerHeight(70);
        this.setMaximumSize(new Dimension(20000000, dataset.getProperties().getLayerHeight()));
        this.setPreferredSize(new Dimension(0, dataset.getProperties().getLayerHeight()));
    }

    public BaseLayerPluginPreferences getLayerPluginPreferencesDialog()
    {
        BaseLayerPluginPreferences prep = new TimeScaleLayerPluginPreferences();
//        List<String> groups = dataset.getGroups();
//        for (int i = 0; i < groups.size(); i++) {
//            prep.initValueForLayer(dataset.getProperties(), groups.get(i));
//        }
        prep.initValueForLayer(dataset.getProperties());
        return prep;
    }

    @Override
    protected void setLayerProperties()
    {
        PropertiesIF properties = dataset.getProperties();
        if (properties == null)
        {
            properties = new PropertiesGts();
            dataset.setProperties(properties);
        }
        try
        {
            List<String> groups = dataset.getGroups();
            for (int i = 0; i < groups.size(); i++) {
                String group = groups.get(i);
                if (properties.getFont(group) == null) {
                    setFontProperties(group);
                }
                if (properties.getFontColor(group) == null) {
                    setFontColorProperties(group);
                }
            }
            if (properties.getScaleMain() == null)
            {
                setScaleMainProperties();
            }
            if (properties.getScaleSub() == null)
            {
                setScaleSubProperties();
            }
            if (properties.getBackgroundColor() == null)
            {
                setBackgroundColorProperties();
            }
            if (properties.isDataReduction() == null)
            {
                setDataReductionProperties();
            }
            if (properties.getDataReductionNum() == null)
            {
                setDataReductionNumProperties();
            }
        }
        catch (XPathExpressionException e)
        {
            log.fatal(e.getMessage(), e);
        }
    }

    /**
     * レイヤの高さ変更が可能かどうか
     */
    protected boolean isResizable()
    {
        return false;
    }

    protected boolean canDrawTimeAxis()
    {
        return false;
    }
    
    @Override
    public void drawBackground(Graphics g)
    {
        super.drawBackground(g);
        int y = (getHeight() / 2) + 10;
        g.setColor(Color.LIGHT_GRAY);
        g.drawLine(0, y + 10, 10000, y + 10);
    }

    /**
     * 年表の表示位置を算出している。<br/> 目盛を1本の線とする。<br/> 目盛を間引きして表示する。
     * 
     * @param g
     */
    @Override
    public void paintLayer(Graphics g)
    {
        // setColors();
        int y = (getHeight() / 2) + 10;
        int k = 0;// 前のプロットの文字列の最後の位置
        int prevX = 0;// 前のプロットの位置

        Iterator<RecordPlot> it = getProcRecordPlot();
        BigDecimal csd = data.getCurrentMin();
        BigDecimal ced = data.getCurrentMax();
        while (it.hasNext())
        {
            TimeLinePlot plot = (TimeLinePlot) it.next();
            if (isPaintResolution(plot.record))
            {
                BigDecimal sd = plot.jdStart.getDatetime();
                int posX = julianToPixel(sd, csd, ced);
                if (plot.titleWidth < 0)
                {
                    plot.titleWidth = (int) (Utilities.stringSize(plot.title, g).getWidth() / 2);
                }
                if (k <= (posX - plot.titleWidth))
                {
                    plot.paint(g, y, posX);
                    k = (int) (posX + plot.titleWidth) + 10;
                    if (reduce)
                    {
                        prevX = posX + 5;
                    }
                    else
                    {
                        prevX = posX;
                    }
                }
                else if (prevX < posX)
                {
                    plot.paintLine(g, y, posX);
                    if (reduce)
                    {
                        prevX = posX + 5;
                    }
                }
            }
        }
    }

    protected boolean doReduce(int ps)
    {
        if (dataset.getProperties().getDataReductionNum() < ps && dataset.getProperties().isDataReduction())
        {
            return true;
        }
        else
        {
            return false;
        }
    }

    /**
     * レイヤ上に描画します。
     * 
     * @param g
     */
    public void paintComponent(Graphics g)
    {
        node.setDrawTimeAxis(false);
        super.paintComponent(g);
    }

    /**
     * 「暦表示」、「暦非表示」について非表示にします。
     */
    protected JMenuItem createDrawTimeAxisMenu()
    {
        return null;
    }

    /**
     * イベントオブジェクトの1つずつに各リスナーを登録します。
     * 
     * @param dataset Datasetオブジェクト
     * @param data TimeAxisIFオブジェクト
     */
    protected void parseRecord(DatasetIF dataset, TimeAxisIF data)
    {
        // try
        // {
        // Record[] records = dataset.getRecords(dataset.getStartDatetime(),
        // dataset.getEndDatetime());
        Record[] records = dataset.getRecords();
        int rl = records.length;
        for (int i = 0; i < rl; i++)
        {
            RecordPlot rp = new TimeLinePlot(records[i], data);
            addRecordPlot(rp);
        }
        // }
        // catch (DatasetException e)
        // {
        // log.fatal(e.getMessage(), e);
        // }
    }

    /**
     * 年表の描画やイベントオブジェクトを押下した際の動作などを実装している
     */
    public class TimeLinePlot extends DefaultRecordPlot
    {
        // /**
        // * タイトルの表示、非表示
        // */
        // private boolean stringWrite = true;

        public int titleWidth = -1;

        /**
         * 年表描画に関するコンストラクタ
         * 
         * @param record Recordオブジェクト
         * @param data TimeAxisIFオブジェクト
         */
        public TimeLinePlot(Record record, TimeAxisIF data)
        {
            super(record, data);
        }

        /**
         * 描画処理を実施します。
         * 
         * @param g
         * @param y Y軸の位置
         */
        public void paint(Graphics g, int y, int x)
        {
            g.setColor(dataset.getProperties().getScaleMain().getFillColor());
            // g.fillRect(x, y - 5, 1, 15);
            g.drawLine(x, y - 5, x, y + 10);
            g.setColor(dataset.getProperties().getFontColor(getGroup()));
            Font pf = g.getFont();
            g.setFont(dataset.getProperties().getFont(getGroup()));
            g.drawString(title, x - titleWidth, y - 10);
            // stringWrite = true;
            g.setColor(Color.black);
            if (pf != null) {
            	g.setFont(pf);
            }
        }

        /**
         * 時間線のみ描画処理を実施します。
         * 
         * @param g
         * @param y Y軸の位置
         */
        public void paintLine(Graphics g, int y, int x)
        {
            g.setColor(dataset.getProperties().getScaleSub().getFillColor());
            // g.fillRect(x, y, 1, 10);
            g.drawLine(x, y, x, y + 10);
            // stringWrite = false;
        }
    }

    private ImageIcon icon = new ImageIcon(getClass().getClassLoader().getResource("TimeChartLayer.gif"));

    @Override
    public ImageIcon getIcon()
    {
        return icon;
    }
}
