/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.printer;

import java.awt.Dimension;
import java.awt.Toolkit;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import javax.print.PrintException;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.T2MapFileChooser;
import jp.nihu.gttime.plugin.ImagePrinterPlugin;
import jp.nihu.gttime.plugin.ImageWriterPlugin;
import jp.nihu.gttime.plugin.LayerData;
import jp.nihu.gttime.plugin.LayerPainter;
import jp.nihu.gttime.plugin.PluginManager;
import jp.nihu.gttime.plugin.PluginManagerException;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.time.TimeAxisIF;
import jp.nihu.gttime.utility.ExtensionFileFilter;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * {@code LayerPlugin} オブジェクトの印刷／画像出力を行うプラグインクラス
 */
public class DefaultImageOutputPlugin extends AbstractImageOutputPlugin
{

    private static Log log = LogFactory.getLog(DefaultImageOutputPlugin.class);

    private static PluginResourceManager rmgr = PluginResourceManager.getManager(DefaultImageOutputPlugin.class, T2Map.LOCALE);

    private List<ImagePrinterPlugin> printerList;

    private List<ImageWriterPlugin> writerList;

    private JDialog dialog;

    /**
     * オブジェクトの初期化を行う。
     */
    @Override
    protected void initPlugin()
    {
        // Load Plugins.
        printerList = loadPlugin(ImagePrinterPlugin.class, rmgr.getString("plugin.printer.type"));
        writerList = loadPlugin(ImageWriterPlugin.class, rmgr.getString("plugin.writer.type"));

        // Create new instance.
        dialog = new ImageOutputDialog(getT2Map(), this);
    }

    /**
     * {@code Plugin} オブジェクトのロードを行う。
     */
    private <T> List<T> loadPlugin(Class<T> clazz, String type)
    {
        PluginManager pm = T2Map.getPluginManager();
        List<T> list = new ArrayList<T>();
        ArrayList<?> v = pm.getPlugins(type);
        for (Object obj : v)
        {
            Hashtable<?, ?> h = (Hashtable<?, ?>) obj;
            Object o = h.get("class");
            if (o != null)
            {
                try
                {
                    Object plugin = pm.newInstance(o.toString());
                    if (clazz.isInstance(plugin))
                    {
                        list.add(clazz.cast(plugin));
                    }
                }
                catch (PluginManagerException e)
                {
                    // Log [error]
                    log.error(e, e);
                }
            }
        }
        return list;
    }

    /**
     * 印刷／画像出力用ダイアログを表示する。
     */
    @Override
    public void output()
    {
        if (T2Map.getProject() != null)
        {
            dialog.setVisible(true);
        }
        else
        {
            showDialog(JOptionPane.WARNING_MESSAGE, "output.title", "output.noproject.message");
        }
    }

    /**
     * 現在アプリケーション側で表示している {@code LayerData} オブジェクト配列 を取得する。
     */
    @Override
    public LayerData[] getLayerDatas()
    {
        return DatasetListForImagePrinter.getArray(T2Map.getProject(), getT2Map().getLayerPanel());
    }

    /**
     * 現在アプリケーション側で表示している時間情報を {@code TimeAxisIF} オブジェクトとして取得する。
     */
    @Override
    public TimeAxisIF getTimeAxisData()
    {
        return getT2Map().getTimeAxisData();
    }

    /**
     * 保存可能なファイル拡張子を取得する。
     */
    @Override
    public String[] getWriterFileSuffixes()
    {
        List<String> list = new ArrayList<String>();
        for (ImageWriterPlugin p : writerList)
        {
            String[] strs = p.getWriterFileSuffixes();
            if (strs != null)
            {
                for (String s : strs)
                {
                    list.add(s);
                }
            }
        }
        return list.toArray(new String[0]);
    }

    /**
     * 現在アプリケーション側で表示しているレイヤ用パネルコンポーネントの サイズを取得する。
     */
    @Override
    public Dimension getLayerViewSize()
    {
        return getT2Map().getLayerPanel().getSize();
    }

    /**
     * 指定した条件で描画した画像の印刷を行う。
     * 
     * @param axis 出力する時間範囲を表す {@code TimeAxisIF} オブジェクト
     * @param datas 出力対象となる {@code LayerData} オブジェクト配列
     * @return 出力に成功した場合 {@code true}
     */
    @Override
    public boolean printImage(TimeAxisIF axis, LayerData[] datas)
    {
        boolean b = false;
        ImagePrinterPlugin plugin = null;
        for (ImagePrinterPlugin p : printerList)
        {
            plugin = (ImagePrinterPlugin) p;
            break;
        }
        if (plugin != null)
        {
            if (plugin.isPrintEnabled())
            {
                // Create new instance.
                LayerPainter painter = new DefaultLayerPainter(axis, datas, T2Map.getPluginManager());
                try
                {
                    b = plugin.print(painter);
                }
                catch (PrintException e)
                {
                    showDialog("print.title", "print.error.message", e);
                    // Log [error]
                    log.error(e, e);
                }
            }
            else
            {
                showDialog("print.title", "print.noservice.message");
            }
        }
        else
        {
            showDialog("print.title", "print.noplugin.message");
        }
        return b;
    }

    /**
     * 指定した条件で描画した画像をファイルへ出力する。
     * 
     * @param axis 出力する時間範囲を表す {@code TimeAxisIF} オブジェクト
     * @param datas 出力対象となる {@code LayerData} オブジェクト配列
     * @return 出力に成功した場合 {@code true}
     */
    @Override
    public boolean writeImageFile(TimeAxisIF axis, LayerData[] datas, String type)
    {
        boolean b = false;
        ImageWriterPlugin plugin = null;
        for (ImageWriterPlugin p : writerList)
        {
            if (p.isWritableFileFormat(type))
            {
                plugin = (ImageWriterPlugin) p;
                break;
            }
        }
        if (plugin != null)
        {
            File f = getSaveFile(plugin, type);
            if (f != null)
            {
                // Create new instance.
                LayerPainter painter = new DefaultLayerPainter(axis, datas, T2Map.getPluginManager());
                try
                {
                    b = plugin.write(painter, type, f);
                }
                catch (IOException e)
                {
                    // Log [error]
                    log.error(e, e);
                }
                if (!b)
                {
                    showDialog("file.title", "file.error.message", f);
                }
            }
        }
        else
        {
            showDialog("file.title", "file.noplugin.message");
            // Log [warn]
            log.warn("Can't get " + type + " writer plugin.");
        }
        return b;
    }

    /**
     * 指定した条件で描画した画像をクリップボードへ出力する。
     * 
     * @param axis 出力する時間範囲を表す {@code TimeAxisIF} オブジェクト
     * @param datas 出力対象となる {@code LayerData} オブジェクト
     * @return 出力に成功した場合 {@code true}
     */
    @Override
    public boolean copyToClipboard(TimeAxisIF axis, LayerData[] datas)
    {
        // Create new instance.
        LayerPainter p = new DefaultLayerPainter(axis, datas, T2Map.getPluginManager());
        // Set image.
        Toolkit.getDefaultToolkit().getSystemClipboard().setContents(TransferableDataHandler.getImage(p.getImage()), getT2Map());
        return true;
    }

    /**
     * 指定した条件で描画した画像のプレビューを行う。
     * 
     * @param axis 出力する時間範囲を表す {@code TimeAxisIF} オブジェクト
     * @param datas 出力対象となる {@code LayerData} オブジェクト配列
     */
    @Override
    public void previewImage(TimeAxisIF axis, LayerData[] datas)
    {
        // Create new instance.
        LayerPainter p = new DefaultLayerPainter(axis, datas, T2Map.getPluginManager());

        // Create preview dialog.
        JDialog d = new ImagePreviewDialog(dialog, p.getImage());
        d.setSize(getLayerViewSize());
        d.setVisible(true);
    }

    /**
     * 保存対象となるファイルを取得する。
     */
    private File getSaveFile(ImageWriterPlugin plugin, String type)
    {
        File f = null;
        JFileChooser fc = new T2MapFileChooser();
        String[] sfx = plugin.getWriterFileSuffixAliases(type);
        String dsp = plugin.getWriterFileDescription(type);
        fc.setFileFilter(new ExtensionFileFilter(sfx, dsp));
        while (f == null)
        {
            if (fc.showSaveDialog(getT2Map()) == JFileChooser.APPROVE_OPTION)
            {
                f = fc.getSelectedFile();
                if (f.getName().lastIndexOf('.') <= 0)
                {
                    f = new File(f.getParent(), f.getName() + "." + type);
                }
                if (f.exists())
                {
                    int rval = JOptionPane.showConfirmDialog(getT2Map(), rmgr.format("file.overwrite.message", f), rmgr.getString("file.title"), JOptionPane.YES_NO_OPTION);
                    if (rval != JOptionPane.YES_OPTION)
                    {
                        f = null;
                    }
                }
            }
            else
            {
                break;
            }
        }
        return f;
    }

    /**
     * メッセージ用ダイアログを表示する。
     */
    private void showDialog(String titleKey, String messageKey, Object... o)
    {
        showDialog(JOptionPane.ERROR_MESSAGE, titleKey, messageKey, o);
    }

    /**
     * メッセージ用ダイアログを表示する。
     */
    private void showDialog(int type, String titleKey, String messageKey, Object... o)
    {
        String title = rmgr.getString(titleKey);
        String message = rmgr.format(messageKey, o);
        JOptionPane.showMessageDialog(getT2Map(), message, title, type);
    }
}
