/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.printer;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.image.BufferedImage;
import java.awt.print.PageFormat;
import java.awt.print.Printable;
import java.awt.print.PrinterException;
import java.util.ArrayList;
import java.util.List;
import java.util.TreeMap;

import javax.print.Doc;
import javax.print.DocFlavor;
import javax.print.DocPrintJob;
import javax.print.PrintException;
import javax.print.PrintService;
import javax.print.PrintServiceLookup;
import javax.print.ServiceUI;
import javax.print.SimpleDoc;
import javax.print.attribute.Attribute;
import javax.print.attribute.HashDocAttributeSet;
import javax.print.attribute.HashPrintRequestAttributeSet;
import javax.print.attribute.PrintRequestAttributeSet;
import javax.print.attribute.standard.OrientationRequested;
import javax.print.event.PrintJobEvent;
import javax.print.event.PrintJobListener;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.plugin.ImagePrinterPlugin;
import jp.nihu.gttime.plugin.LayerPainter;
import jp.nihu.gttime.plugin.PluginResourceManager;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * 画像印刷用プラグインクラス
 */
public class DefaultImagePrinterPlugin implements ImagePrinterPlugin, Printable, PrintJobListener
{
    private static Log log = LogFactory.getLog(DefaultImagePrinterPlugin.class);

    private static PluginResourceManager rmgr = PluginResourceManager.getManager(DefaultImagePrinterPlugin.class, T2Map.LOCALE);

    public static final DocFlavor DEFAULT_FLAVOR = DocFlavor.SERVICE_FORMATTED.PRINTABLE;

    private BufferedImage layerImage = null;

    private List<Rectangle> clipList = new ArrayList<Rectangle>();

    /**
     * 画像印刷を行う。
     * 
     * @param p 画像描画用オブジェクト
     * @return 印刷に成功した場合 {@code true}
     * @exception PrintException この例外によって、原因をより正確に記述する インタフェースが実装されることもある
     */
    public boolean print(LayerPainter p) throws PrintException
    {
        boolean b = false;
        // Get print image.
        setLayerImage(p.getImage());
        // Show print dialog.
        PrintRequestAttributeSet attrSet = getAttributeSet();
        PrintService service = setUp(attrSet);
        if (service != null)
        {
            DocPrintJob job = service.createPrintJob();
            Doc doc = new SimpleDoc(this, getFlavor(), new HashDocAttributeSet());
            job.addPrintJobListener(this);
            job.print(doc, attrSet);
            b = true;
        }
        return b;
    }

    /**
     * 現在登録されているプリンタで画像印刷が可能かを判定する。
     * 
     * @return 印刷可能な場合 {@code true}
     */
    public boolean isPrintEnabled()
    {
        boolean b = false;
        PrintService[] services = PrintServiceLookup.lookupPrintServices(getFlavor(), getAttributeSet());
        if (services != null && services.length > 0)
        {
            // **************************************************************//
            // 一部の Linux ディトリビューション（Ubuntu 7.10 等）においては、
            // Java Print Service の印刷ダイアログ使用時に例外が発生する.
            // よって OrientationRequested アトリビュートをサポートしない
            // 印刷サービスが存在する場合は、印刷機能を利用不可とした.
            // **************************************************************//
            for (PrintService s : services)
            {
                Class<? extends Attribute> clazz = OrientationRequested.class;
                if (!(b = s.isAttributeCategorySupported(clazz)))
                {
                    // Log [warn]
                    // String format = "Printer[%s] not support attribute %s.";
                    // log.warn(String.format(format, s, clazz));
                    log.warn("Printer not support attribute.");
                    break;
                }
            }
        }
        return b;
    }

    /**
     * 印刷用ダイアログの表示位置を取得する。
     */
    private Point getDialogLocation()
    {
        Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
        Point p = new Point(d.width / 2, d.height / 2);
        p.translate(rmgr.getInt("location.translate.x"), rmgr.getInt("location.translate.y"));
        return p;
    }

    /**
     * 印刷に使用する {@code DocFlavor} を取得する。
     */
    private DocFlavor getFlavor()
    {
        return DEFAULT_FLAVOR;
    }

    /**
     * 印刷に使用する {@code PrintRequestAttributeSet} を取得する。
     */
    private PrintRequestAttributeSet getAttributeSet()
    {
        PrintRequestAttributeSet attr = new HashPrintRequestAttributeSet();
        return attr;
    }

    /**
     * 印刷用ダイアログを表示する。
     */
    private PrintService setUp(PrintRequestAttributeSet attr)
    {
        Point location = getDialogLocation();
        PrintService[] services = PrintServiceLookup.lookupPrintServices(getFlavor(), attr);
        // プリンタ名でのソート
        if (services != null)
        {
            TreeMap<String, PrintService> map = new TreeMap<String, PrintService>();
            for (PrintService s : services)
            {
                map.put(s.getName(), s);
            }
            services = map.values().toArray(new PrintService[0]);
        }
        return ServiceUI.printDialog(null, location.x, location.y, services, PrintServiceLookup.lookupDefaultPrintService(), getFlavor(), attr);
    }

    /**
     * 指定されたインデックスにあるページを、指定された書式で、指定された {@code Graphics} コンテキストに印刷する。
     * 
     * @param graphics ページが描画されるコンテキスト
     * @param pageFormat 描画されるページのサイズと印刷方向
     * @param pageIndex 描画されるページのゼロから始まるインデックス
     * @return ページの描画に成功した場合は {@code PAGE_EXISTS}、 {@code pageIndex}
     *         で指定されたページが存在しない場合は {@code NO_SUCH_PAGE}
     * @exception PrinterException 印刷が中断された場合
     */
    public int print(Graphics graphics, PageFormat pageFormat, int pageIndex) throws PrinterException
    {
        int sts = NO_SUCH_PAGE;
        if (isEmptyClipImage())
        {
            Dimension d = new Dimension();
            d.setSize(pageFormat.getImageableWidth(), pageFormat.getImageableHeight());
            setClipImage(d);
        }
        if (pageIndex < getClipImageSize())
        {
            BufferedImage img = getClipImage(pageIndex);
            if (img != null)
            {
                Graphics2D g = (Graphics2D) graphics;
                g.translate(pageFormat.getImageableX(), pageFormat.getImageableY());
                g.drawImage(img, 0, 0, null);
                sts = PAGE_EXISTS;
            }
        }
        else
        {
            clearClipImage();
        }
        return sts;
    }

    /**
     * 印刷対象画像を設定する。
     */
    private void setLayerImage(BufferedImage image)
    {
        clearClipImage();
        layerImage = image;
    }

    /**
     * インデックス値に対応した分割画像を取得する。
     */
    private BufferedImage getClipImage(int index)
    {
        BufferedImage img = null;
        if (index < clipList.size())
        {
            Rectangle r = clipList.get(index);
            if (r != null)
            {
                img = layerImage.getSubimage(r.x, r.y, r.width, r.height);
            }
        }
        return img;
    }

    /**
     * 画像を印刷ページ毎に分割する。
     */
    private void setClipImage(Dimension d)
    {
        if (layerImage != null && clipList.isEmpty())
        {
            Point p = new Point();
            Rectangle r = new Rectangle(p, d);
            Rectangle imgRect = new Rectangle(0, 0, layerImage.getWidth(), layerImage.getHeight());

            while (imgRect.contains(p))
            {
                while (imgRect.contains(p))
                {
                    clipList.add(r.intersection(imgRect));
                    // move Y position.
                    p.setLocation(r.getX(), p.getY() + r.getHeight());
                    r.setLocation(p);
                }
                // move X position.
                p.setLocation(p.getX() + r.getWidth(), 0);
                r.setLocation(p);
            }
            // Log [debug]
            log.debug("PaperArea=" + d + " : PrintArea=" + clipList);
        }
    }

    /**
     * 分割画像数を取得する。
     */
    private int getClipImageSize()
    {
        return clipList.size();
    }

    /**
     * 分割画像が存在するかを判定する。
     */
    private boolean isEmptyClipImage()
    {
        return clipList.isEmpty();
    }

    /**
     * 分割画像を削除する。
     */
    private void clearClipImage()
    {
        clipList.clear();
    }

    /**
     * データが正常に印刷サービスに転送され、クライアントがそのデータに 割り当てたローカルリソースを解放できることをクライアントに通知する
     * ために呼び出される。
     */
    public void printDataTransferCompleted(PrintJobEvent pje)
    {
        // Log [info]
        log.info("Print data transfer conpleted.");
    }

    /**
     * ジョブがユーザまたはプログラムによって取り消されたことをクライアントに 通知するために呼び出される。
     */
    public void printJobCanceled(PrintJobEvent pje)
    {
        // Log [warn]
        log.warn("Print job canceled.");
    }

    /**
     * ジョブが正常に完了したことをクライアントに通知するために呼び出される。
     */
    public void printJobCompleted(PrintJobEvent pje)
    {
        // Log [info]
        log.info("Print job completed.");
    }

    /**
     * ジョブが正常に完了しなかったこと、および再送信が必要であることを クライアントに通知するために呼び出される。
     */
    public void printJobFailed(PrintJobEvent pje)
    {
        // Log [error]
        log.error("Print job failed.");
    }

    /**
     * イベントの送信がこれ以上ないことをクライアントに通知するために 呼び出される。
     */
    public void printJobNoMoreEvents(PrintJobEvent pje)
    {
        // Log [debug]
        log.debug("Print job no more events.");
    }

    /**
     * ユーザが修復できるエラーが発生したことをクライアントに通知するために 呼び出される。
     */
    public void printJobRequiresAttention(PrintJobEvent pje)
    {
        // Log [warn]
        log.warn("Print job requires attention.");
    }
}
