/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.printer;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.T2MapProgressBarDialog;
import jp.nihu.gttime.plugin.LayerData;
import jp.nihu.gttime.plugin.LayerPainter;
import jp.nihu.gttime.plugin.LayerPlugin;
import jp.nihu.gttime.plugin.PluginManager;
import jp.nihu.gttime.plugin.PluginManagerException;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.time.TimeAxisIF;
import jp.nihu.gttime.tree.TreeCheckNode;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * {@code LayerPlugin} オブジェクトの描画用クラス
 */
public class DefaultLayerPainter implements LayerPainter
{

    private static Log log = LogFactory.getLog(DefaultLayerPainter.class);

    private static PluginResourceManager rmgr = PluginResourceManager.getManager(DefaultLayerPainter.class, T2Map.LOCALE);

    private static final String LAYER_DEFAULT_TYPE = rmgr.getString("layer.default.type");

    private PluginManager manager;

    private List<LayerPlugin> list;

    private Dimension size;

    private boolean drawBorder;

    /**
     * {@code LayerPlugin} オブジェクトの描画する {@code LayerPainter} を 生成する。
     * 
     * @param axis 時間範囲を表す {@code TimeAxisIF} オブジェク
     * @param datas 描画対象となる {@code LayerData} オブジェクト配列
     * @param manager {@code PluginManager} オブジェクト
     */
    public DefaultLayerPainter(TimeAxisIF axis, LayerData[] datas, PluginManager manager)
    {
        setPluginManager(manager);
        list = getLayerPlugins(axis, datas);
        size = getLayerSize(list);
        setDrawBorder(true);
    }

    /**
     * {@code LayerPlugin} オブジェクトを描画する。
     * 
     * @param g 描画に使用するグラフィックコンテキスト
     */
    public void paint(Graphics g, boolean repaintLayer)
    {
        // Frame frame = (Frame) SwingUtilities.getRoot();
        String title = rmgr.getString("progress.title");
        String message = rmgr.getString("progress.message");
        T2MapProgressBarDialog<Object> progress = new T2MapProgressBarDialog<Object>(null, true, title, message);
        Callable<Object> callable = (Callable<Object>) new PaintImage(g, repaintLayer);
        FutureTask<Object> f = (FutureTask<Object>) progress.submit(callable);
        progress.setVisible(true);
        try
        {
            f.get();
        }
        catch (CancellationException e)
        {
        }
        catch (InterruptedException e)
        {
            log.fatal(e.getMessage(), e);
        }
        catch (ExecutionException e)
        {
            log.fatal(e.getMessage(), e);
        }

        // for (LayerPlugin layer : getLayerPlugins( )) {
        // paint(g, layer, repaintLayer);
        // }
    }

    private class PaintImage implements Callable<Object>
    {
        private Graphics g;

        private boolean repaintLayer;

        public PaintImage(Graphics g, boolean repaintLayer)
        {
            this.g = g;
            this.repaintLayer = repaintLayer;
        }

        public Object call() throws Exception
        {
            for (LayerPlugin layer : getLayerPlugins())
            {
                paint(g, layer, repaintLayer);
            }
            return null;
        }
    }

    /**
     * {@code LayerPlugin} オブジェクトを描画する。
     * 
     * @param g 描画に使用するグラフィックコンテキスト
     * @param layer 描画対象となる {@code LayerPlugin} オブジェクト
     */
    public void paint(Graphics g, LayerPlugin layer, boolean repaintLayer)
    {
        // Draw LayerImage.
        if (repaintLayer)
        {
            layer.setPaintAsNotImage();
        }
        layer.paintComponent(g);
        // Translate 'y' position.
        log.debug(layer.getName() + " :: " + layer.getHeight());
        g.translate(0, layer.getHeight());
        // Draw border line.
        drawBorder(g);
    }

    /**
     * 各 {@code LayerPlugin} の間に罫線を描画する。
     */
    private void drawBorder(Graphics g)
    {
        if (isDrawBorder())
        {
            g.setColor(java.awt.Color.BLACK);
            g.drawLine(0, 0, getSize().width, 0);
        }
    }

    /**
     * {@code LayerPlugin} オブジェクトのイメージデータを取得する。
     */
    public BufferedImage getImage()
    {
        return getImage(BufferedImage.TYPE_INT_RGB);
    }

    /**
     * {@code LayerPlugin} オブジェクトのイメージデータを取得する。 指定したイメージ形式で構築した
     * {@code BufferedImage} に対して、 {@code LayerPlugin} オブジェクトの描画を行う。
     * 
     * @param imageType 取得するイメージの形式
     */
    public BufferedImage getImage(int imageType)
    {
        // Create image instance.
        BufferedImage image = new BufferedImage(getSize().width, getSize().height, imageType);
        // Create main graphics.
        Graphics g = image.createGraphics();
        // Draw layer.
        paint(g, false);
        g.dispose();
        return image;
    }

    /**
     * 描画対象となる {@code LayerPlugin} オブジェクトを取得する。
     * 
     * @return 描画対象の {@code LayerPlugin} オブジェクト配列
     */
    public LayerPlugin[] getLayerPlugins()
    {
        return list.toArray(new LayerPlugin[0]);
    }

    /**
     * 描画サイズを取得する。
     * 
     * @return 描画サイズを示す {@code Dimension} オブジェクト
     */
    public Dimension getSize()
    {
        return size;
    }

    /**
     * 各 {@code LayerPlugin} の間に罫線を描画するかを判定する。
     * 
     * @return 罫線を描画する場合 {@code true}
     */
    public boolean isDrawBorder()
    {
        return drawBorder;
    }

    /**
     * 各 {@code LayerPlugin} の間に罫線を描画するかを指定する。
     * 
     * @param b 罫線を描画する場合 {@code true}
     */
    public void setDrawBorder(boolean b)
    {
        drawBorder = b;
    }

    /**
     * {@code PluginManager} を取得する。
     */
    private PluginManager getPluginManager()
    {
        return manager;
    }

    /**
     * {@code PluginManager} を設定する。
     */
    private void setPluginManager(PluginManager manager)
    {
        this.manager = manager;
    }

    /**
     * {@code LayerData} に対応した {@code LayerPlugin} を取得する。
     */
    private List<LayerPlugin> getLayerPlugins(TimeAxisIF axis, LayerData[] datas)
    {
        List<LayerPlugin> list = new ArrayList<LayerPlugin>();
        if (datas != null)
        {
            for (LayerData d : datas)
            {
                // Get LayerPlugin.
                LayerPlugin layer = getLayerPlugin(d);
                if (layer != null)
                {
                    // Create new instance.
                    // parameters :dataset, allow child = false, enable = true
                    TreeCheckNode node = new TreeCheckNode(d.getDataset(), false);
                    // node.setSelected(true);
                    // Set draw scale.
                    node.setDrawTimeAxis(d.isDrawTimeAxis());
                    // Set image size.
                    // node.setHeight(d.getSize( ).height);
                    d.getDataset().getProperties().setLayerHeight(d.getSize().height);
                    // Setting for layer.
                    layer.setSize(d.getSize());
                    layer.initForPrint(axis, node);
                    list.add(layer);
                }
                else
                {
                    // Log [warn]
                    log.warn("Can't get LayerPlugin object.");
                }
            }
        }
        return list;
    }

    /**
     * {@code LayerData} に対応した {@code LayerPlugin} を取得する。
     */
    private LayerPlugin getLayerPlugin(LayerData data)
    {
        LayerPlugin layer = null;
        String type = data.getType();
        if (type == null || type.trim().length() == 0)
        {
            type = LAYER_DEFAULT_TYPE;
        }
        PluginManager pm = getPluginManager();
        Hashtable<?, ?> h = pm.getPlugins("LayerPlugin", type);
        if (h != null)
        {
            try
            {
                layer = (LayerPlugin) pm.newInstance(h.get("class").toString());
            }
            catch (PluginManagerException e)
            {
                // Log [error]
                log.error(e, e);
            }
        }
        return layer;
    }

    /**
     * {@code LayerPlugin} の描画サイズを算出する。
     */
    private Dimension getLayerSize(List<LayerPlugin> list)
    {
        // Create new instance.
        Dimension d = new Dimension();
        if (list != null)
        {
            for (LayerPlugin l : list)
            {
                if (d.width < l.getWidth())
                {
                    d.width = l.getWidth();
                }
                d.height += l.getHeight();
            }
        }
        return d;
    }
}
