/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.printer;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dialog;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JDialog;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.JTextField;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ListSelectionEvent;
import javax.swing.event.ListSelectionListener;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.plugin.LayerData;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.time.TimeAxisIF;
import jp.nihu.gttime.time.TimeLineData;
import jp.nihu.gttime.time.utility.ISO8601Exception;
import jp.nihu.gttime.time.utility.JulianGregorianDate;
import jp.nihu.gttime.utility.TextValidator;
import jp.nihu.gttime.utility.TextValidator.DATE;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * 印刷／画像出力用ダイアログクラス
 */
@SuppressWarnings("serial")
public class ImageOutputDialog extends JDialog implements ActionListener, ListSelectionListener
{
    private static Log log = LogFactory.getLog(ImageOutputDialog.class);

    private static PluginResourceManager rmgr = PluginResourceManager.getManager(ImageOutputDialog.class, T2Map.LOCALE);

    private AbstractImageOutputPlugin plugin;

    private List<JButton> btnList;

    private JList layerList;

    private JTextField timeFromField;

    private JTextField timeToField;

    private ButtonGroup scaleBtnGrp;

    private ButtonGroup targetBtnGrp;

    private JComboBox targetFileCmb;

    private JSpinner widthSp;

    /**
     * 印刷／画像出力用ダイアログを生成する。
     * 
     * @param owner ダイアログを表示する {@code Frame}
     * @param plugin 呼出元となるプラグインオブジェクト
     */
    public ImageOutputDialog(Frame owner, AbstractImageOutputPlugin plugin)
    {
        super(owner, true);
        initDialog(plugin);
    }

    /**
     * 印刷／画像出力用ダイアログを生成する。
     * 
     * @param owner ダイアログを表示する {@code Dialog}
     * @param plugin 呼出元となるプラグインオブジェクト
     */
    public ImageOutputDialog(Dialog owner, AbstractImageOutputPlugin plugin)
    {
        super(owner, true);
        initDialog(plugin);
    }

    /**
     * 呼出元となるプラグインオブジェクトを取得する。
     */
    private AbstractImageOutputPlugin getPlugin()
    {
        return plugin;
    }

    /**
     * 呼出元となるプラグインオブジェクトを設定する。
     */
    private void setPlugin(AbstractImageOutputPlugin plugin)
    {
        this.plugin = plugin;
    }

    /**
     * ダイアログの構築を行う。
     */
    private void initDialog(AbstractImageOutputPlugin plugin)
    {
        setPlugin(plugin);

        // Create new instance. [global]
        btnList = new ArrayList<JButton>();
        layerList = new JList();
        timeFromField = new JTextField(8);
        timeToField = new JTextField(8);
        scaleBtnGrp = new ButtonGroup();
        targetBtnGrp = new ButtonGroup();
        targetFileCmb = new JComboBox();
        widthSp = new JSpinner();
        // Create new instance. [local]
        JPanel itemPane = new JPanel();
        JPanel buttonPane = new JPanel();
        JPanel scalePane = new JPanel();
        JPanel targetPane = new JPanel();
        JScrollPane layerScrll = new JScrollPane();
        layerScrll.getViewport().add(layerList);
        layerScrll.setPreferredSize(new Dimension(rmgr.getInt("item.layer.width"), rmgr.getInt("item.layer.height")));

        // Set data.
        LayerData[] datas = getPlugin().getLayerDatas();
        TimeAxisIF axis = getPlugin().getTimeAxisData();
        Dimension lsize = getPlugin().getLayerViewSize();
        String[] sfxes = getPlugin().getWriterFileSuffixes();
        setLayerList(layerList, datas);
        timeFromField.setText(toGregorianDate(axis.getCurrentMin()));
        timeToField.setText(toGregorianDate(axis.getCurrentMax()));
        setTargetFileSuffixes(targetFileCmb, sfxes);
        setWidthSpinner(widthSp, lsize.width / 10 * 10);

        // Add Components. [scalePane, targetPane]
        JRadioButton radio;
        radio = getRadioButton("item.scale", "true", false);
        scaleBtnGrp.add(radio);
        scalePane.add(radio);
        radio = getRadioButton("item.scale", "false", false);
        radio.setSelected(true);
        scaleBtnGrp.add(radio);
        scalePane.add(radio);
        
        if (plugin instanceof DefaultPrintOutputPlugin) {
	        radio = getRadioButton("item.target", "print", true);
	        targetBtnGrp.add(radio);
	        targetPane.add(radio);
        } else if (plugin instanceof DefaultImageOutputPlugin) {
	        radio = getRadioButton("item.target", "clip", true);
	        targetBtnGrp.add(radio);
	        targetPane.add(radio);
	        
	        radio = getRadioButton("item.target", "file", true);
	        targetBtnGrp.add(radio);
	        targetPane.add(radio);
        }

        // Set Layout.
        itemPane.setLayout(new GridBagLayout());
        // Set Border.
        itemPane.setBorder(BorderFactory.createTitledBorder(rmgr.getString("border.title")));

        // Add Components. [itemPane]
        GridBagConstraints c = new GridBagConstraints();
        c.gridy = 0;
        c.ipadx = c.ipady = 2;
        setGridBag(c, "item.layer.label", itemPane, layerScrll);
        setGridBag(c, "item.time.label", itemPane, timeFromField, new JLabel(rmgr.getString("item.time.delim")), timeToField);
        setGridBag(c, "item.scale.label", itemPane, scalePane);
        if (plugin instanceof DefaultImageOutputPlugin) {
        	setGridBag(c, "item.target.label", itemPane, targetPane, targetFileCmb);
        }
        setGridBag(c, "item.width.label", itemPane, widthSp, new JLabel(rmgr.getString("item.width.unit")));

        // Add Component. [buttonPane]
        JButton btn;
        btn = getButton("button", "output");
        buttonPane.add(btn);
        btnList.add(btn);
        btn = getButton("button", "preview");
        buttonPane.add(btn);
        btnList.add(btn);
        btn = getButton("button", "cancel");
        buttonPane.add(btn);
        setButtonEnabled(!layerList.isSelectionEmpty());

        // Set Title.
        if (plugin instanceof DefaultPrintOutputPlugin) {
        	setTitle(rmgr.getString("dialog.title.print"));
        } else if (plugin instanceof DefaultImageOutputPlugin) {
        	setTitle(rmgr.getString("dialog.title.image"));
        }
        
        // Set Layout.
        getContentPane().setLayout(new BorderLayout());
        // Add Components.
        getContentPane().add(itemPane, BorderLayout.CENTER);
        getContentPane().add(buttonPane, BorderLayout.SOUTH);
        pack();
        setResizable(false);
        setLocationRelativeTo(getOwner());
    }

    /**
     * {@code JList} に対してレイヤ情報を設定する。
     */
    private void setLayerList(JList list, LayerData[] datas)
    {
        list.addListSelectionListener(this);
        list.setListData(datas);
        if (datas != null && datas.length > 0)
        {
            list.setListData(datas);
            List<Integer> indexList = new ArrayList<Integer>();
            for (int i = 0; i < datas.length; i++)
            {
                if (datas[i].isVisible())
                {
                    indexList.add(i);
                }
            }
            if (indexList.size() > 0)
            {
                int indices[] = new int[indexList.size()];
                for (int i = 0; i < indexList.size(); i++)
                {
                    indices[i] = indexList.get(i);
                }
                list.setSelectedIndices(indices);
            }
        }
    }

    /**
     * {@code JComboBox} に対してファイル拡張子を設定する。
     */
    private void setTargetFileSuffixes(JComboBox cb, String[] suffixes)
    {
        for (String s : suffixes)
        {
            cb.addItem(s);
        }
        Dimension d = cb.getPreferredSize();
        d.width += 12;
        cb.setPreferredSize(d);
        cb.setEditable(false);
        cb.setEnabled(true);
    }

    /**
     * {@code JSpinner} に対して描画幅を設定する。
     */
    private void setWidthSpinner(JSpinner sp, int width)
    {
        JSpinner.NumberEditor editor = new JSpinner.NumberEditor(sp, "0");
        SpinnerNumberModel model = new SpinnerNumberModel();
        editor.getTextField().setColumns(5);
        editor.getTextField().setEditable(true);
        editor.getTextField().setHorizontalAlignment(JTextField.LEFT);
        editor.getTextField().setValue(width);
        model.setValue(width);
        model.setMinimum(rmgr.getInt("item.width.min"));
        model.setMaximum(rmgr.getInt("item.width.max"));
        model.setStepSize(rmgr.getInt("item.width.stepsize"));
        sp.setEditor(editor);
        sp.setModel(model);
    }

    /**
     * 指定した条件で {@code JButton} を生成する。
     */
    private JButton getButton(String keyPrefix, String cmd)
    {
        JButton btn = new JButton();
        StringBuffer buf = new StringBuffer(keyPrefix);
        buf.append(".").append(cmd);
        btn.setText(rmgr.getString(buf.toString()));
        btn.setActionCommand(cmd);
        btn.addActionListener(this);
        return btn;
    }

    /**
     * 指定した条件で {@code JRadioButton} を生成する。
     */
    private JRadioButton getRadioButton(String keyPrefix, String cmd, boolean addListener)
    {
        JRadioButton btn = new JRadioButton();
        StringBuffer buf = new StringBuffer(keyPrefix);
        buf.append(".").append(cmd).append(".name");
        btn.setText(rmgr.getString(buf.toString()));
        btn.setActionCommand(cmd);
//        btn.setSelected(rmgr.getString(keyPrefix + ".selected").equals(cmd));
        
        if ("print".equals(cmd) || "file".equals(cmd)) {
        	btn.setSelected(true);
        }
        
        if (addListener)
        {
            btn.addActionListener(this);
        }
        return btn;
    }

    /**
     * {@code GridBagLayout} を持つパネルに対してコンポーネントの設定を行う。
     */
    private void setGridBag(GridBagConstraints gbc, String labelKey, Container parent, Component... compornents)
    {
        JPanel panel = new JPanel();
        for (Component c : compornents)
        {
            if (c instanceof JPanel)
            {
                panel = (JPanel) c;
            }
            else
            {
                panel.add(c);
            }
        }
        gbc.gridx = 0;
        gbc.anchor = GridBagConstraints.EAST;
        parent.add(new JLabel(rmgr.getString(labelKey)), gbc);
        gbc.gridx = 1;
        gbc.anchor = GridBagConstraints.WEST;
        parent.add(panel, gbc);
        gbc.gridy++;
    }

    /**
     * グレゴリオ暦をユリウス暦に変換する。
     */
    private static BigDecimal toJulianDate(String s)
    {
        BigDecimal val = BigDecimal.valueOf(0);
        try
        {
            val = JulianGregorianDate.toJulian(s);
        }
        catch (ISO8601Exception e)
        {
            // Log [error]
            log.error(e, e);
        }
        return val;
    }

    /**
     * ユリウス暦をグレゴリオ暦に変換する。
     */
    private static String toGregorianDate(BigDecimal d)
    {
        return JulianGregorianDate.toGregorian(d.setScale(0, BigDecimal.ROUND_HALF_UP).toBigInteger());
    }

    /**
     * アクションが発生すると呼び出される。
     */
    public void actionPerformed(ActionEvent evt)
    {
        Object o = evt.getSource();
        String cmd = evt.getActionCommand();
        if (o instanceof JRadioButton)
        {
            targetFileCmb.setEnabled("file".equals(cmd));
        }
        else
        {
            boolean b = false;
            if ("cancel".equals(cmd))
            {
                b = true;
            }
            else
            {
                // Check input values.
                if (validateText())
                {
                    if ("output".equals(cmd))
                    {
                        if (targetBtnGrp.getSelection() != null)
                        {
                            cmd = targetBtnGrp.getSelection().getActionCommand();
                        }
                    }
                    b = outputImage(cmd);
                }
            }
            if (b)
            {
                dispose();
            }
        }
    }

    /**
     * 指定した {@code cmd} の内容沿って画像出力を行う。
     */
    private boolean outputImage(String cmd)
    {
        boolean b = false;
        TimeAxisIF axis = getTimeAxisData();
        LayerData[] datas = getSelectedLayer();
        if ("preview".equals(cmd))
        {
            getPlugin().previewImage(axis, datas);
        }
        else if ("print".equals(cmd))
        {
            b = getPlugin().printImage(axis, datas);
        }
        else if ("clip".equals(cmd))
        {
            b = getPlugin().copyToClipboard(axis, datas);
        }
        else if ("file".equals(cmd))
        {
            b = getPlugin().writeImageFile(axis, datas, getFileType());
        }
        return b;
    }

    /**
     * 出力する時間範囲を取得する。
     */
    private TimeAxisIF getTimeAxisData()
    {
        // Create new instance.
        TimeAxisIF axis = new TimeLineData();
        BigDecimal from = toJulianDate(timeFromField.getText());
        BigDecimal to = toJulianDate(timeToField.getText());
        axis.setApplicationMin(from);
        axis.setApplicationMax(to);
        axis.setMin(from);
        axis.setMax(to);
        axis.setCurrentMin(from);
        axis.setCurrentMax(to);
        return axis;
    }

    /**
     * 出力するレイヤを取得する。
     */
    private LayerData[] getSelectedLayer()
    {
        List<LayerData> list = new ArrayList<LayerData>();
        Object[] objs = layerList.getSelectedValues();
        if (objs != null)
        {
            int width = getImageWidth();
            boolean scale = isDrawScale();
            for (Object o : objs)
            {
                if (o instanceof LayerData)
                {
                    LayerData d = (LayerData) o;
                    d.setWidth(width);
                    d.setDrawTimeAxis(scale);
                    list.add(d);
                }
            }
        }
        return list.toArray(new LayerData[0]);
    }

    /**
     * 出力する画像幅を取得する。
     */
    private int getImageWidth()
    {
        int width = 0;
        try
        {
            width = Integer.parseInt(widthSp.getValue().toString());
        }
        catch (Exception e)
        {
            // Log [error]
            log.error(e, e);
        }
        width = Math.max(width, rmgr.getInt("item.width.min"));
        width = Math.min(width, rmgr.getInt("item.width.max"));
        widthSp.setValue(width);
        return width;
    }

    /**
     * 出力画像に時間目盛を描画するかを判定する。
     */
    private boolean isDrawScale()
    {
        boolean b = false;
        if (scaleBtnGrp.getSelection() != null)
        {
            String s = scaleBtnGrp.getSelection().getActionCommand();
            b = Boolean.valueOf(s).booleanValue();
        }
        return b;
    }

    /**
     * 出力する画像のファイル形式を取得する。
     */
    private String getFileType()
    {
        return (String) targetFileCmb.getSelectedItem();
    }

    /**
     * 選択範囲内の値が変更されたときに呼び出される。
     */
    public void valueChanged(ListSelectionEvent evt)
    {
        if (evt.getValueIsAdjusting())
        {
            Object o = evt.getSource();
            if (o instanceof JList)
            {
                setButtonEnabled(!((JList) o).isSelectionEmpty());
            }
        }
    }

    /**
     * {@code JButton} の使用可／不可を設定する。
     */
    private void setButtonEnabled(boolean b)
    {
        for (JButton button : btnList)
        {
            button.setEnabled(b);
        }
    }

    /**
     * {@code JTextField} に入力した値を検証する。
     */
    private boolean validateText()
    {
        TextValidator[] vs = { new DATE(timeFromField, "y-MM-dd", true), new DATE(timeToField, "y-MM-dd", true) };
        // 入力値検証
        boolean b = TextValidator.validate(vs);
        if (b)
        {
            // 「開始日 < 終了日」検証
            BigDecimal d1 = toJulianDate(timeFromField.getText());
            BigDecimal d2 = toJulianDate(timeToField.getText());
            if (!(b = (d1.compareTo(d2) < 0)))
            {
                timeFromField.setBackground(TextValidator.NG);
                timeToField.setBackground(TextValidator.NG);
            }
        }
        return b;
    }
}
