/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.plugin.printer;

import java.awt.Image;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.datatransfer.UnsupportedFlavorException;
import java.io.File;
import java.io.IOException;
import java.util.List;

/**
 * 指定したオブジェクトを転送する際に必要な能力を実装する {@code Transferable} を提供するクラス。
 */
public class TransferableDataHandler implements Transferable
{
    private Object obj;

    private DataFlavor[] flavors;

    /**
     * 指定したオブジェクトを転送可能な {@code Transferable} を作成する。
     * 
     * @param obj 転送されるデータを表すオブジェクト
     * @param flavor このデータを転送できるデータフレーバ
     */
    private TransferableDataHandler(Object obj, DataFlavor flavor)
    {
        this.obj = obj;
        this.flavors = new DataFlavor[1];
        this.flavors[0] = flavor;
    }

    /**
     * 転送されるデータを表すオブジェクトを返却する。 返されるオブジェクトのクラスは、フレーバの表現クラスで定義される。
     * 
     * @param flavor データに対して要求されたフレーバ
     * @exception IOException データが、要求されたフレーバで使用不能である場合
     * @exception UnsupportedFlavorException 要求したデータフレーバが サポートされていない場合
     */
    public Object getTransferData(DataFlavor flavor) throws UnsupportedFlavorException, IOException
    {
        if (isDataFlavorSupported(flavor))
        {
            return obj;
        }
        else
        {
            throw new UnsupportedFlavorException(flavor);
        }
    }

    /**
     * データを提供することができるフレーバを示す {@code DataFlavor} オブジェクトの配列を返却する。
     * 
     * @return このデータを転送できるデータフレーバの配列
     */
    public DataFlavor[] getTransferDataFlavors()
    {
        return flavors;
    }

    /**
     * 指定されたデータフレーバが、このオブジェクトに対してサポートされて いるかを返却する。
     * 
     * @param flavor データに対して要求されたフレーバ
     * @return データフレーバがサポートされている場合 {@code true}
     */
    public boolean isDataFlavorSupported(DataFlavor flavor)
    {
        boolean b = false;
        for (DataFlavor f : getTransferDataFlavors())
        {
            if (b = f.equals(flavor))
            {
                break;
            }
        }
        return b;
    }

    /**
     * 指定したオブジェクトを転送可能な {@code Transferable} を返却する。
     * 
     * @param obj 転送されるデータを表すオブジェクト
     * @param flavor このデータを転送できるデータフレーバ
     */
    public static Transferable get(Object obj, DataFlavor flavor)
    {
        return new TransferableDataHandler(obj, flavor);
    }

    /**
     * 指定したファイルのリストを転送可能な {@code Transferable} を返却する。
     * 
     * @param list 転送する {@code List<File>} オブジェクト
     */
    public static Transferable getFileList(List<File> list)
    {
        return get(list, DataFlavor.javaFileListFlavor);
    }

    /**
     * 指定した {@code Image} を転送可能な {@code Transferable} を返却する。
     * 
     * @param image 転送する {@code Image} オブジェクト
     */
    public static Transferable getImage(Image image)
    {
        return get(image, DataFlavor.imageFlavor);
    }

    /**
     * 指定した {@code String} を転送可能な {@code Transferable} を返却する。
     * 
     * @param s 転送する {@code String} オブジェクト
     */
    public static Transferable getString(String s)
    {
        return get(s, DataFlavor.stringFlavor);
    }
}
