/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.rdf;

import java.awt.Frame;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.concurrent.Callable;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.FutureTask;

import javax.swing.JFileChooser;
import javax.swing.SwingUtilities;
import javax.swing.filechooser.FileFilter;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.T2MapProgressBarDialog;
import jp.nihu.gttime.plugin.PluginResourceManager;
import jp.nihu.gttime.utility.ExtensionFileFilter;
import jp.nihu.gttime.utility.ExtensionFileSaveChooser;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.apache.jena.riot.RDFDataMgr;
import org.apache.jena.riot.RDFFormat;

import com.hp.hpl.jena.rdf.model.Model;

/**
 * csv出力時のダイアログ及びgtmファイルを作成します。
 */
public class RDFSave
{
    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(RDFSave.class);

    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(RDFSave.class, T2Map.LOCALE);
    
    private T2Map t2map;
    
    public RDFSave(T2Map t2map) {
    	this.t2map = t2map;
    }

    /**
     * csvファイル保存ダイアログの表示。
     * 
     * @param dataset Datasetオブジェクト
     * @param t2map T2Mapオブジェクト
     */
    public void save(Model model, String name) {
        String currentDirPath = T2Map.properties.getProperty("jp.nihu.gttime.T2Map.CURRENT_DIR");
        ExtensionFileSaveChooser chooser = new ExtensionFileSaveChooser(currentDirPath + File.separator + name);
        
        Object[][] exts = {
        		{ "ttl", "Turtle", RDFFormat.TURTLE },
        		{ "nt", "N-Triple", RDFFormat.NTRIPLES },
        		{ "nq", "N-Quads", RDFFormat.NQUADS },
        		{ "trig", "TriG", RDFFormat.TRIG },
        		{ "rdf", "RDF/XML", RDFFormat.RDFXML },
        		{ "rdf", "RDF/XML Abbrev.", RDFFormat.RDFXML_ABBREV },
        		{ "rj", "RDF/JSON", RDFFormat.RDFJSON },
        		{ "jsonld", "JSON-LD", RDFFormat.JSONLD }
        };
        
        for (Object[] ext : exts) {
        	chooser.addChoosableFileFilter(new ExtensionFileFilter((String)ext[0], (String)ext[1]));
        }

        int returnVal = chooser.showSaveDialog(SwingUtilities.getRoot(t2map));
        // 保存ボタンが押された時
        if (returnVal == JFileChooser.APPROVE_OPTION) {
            File file = chooser.getSelectedFile();
            FileFilter filter = chooser.getFileFilter();
            RDFFormat format = RDFFormat.RDFXML;
            for (Object[] ext : exts) {
                if (filter.getDescription().equals(ext[1])) {
                	format = (RDFFormat)ext[2];
                	break;
                }
            }
            Frame frame = (Frame) SwingUtilities.getRoot(t2map);
            String title = rmgr.getString("progress.title");
            String message = rmgr.format("progress.message", file.getAbsolutePath());
            T2MapProgressBarDialog<Object> progress = new T2MapProgressBarDialog<Object>(frame, true, title, message);
            Callable<Object> callable = (Callable<Object>) new SaveRdf(model, format, file.getParentFile(), file.getName());
            FutureTask<Object> f = (FutureTask<Object>) progress.submit(callable);
            progress.setVisible(true);
            
            try {
                f.get();
            } catch (CancellationException e) {
            } catch (InterruptedException e) {
                log.fatal(e.getMessage(), e);
            } catch (ExecutionException e) {
                log.fatal(e.getMessage(), e);
            }
            
            try {
                T2Map.properties.store();
            } catch (FileNotFoundException e) {
                log.fatal(e.getMessage(), e);
            } catch (IOException e) {
                log.fatal(e.getMessage(), e);
            }
    	} else {
    		log.fatal("エクスポートするオブジェクトがDatasetではありません。");
    	}
    }


    private class SaveRdf implements Callable<Object> {
      private Model model;
      
      private RDFFormat format;

      private File base;

      private String fileName;

      public SaveRdf(Model model, RDFFormat format, File base, String fileName) {
          this.model = model;
          this.format = format;
          this.base = base;
          this.fileName = fileName;
      }

      public Object call() throws Exception {
          saveModel(model, format, base, fileName);
          return model;
      }
    }

    private void saveModel(Model model, RDFFormat format, File base, String fileName) throws FileNotFoundException {
    	FileOutputStream out =  new FileOutputStream(new File(base, fileName));
    	RDFDataMgr.write(out, model, format);
    }
}
