/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.tree;

import javax.swing.ImageIcon;
import javax.swing.tree.DefaultMutableTreeNode;

import jp.nihu.gttime.data.DatasetIF;
import jp.nihu.gttime.data.ProjectIF;
import jp.nihu.gttime.plugin.LayerPlugin;

/**
 * プロジェクトとデータセットなどのツリー情報を管理するクラスです。
 */
public class TreeCheckNode extends DefaultMutableTreeNode
{
    // private static Log log = LogFactory.getLog(TreeCheckNode.class);

    /**
     * シリアルバーションID
     */
    private static final long serialVersionUID = 1L;

    /**
     * このノードに対応するレイヤ
     */
    private LayerPlugin layer = null;

    private TreeDragDrop treeView = null;

    /**
     * 時間軸の表示状態
     */
    private boolean drawTimeAxis = false;

    /**
     * 表示スケールの種別
     */
    private String selectScale = "DefaultScale";

    /**
     * Y軸（目盛り値）の最小数値指定
     */
    private float minTextH = 0;

    /**
     * Y軸（目盛り値）の最大数値指定
     */
    private float maxTextH = 0;

    /**
     * Y軸（目盛り値）の数値指定を「apply/cancel」の状態
     */
    private int buttonScaleSts = 0;

    /**
     * X軸の補助線の表示/非表示の状態
     */
    private boolean borderDrow = true;

    /**
     * Y軸の最大値
     */
    private float scaleMax = 0;

    /**
     * Y軸の最小値
     */
    private float scaleMin = 0;

    private ImageIcon icon = null;

    public void setIcon(ImageIcon icon)
    {
        this.icon = icon;
    }

    public ImageIcon getIcon()
    {
        return icon;
    }

    public void setTreeView(TreeDragDrop treeView)
    {
        this.treeView = treeView;
    }

    public TreeDragDrop getTreeView()
    {
        return treeView;
    }

    public void setLayer(LayerPlugin layer)
    {
        this.layer = layer;
    }

    public LayerPlugin getLayer()
    {
        return layer;
    }

    /**
     * Y軸の最小値を設定します
     * 
     * @param scaleMin Y軸の最小値
     */
    public void setScaleMin(float scaleMin)
    {
        this.scaleMin = scaleMin;
    }

    /**
     * Y軸の最小値を取得します
     * 
     * @return Y軸の最小値
     */
    public float getScaleMin()
    {
        return scaleMin;
    }

    /**
     * Y軸の最大値を設定します
     * 
     * @param scaleMax Y軸の最大値
     */
    public void setScaleMax(float scaleMax)
    {
        this.scaleMax = scaleMax;
    }

    /**
     * Y軸の最大値を取得します
     * 
     * @return Y軸の最大値
     */
    public float getScaleMax()
    {
        return scaleMax;
    }

    /**
     * X軸の補助線の表示/非表示の状態を取得します。
     * 
     * @return X軸の補助線の表示/非表示の状態
     */
    public boolean isBorderDrow()
    {
        return borderDrow;
    }

    /**
     * X軸の補助線の表示/非表示の状態を設定します。
     * 
     * @param borderSlected X軸の補助線の表示/非表示の状態
     * @return X軸の補助線の表示/非表示の状態
     */
    public void setBorderDrow(boolean borderDrow)
    {
        this.borderDrow = borderDrow;
    }

    /**
     * Y軸（目盛り値）の最小数値を設定します。
     * 
     * @param minTextH Y軸（目盛り値）の最小数値
     */
    public void setInputMin(float minTextH)
    {
        this.minTextH = minTextH;
    }

    /**
     * Y軸（目盛り値）の最小数値を取得します。
     * 
     * @return Y軸（目盛り値）の最小数値
     */
    public float getInputMin()
    {
        return minTextH;
    }

    /**
     * Y軸（目盛り値）の最大数値を設定します。
     * 
     * @param maxTextH Y軸（目盛り値）の最大数値
     */
    public void setInputMax(float maxTextH)
    {
        this.maxTextH = maxTextH;
    }

    /**
     * Y軸（目盛り値）の最大数値を取得します。
     * 
     * @return Y軸（目盛り値）の最大数値
     */
    public float getInputMax()
    {
        return maxTextH;
    }

    /**
     * Y軸（目盛り値）の数値指定実行について「apply/cancel」の状態を設定します。
     * 
     * @param btnScaleSts Y軸（目盛り値）の数値指定実行について「apply/cancel」の状態
     */
    public void setButtonScaleSts(int btnScaleSts)
    {
        this.buttonScaleSts = btnScaleSts;
    }

    /**
     * Y軸（目盛り値）の数値指定実行について「apply/cancel」の状態を取得します
     * 
     * @return Y軸（目盛り値）の数値指定実行について「apply/cancel」の状態
     */
    public int getButtonScaleSts()
    {
        return buttonScaleSts;
    }

    /**
     * 時間軸の表示状態を取得します。
     * 
     * @return 時間軸の表示状態
     */
    public boolean isDrawTimeAxis()
    {
        return drawTimeAxis;
    }

    /**
     * 時間軸の表示状態を設定します。
     * 
     * @param drawAxis 時間軸の表示状態
     */
    public void setDrawTimeAxis(boolean drawAxis)
    {
        this.drawTimeAxis = drawAxis;
    }

    /**
     * 表示スケールの種別を取得します
     * 
     * @return 表示スケールの種別
     */
    public String getSelectScale()
    {
        return selectScale;
    }

    /**
     * 表示スケールの種別を設定します
     * 
     * @param selectScale 表示スケールの種別
     */
    public void setSelectScale(String selectScale)
    {
        this.selectScale = selectScale;
    }

    // /**
    // * レイヤの高さを取得します。
    // *
    // * @return レイヤの高さ
    // */
    // public int getHeight()
    // {
    // return height;
    // }

    // /**
    // * レイヤの高さを設定します。
    // *
    // * @param height レイヤの高さ
    // */
    // public void setHeight(int height)
    // {
    // if (height > 70)
    // {
    // this.height = 70;
    // }
    // else
    // {
    // this.height = height;
    // }
    // }

    // /**
    // * TreeCheckNodeのコンストラクター<br/>
    // *
    // * @param userObject Objectオブジェクト
    // */
    // public TreeCheckNode(Object userObject)
    // {
    // this(userObject, true, false);
    // }

    /**
     * 空のコンストラクター
     */
    public TreeCheckNode()
    {
    }

    /**
     * TreeCheckNodeのコンストラクター
     * 
     * @param nodeObj Objectオブジェクト
     * @param allowsChildren
     * @param selected
     */
    public TreeCheckNode(Object nodeObj, boolean allowsChildren)
    {
        super(nodeObj, allowsChildren);
        // setSelected(selected);
        // Object obj = getUserObject();
        // if (obj instanceof ProjectIF)
        // {
        // ((ProjectIF)obj).setVisible(selected);
        // }
        // else if (obj instanceof DatasetIF)
        // {
        // ((DatasetIF)obj).setVisible(selected);
        // }
    }

    /**
     * リーフとなる子要素の存在確認
     */
    public boolean isLeaf()
    {
        boolean leaf = false;
        Object nodeObj = getUserObject();
        if (nodeObj instanceof ProjectIF)
        {
            leaf = false;
        }
        else if (nodeObj instanceof DatasetIF)
        {
            leaf = true;
        }
        return leaf;
    }

    /**
     * ノードの選択状態を設定します。
     * 
     * @param selected ノードの選択状態
     */
    public void setSelected(boolean selected)
    {
        Object obj = getUserObject();
        if (obj instanceof ProjectIF)
        {
            ((ProjectIF) obj).setVisible(selected);
        }
        else if (obj instanceof DatasetIF)
        {
            ((DatasetIF) obj).setVisible(selected);
        }
    }

    /**
     * ノードが有効かどうか確認します。
     * 
     * @return ノードの状態
     */
    protected boolean isEnabled()
    {
        boolean enabled = true;
        if (!isSelected())
        {
            enabled = false;
        }
        else
        {
            enabled = isParentEnabled();
        }
        return enabled;
    }

    /**
     * 親ノードが有効かどうかの確認します。
     * 
     * @return 親ノードの状態
     */
    protected boolean isParentEnabled()
    {
        TreeCheckNode parent = (TreeCheckNode) getParent();
        if (parent != null && !parent.isEnabled())
        {
            return false;
        }
        return true;
    }

    /**
     * チェックボックス選択状態を取得します。
     * 
     * @return チェックボックス選択状態
     */
    public boolean isSelected()
    {
        Object obj = getUserObject();
        if (obj instanceof ProjectIF)
        {
            return ((ProjectIF) obj).isVisible();
        }
        else if (obj instanceof DatasetIF)
        {
            return ((DatasetIF) obj).isVisible();
        }
        return false;
    }
    
    private boolean clicked = false;
    
    public void setClicked(boolean clicked) {
    	this.clicked = clicked;
    }
    
    public boolean isClicked()
    {
        return clicked;
    }
}
