/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.tree;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import javax.swing.ImageIcon;
import javax.swing.JCheckBox;
import javax.swing.JPanel;
import javax.swing.JTree;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.tree.TreeCellRenderer;
import javax.swing.tree.TreePath;

import jp.nihu.gttime.data.DatasetIF;
import jp.nihu.gttime.data.ProjectIF;

/**
 * ツリーを描画するクラスである。
 */
public class TreeCheckRenderer extends JPanel implements TreeCellRenderer, MouseMotionListener, MouseListener, KeyListener
{
    /**
     * シリアルバーションID
     */
    private static final long serialVersionUID = 1L;

    /**
     * JCheckBoxオブジェクト
     */
    private JCheckBox check = null;
    
//    public JCheckBox getCheckBox() {
//    	return check;
//    }

    /**
     * TreeDrawオブジェクト
     */
    private TreeDraw label = null;

    /**
     * ターゲットノードをチェックする際に使用する
     */
    private boolean isTargetNode;

    /**
     * ターゲットリーフをチェックする際に使用する
     */
    private boolean isTargetNodeLeaf;

    /**
     * マウスがポイントしている行番号
     */
    private int rollOverRowIndex = -1;
    
    private int clickRowIndex = -1;

    /**
     * 行がマウスにポイントされているときの背景色
     */
    private static final Color rollOverRowColor = new Color(220, 240, 255);

    /**
     * ツリーオブジェクト
     */
    private final JTree tree;

    /**
     * ツリーの描画前処理
     */
    public TreeCheckRenderer(JTree tree)
    {
        super();

        this.tree = tree;
        tree.addMouseMotionListener(this);
        tree.addMouseListener(this);
        tree.addKeyListener(this);

        setLayout(null);
        add(check = new JCheckBox());
        add(label = new TreeDraw());
        check.setBackground(Color.WHITE);
        // label.setForeground(UIManager.getColor("Tree.textForeground"));
    }

    /**
     * ツリーセルを描画（ドラッグ時にノードやリーフに色枠線を描画）する
     * 
     * @see javax.swing.tree.TreeCellRenderer#getTreeCellRendererComponent(javax.swing.JTree,
     *      java.lang.Object, boolean, boolean, boolean, int, boolean)
     */
    public Component getTreeCellRendererComponent(JTree tree, Object value, boolean selected, boolean expanded, boolean leaf, int row, boolean hasFocus)
    {
        TreeCheckNode node = (TreeCheckNode) value;
        String stringValue = tree.convertValueToText(value, selected, expanded, leaf, row, hasFocus);

        check.setSelected(node.isSelected());
        label.setText(stringValue);
        
        node.setClicked(false);

        Object obj = node.getUserObject();
        if (obj instanceof ProjectIF)
        {
            ((ProjectIF) obj).setOpen(expanded);
        }

        if (node.isParentEnabled())
        {
            check.setEnabled(true);
        }
        else
        {
            check.setEnabled(false);
        }

        if (row == rollOverRowIndex)
        {
            label.setOpaque(true);
            label.setBackground(rollOverRowColor);
        }
        else if (row == clickRowIndex)
        {
        	if (obj instanceof DatasetIF)
            {
        		node.setClicked(true);
            	label.setOpaque(true);
                label.setBackground(Color.lightGray);
            }
        	else
        	{
        		label.setOpaque(false);
        	}
        }
        else
        {
            label.setOpaque(false);
        }

        treeIcon(expanded, leaf, node);

        // 選択線
        if (tree instanceof TreeDragDrop)
        {
            isTargetNode = (value.equals(((TreeDragDrop) tree).getDropTargetNode()));
            isTargetNodeLeaf = (isTargetNode && node.isLeaf());
        }

        return this;
    }

    /**
     * ノードやリーフのアイコンの設定を行います。
     * 
     * @param expanded ノード開閉状態
     * @param leaf リーフ
     */
    private void treeIcon(boolean expanded, boolean leaf, TreeCheckNode node)
    {
        if (leaf)
        {
            ImageIcon icon = node.getIcon();
            if (icon != null)
            {
                label.setIcon(icon);
            }
            else
            {
                label.setIcon(UIManager.getIcon("Tree.leafIcon"));
            }
        }
        else if (expanded)
        {
            label.setIcon(UIManager.getIcon("Tree.openIcon"));
        }
        else
        {
            label.setIcon(UIManager.getIcon("Tree.closedIcon"));
        }
    }

    /**
     * ノードとリーフのサイズを取得します。
     * 
     * @return ノードとリーフのサイズ
     */
    public Dimension getPreferredSize()
    {
        Dimension dCheck = check.getPreferredSize();
        Dimension dLabel = label.getPreferredSize();
        return new Dimension(dCheck.width + dLabel.width, (dCheck.height < dLabel.height ? dLabel.height : dCheck.height));
    }

    /**
     * レイアウトを設定します。
     */
    public void doLayout()
    {
        Dimension dCheck = check.getPreferredSize();
        Dimension dLabel = label.getPreferredSize();
        int yCheck = 0;
        int yLabel = 0;
        if (dCheck.height < dLabel.height)
        {
            yCheck = (dLabel.height - dCheck.height) / 2;
        }
        else
        {
            yLabel = (dCheck.height - dLabel.height) / 2;
        }
        check.setLocation(0, yCheck);
        check.setBounds(0, yCheck, dCheck.width, dCheck.height - 2);
        label.setLocation(dCheck.width, yLabel);
        label.setBounds(dCheck.width, yLabel, dLabel.width, dLabel.height - 2);
    }

    /**
     * バックグラウンドを設定します。
     * 
     * @param color バックグラウンド色の設定
     */
    @Override
    public void setBackground(Color color)
    {
        super.setBackground(Color.WHITE);
    }

    /**
     * 選択線を描画します。
     * 
     * @param g
     */
    public void paintComponent(Graphics g)
    {
        super.paintComponent(g);
        if (isTargetNode)
        {
            if (isTargetNodeLeaf)
            {
                g.setColor(Color.red);
                g.drawLine(0, 0, getSize().width, 0);
            }
            else
            {
                g.setColor(Color.orange);
                g.drawRect(0, 0, getSize().width - 1, getSize().height - 1);
            }
        }
    }
    
    
    
    
// ---------------------------------- mouse event ----------------------------------
    
    /**
     * マウス押下時の動作を実装します。
     * 
     * @see java.awt.event.MouseListener#mouseClicked(java.awt.event.MouseEvent)
     */
    public void mouseClicked(MouseEvent e)
    {
//    	int row = tree.getRowForLocation(e.getX(), e.getY());
//        if (row != clickRowIndex)
//        {
//        	clickRowIndex = row;
//            tree.repaint();
//        }
    }

    /**
     * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
     */
    public void mousePressed(MouseEvent e)
    {
    }

    /**
     * @see java.awt.event.MouseListener#mouseEntered(java.awt.event.MouseEvent)
     */
    public void mouseEntered(MouseEvent e)
    {
    }

    /**
     * @see java.awt.event.MouseListener#mouseExited(java.awt.event.MouseEvent)
     */
    public void mouseExited(MouseEvent e)
    {
    }

    /**
     * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
     */
    public void mouseReleased(MouseEvent e)
    {
    }

    /**
     * @see java.awt.event.MouseMotionListener#mouseMoved(java.awt.event.MouseEvent)
     */
    public void mouseMoved(MouseEvent e)
    {
        int row = tree.getRowForLocation(e.getX(), e.getY());
        if (row != rollOverRowIndex)
        {
            rollOverRowIndex = row;
            tree.repaint();
        }
    }

    /**
     * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
     */
    public void mouseDragged(MouseEvent e)
    {
    }
    
    
// ---------------------------------- key event ----------------------------------

	public void keyPressed(KeyEvent e) {
	}

	public void keyReleased(KeyEvent e) {
	}

	public void keyTyped(KeyEvent e) {
//		if (e.getKeyChar() == '') {
//		}
	}
}
