/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.tree;

import java.awt.Component;
import java.awt.Container;

import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.data.DatasetIF;
import jp.nihu.gttime.data.ProjectIF;
import jp.nihu.gttime.layer.LayerViewPart;
import jp.nihu.gttime.plugin.LayerPlugin;
import jp.nihu.gttime.plugin.PluginManager;
import jp.nihu.gttime.plugin.PluginManagerException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * ツリー内のノード操作とレイヤを連動させる為に管理するクラスです。
 */
public class TreeViewPart extends JScrollPane
{
    /**
     * シリアルバーションID
     */
    private static final long serialVersionUID = 1L;

    /**
     * ログ用オブジェクト
     */
    private static final Log log = LogFactory.getLog(TreeViewPart.class);

    /**
     * TreeDragDropオブジェクト
     */
    private TreeDragDrop treeView;

    /**
     * 初期化処理を呼び出すコンストラクター
     */
    public TreeViewPart()
    {
    }

    /**
     * ツリー機能を初期化します。<br/>
     * <ul>
     * <li>自クラスのTreeDragDropを指定します。</li>
     * <li>ツリー関連のオブジェクトを生成します。</li>
     * </ul>
     */
    private void init()
    {
        treeView = new TreeDragDrop(this);

        TreeCheckNode node = new TreeCheckNode();
        DefaultTreeModel mod = new DefaultTreeModel(node);
        treeView.setModel(mod);
        treeView.setShowsRootHandles(false);
        treeView.setRootVisible(false);

        treeView.getSelectionModel().setSelectionMode(TreeSelectionModel.SINGLE_TREE_SELECTION);
        treeView.putClientProperty("JTree.lineStyle", "Angled");

        setViewportView(treeView);
    }

    /**
     * プロジェクトをロードします。
     * 
     * @param project Projectオブジェクト
     */
    public void load(ProjectIF project)
    {
        init();
        treeView.setShowsRootHandles(true);
        treeView.setRootVisible(true);
        createLayers(project);
    }

    /**
     * レイヤリストの再構築
     */
    public void rebuildLayerList()
    {
        count = 0;
        if (treeView != null)
        {
            TreeModel model = treeView.getModel();
            TreeCheckNode node = (TreeCheckNode) model.getRoot();
            rebuildLayerList(node);
            removeLayer();

            JPanel layerList = getLayerList();
            if (layerList != null)
            {
                layerList.updateUI();
            }
            treeView.updateUI();
        }
    }

    /**
     * レイヤリストの再構築
     * 
     * @param project Projectオブジェクト
     */
    public void rebuildLayerList(ProjectIF project)
    {
//    	log.debug("JJJJJJJJJJJJJJJJJJJJJJ 0");
        count = 0;
        TreeModel model = treeView.getModel();
        TreeCheckNode node = (TreeCheckNode) model.getRoot();
        node = updateTree(project, node);
        rebuildLayerList(node);
        removeLayer();
        JPanel layerList = getLayerList();
        layerList.updateUI();
        treeView.updateUI();
    }

    private void removeLayer()
    {
        JPanel layerList = getLayerList();
        Component[] comps = layerList.getComponents();
        int cl = comps.length;
        for (int i = count; i < cl; i++)
        {
            layerList.remove(count);
        }
    }

    /**
     * レイヤリストの再構築
     * 
     * @param node 基点となるノード
     */
    private void rebuildLayerList(TreeCheckNode node)
    {
        Object nodeObj = node.getUserObject();
        if (nodeObj instanceof ProjectIF)
        {
            for (int i = 0; i < node.getChildCount(); i++)
            {
                rebuildLayerList((TreeCheckNode) node.getChildAt(i));
            }
        }
        else if (nodeObj instanceof DatasetIF)
        {
            setLayerToVisible(node);
        }
    }

    private int count = 0;

    /**
     * ツリーの再構築
     * 
     * @param node TreeCheckNodeオブジェクト
     */
    private void setLayerToVisible(TreeCheckNode node)
    {
        boolean flag = true;
        JPanel layerList = getLayerList();
        Component[] comps = layerList.getComponents();
        for (int i = 0; i < comps.length; i++)
        {
            if (comps[i] instanceof LayerPlugin)
            {
                if (((LayerPlugin) comps[i]).getTreeCheckNode().equals(node))
                {
                    flag = false;
                    if (node.isParentEnabled())
                    {
                        Component com = comps[i];
                        layerList.remove(com);
                        layerList.add(com, count);
                        com.setVisible(node.isSelected());
                    }
                    else
                    {
                        comps[i].setVisible(false);
                    }
                    count++;
                }
            }
        }
        if (flag)// CSVなどをインポートしたときに実行される。
        {
            LayerViewPart layerView = getLayerPanel();
            layerView.addLayer(node, count);
            count++;
        }
    }

    /**
     * 数値レイヤの表示に使用するレイヤプラグインを変更します。
     * 
     * @param dataset Dataseteオブジェクト
     * @param node TreeCheckNodeオブジェクト
     * @param name レイヤプラグイン名
     * @param clazz レイヤプラグインクラス名
     */
    public void changeLayerPlugin(DatasetIF dataset, TreeCheckNode node, String name, String clazz)
    {
        LayerViewPart layerViewPart = getLayerPanel();
        JPanel layerList = getLayerList();
        Component[] comps = layerList.getComponents();
        try
        {
            dataset.setLayerType(name);
            node.setUserObject(dataset);
            PluginManager pluginManager = T2Map.getPluginManager();
            LayerPlugin layer;
            layer = (LayerPlugin) pluginManager.newInstance(clazz);

            layer.init(getT2Map().getTimeAxisData(), node);
            for (int j = 0; j < comps.length; j++)
            {
                if (comps[j] instanceof LayerPlugin)
                {
                    LayerPlugin prev = (LayerPlugin) comps[j];
                    if (prev.getTreeCheckNode().equals(node))
                    {
                        layerViewPart.removeLayer(j);
                        layerViewPart.addLayer(layer, j);
                        node.setLayer(layer);
                        dataset.applyStyleChanges(layer.getPluginName(), layer.getStyleName());
                        count = j;
                    }
                }
            }
            node.setIcon(layer.getIcon());
        }
        catch (PluginManagerException e)
        {
            log.fatal(e.getMessage(), e);
        }
        layerList.updateUI();
        treeView.updateUI();
    }

    /**
     * レイヤの新規生成
     * 
     * @param project
     */
    private void createLayers(ProjectIF project)
    {
        getLayerPanel().createLayerList();
        TreeCheckNode node = createTree(project);
        DefaultTreeModel mod = new DefaultTreeModel(node);
        treeView.setModel(mod);
        expandAll();
    }

    /**
     * ツリーを構築します
     * 
     * @param project Projectオブジェクト
     * @param layerList レイヤのパネル
     * @return TreeCheckNodeオブジェクトを取得します。
     */
    private TreeCheckNode createTree(ProjectIF project)
    {
//        TreeCheckNode node = new TreeCheckNode(project, true);
//
//        DatasetIF[] ds = project.getDatasets();
//        LayerViewPart layerView = getLayerPanel();
//        for (int i = 0; i < ds.length; i++)
//        {
//            TreeCheckNode cnode = new TreeCheckNode(ds[i], false);
//            createNode(node, cnode);
//
//            log.debug("Store Dataset HashCode: " + ds[i].hashCode() + ", Record #:" + ds[i].getRecords().length);
//
//            cnode.setTreeView(treeView);
//            layerView.addLayer(cnode);
//        }
//
//        ProjectIF[] projects = project.getProjects();
//        if (projects != null)
//        {
//            for (int j = 0; j < projects.length; j++)
//            {
//                TreeCheckNode cnode = createTree(projects[j]);
//                createNode(node, cnode);
//            }
//        }
    	
    	TreeCheckNode node = new TreeCheckNode(project, true);

        Object[] objs = project.getNodes();
        if (objs != null) {
            LayerViewPart layerView = getLayerPanel();
            for (int i = 0; i < objs.length; i++)
            {
            	if (objs[i] instanceof DatasetIF)
            	{
            		DatasetIF ds = (DatasetIF)objs[i];
                    TreeCheckNode cnode = new TreeCheckNode(objs[i], false);
                    createNode(node, cnode);

                    log.debug("Store Dataset HashCode: " + ds.hashCode() + ", Record #:" + ds.getRecords().length);

                    cnode.setTreeView(treeView);
                    layerView.addLayer(cnode);
            	}
            	else if (objs[i] instanceof ProjectIF)
            	{
            		ProjectIF p = (ProjectIF)objs[i];
            		TreeCheckNode cnode = createTree(p);
                    createNode(node, cnode);
            	}
            }
        }
        return node;
    }

    /**
     * 追加されるデータセットをデータセットの最後に追加します。
     * 
     * @param project Projectオブジェクト
     * @param node TreeCheckNodeオブジェクト
     * @return TreeCheckNodeオブジェクトを取得します。
     */
    private TreeCheckNode updateTree(ProjectIF project, TreeCheckNode node)
    {
//        DatasetIF[] ds = project.getDatasets();
//        ProjectIF[] projects = project.getProjects();
//        int nodeLen = 0;
//        if (projects != null)
//        {
//            nodeLen = node.getChildCount() - projects.length;
//        }
//        else
//        {
//            nodeLen = node.getChildCount();
//        }
//        for (int i = 0; i < ds.length; i++)
//        {
//            if (i >= nodeLen)
//            {
//                TreeCheckNode cnode = new TreeCheckNode(ds[i], false);
//                cnode.setTreeView(treeView);
//                node.insert(cnode, i);
//            }
//        }
    	
//    	log.debug("SSSSSSSSSSSSS 1");
    	
        Object[] objs = project.getNodes();
        int nodeLen = node.getChildCount();
        
        for (int i = 0; objs != null && i < objs.length; i++)
        {
//        	log.debug("SSSSSSSSSSSSS 2");
            if (i >= nodeLen)
            {
//            	log.debug("SSSSSSSSSSSSS 3");
            	if (objs[i] instanceof DatasetIF)
            	{
//            		log.debug("SSSSSSSSSSSSS 4");
            		DatasetIF ds = (DatasetIF)objs[i];
                    TreeCheckNode cnode = new TreeCheckNode(ds, false);
                    createNode(node, cnode);
                    log.debug("Store Dataset HashCode: " + ds.hashCode() + ", Record #:" + ds.getRecords().length);
                    cnode.setTreeView(treeView);
//                    layerView.addLayer(cnode);
            	}
            	else if (objs[i] instanceof ProjectIF)
            	{
//            		log.debug("SSSSSSSSSSSSS 5");
            		ProjectIF p = (ProjectIF)objs[i];
            		TreeCheckNode cnode = createTree(p);
                    createNode(node, updateTree(p, cnode));
            	}
            }
            if (objs[i] instanceof ProjectIF)
        	{
//        		log.debug("SSSSSSSSSSSSS 6");
            	updateTree((ProjectIF)objs[i], (TreeCheckNode)node.getChildAt(i));
        	}
        }
        
        return node;
    }

    /**
     * ノードを構築します。
     * 
     * @param node TreeCheckNodeオブジェクト
     * @param cnode TreeCheckNodeオブジェクト
     */
    private void createNode(TreeCheckNode node, TreeCheckNode cnode)
    {
        node.add(cnode);
    }

    /**
     * ツリーを展開します。
     */
    private void expandAll()
    {
        for (int j = 0; j < treeView.getRowCount(); j++)
        {
            treeView.expandRow(j);
        }
    }

    /**
     * T2Mapオブジェクトを取得します。
     * 
     * @return T2Mapオブジェクト
     */
    private T2Map getT2Map()
    {
        for (Container cont = getParent(); cont != null; cont = cont.getParent())
            if (cont instanceof T2Map)
            {
                return (T2Map) cont;
            }
        return null;
    }

    /**
     * LayerViewPartオブジェクトを取得します。
     * 
     * @return LayerViewPartオブジェクト
     */
    private LayerViewPart getLayerPanel()
    {
        return getT2Map().getLayerPanel();
    }

    /**
     * LayerListを取得します。
     * 
     * @return LayerList
     */
    private JPanel getLayerList()
    {
        return getLayerPanel().getLayerList();
    }

    /**
     * 全てチェックボックスを選択状態にします。
     */
    public void checkAll()
    {
        if (treeView != null)
        {
            for (int j = 0; j < treeView.getRowCount(); j++)
            {
                TreePath path = treeView.getPathForRow(j);
                TreeCheckNode node = (TreeCheckNode) path.getLastPathComponent();
                node.setSelected(true);
            }
            this.rebuildLayerList();
            updateUI();
        }
    }

    /**
     * 全てチェックボックスを非選択状態にします。
     */
    public void uncheckAll()
    {
        if (treeView != null)
        {
            for (int j = 0; j < treeView.getRowCount(); j++)
            {
                TreePath path = treeView.getPathForRow(j);
                TreeCheckNode node = (TreeCheckNode) path.getLastPathComponent();
                node.setSelected(false);
            }
            this.rebuildLayerList();
            updateUI();
        }
    }
}
