/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.utility;

import java.awt.Component;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.File;

import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.SwingUtilities;
import javax.swing.filechooser.FileFilter;
import javax.swing.plaf.FileChooserUI;
import javax.swing.plaf.basic.BasicFileChooserUI;

import jp.nihu.gttime.T2Map;
import jp.nihu.gttime.T2MapFileChooser;
import jp.nihu.gttime.plugin.PluginResourceManager;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * ファイルを保存する際の実行クラスです。
 */
public class ExtensionFileSaveChooser extends T2MapFileChooser implements PropertyChangeListener
{
    /**
     * シリアルバーションID
     */
    private static final long serialVersionUID = 1L;

    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(ExtensionFileSaveChooser.class);

    /**
     * リソースマネージャーオブジェクト
     */
    private static PluginResourceManager rmgr = PluginResourceManager.getManager(ExtensionFileSaveChooser.class, T2Map.LOCALE);

    /**
     * ファイルに保存するダイアログのデフォルト情報をセットします。
     * 
     * @param defaultFileName ファイル名の初期表示文字列
     */
    public ExtensionFileSaveChooser(String defaultFileName)
    {
        log.debug("Initialize ExtensionFileSaveChooser.");
        setLocale(T2Map.LOCALE);
        this.defaultFileName = defaultFileName;
        setAcceptAllFileFilterUsed(false);
        addPropertyChangeListener(this);
    }

    /**
     * ファイルに保存するダイアログのデフォルト情報をセットします。
     * 
     * @param file Fileオブジェクト
     */
    public ExtensionFileSaveChooser(File file)
    {
        this(file.getName());
        if (file != null)
        {
            setSelectedFile(file);
        }
        // this.defaultFileName = file.getName();
        // setAcceptAllFileFilterUsed(false);
        // addPropertyChangeListener(this);
    }

    /**
     * ファイル名が既に存在するか確認します。
     */
    public void approveSelection()
    {
        if (this.getSelectedFile() != null)
        {
            if (this.getSelectedFile().exists() == true)
            {
                int answer = JOptionPane.showConfirmDialog((Component) this, rmgr.getString("superscription.message"), rmgr.getString("superscription.message2"), JOptionPane.YES_NO_OPTION);
                if (answer == JOptionPane.YES_OPTION)
                {
                    super.approveSelection();
                }
            }
            else
            {
                super.approveSelection();
            }
        }
    }

    /**
     * ファイル名の初期表示名
     */
    private String defaultFileName = "default";

    /**
     * ファイルの保存先のディレクトリ
     */
    private String selectedFileDirectory = null;

    /**
     * 保存するファイルのファイル名
     */
    private String selectedFileName = null;

    /**
     * 選択されたファイルを設定します。
     * 
     * @param evt PropertyChangeEventオブジェクト
     */
    public void propertyChange(PropertyChangeEvent evt)
    {
        FileChooserUI ui = getUI();
        if (ui instanceof BasicFileChooserUI)
        {
            String name = ((BasicFileChooserUI) ui).getFileName();
            if (name != null && !name.equals(""))
            {
                int lastDot = name.lastIndexOf(".");
                if (lastDot > -1)
                {
                    selectedFileName = name.substring(0, lastDot);
                }
            }
            else
            {
                selectedFileName = defaultFileName;
            }
        }
        if (evt.getPropertyName().equals(JFileChooser.FILE_FILTER_CHANGED_PROPERTY))
        {
            selectedFileDirectory = getCurrentDirectory().getAbsolutePath();
            final String fileName = selectedFileDirectory + File.separator + selectedFileName + "." + getExt();
            final File selectedFile = new File(fileName);
            SwingUtilities.invokeLater(new Runnable()
            {
                public void run()
                {
                    setSelectedFile(selectedFile);
                }
            });
        }
        else if (evt.getPropertyName().equals(JFileChooser.SELECTED_FILE_CHANGED_PROPERTY))
        {
            File selectedFile = getSelectedFile();
            if (selectedFile == null)
            {
                if (selectedFileDirectory != null)
                {
                    String name = selectedFileDirectory + File.separator + selectedFileName + "." + getExt();
                    final File recoveredSelectedFile = new File(name);
                    SwingUtilities.invokeLater(new Runnable()
                    {
                        public void run()
                        {
                            setSelectedFile(recoveredSelectedFile);
                        }
                    });
                }
            }
            else
            {
                selectedFileDirectory = getCurrentDirectory().getAbsolutePath();
            }
        }
        else if (evt.getPropertyName().equals(JFileChooser.DIRECTORY_CHANGED_PROPERTY))
        {
            File newValue = (File) evt.getNewValue();
            if (newValue != null)
            {
                selectedFileDirectory = newValue.getAbsolutePath();
            }
            if (selectedFileDirectory != null)
            {
                final String fileName = selectedFileDirectory + File.separator + selectedFileName + "." + getExt();
                final File selectedFile = new File(fileName);
                SwingUtilities.invokeLater(new Runnable()
                {
                    public void run()
                    {
                        setSelectedFile(selectedFile);
                    }
                });
            }
        }
    }
    
    private String getExt() {
    	FileFilter fileFilter = this.getFileFilter();
        String ext = "";
        if (fileFilter instanceof ExtensionFileFilter) {
        	ext = ((ExtensionFileFilter)fileFilter).getExt();
        } else {
        	ext = fileFilter.getDescription().toLowerCase();
        }
        return ext;
    }
}
