/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.utility;

import java.io.IOException;
import java.io.StringReader;
import java.text.SimpleDateFormat;
import java.util.Date;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.xpath.XPathExpressionException;

import jp.nihu.gttime.data.DatasetException;
import jp.nihu.gttime.data.DatasetIF;
import jp.nihu.gttime.time.utility.JulianGregorianDate;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.w3c.dom.Document;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

public class TMM
{
    /**
     * ログ用オブジェクト
     */
    private static Log log = LogFactory.getLog(TMM.class);

    /**
     * 現在のデータセットの状態をtmmファイルに記入する。
     * 
     * @param title データセット名
     * @param layertype レイヤタイプ
     * @param dataset Datasetオブジェクト
     * @return Documentオブジェクト
     */
    public static Document create(String title, String layertype, DatasetIF dataset, String format)
    {
        StringBuffer sb = new StringBuffer();

        Date date = new Date();
        SimpleDateFormat uniqName = new SimpleDateFormat("yyyyMMddhhmmssSSS");
        String sDate = null;
        String eDate = null;
        Document domTmm = null;
        try
        {
            if (dataset.getStartDatetime() != null)
            {
                sDate = JulianGregorianDate.toGregorian(dataset.getStartDatetime(), "Z");
            }
            if (dataset.getEndDatetime() != null)
            {
                eDate = JulianGregorianDate.toGregorian(dataset.getEndDatetime(), "Z");
            }

            Document mdom = dataset.getDomDataset();
            sb.append("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n");
            sb.append("<METADATA>\n");
            sb.append(getMETA(mdom, "dc.title", title));
            sb.append(getMETA(mdom, "dc.description"));
            sb.append(getMETA(mdom, "dc.pulisher"));
            sb.append(getMETA(mdom, "dc.date", uniqName.format(date)));
            sb.append(getMETA(mdom, "dc.type"));
            sb.append(getMETA(mdom, "dc.format", format));
            sb.append(getMETA(mdom, "dc.language"));
            sb.append(getMETA(mdom, "dc.coverage.x.min"));
            sb.append(getMETA(mdom, "dc.coverage.x.max"));
            sb.append(getMETA(mdom, "dc.coverage.y.min"));
            sb.append(getMETA(mdom, "dc.coverage.y.max"));
            sb.append(getMETA(mdom, "dc.coverage.t.early"));
            if (sDate != null)
            {
                sb.append(getMETA(mdom, "dc.coverage.t.early", sDate));
            }
            if (eDate != null)
            {
                sb.append(getMETA(mdom, "dc.coverage.t.late", eDate));
            }
            sb.append(getMETA(mdom, "dc.rights"));
            sb.append(getMETA(mdom, "dc.subject.domain"));
            sb.append(getMETA(mdom, "ecai.team"));
            sb.append(getMETA(mdom, "tm.Server.Type"));
            sb.append(getMETA(mdom, "tm.Spatial.MapObjType"));
            sb.append(getMETA(mdom, "tm.Time.TimeVarying"));
            sb.append(getMETA(mdom, "tm.Filter.AccessibleFields"));
            sb.append(getMETA(mdom, "tm.TableData.TableName"));
            sb.append(getMETA(mdom, "tm.TableData.KeyIDfield"));
            sb.append(getMETA(mdom, "tm.TableData.Labelfield"));
            sb.append(getMETA(mdom, "tm.sys.Local.GISType"));
            sb.append(getMETA("t2map.layertype", null, null, layertype));
            sb.append("</METADATA>\n");

            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            DocumentBuilder db = dbf.newDocumentBuilder();
            domTmm = db.parse(new InputSource(new StringReader(sb.toString())));
        }
        catch (DatasetException e)
        {
            log.fatal(e.getMessage(), e);
        }
        catch (XPathExpressionException e)
        {
            log.fatal(e.getMessage(), e);
        }
        catch (ParserConfigurationException e)
        {
            log.fatal(e.getMessage(), e);
        }
        catch (SAXException e)
        {
            log.fatal(e.getMessage(), e);
        }
        catch (IOException e)
        {
            log.fatal(e.getMessage(), e);
        }
        return domTmm;
    }

    private static String getMETA(Document mdom, String name, String content) throws XPathExpressionException
    {
        return getMETA(name, getTMCODE(mdom, name), getSCHEMECODE(mdom, name), content);
    }

    private static String getMETA(Document mdom, String name) throws XPathExpressionException
    {
        return getMETA(name, getTMCODE(mdom, name), getSCHEMECODE(mdom, name), getCONTENT(mdom, name));
    }

    private static String getMETA(String name, String tmcode, String schemecode, String content)
    {
        StringBuffer sb = new StringBuffer();
        sb.append("<META NAME=\"").append(name).append("\" ");
        if (tmcode != null)
        {
            sb.append("TMCODE=\"").append(tmcode).append("\" ");
        }
        if (schemecode != null)
        {
            sb.append("SCHEMECODE=\"").append(schemecode).append("\" ");
        }
        sb.append("CONTENT=\"").append(content).append("\"/>\n");
        return sb.toString();
    }

    private static String getTMCODE(Document mdom, String name) throws XPathExpressionException
    {
        return XMLUtil.getValue(mdom, "/METADATA/META[@NAME='" + name + "']/@TMCODE", null);
    }

    private static String getSCHEMECODE(Document mdom, String name) throws XPathExpressionException
    {
        return XMLUtil.getValue(mdom, "/METADATA/META[@NAME='" + name + "']/@SCHEMECODE", null);
    }

    private static String getCONTENT(Document mdom, String name) throws XPathExpressionException
    {
        return XMLUtil.getValue(mdom, "/METADATA/META[@NAME='" + name + "']/@CONTENT", null);
    }
}
