/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.utility;

import java.awt.Color;
import java.text.DecimalFormat;
import java.text.Format;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.List;
import java.util.regex.PatternSyntaxException;

import javax.swing.text.JTextComponent;

import jp.nihu.gttime.time.utility.ISO8601Exception;
import jp.nihu.gttime.time.utility.JulianGregorianDate;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * {@code JTextComponent} の入力値を検証するユーティリティクラス.
 */
public abstract class TextValidator
{

    /**
     * 日時を検証する.
     */
    public static class DATE extends TextValidator
    {

        /**
         * 指定したパターンで日時検証を行う {@code TextValidator} を作成する. 必須入力検証は行わない.
         * 
         * @param c 検証対象となる {@code JTextComponent}
         * @param pattern 入力値を検証するための {@code DateFormat} パターン
         */
        public DATE(JTextComponent c, String pattern)
        {
            this(c, pattern, false);
        }

        /**
         * 指定したパターンで日時検証を行う {@code TextValidator} を作成する.
         * 
         * @param c 検証対象となる {@code JTextComponent}
         * @param pattern 入力値を検証するための {@code DateFormat} パターン
         * @param required 必須入力検証を行う場合は {@code true}
         */
        public DATE(JTextComponent c, String pattern, boolean required)
        {
            super(c, pattern, required);
        }

        /**
         * 指定したパターンを初期値に持つ {@code DateFormat} を返却する.
         */
        @Override
        protected Format createFormat(String pattern)
        {
            return new SimpleDateFormat(pattern);
        }

        /**
         * パターン検証以外の検証
         */
        @Override
        protected boolean validate(String s)
        {
            boolean b = false;
            try
            {
                // DateFormat で認識可能な値でも、JulianGregorianDate では
                // 例外となるため.
                // JulianGregorianDate が DateFormat に対応した場合は、
                // 本メソッドを削除すること.
                JulianGregorianDate.toJulian(s);
                b = true;
            }
            catch (ISO8601Exception e)
            {
                // Log [debug]
                log.debug(s + ":" + e.getMessage());
            }
            return b;
        }
    }

    /**
     * 数値を検証する.
     */
    public static class NUMBER extends TextValidator
    {
        private Number min;

        private Number max;

        /**
         * 指定したパターンで数値検証を行う {@code TextValidator} を作成する. 必須入力検証は行わない.
         * 
         * @param c 検証対象となる {@code JTextComponent}
         * @param pattern 入力値を検証するための {@code DecimalFormat} パターン
         */
        public NUMBER(JTextComponent c, String pattern)
        {
            this(c, pattern, false);
        }

        /**
         * 指定したパターンで数値検証を行う {@code TextValidator} を作成する.
         * 
         * @param c 検証対象となる {@code JTextComponent}
         * @param pattern 入力値を検証するための {@code DecimalFormat} パターン
         * @param required 必須入力検証を行う場合は {@code true}
         */
        public NUMBER(JTextComponent c, String pattern, boolean required)
        {
            super(c, pattern, required);
        }

        /**
         * 指定したパターンで数値検証を行う {@code TextValidator} を作成する. 必須入力検証は行わない.
         * 
         * @param c 検証対象となる {@code JTextComponent}
         * @param pattern 入力値を検証するための {@code DecimalFormat} パターン
         * @param min 最小値. 検証を行わない場合は {@code null}
         * @param max 最大値. 検証を行わない場合は {@code null}
         */
        public NUMBER(JTextComponent c, String pattern, Number min, Number max)
        {
            this(c, pattern, false, min, max);
        }

        /**
         * 指定したパターンで数値検証を行う {@code TextValidator} を作成する.
         * 
         * @param c 検証対象となる {@code JTextComponent}
         * @param pattern 入力値を検証するための {@code DecimalFormat} パターン
         * @param required 必須入力検証を行う場合は {@code true}
         * @param min 最小値. 検証を行わない場合は {@code null}
         * @param max 最大値. 検証を行わない場合は {@code null}
         */
        public NUMBER(JTextComponent c, String pattern, boolean required, Number min, Number max)
        {
            this(c, pattern, required);
            this.min = min;
            this.max = max;
        }

        /**
         * 指定したパターンを初期値に持つ {@code DecimalFormat} を返却する.
         */
        @Override
        protected Format createFormat(String pattern)
        {
            return new DecimalFormat(pattern);
        }

        /**
         * 範囲の検証.
         */
        @Override
        protected boolean validate(String s)
        {
            boolean b = true;
            try
            {
                double d = Double.parseDouble(s);
                if (min != null)
                {
                    b &= (min.doubleValue() <= d);
                }
                if (max != null)
                {
                    b &= (d <= max.doubleValue());
                }
            }
            catch (Exception e)
            {
                b = false;
            }
            return b;
        }
    }

    /**
     * 任意の入力値を正規表現を用いて検証する.
     */
    public static class REGEX extends TextValidator
    {

        /**
         * 指定した正規表現での検証を行う {@code TextValidator} を作成する.
         * 
         * @param c 検証対象となる {@code JTextComponent}
         * @param pattern 入力値を検証するため正規表現
         */
        public REGEX(JTextComponent c, String pattern)
        {
            this(c, pattern, false);
        }

        /**
         * 指定した正規表現での検証を行う {@code TextValidator} を作成する.
         * 
         * @param c 検証対象となる {@code JTextComponent}
         * @param pattern 入力値を検証するため正規表現
         * @param required 必須入力検証を行う場合は {@code true}
         */
        public REGEX(JTextComponent c, String pattern, boolean required)
        {
            super(c, pattern, required);
        }

        /**
         * 正規表現での検証
         */
        @Override
        protected boolean validate(String s)
        {
            boolean b = false;
            try
            {
                b = s.matches(getPattern());
            }
            catch (PatternSyntaxException e)
            {
                // Log [warn]
                log.warn(e, e);
            }
            return b;
        }
    }

    /**
     * 必須入力値を検証する.
     */
    public static class REQUIRED extends TextValidator
    {
        /**
         * 必須入力検証を行う {@code TextValidator} を作成する.
         * 
         * @param c 検証対象となる {@code JTextComponent}
         */
        public REQUIRED(JTextComponent c)
        {
            super(c, null, true);
        }
    }

    /**
     * 入力文字の長さを検証する.
     */
    public static class LENGTH extends TextValidator
    {
        private int length;

        /**
         * 文字列長の検証を行う {@code TextValidator} を作成する. 必須入力検証は行わない.
         * 
         * @param c 検証対象となる {@code JTextComponent}
         * @param length 最大文字列長
         */
        public LENGTH(JTextComponent c, int length)
        {
            this(c, false, length);
        }

        /**
         * 文字列長の検証を行う {@code TextValidator} を作成する.
         * 
         * @param c 検証対象となる {@code JTextComponent}
         * @param required 必須入力検証を行う場合は {@code true}
         * @param length 最大文字列長
         */
        public LENGTH(JTextComponent c, boolean required, int length)
        {
            super(c, null, required);
            this.length = length;
        }

        /**
         * 文字列長の検証
         */
        @Override
        protected boolean validate(String s)
        {
            return (s.length() <= length);
        }
    }

    private static Log log = LogFactory.getLog(TextValidator.class);

    public static final Color OK = Color.WHITE;

    public static final Color NG = Color.RED;

    private JTextComponent c;

    private String pattern;

    private Format format;

    private boolean required;

    /**
     * 指定したパターンで検証を行う {@code TextValidator} を作成する.
     * 
     * @param c 検証対象となる {@code JTextComponent}
     * @param pattern 検証に使用するパターン文字列
     * @param required 必須入力検証を行う場合は {@code true}
     */
    protected TextValidator(JTextComponent c, String pattern, boolean required)
    {
        this.c = c;
        this.pattern = pattern;
        this.required = required;
        this.format = createFormat(pattern);
    }

    /**
     * 検証に使用する {@code Format} を作成する.
     */
    protected Format createFormat(String pattern)
    {
        return null;
    }

    /**
     * 検証対象となる {@code JTextComponent} を返却する.
     */
    protected JTextComponent getComponent()
    {
        return c;
    }

    /**
     * 検証に使用するパターン文字列を返却する.
     */
    protected String getPattern()
    {
        return pattern;
    }

    /**
     * 検証に使用する {@code Format} を返却する.
     */
    protected Format getFormat()
    {
        return format;
    }

    /**
     * 必須入力検証を行う場合は {@code true} を返却する.
     */
    protected boolean isRequierd()
    {
        return required;
    }

    /**
     * {@code JTextComponent} の入力値検証を行う.
     */
    public boolean validate()
    {
        boolean b = true;
        String s = getComponent().getText();
        if (s != null && s.trim().length() > 0)
        {
            Format f = getFormat();
            if (f != null)
            {
                try
                {
                    f.parseObject(s);
                }
                catch (ParseException e)
                {
                    b = false;
                    // Log [debug]
                    log.debug(s + "invalid value.", e);
                }
            }
            if (b)
            {
                b = validate(s);
            }
        }
        else
        {
            b = !isRequierd();
        }
        getComponent().setBackground((b) ? OK : NG);
        return b;
    }

    /**
     * サブクラス側で独自に検証が必要な場合は、本メソッドをオーバーライドして 検証結果を返却する。
     */
    protected boolean validate(String s)
    {
        return true;
    }

    /**
     * {@code JTextComponent} の入力値検証を行う.
     * 
     * @param list {@code JTextComponent} を要素に持つ {@code List}
     * @return 全ての検証が真の場合 true.
     */
    public static boolean validate(List<TextValidator> list)
    {
        return validate(list.toArray(new TextValidator[0]));
    }

    /**
     * {@code JTextComponent} の入力値検証を行う.
     * 
     * @param validators {@code JTextComponent} 配列
     * @return 全ての検証が真の場合 true.
     */
    public static boolean validate(TextValidator[] validators)
    {
        boolean b = true;
        for (TextValidator v : validators)
        {
            b &= v.validate();
        }
        return b;
    }
}
