/* ===========================================================
 * GT-Time
 * ===========================================================
 *
 * Copyright (C) 2010 National Institutes for the Humanities. 
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3.0 of the License, or
 * (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public
 * License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110, USA.
 *
 * [Java is a trademark or registered trademark of Sun Microsystems, Inc.
 * in the United States and other countries.]
 */

package jp.nihu.gttime.utility;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.Iterator;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.xml.XMLConstants;
import javax.xml.namespace.NamespaceContext;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;

import org.apache.xerces.dom.DOMImplementationImpl;
import org.w3c.dom.DOMImplementation;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.ls.DOMImplementationLS;
import org.w3c.dom.ls.LSOutput;
import org.w3c.dom.ls.LSSerializer;

/**
 * XPathを実行するユーティリティークラス
 */
public class XMLUtil
{
    private static final GtsNamespaceContext gtsNs = new GtsNamespaceContext();

    /**
     * 引数のドキュメントに対して、引数のXPathを実行します。
     * 
     * @param doc ドキュメント
     * @param pathExpr XPath
     * @return XPath実行結果
     * @throws XPathExpressionException
     */
    public static String getValue(Object doc, String pathExpr, NamespaceContext context) throws XPathExpressionException
    {
        XPathFactory factory = XPathFactory.newInstance();
        javax.xml.xpath.XPath xpath = factory.newXPath();
        if (context != null)
        {
            xpath.setNamespaceContext(context);
        }
        XPathExpression expr = xpath.compile(pathExpr);
        return (String) expr.evaluate(doc, XPathConstants.STRING);
    }

    public static Element getElement(Object doc, String xpathStr, NamespaceContext context) throws XPathExpressionException
    {
        Element elem = null;
        XPathFactory factory = XPathFactory.newInstance();
        XPath xpath = factory.newXPath();
        xpath.setNamespaceContext(XMLUtil.getGtsNsContext());
        XPathExpression expr = xpath.compile(xpathStr);
        elem = (Element) expr.evaluate(doc, XPathConstants.NODE);
        return elem;
    }
    
    public static NodeList getElements(Object doc, String xpathStr, NamespaceContext context) throws XPathExpressionException
    {
        NodeList elems = null;
        XPathFactory factory = XPathFactory.newInstance();
        XPath xpath = factory.newXPath();
        xpath.setNamespaceContext(XMLUtil.getGtsNsContext());
        XPathExpression expr = xpath.compile(xpathStr);
        elems = (NodeList) expr.evaluate(doc, XPathConstants.NODESET);
        return elems;
    }

    public static void saveXml(Document doc, File file) throws IOException
    {
        File parent = file.getParentFile();
        if (parent.exists() || parent.mkdirs())
        {
            DOMImplementation domImpl = new DOMImplementationImpl();
            DOMImplementationLS domImplLS = (DOMImplementationLS) domImpl.getFeature("LS", "3.0");
            LSOutput lsOutput = domImplLS.createLSOutput();
            LSSerializer lsSerializer = domImplLS.createLSSerializer();
            FileOutputStream fos = new FileOutputStream(file);
            lsOutput.setByteStream(fos);
            lsSerializer.write(doc, lsOutput);
            fos.close();
        }
    }
    
    public static byte[] getByteXml(Document doc) throws IOException
    {
        DOMImplementation domImpl = new DOMImplementationImpl();
        DOMImplementationLS domImplLS = (DOMImplementationLS) domImpl.getFeature("LS", "3.0");
        LSOutput lsOutput = domImplLS.createLSOutput();
        LSSerializer lsSerializer = domImplLS.createLSSerializer();
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        lsOutput.setByteStream(out);
        lsSerializer.write(doc, lsOutput);
        byte[] b = out.toByteArray();
        out.close();
        return b;
    }
    
    public static ByteArrayInputStream getByteInputStream(Document doc) throws IOException
    {
        DOMImplementation domImpl = new DOMImplementationImpl();
        DOMImplementationLS domImplLS = (DOMImplementationLS) domImpl.getFeature("LS", "3.0");
        LSOutput lsOutput = domImplLS.createLSOutput();
        LSSerializer lsSerializer = domImplLS.createLSSerializer();
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        lsOutput.setByteStream(out);
        lsSerializer.write(doc, lsOutput);
        byte[] b = out.toByteArray();
        out.close();
        return new ByteArrayInputStream(b);
    }
    
    public static void printXml(Document doc) throws IOException
    {
//        DOMImplementation domImpl = new DOMImplementationImpl();
//        DOMImplementationLS domImplLS = (DOMImplementationLS) domImpl.getFeature("LS", "3.0");
//        LSOutput lsOutput = domImplLS.createLSOutput();
//        LSSerializer lsSerializer = domImplLS.createLSSerializer();
//        lsOutput.setByteStream(System.out);
//        lsSerializer.write(doc, lsOutput);
    }

    public static GtsNamespaceContext getGtsNsContext()
    {
        return gtsNs;
    }

    private static class GtsNamespaceContext implements NamespaceContext
    {
        public String getNamespaceURI(String prefix)
        {
            if (prefix == null)
            {
                throw new NullPointerException("Null prefix");
            }
            else if ("hgis".equals(prefix))
            {
                return "http://www.h-gis.org/hgis";
            }
            else if ("gt".equals(prefix))
            {
                return "http://www.h-gis.org/gt";
            }
            else if ("xml".equals(prefix))
            {
                return XMLConstants.XML_NS_URI;
            }
            else if ("xsi".equals(prefix))
            {
                return XMLConstants.W3C_XML_SCHEMA_INSTANCE_NS_URI;
            }
            return XMLConstants.NULL_NS_URI;
        }

        public String getPrefix(String namespaceURI)
        {
            throw new UnsupportedOperationException();
        }

        public Iterator<Object> getPrefixes(String namespaceURI)
        {
            throw new UnsupportedOperationException();
        }
    }
    
    public static String getElementContentsAsString(Node e) {
		StringBuilder c = new StringBuilder();
		NodeList nl = e.getChildNodes();
		for (int i = 0; i < nl.getLength(); i++) {
			Node n = nl.item(i);
			if (n.getNodeType() == Node.ELEMENT_NODE) {
				c.append("<").append(n.getNodeName());
				NamedNodeMap as = n.getAttributes();
				for (int j = 0; j < as.getLength(); j++) {
					Node a = as.item(j);
					c.append(" ").append(a.getNodeName()).append("=\"").append(a.getNodeValue()).append("\"");
				}
				c.append(">");
				c.append(getElementContentsAsString(n));
				c.append("</").append(n.getNodeName()).append(">");
			} else if (n.getNodeType() == Node.TEXT_NODE && 0 < n.getNodeValue().trim().length()) {
				String val = n.getNodeValue().trim();
				//val = replaceUrlToLink(val);
				c.append(val);
			} else if (n.getNodeType() == Node.CDATA_SECTION_NODE) {
				String val = n.getNodeValue().trim();
//				val = val.replaceAll("&", "&amp;").replaceAll("<", "&lt;").replaceAll(">", "&gt;").replaceAll("\"", "&quot;");
				val = val.replaceAll("&lt;", "&amp;lt;").replaceAll("&gt;", "&amp;gt;").replaceAll("&quot;", "&amp;quot;").replaceAll("<", "&lt;").replaceAll(">", "&gt;").replaceAll("\"", "&quot;");
				val = replaceUrlToLinkInCdata(val);
				c.append(val);
			}
		}
		return replaceUrlToLink(c.toString());
	}
    
    private static String regex = "https?://[\\w!/:;%#\\$&\\?\\(\\)~\\.=\\+\\-@,]+";
    
    private static String replaceUrlToLink(String val) {
    	if (val.trim().matches("^" + regex + "$")) {
    		return "<a href=\"" + val + "\">" + val + "</a>";
    	}
    	return val;
    }
    
    private static String replaceUrlToLinkInCdata(String val) {
    	String[] vals = val.split("&quot;");
    	boolean a = val.endsWith("&quot;");
    	val = "";
        for (int i = 0; i < vals.length; i++) {
        	String[] valss = vals[i].split("&lt;");
        	boolean b = vals[i].endsWith("&lt;");
        	vals[i] = "";
        	for (int j = 0; j < valss.length; j++) {
        		String[] valsss = valss[j].split("&gt;");
        		boolean c = valss[j].endsWith("&gt;");
        		valss[j] = "";
            	for (int k = 0; k < valsss.length; k++) {
                    Pattern p = Pattern.compile(regex);
                    Matcher m = p.matcher(valsss[k]);
                    valsss[k] = m.replaceAll("<a href=\"$0\">$0</a>");
            		if (k != 0) {
            			valss[j] += "&gt;";
                    }
            		valss[j] += valsss[k];
            	}
            	if (c) valss[j] += "&gt;";
        		if (j != 0) {
        			vals[i] += "&lt;";
                }
        		vals[i] += valss[j];
        	}
        	if (b) vals[i] += "&lt;";
            if (i != 0) {
                val += "&quot;";
            }
            val += vals[i];
        }
        if (a) val += "&quot;";
		return val;
    }
}
